/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: fifocsv.c,v 1.15 2005/11/17 23:03:40 phk Exp $
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <err.h>
#include <time.h>
#include <sys/endian.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <sys/sbuf.h>
#include "../measured_proto.h"
#include "../fifolog.h"

static int do_debug = 0;
static int do_meta = 0;
static int do_record = 0;

struct col {
	unsigned		id;
	int			flags;
#define HAVE_VAL	(1 << 0)
#define HAVE_META	(1 << 1)
#define HAVE_RECORD	(1 << 2)
	TAILQ_ENTRY(col)	list;
	char			*val;
	struct sbuf		*txt;
	char			*p;
	struct sbuf		*record;
	char			*q;
};

static TAILQ_HEAD(, col)	collist =
    TAILQ_HEAD_INITIALIZER(collist);

static struct col	*printfcol;

static unsigned rows;

static int
rd32(FILE *fi, uint32_t *u)
{
	unsigned char buf[4];
	int i;

	if (feof(fi))
		return (0);
	i = fread(buf, 4, 1, fi);
	*u = be32dec(buf);
	return (i);
}

static uint32_t
rd8(FILE *fi)
{
	int i;

	i = fgetc(fi);
	return (i);
}

static void
swallow_d(FILE *fi)
{
	int i;
	double d;

	i = fread(&d, sizeof d, 1, fi);
}

static void
swallow_n(FILE *fi, int len)
{
	int i;

	while (len--) {
		i = fgetc(fi);
	}
}

static void
swallow_s(FILE *fi)
{
	int i;

	do {
		i = fgetc(fi);
	} while(i != '\0');
}

static void
Pass1(FILE *fi)
{
	uint32_t u, id, v;
	struct col *cp, *cp2;
	int flag;

	id = 0;
	while (rd32(fi, &u) == 1) {
		if (u == FIFO_TIMESTAMP) {
			swallow_n(fi, sizeof(time_t));
			if (feof(fi))
				break;
			continue;
		}
		if (u == FIFO_PRINTF) {
			swallow_s(fi);
			if (feof(fi))
				break;
			continue;
		}
		if (u == FIFO_DEBUG) {
			swallow_s(fi);
			if (feof(fi))
				break;
			continue;
		} 

		flag = 0;
		id = u >> 2;
		switch(u & 3) {
		case FIFO_VALUE:
			flag |= HAVE_VAL;
			swallow_d(fi);
			break;
		case FIFO_LIMIT:
			flag |= HAVE_META;
			swallow_n(fi, 5);
			break;
		case FIFO_CONFIG:
			flag |= HAVE_META;
			v = rd8(fi);
			if (!(v & FIFO_ATTRIB)) {
				switch (v) {
				case E_NEW_GROUP:
					swallow_s(fi);
					break;
				case E_NEW_POINT:
					swallow_n(fi, 1);
					break;
				default:
					break;
				}
			} else  {
				switch(v & ~FIFO_ATTRIB) {
				case A_LABEL:
				case A_FORMAT:
				case A_UNIT:
					swallow_s(fi);
					break;
				default:
					swallow_d(fi);
					break;
				}
			}
			break;
		case FIFO_RECORD:
			flag |= HAVE_RECORD;
			v = rd8(fi);
			swallow_n(fi, v);
		}

		if (feof(fi))
			break;

		TAILQ_FOREACH(cp, &collist, list)
			if (cp->id >= id)
				break;

		if (cp != NULL && cp->id == id) {
			cp->flags |= flag;
			continue;
		}

		cp2 = calloc(sizeof *cp2, 1);
		cp2->id = id;
		cp2->flags |= flag;
		if (cp == NULL)
			TAILQ_INSERT_TAIL(&collist, cp2, list);
		else
			TAILQ_INSERT_BEFORE(cp, cp2, list);
	}
	u = 0;
	id = 0;
	TAILQ_FOREACH(cp, &collist, list) {
		id++;
		if (cp->flags & HAVE_VAL) {
			u++;
			asprintf(&cp->val, "%03u.%03u.%03u",
			    SITE_OF(cp->id), GROUP_OF(cp->id),
			    POINT_OF(cp->id));
		}

		if (do_meta && (cp->flags & HAVE_META))
			u++;
		cp->txt = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
		sbuf_printf(cp->txt, "\n%03u.%03u.%03u\nMETA",
		    SITE_OF(cp->id), GROUP_OF(cp->id),
		    POINT_OF(cp->id));

		if (do_record && (cp->flags & HAVE_RECORD))
			u++;
		cp->record = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
		sbuf_printf(cp->record, "\n%03u.%03u.%03u\nRECORD",
		    SITE_OF(cp->id), GROUP_OF(cp->id),
		    POINT_OF(cp->id));
	}
	fprintf(stderr, "%u ids\n", id);
	fprintf(stderr, "%u columns\n", 4 + u);
}

static void
line(time_t t)
{
	struct col *cp;
	char *p, *q;
	int i;
	char buf[100];
	struct tm tm;

	if (t != 0) {
		gmtime_r(&t, &tm);
		strftime(buf, sizeof buf, "\"%Y-%m-%d\",\"%H:%M:%S\"", &tm);
	} else {
		strcpy(buf, "\"UTC DATE\",\"UTC TIME\"");
	}

	i = 0;
	TAILQ_FOREACH(cp, &collist, list) {
		if ((cp->flags & HAVE_VAL) && cp->val != NULL)
			i++;
		if (do_meta && (cp->flags & HAVE_META)) {
			sbuf_finish(cp->txt);
			if (sbuf_len(cp->txt) > 0) {
				i++;
				cp->p = sbuf_data(cp->txt) + 1;
			} else {
				sbuf_clear(cp->txt);
				cp->p = NULL;
			}
		} 
		if (do_record && (cp->flags & HAVE_RECORD)) {
			sbuf_finish(cp->record);
			if (sbuf_len(cp->record) > 0) {
				i++;
				cp->q = sbuf_data(cp->record) + 1;
			} else {
				sbuf_clear(cp->record);
				cp->q = NULL;
			}
		}
	}
	while (i) {
		rows++;
		i = 0;
		printf("%s", buf);
		TAILQ_FOREACH(cp, &collist, list) {
			if (cp->flags & HAVE_VAL) {
				if (cp->val == NULL) {
					printf(",");
				} else {
					printf(",\"%s\"", cp->val);
					free(cp->val);	
					cp->val = NULL;
				}
			}
			if (do_meta && (cp->flags & HAVE_META)) {
				if (cp->p == NULL) {
					printf(",");
				} else {
					p = strchr(cp->p, '\n');
					if (p != NULL) {
						i++;
						*p = '\0';
					}
					printf(",\"%s\"", cp->p);
					if (p == NULL) {
						sbuf_clear(cp->txt);
						cp->p = NULL;
					} else {
						cp->p = p + 1;
					}
				}
			} else {
				sbuf_clear(cp->txt);
			}
			if (do_record && (cp->flags & HAVE_RECORD)) {
				if (cp->q == NULL) {
					printf(",");
				} else {
					q = strchr(cp->q, '\n');
					if (q != NULL) {
						i++;
						*q = '\0';
					}
					printf(",\"%s\"", cp->q);
					if (q == NULL) {
						sbuf_clear(cp->record);
						cp->q = NULL;
					} else {
						cp->q = q + 1;
					}
				}
			} else {
				sbuf_clear(cp->record);
			}
		}
		printf("\r\n");
	}
}

static char *
get_s(FILE *fi)
{
	static char buf[BUFSIZ];
	char *cp;
	int i;

	cp = buf;
	while (1) {
		i = fgetc(fi);
		if (i == EOF)
			i = '\0';
		*cp++ = i;
		if (i == 0)
			break;
	}
	return (buf);
}

static double
get_d(FILE *fi)
{
	double d;
	int i;

	i = fread(&d, sizeof d, 1, fi);
	if (i != 1)
		d = NAN;
	return (d);
}

static void
Pass2(FILE *fi)
{
	uint32_t u, id, v, x;
	struct col *cp;
	time_t t;

	t = 0;
	line(t);
	while (rd32(fi, &u) == 1) {
		if (u == FIFO_TIMESTAMP) {
			line(t);
			rd32(fi, &t);
			continue;
		} 
		if (u == FIFO_PRINTF) {
			if (printfcol->val != NULL)
				line(t);
			printfcol->val = strdup(get_s(fi));
			continue;
		}
		if (u == FIFO_DEBUG) {
			sbuf_printf(printfcol->txt, "\n%s", get_s(fi));
			continue;
		}
		id = u >> 2;
		TAILQ_FOREACH(cp, &collist, list)
			if (cp->id == id)
				break;
		if (cp == NULL) {
			fprintf(stderr, "%03u.%03u.%03u not found\n",
			    SITE_OF(id), GROUP_OF(id), POINT_OF(id));
			assert(cp != NULL);
		}
		switch(u & 3) {
		case FIFO_VALUE:
			if (cp->val != NULL)
				line(t);
			asprintf(&cp->val, "%g", get_d(fi));
			break;
		case FIFO_LIMIT:
			v = rd8(fi);
			rd32(fi, &x);
			if (feof(fi))
				break;
			switch(v) {
			case L_LOW:		sbuf_printf(cp->txt, "\nLimit=LOW"); break;
			case L_SAG:		sbuf_printf(cp->txt, "\nLimit=SAG"); break;
			case L_RAISE:		sbuf_printf(cp->txt, "\nLimit=RAISE"); break;
			case L_HIGH:		sbuf_printf(cp->txt, "\nLimit=HIGH"); break;
			case L_TIMEOUT:		sbuf_printf(cp->txt, "\nLimit=TIMEOUT"); break;
			case L_NORMAL:		sbuf_printf(cp->txt, "\nLimit=NORMAL"); break;
			default:
				fprintf(stderr, "%03u.%03u.%03u limit %d\n",
				    SITE_OF(id), GROUP_OF(id), POINT_OF(id), v);
				assert(1 == 0);
				break;
			}
			sbuf_printf(cp->txt, "\nTrap=%u", x);
			break;
		case FIFO_CONFIG:
			v = rd8(fi);
			if (feof(fi))
				break;
			if (!(v & FIFO_ATTRIB)) {
				switch (v) {
				case E_NEW_SITE:
					sbuf_printf(cp->txt, "\nNEW_SITE");
					break;
				case E_DESTROY_SITE:
					sbuf_printf(cp->txt, "\nDESTROY_SITE");
					break;
				case E_NEW_GROUP:
					sbuf_printf(cp->txt, "\nNEW_GROUP=%s", get_s(fi));
					break;
				case E_DESTROY_GROUP:
					sbuf_printf(cp->txt, "\nDESTROY_POINT");
					break;
				case E_NEW_POINT:
					x = rd8(fi);
					if (feof(fi))
						break;
					sbuf_printf(cp->txt, "\nCREATE_POINT");
					switch(x) {
					case UNDEF:	sbuf_printf(cp->txt, "\nTYPE=UNDEF"); break;
					case AIN:	sbuf_printf(cp->txt, "\nTYPE=AIN"); break;
					case DIN:	sbuf_printf(cp->txt, "\nTYPE=DIN"); break;
					case AOUT:	sbuf_printf(cp->txt, "\nTYPE=AOUT"); break;
					case DOUT:	sbuf_printf(cp->txt, "\nTYPE=DOUT"); break;
					case COUNT:	sbuf_printf(cp->txt, "\nTYPE=COUNTER"); break;
					default:
						fprintf(stderr, "LINE %d %d\n", __LINE__, x);
						assert(1 == 0);
						break;
					}
					break;
				case E_DESTROY_POINT:
					sbuf_printf(cp->txt, "\nDESTROY_POINT");
					break;
				default: 
					fprintf(stderr, "%03u.%03u.%03u config %d\n",
					    SITE_OF(id), GROUP_OF(id), POINT_OF(id), v);
					assert(1 == 0);
					break;
				}
			} else {
				switch(v & ~FIFO_ATTRIB) {
				case A_LABEL:	   sbuf_printf(cp->txt, "\nLABEL=%s", get_s(fi)); break;
				case A_FORMAT:	   sbuf_printf(cp->txt, "\nFORMAT=%s", get_s(fi)); break;
				case A_UNIT:	   sbuf_printf(cp->txt, "\nUNIT=%s", get_s(fi)); break;
				case A_LOW:	   sbuf_printf(cp->txt, "\nLOW=%g", get_d(fi)); break;
				case A_SAG:	   sbuf_printf(cp->txt, "\nSAG=%g", get_d(fi)); break;
				case A_RAISE:	   sbuf_printf(cp->txt, "\nRAISE=%g", get_d(fi)); break;
				case A_HIGH:	   sbuf_printf(cp->txt, "\nHIGH=%g", get_d(fi)); break;
				case A_LIMITDELAY: sbuf_printf(cp->txt, "\nLIMITDELAY=%g", get_d(fi)); break;
				case A_HYSTERESIS: sbuf_printf(cp->txt, "\nHYSTERESIS=%g", get_d(fi)); break;
				default:
					fprintf(stderr, "%03u.%03u.%03u attrib %d\n",
					    SITE_OF(id), GROUP_OF(id), POINT_OF(id), v);
					assert(1 == 0);
					break;
				}
			}
			break;
		case FIFO_RECORD:
			x = rd8(fi);
			if (do_record)
				line(t);
			sbuf_printf(cp->record, "\n0x");
			while (x > 0) {
				v = rd8(fi);
				sbuf_printf(cp->record, "%02x", v);
				x--;
			}
			if (do_record)
				line(t);
			break;
		}
	}
	line(t);
	fprintf(stderr, "%u rows\n", rows);
}

int
main(int argc, char **argv)
{
	FILE *fi;
	int ch;

	while ((ch = getopt(argc, argv, "dmr")) != -1) {
		switch (ch) {
		case 'd':
			do_debug = !do_debug;
			break;
		case 'm':
			do_meta = !do_meta;
			break;
		case 'r':
			do_record = !do_record;
			break;
		default:
			errx(2, 
			    "Usage: %s [-dm] [input_file]", argv[0]);
		}
	}
	argc -= optind;
	argv += optind;
	if (argc == 0)
		errx(2, "No file argument");
	if (argc != 1)
		errx(2, "Too many file arguments");

	fi = fopen(argv[0], "r");
	if (fi == NULL)
		err(2, "Cannot open %s", argv[0]);

	Pass1(fi);
	fclose(fi);

	printfcol = calloc(sizeof *printfcol, 1);
	printfcol->id = FIFO_PRINTF;
	printfcol->txt = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
	printfcol->record = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
	printfcol->flags |= HAVE_VAL;
	if (do_debug)
		printfcol->flags |= HAVE_META;
	sbuf_printf(printfcol->txt, "\nDEBUG");
	asprintf(&printfcol->val, "PRINTF");
	TAILQ_INSERT_HEAD(&collist, printfcol, list);

	fi = fopen(argv[0], "r");
	if (fi == NULL)
		err(2, "Cannot reopen %s", argv[0]);
	Pass2(fi);
	fclose(fi);
	return (0);
}
