/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: fiforead.c,v 1.26 2008/03/22 16:46:21 phk Exp $
 *
 */

#include <assert.h>
#include <err.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <vis.h>

#include <sys/queue.h>

#include "fifolog.h"
#include "libfifolog.h"
#include "../pattern.h"

static int opt_a;
static time_t opt_B;
static int opt_c;
static int opt_d = 1;
static time_t opt_E;
static time_t opt_i;
static const char *opt_F = "H";
static const char *opt_T;
static const char *opt_o;
static const char *opt_p;
static const char *opt_R;
static regex_t R;
static int opt_r = 0;

static FILE *fo;
static struct pattern *pattern;

struct lastv {
	TAILQ_ENTRY(lastv)	list;
	unsigned		u;
	char 			*last;
};

static TAILQ_HEAD(, lastv)	lasthdr =
    TAILQ_HEAD_INITIALIZER(lasthdr);

static void
Render(void *priv __unused, time_t now, unsigned flag, const unsigned char *p, unsigned l)
{
	static struct tm utc;
	char buf[128];
	static time_t t_last;
	char *x;
	int i;
	uint32_t u;
	struct lastv *lv, *lv2;

	if (now < opt_B || now > opt_E)
		return;

	if (flag & FIFOLOG_LENGTH) {
		if (!opt_r)
			return;
	} else {
		if (!opt_d)
			return;
		if (opt_c && memcmp(p, "P VALUE ", 8))
			return;
	}

	if (pattern != NULL && !MatchPattern(pattern, flag))
		return;

	if (opt_R != NULL && regexec(&R, (const char *)p, 0, NULL, 0))
		return;

	if (opt_i && now / opt_i == t_last)
		return;
	if (opt_i)
		t_last = now / opt_i;

	if (opt_T != NULL) {
		(void)gmtime_r(&now, &utc);
		i = strftime(buf, sizeof buf, opt_T, &utc);
		assert(i > 0);
	} else {
		sprintf(buf, "%12ld", (long)now);
	}

	sprintf(strchr(buf, '\0'),
	    " %03u.%03u.%03u %s",
	    (flag >> 20) & 0x3ff,
	    (flag >> 10) & 0x3ff,
	    (flag >> 0) & 0x3ff,
	    (flag & FIFOLOG_LENGTH) ? "r " : "");
		
	if (flag & FIFOLOG_LENGTH) {
		if (*opt_F == 'V') {
			x = malloc(l * 4 + 1);
			assert(x != NULL);
			strvisx(x, p, l, VIS_CSTYLE | VIS_NL | VIS_TAB);
			fprintf(fo, "%s%s\n", buf, x);
			free(x);
			return;
		}
		fprintf(fo, "%s", buf);
		for (; l > 0; l--) {
			if (opt_a && *p >= ' ' && *p <= '~')
				fprintf(fo, "  %c", *p++);
			else
				fprintf(fo, " %02x", *p++);
		}
		fprintf(fo, "\n");
		return;
	} 

	if (opt_c) {
		u &= ~FIFOLOG_TIMESTAMP;
		TAILQ_FOREACH(lv, &lasthdr, list)
			if (lv->u >= u)
				break;
		if (lv == NULL || lv->u > u) {
			lv2 = calloc(sizeof *lv2, 1);
			assert(lv2 != NULL);
			lv2->u = u;
			lv2->last = strdup(p);
			assert(lv2->last != NULL);
			if (lv != NULL)
				TAILQ_INSERT_BEFORE(lv, lv2, list);
			else
				TAILQ_INSERT_TAIL(&lasthdr, lv2, list);
		} else if (strcmp(lv->last, p)) {
			lv->last = realloc(lv->last, strlen(p) + 1);
			assert(lv->last != NULL);
			strcpy(lv->last, p);
		} else
			return;
	}

	fprintf(fo, "%s%s\n", buf, p);
}

/*--------------------------------------------------------------------*/

static void
Usage(void)
{
	fprintf(stderr,
		"Usage: fiforead [options] fifofile\n"
		"\t-b <start time integer>\n"
		"\t-B <start time>\n"
		"\t-c # toggle 'show only changes'\n"
		"\t-d # toggle data display\n"
		"\t-e <end time integer>\n"
		"\t-E <end time>\n"
		"\t-F <record formatting [V/h]>\n"
		"\t-i <interval>\n"
		"\t-o <output file>\n"
		"\t-p <pattern>\n"
		"\t-r # toggle record display\n"
		"\t-R <regexp> # match regexp\n"
		"\t-t # format timestamps as %%Y%%m%%d%%H%%M%%S\n"
		"\t-T <timestamp format>\n"
	);
	exit (2);
}

int
main(int argc, char **argv)
{
	int ch, i;
	off_t o;
	struct fifolog_reader *fl;
	const char *progname;

	progname=argv[0];

	time(&opt_E);
	opt_o = "-";
	while ((ch = getopt(argc, argv, "ab:cB:de:E:F:i:o:p:rR:tT:")) != -1) {
		switch (ch) {
		case 'a':
			opt_a = !opt_a;
			break;
		case 'b':
			opt_B = strtoul(optarg, NULL, 0);
			break;
		case 'B':
			opt_B = get_date(optarg);
			if (opt_B == -1)
				errx(1, "Didn't understand \"%s\"", optarg);
			break;
		case 'c':
			opt_c = !opt_c;
			break;
		case 'd':
			opt_d = !opt_d;
			break;
		case 'e':
			opt_E = strtoul(optarg, NULL, 0);
			break;
		case 'E':
			opt_E = get_date(optarg);
			if (opt_E == -1)
				errx(1, "Didn't understand \"%s\"", optarg);
			break;
		case 'F':
			opt_F = optarg;
			break;
		case 'i':
			opt_i = strtoul(optarg, NULL, 0);
			break;
		case 'o':
			opt_o = optarg;
			break;
		case 'p':
			opt_p = optarg;
			break;
		case 'r':
			opt_r = !opt_r;
			break;
		case 'R':
			opt_R = optarg;
			break;
		case 't':
			opt_T = "%Y%m%d%H%M%S";
			break;
		case 'T':
			opt_T = optarg;
			break;
		default:
			Usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (opt_R != NULL) {
		i = regcomp(&R, opt_R, REG_NOSUB);
		if (i != 0) {
			char buf[BUFSIZ];
			regerror(i, &R, buf, sizeof buf);
			fprintf(stderr, "-R argument: %s\n", buf);
			exit (1);
		}
	}

	fprintf(stderr, "From\t%ld %s", (long)opt_B, ctime(&opt_B));
	fprintf(stderr, "To\t%ld %s", (long)opt_E, ctime(&opt_E));
	if (opt_B >= opt_E)
		errx(1, "Begin time not before End time");
	if (!(opt_d || opt_r))
		errx(1, "Neither Data nor Record selected");
	if (opt_d)
		fprintf(stderr, "Data\tselected\n");
	if (opt_r)
		fprintf(stderr, "Record\tselected\n");
	if (opt_c)
		fprintf(stderr, "Show only changes\n");

	if (argv[0] == NULL)
		errx(1, "Usage: %s [options] fifolog", progname);
	fl = fifolog_reader_open(argv[0]);
	
	if (!strcmp(opt_o, "-"))
		fo = stdout;
	else {
		fo = fopen(opt_o, "w");
		if (fo == NULL)
			err(1, "Cannot open: %s", argv[1]);
	}

	if (opt_p != NULL)
		pattern = CompilePattern(opt_p);

	o = fifolog_reader_seek(fl, opt_B);
	fifolog_reader_process(fl, o, Render, NULL, opt_E);
	return (0);
}
