/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: alcatel_ioo.c,v 1.12 2008/02/17 09:55:54 phk Exp $
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <configkit.h>
#include <assert.h>
#include <poll.h>
#include <unistd.h>
#include <string.h>
#include <regex.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/types.h>
#include <md5.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <paths.h>

#include "measured.h"
#include "phk_tcp.h"
#include "configkit_conf.h"

#define MAX_RULE			200

#define DEFAULT_KEEPALIVE		10

enum notificationType {
	alarmNew = 0,
	alarmRaise,
	alarmClear,
	alarmList,
	alarmPurge
};

struct alcatel_ioo_rule {
	unsigned			magic;
#define ALCATEL_IOO_RULE_MAGIC		0x60c81d0d
	struct ck_instance		cki;
	unsigned			rule;
	struct alcatel_ioo		*sc;
	char				*match;
	regex_t				regex;
	unsigned			count;
	struct point			*point;
};

struct alcatel_ioo_alarm {
	unsigned			magic;
#define ALCATEL_IOO_ALARM_MAGIC		0x445e157a
	intmax_t			alarmId;
	unsigned			rule;
	enum notificationType		state;
	TAILQ_ENTRY(alcatel_ioo_alarm)	list;
};

struct alcatel_ioo {
	unsigned			magic;
#define ALCATEL_IOO_MAGIC 		0x0c7fde55
	struct ck_instance		cki;
	struct group			*gp;
	pthread_t			tp;
	pthread_mutex_t			lock;
	unsigned			debug;
	struct alcatel_ioo_rule		*rule[MAX_RULE];
	char				*port;
	char				*host;
	char				*password;
	struct timespec			last;
	unsigned			keepalive;
	int				running;
	int				fd;
	double				retry;
	enum {
	    IOO_CONN,
	    IOO_START_ALARMS,
	    IOO_RUN
	}				state;
	char				buf[BUFSIZ];
	int				bufptr;
	TAILQ_HEAD(,alcatel_ioo_alarm)	alarms;
};


/*--------------------------------------------------------------------*/

static void
alcatel_ioo_purgestate(struct alcatel_ioo *sc)
{
	struct alcatel_ioo_alarm *ap;
	int i;

	AZ(pthread_mutex_lock(&sc->lock));
	while (!TAILQ_EMPTY(&sc->alarms)) {
		ap = TAILQ_FIRST(&sc->alarms);
		TAILQ_REMOVE(&sc->alarms, ap, list);
		free(ap);
	}
	for (i = 1; i < MAX_RULE; i++) {
		if (sc->rule[i] == NULL || sc->rule[i]->match == NULL)
			continue;
		sc->rule[i]->count = 0;
		SetVal(sc->rule[i]->point, 0);
	}
	AZ(pthread_mutex_unlock(&sc->lock));
}

static void
alcatel_ioo_alarm(struct alcatel_ioo *sc, intmax_t nbr, enum notificationType state, char *p)
{
	struct alcatel_ioo_alarm *ap;
	struct alcatel_ioo_rule *rp;
	int i, j;

	TAILQ_FOREACH(ap, &sc->alarms, list)
		if (ap->alarmId == nbr)
			break;
	if (ap == NULL) {
		ap = calloc(sizeof *ap, 1);
		assert(ap != NULL);
		ap->magic = ALCATEL_IOO_ALARM_MAGIC;
		ap->alarmId = nbr;
		TAILQ_INSERT_TAIL(&sc->alarms, ap, list);
		for (i = 1; i < MAX_RULE; i++) {
			if (sc->rule[i] == NULL || sc->rule[i]->match == NULL)
				continue;
			j = regexec(&sc->rule[i]->regex, p, 0, NULL, 0);
			printf("r%d %Q -> %d\n",
				i, sc->rule[i]->match, j);
			if (j == 0) {
				ap->rule = i;
				break;
			}
		}
	}
	printf("%d -> %jd r%d <%s>\n", state, nbr, ap->rule, p);
	rp = sc->rule[ap->rule];
	if (rp == NULL)
		return;
	switch (ap->state) {
	case alarmNew:
		assert (state == alarmRaise || state == alarmList);
		rp->count++;
		SetVal(rp->point, rp->count);
		break;
	case alarmClear:
		assert (state == alarmRaise || state == alarmPurge);
		if (state == alarmRaise) {
			rp->count++;
			SetVal(rp->point, rp->count);
		}
		break;
	case alarmRaise:
	case alarmList:
		assert (state == alarmClear || state == alarmPurge);
		rp->count--;
		SetVal(rp->point, rp->count);
		break;
	default:
		assert(0 == __LINE__);
	}
	ap->state = state;
	if (state == alarmPurge) {
		TAILQ_REMOVE(&sc->alarms, ap, list);
		ap->magic = 0;
		free(ap);
	}
}


static char *
skip_pfx(char *p, const char *prefix)
{
	int i;

	i = strlen(prefix);
	if (memcmp(p, prefix, i))
		return (NULL);
	return (p + i);
}

static void
alcatel_ioo_alarmnotif(struct alcatel_ioo *sc, char *p)
{
	char *q, *r;
	enum notificationType state;
	intmax_t nbr;

	if (sc->debug & 4)
		SCDebug(sc, "alarm <%s>", p);

	q = skip_pfx(p, "notificationType=alarm");
	if (q == NULL) {
		SCLog(sc, "syntax: ! %s <%s>", "notificationType", p);
		return;
	}
	if ((r = skip_pfx(q, "Raise|")) != NULL) {
		state = alarmRaise;
	} else if ((r = skip_pfx(q, "Clear|")) != NULL) {
		state = alarmClear;
	} else if ((r = skip_pfx(q, "Purge|")) != NULL) {
		state = alarmPurge;
	} else if ((r = skip_pfx(q, "List|")) != NULL) {
		state = alarmList;
	} else {
		SCLog(sc, "syntax: ? %s <%s>", "notificationType", p);
		return;
	}
	printf("%.40s\n", r);
	if (state != alarmPurge) {
		q = skip_pfx(r, "currentAlarmId=");
		if (q == NULL) {
			SCLog(sc, "syntax: ? %s <%s>", "currentAlarmId=", p);
			return;
		}
		nbr = strtoimax(q, &r, 10);
		if (r == NULL || *r != '|') {
			SCLog(sc, "syntax: ? alarmid <%s>", p);
			return;
		}
	} else {
		q = skip_pfx(r, "purgeAlarmsList={");
		if (q == NULL) {
			SCLog(sc, "syntax: ? %s <%s>", "purgeAlarmsList={", p);
			return;
		}
		nbr = strtoimax(q, &r, 10);
		if (r == NULL || *r != '}') {
			SCLog(sc, "syntax: ? alarmid <%s>", p);
			return;
		}
	}
	AZ(pthread_mutex_lock(&sc->lock));
	alcatel_ioo_alarm(sc, nbr, state, r + 1);
	AZ(pthread_mutex_unlock(&sc->lock));
}

static void
alcatel_ioo_unit(struct alcatel_ioo *sc, int len)
{
	char *q, *s;
	int i;

	if (sc->debug & 2)
		SCDebug(sc, "unit <%.*s>", len, sc->buf);
	q = strchr(sc->buf, '[');
	assert(q != NULL);
	*q++ = '\0';
	if (sc->state == IOO_CONN && !strcmp(sc->buf, "CON_CONF")) {
		(void)fdprintf(sc->fd, "START_UNSOL_ALARMS_REQ[]");
		sc->state = IOO_START_ALARMS;
		return;
	}
	if (sc->state == IOO_START_ALARMS &&
	    !strcmp(sc->buf, "START_UNSOL_ALARMS_CONF")) {
		sc->state = IOO_RUN;
		return;
	}
	if (!strcmp(sc->buf, "HEARTBEAT_CONF")) {
		AZ(pthread_mutex_lock(&sc->lock));
		for (i = 1; i < MAX_RULE; i++) {
			if (sc->rule[i] == NULL || sc->rule[i]->match == NULL)
				continue;
			SetVal(sc->rule[i]->point, sc->rule[i]->count);
		}
		AZ(pthread_mutex_unlock(&sc->lock));
		return;
	}
	assert(sc->state == IOO_RUN);
	if (!strcmp(sc->buf, "ALARM_NOTIF")) {
		alcatel_ioo_alarmnotif(sc, q);
		return;
	}
	printf("%s\n", sc->buf);
	while (1) {
		s = strchr(q, '|');
		if (s != NULL) {
			printf("\t%.*s\n", s - q, q);
			q = s + 1;
			continue;
		}
		printf("\t%s\n", q);
		break;
	}
	
}


static void
alcatel_ioo_rx(struct alcatel_ioo *sc)
{
	char *p;
	int i, j;

	sc->buf[sc->bufptr] = '\0';
	if (sc->debug & 1)
		SCDebug(sc, "buf <%s>", sc->buf);
	for (j = 0; j < sc->bufptr; j += i) {
		i = sc->bufptr - j;
		if (i > 255)
			i = 255;
		Record(GI(sc->gp) + 0, i, sc->buf + j);
	}

	p = strchr(sc->buf, ']');
	if (p == NULL)
		return;
	*p = '\0';
	j = ++p - sc->buf;
	alcatel_ioo_unit(sc, j);
	memcpy(sc->buf, p, sc->bufptr - j);
	sc->bufptr -= j;
	alcatel_ioo_rx(sc);
}

static int
alcatel_ioo_thread_loop(struct alcatel_ioo *sc)
{
	struct pollfd pfd[1];
	int i, j;
	struct timespec now;

	pfd[0].fd = sc->fd; /* XXX */
	pfd[0].events = POLLIN | POLLERR | POLLHUP;
	i = poll(pfd, 1, 10000);
	now = Now();
	if (i < 0) {
		Log(GI(sc->gp), "Poll failed (%s)",
		    strerror(errno));
		return (1);
	} else if (i == 0) {
		if (sc->last.tv_sec + 3.0 * sc->keepalive <
		     now.tv_sec) {
			Log(GI(sc->gp), "Connection timed out (%s)",
			    strerror(errno));
			return(1);
		}
		if (fdprintf(sc->fd, "HEARTBEAT_REQ[]")) {
			Log(GI(sc->gp), "Connection write failed (%s)",
			    strerror(errno));
			return(1);
		}
	} else {
		j = sizeof (sc->buf) - sc->bufptr;
		j = read(sc->fd, sc->buf + sc->bufptr, j - 1);
		if (j <= 0) {
			Log(GI(sc->gp), "Read failed (%d/%s)", j,
			    strerror(errno));
			return(1);
		}
		sc->last = now;
		sc->bufptr += j;
		alcatel_ioo_rx(sc);
	}
	return (0);
}

static void *
alcatel_ioo_thread(void *priv)
{
	struct alcatel_ioo *sc;
	struct timeval tv;
	struct linger linger;

	CAST_OBJ_NOTNULL(sc, priv, ALCATEL_IOO_MAGIC);

	printf("Start alcatel_ioo thread\n");
	while (1) {
		if (sc->fd >= 0) {
			close(sc->fd);
			sc->fd = -1;
		}
		sc->fd = PhkTcpOpen(sc->host,
		    sc->port == NULL ? "3001" : sc->port);
		if (sc->fd < 0) {
			Log(GI(sc->gp),
			    "Connect failed, trying again in %g sec",
			    sc->retry);
			usleep((int)(sc->retry * 1e6));
			continue;
		}

		Log(GI(sc->gp), "Connected");

		sc->last = Now();

		alcatel_ioo_purgestate(sc);

		tv.tv_sec = sc->keepalive;
		tv.tv_usec = 0;
		assert(0 == setsockopt(sc->fd, SOL_SOCKET, SO_RCVTIMEO,
		    &tv, sizeof tv));
		assert(0 == setsockopt(sc->fd, SOL_SOCKET, SO_SNDTIMEO,
		    &tv, sizeof tv));

		linger.l_onoff = 0;
		linger.l_linger = 0;
		assert(0 == setsockopt(sc->fd, SOL_SOCKET, SO_LINGER,
		    &linger, sizeof linger));

		if (fdprintf(sc->fd, "CON_REQ[KEY=%s]", sc->password))
			continue;
		sc->state = IOO_CONN;
		sc->bufptr = 0;

		while (!alcatel_ioo_thread_loop(sc))
			continue;
	}
}

static void
alcatel_ioo_reconf(CK_ARG, struct alcatel_ioo *sc)
{
	void *vp;

	if (sc->running) {
		AZ(pthread_mutex_lock(&sc->lock));
		AZ(pthread_cancel(sc->tp));
		AZ(pthread_join(sc->tp, &vp));
		AZ(pthread_mutex_unlock(&sc->lock));
		if (sc->fd >= 0)
			close(sc->fd);
		sc->fd = -1;
		sc->running = 0;
	}
	if (sc->host != NULL) {
		AZ(pthread_create(&sc->tp, NULL, alcatel_ioo_thread, sc));
		sc->running = 1;
	}
}

static void
alcatel_ioo_init(CK_ARG, struct alcatel_ioo *sc)
{

	sc->fd = -1;
	sc->retry = 3;
	sc->keepalive = DEFAULT_KEEPALIVE;
	TAILQ_INIT(&sc->alarms);

}

static void
alcatel_ioo_deinit(CK_ARG, struct alcatel_ioo *sc)
{

	free(sc->host);
	sc->host = NULL;
	free(sc->port);
	sc->port = NULL;
	alcatel_ioo_reconf(CK_PASS, sc);
}

int
cfg_alcatel_ioo_match(CK_ARG, const char **arg_0)
{
	struct alcatel_ioo_rule *rp;
	char buf[BUFSIZ];
	int i;

	CAST_OBJ_NOTNULL(rp, cfa->priv, ALCATEL_IOO_RULE_MAGIC);
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (rp->match != NULL)
			ck_printf(CK_PASS, "%s %Q\n", cfa->tree, rp->match);
		return (0);
	}
	assert(arg_0 != NULL);
	assert(*arg_0 != NULL);
	AZ(pthread_mutex_lock(&rp->sc->lock));
	if (rp->match != NULL) {
		free(rp->match);
		rp->match = NULL;
		regfree(&rp->regex);
	}
	if (!(cfa->flag & (CK_NO))) {
		rp->match = strdup(*arg_0);
		i = regcomp(&rp->regex, rp->match, REG_EXTENDED);
		if (i != 0) {
			free(rp->match);
			rp->match = NULL;
			(void)regerror(i, &rp->regex, buf, sizeof buf);
			ck_printf(CK_PASS, "Invalid regexp: %s", buf);
			regfree(&rp->regex);
		}
	}
	AZ(pthread_mutex_unlock(&rp->sc->lock));
	return (0);
}

struct ck_instance *
cfg_alcatel_ioo_rule_new(CK_ARG, unsigned u)
{
	struct alcatel_ioo *sc;
	struct alcatel_ioo_rule *rp;

	CAST_OBJ_NOTNULL(sc, cfa->priv, ALCATEL_IOO_MAGIC);
	if (u < 1 || u >= MAX_RULE) {
		ck_printf(CK_PASS, "Illegal rule number %u [1...%u]",
		    u, MAX_RULE - 1);
		return (NULL);
	}
	AZ(pthread_mutex_lock(&sc->lock));
	rp = sc->rule[u];
	if (rp == NULL) {
		ALLOC_OBJ(rp, ALCATEL_IOO_RULE_MAGIC);
		rp->rule = u;
		sc->rule[u] = rp;
	}
	rp->cki.obj = rp;
	rp->sc = sc;
	rp->point = CreatePoint(sc->gp, u, AIN);
	AZ(pthread_mutex_unlock(&sc->lock));
	return (&rp->cki);
}

int
cfg_alcatel_ioo_rule_destroy(CK_ARG)
{
	struct alcatel_ioo_rule *rp;
	struct alcatel_ioo *sc;

	CAST_OBJ(rp, cfa->priv, ALCATEL_IOO_RULE_MAGIC);
	if (rp == NULL)
		return (0);

	sc = rp->sc;
	AZ(pthread_mutex_lock(&sc->lock));
	DestroyPoint(rp->point);
	sc->rule[rp->rule] = NULL;
	AZ(pthread_mutex_unlock(&sc->lock));
	return (0);
}

WordVar(alcatel_ioo, ALCATEL_IOO_MAGIC, port, port,
    alcatel_ioo_reconf(CK_PASS, sc));
WordVar(alcatel_ioo, ALCATEL_IOO_MAGIC, host, host,
    alcatel_ioo_reconf(CK_PASS, sc));
WordVar(alcatel_ioo, ALCATEL_IOO_MAGIC, password, password,
    alcatel_ioo_reconf(CK_PASS, sc));
UintVar(alcatel_ioo, ALCATEL_IOO_MAGIC, keepalive, keepalive,
    DEFAULT_KEEPALIVE, alcatel_ioo_reconf(CK_PASS, sc));
UintVar(alcatel_ioo, ALCATEL_IOO_MAGIC, debug, debug, 0, );

MODULE_GROUP(alcatel_ioo, ALCATEL_IOO_MAGIC);
