/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: circutor.c,v 1.21 2008/03/22 16:01:03 phk Exp $
 *
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <configkit.h>
#include <pthread.h>
#include <unistd.h>
#include <termios.h>
#include <sys/endian.h>

#include "measured.h"
#include "modbus.h"
#include "configkit_conf.h"

struct modbus_pt {
	unsigned		reg;
	enum point_type		type;
	const char		*fmt;
	double 			scale;
	const char		*unit;
	const char		*label;
};

struct modbus_dev {
	unsigned		magic;
#define MODBUS_DEV_MAGIC	0x0a44708f
	unsigned		r0;
	unsigned		r1;
	unsigned		nreg;
	struct modbus_pt	*regs;
};

/* CVMk-B ------------------------------------------------------------*/

static struct modbus_pt  cvmkb_pt[] = {
	{  2, AIN,	"%.0f",	1,	"V", 	"Voltage (1)" },
	{  4, AIN,	"%.3f",	1e-3,	"A", 	"Current (1)" },
	{  6, AIN,	"%.0f",	1,	"W", 	"Active Power (1)" },
	{  8, AIN,	"%.1f",	1,	"VA", 	"varL (1)" },
	{ 10, AIN,	"%.0f",	1,	"VA", 	"varC (1)" },
	{ 12, AIN,	"%.0f",	1,	"%%", 	"PF (1)" },

	{ 14, AIN,	"%.0f",	1,	"V", 	"Voltage (2)" },
	{ 16, AIN,	"%.3f",	1e-3,	"A", 	"Current (2)" },
	{ 18, AIN,	"%.0f",	1,	"W", 	"Active Power (2)" },
	{ 20, AIN,	"%.1f",	1,	"VA", 	"varL (2)" },
	{ 22, AIN,	"%.0f",	1,	"VA", 	"varC (2)" },
	{ 24, AIN,	"%.0f",	1,	"%%", 	"PF (2)" },

	{ 26, AIN,	"%.0f",	1,	"V", 	"Voltage (3)" },
	{ 28, AIN,	"%.3f",	1e-3,	"A", 	"Current (3)" },
	{ 30, AIN,	"%.0f",	1,	"W", 	"Active Power (3)" },
	{ 32, AIN,	"%.1f",	1,	"VA", 	"varL (3)" },
	{ 34, AIN,	"%.0f",	1,	"VA", 	"varC (3)" },
	{ 36, AIN,	"%.0f",	1,	"%%", 	"PF (3)" },

	{ 50, AIN,	"%.1f",	.1,	"Hz", 	"Frequency" },

	{ 54, AIN,	"%.0f", 1,	"V",	"Voltage (1-2)"},
	{ 56, AIN,	"%.0f", 1,	"V",	"Voltage (2-3)"},
	{ 58, AIN,	"%.0f", 1,	"V",	"Voltage (3-1)"},

	{ 62, COUNT,	"%.0f", 1,	"W",	"+Energy" },
	{ 64, COUNT,	"%.0f", 1,	"VA",	"+varL" },
	{ 66, COUNT,	"%.0f", 1,	"VA",	"+varC" },
	{ 70, COUNT,	"%.0f", 1,	"W",	"-Energy" },
	{ 72, COUNT,	"%.0f", 1,	"VA",	"-varL" },
	{ 74, COUNT,	"%.0f", 1,	"VA",	"-varC" },
};

static struct modbus_dev cvmkb_dev = {
	.magic =		MODBUS_DEV_MAGIC,
	.r0 =			2,
	.r1 =			75,
	.nreg =			sizeof(cvmkb_pt)/sizeof(cvmkb_pt[0]),
	.regs =			cvmkb_pt
};

/* MK-LCD ------------------------------------------------------------*/

static struct modbus_pt  mklcd_pt[] = {
	{  0, AIN,	"%.1f",	0.1,	"V", 	"Voltage" },
	{  2, AIN,	"%.3f",	1e-3,	"A", 	"Current" },
	{  4, AIN,	"%.0f",	1,	"W", 	"Active Power" },
	{  6, AIN,	"%.1f",	0.1,	"Hz", 	"Frequency" },
	{  8, COUNT,	"%.0f",	1,	"Wh", 	"Active Energy" },
	{ 10, COUNT,	"%.0f",	1,	"Wh", 	"Partial Active Energy" },
};

static struct modbus_dev mklcd_dev = {
	.magic =		MODBUS_DEV_MAGIC,
	.r0 =			0,
	.r1 =			11,
	.nreg =			sizeof(mklcd_pt)/sizeof(mklcd_pt[0]),
	.regs =			mklcd_pt
};

/* Softc -------------------------------------------------------------*/

struct circutor {
	unsigned		magic;
#define CIRCUTOR_MAGIC		0x683486e5
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;

	unsigned		debug;
	struct point		**pp;

	unsigned		modbus;
	unsigned		unit;
	char 			*model;
	struct modbus_req	*req0;
};

/* ConfigKit routines ------------------------------------------------*/

static void
circutor_modbus_cb(void *priv1, void *priv2, unsigned char *buf, unsigned l)
{
	struct circutor *sc;
	struct modbus_dev *md;
	unsigned u, v;

	CAST_OBJ_NOTNULL(sc, priv1, CIRCUTOR_MAGIC);
	CAST_OBJ_NOTNULL(md, priv2, MODBUS_DEV_MAGIC);

	if (buf[1] & 0x80) {
		printf("Error: %02x %02x %02x\n",
			buf[0], buf[1], buf[2]);
		return;
	}

	for (u = 0; u < md->nreg; u++) {
		v = 3 + (md->regs[u].reg - md->r0) * 2;
		if (v + 4 <= 5 + l * 2 - 2) {
			SetVal(sc->pp[u],
			    (int)be32dec(buf + v) * md->regs[u].scale);
		}
	}
}

static void
circutor_reg(struct circutor *sc, struct modbus_req **rp, struct modbus_dev *md)
{
	unsigned u;

	if (*rp != NULL) {
		ModBus_Unregister(*rp);
		CleanGroup(sc->gp);
		*rp = NULL;
	}
	if (md == NULL)
		return;

	sc->pp = calloc(sizeof *sc->pp, md->nreg);
	assert(sc->pp != NULL);
	for (u = 0; u < md->nreg; u++) {
		sc->pp[u] = CreatePoint(sc->gp, u + 1, md->regs[u].type);
		SetPointLabel(sc->pp[u], md->regs[u].label);
		SetPointUnit(sc->pp[u], md->regs[u].unit);
		SetPointFormat(sc->pp[u], md->regs[u].fmt);
	}
	
	*rp = ModBus_Register(
	    sc->modbus, sc->unit, md->r0, 1 + md->r1 - md->r0,
	    circutor_modbus_cb, sc, md);
}

int
cfg_circutor_modbus(CK_ARG, unsigned *bus, unsigned *unit, const char **model)
{
	struct circutor *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, CIRCUTOR_MAGIC);
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		ck_printf(CK_PASS, "%s %u %u %s\n",
		    cfa->tree, sc->modbus, sc->unit, sc->model);
		return (0);
	}
	(void)ck_UintVar(CK_PASS, bus, &sc->modbus, 0);
	(void)ck_UintVar(CK_PASS, unit, &sc->unit, 0);
	(void)ck_WordVar(CK_PASS, model, &sc->model);
	if (sc->modbus != 0 && !strcmp(sc->model, "mk_lcd"))
		circutor_reg(sc, &sc->req0, &mklcd_dev);
	else if (sc->modbus != 0 && !strcmp(sc->model, "cvmk_b"))
		circutor_reg(sc, &sc->req0, &cvmkb_dev);
	else
		circutor_reg(sc, &sc->req0, NULL);
	return (0);
}

static void
circutor_init(CK_ARG, struct circutor *sc __unused)
{

}

static void
circutor_deinit(CK_ARG, struct circutor *sc __unused)
{

}

UintVar(circutor, CIRCUTOR_MAGIC, debug, debug, 0, );
MODULE_GROUP(circutor, CIRCUTOR_MAGIC);
