/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: fernau_rciu.c,v 1.11 2008/02/17 09:55:54 phk Exp $
 *
 */

#include <stdio.h>
#include <pthread.h>
#include <configkit.h>
#include <termios.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "measured.h"
#include "configkit_conf.h"

/*--------------------------------------------------------------------*/

/*
 * We react to changed values immediately and log values identical to
 * the last only ever DEF_LOGRATE seconds.
 */
#define DEF_LOGRATE		5

/*
 * How many seconds we "hold the button down" when we send commands.
 */
#define DEF_HOLDTIME	3

/*
 * The idle state of the bits we transmit.
 * 0x40 is the "Key Sync" and we must send it high in order to not
 * disturb the IDENT generation from other sources.
 */
#define DEF_SENDBITS	0x40

/*--------------------------------------------------------------------*/

#define I(byte, bit)	(byte * 5 + bit)

static struct ptdesc fernau_rciu_pt[] = {
    { .idx = I(0, 0), .type = DIN, .label = "Beacon Shutdown", },
    { .idx = I(0, 1), .type = DIN, .label = "Beacon Transfer", },
    { .idx = I(0, 2), .type = DIN, .label = "Beacon Normal", },

    { .idx = I(0, 3), .type = DIN, .label = "Power 1 Fault", },
    { .idx = I(1, 0), .type = DIN, .label = "Power 1 Mains", },
    { .idx = I(3, 3), .type = DIN, .label = "Power 1 Battery", },
    { .idx = I(3, 4), .type = DIN, .label = "Power 1 Low Battery", },

    { .idx = I(0, 4), .type = DIN, .label = "Power 2 Fault", },
    { .idx = I(4, 1), .type = DIN, .label = "Power 2 Mains", },
    { .idx = I(5, 2), .type = DIN, .label = "Power 2 Battery", },
    { .idx = I(5, 3), .type = DIN, .label = "Power 2 Low Battery", },


    { .idx = I(1, 1), .type = DIN, .label = "Mon 1 Fault", },
    { .idx = I(1, 2), .type = DIN, .label = "Mon 1 Standby", },
    { .idx = I(1, 3), .type = DIN, .label = "Mon 1 Live", },

    { .idx = I(1, 4), .type = DIN, .label = "Mon 2 Fault", },
    { .idx = I(2, 0), .type = DIN, .label = "Mon 2 Standby", },
    { .idx = I(2, 1), .type = DIN, .label = "Mon 2 Live", },

    { .idx = I(2, 2), .type = DIN, .label = "Tx 1 Fault", },
    { .idx = I(2, 3), .type = DIN, .label = "Tx 1 Standby", },
    { .idx = I(2, 4), .type = DIN, .label = "Tx 1 Live", },

    { .idx = I(3, 0), .type = DIN, .label = "Tx 2 Fault", },
    { .idx = I(3, 1), .type = DIN, .label = "Tx 2 Standby", },
    { .idx = I(3, 2), .type = DIN, .label = "Tx 2 Live", },

    { .idx = I(4, 0), .type = DIN, .label = "Local/Remote", },
    { .idx = I(4, 2), .type = DIN, .label = "Comms Fault", },
    { .idx = I(4, 3), .type = DIN, .label = "Audio Min Alarm", },
    { .idx = I(4, 4), .type = DIN, .label = "Audio Maj Alarm", },

    { .idx = I(5, 0), .type = DIN, .label = "Key", },
    { .idx = I(5, 1), .type = DIN, .label = "CPU Fault", },
    { .idx = I(5, 4), .type = DIN, .label = "Status", },

    { .idx = I(6, 1), .type = DOUT, .label = "Transfer Transponder", },
    { .idx = I(6, 2), .type = DOUT, .label = "DME Off", },
    { .idx = I(6, 3), .type = DOUT, .label = "DME On (or reset)", },
    { .idx = I(6, 4), .type = DOUT, .label = "Change Main transponder", },
    { .idx = I(6, 0), .type = DOUT, .label = "Reset Audio Alarm", },
#if 0
    { .idx = I(6, 5), .type = DOUT, .label = "Runway End", },
    { .idx = I(6, 6), .type = DOUT, .label = "Key Sync", },
#endif
    { .label = NULL, }
};

/*--------------------------------------------------------------------*/

struct fernau_rciu {
	unsigned		magic;
#define FERNAU_RCIU_MAGIC 		0xe9c53ff0
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	double			lograte;
	double			holdtime;

	unsigned char		send;
	unsigned char		lsend;
	struct timespec		esend;

	unsigned char		rcv[6];
	struct timespec		last[6];

	unsigned		debug;
	struct serial		*serial;
	struct point		**p;
};


static void *
fernau_rciu_thread(void *priv)
{
	struct termios t;
	struct fernau_rciu *sc;
	unsigned char buf;
	unsigned u, b;
	int i;
	struct timespec now;

	CAST_OBJ_NOTNULL(sc, priv, FERNAU_RCIU_MAGIC);

	AZ(tcgetattr(serial_fd(sc->serial), &t));
	AZ(cfsetspeed(&t, B1200));
	cfmakeraw(&t);
	t.c_cflag |= CLOCAL;
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 10;
	AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));
	
	/* Ignore the first bunch of characters */
	for (i = 0; i < 10; i++)
		read(serial_fd(sc->serial), &buf, 1);
	while (1) {
		i = read(serial_fd(sc->serial), &buf, 1);
		if (sc->debug & 2)
			SCDebug(sc, "read %d %02x", i, buf);
		if (i == 0) {
			GLog(sc->gp, "read timeout");
			continue;
		}
		assert(i == 1);

		now = Now();

		u = buf >> 5;
		if (u > 5) {
			/* This is an invalid tag value */
			Record(GI(sc->gp) + 0, 1, &buf);
			continue;
		}
		if (sc->rcv[u] != buf ||
		    sc->lograte < delta_t(&now, &sc->last[u])) {
			Record(GI(sc->gp) + 1, 1, &buf);
			for (b = 0; b < 5; b++)
				SetValBool(sc->p[I(u, b)], buf & (1 << b));
			sc->last[u] = now;
			sc->rcv[u] = buf;
		}
		if (u != 0)
			continue;
		if (sc->holdtime > delta_t(&now, &sc->esend)) {
			if (sc->send != sc->lsend)
				Record(GI(sc->gp) + 2, 1, &sc->send);
			if (sc->debug & 1)
				SCDebug(sc, "write 0x%02x", sc->send);
			i = write(serial_fd(sc->serial), &sc->send, 1);
			assert(i == 1);
			sc->lsend = sc->send;
		} else if (sc->send != DEF_SENDBITS) {
			sc->send = DEF_SENDBITS;
			sc->esend = now;
			for (u = 0; u < 5; u++)
				SetValBool(sc->p[I(6,u)], 0);
		}
	}
}

static void
fernau_rciu_control(struct point *p, void *priv1, void *priv2, double val)
{
	struct fernau_rciu *sc = priv1;
	unsigned char *d = priv2;

	if (val > .5) {
		sc->send |= *d;
		sc->esend = Now();
		SetValBool(p, 1);
	}
}

static unsigned ctrl_bits[8] = {
	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
};

static void
fernau_rciu_init(CK_ARG, struct fernau_rciu *sc)
{
	unsigned u;

	sc->lograte = DEF_LOGRATE;
	sc->holdtime = DEF_HOLDTIME;
	sc->send = DEF_SENDBITS;
	sc->esend = Now();
	u = CreatePoints(sc->gp, &sc->p, fernau_rciu_pt);
	for (u = 0; u < 5; u++) {
		SetPointControl(sc->p[I(6,u)], fernau_rciu_control,
		    sc, &ctrl_bits[u]);
		SetValBool(sc->p[I(6,u)], 0);
	}
	serial_init(CK_PASS, &sc->serial, fernau_rciu_thread, sc);
}

static void
fernau_rciu_deinit(CK_ARG, struct fernau_rciu *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	free(sc->p);
}

SerialFunc(fernau_rciu, FERNAU_RCIU_MAGIC, serial);
UintVar(fernau_rciu, FERNAU_RCIU_MAGIC, debug, debug, 0, );
RealVar(fernau_rciu, FERNAU_RCIU_MAGIC, lograte, lograte, DEF_LOGRATE, "%.1f", );
RealVar(fernau_rciu, FERNAU_RCIU_MAGIC, holdtime, holdtime, DEF_HOLDTIME, "%.1f", );
MODULE_GROUP(fernau_rciu, FERNAU_RCIU_MAGIC);
