/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: fernau_rmm.c,v 1.24 2008/02/17 09:55:54 phk Exp $
 *
 */

#include <stdio.h>
#include <pthread.h>
#include <configkit.h>
#include <termios.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "measured.h"
#include "configkit_conf.h"

/*--------------------------------------------------------------------*/

#define PIDX(mon, nbr)	.idx = mon * 100 + nbr, .point = mon * 100 + nbr
#define PERMONPT(mn) \
    { PIDX(mn, 0), .type = AIN, .label = "Pulse Delay mon" #mn, },	\
    { PIDX(mn, 1), .type = AIN, .label = "Reply Efficiency mon" #mn, }, \
    { PIDX(mn, 2), .type = AIN, .label = "TX Power mon" #mn, },	\
    { PIDX(mn, 3), .type = AIN, .label = "VSWR mon" #mn, },		\
    { PIDX(mn, 4), .type = AIN, .label = "History msgs" #mn, },	\
    { PIDX(mn, 5), .type = AIN, .label = "Clock offset" #mn, },	\
    { PIDX(mn, 6), .type = AIN, .label = "Clock reading" #mn, },	\

static struct ptdesc fernau_rmm_pt[] = {
    PERMONPT(1)
    PERMONPT(2)
    { .label = NULL, }
};

/*--------------------------------------------------------------------*/

struct fernau_rmm {
	unsigned		magic;
#define FERNAU_RMM_MAGIC 		0xe43fc21c
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	unsigned		debug;
	struct serial		*serial;
	struct point		**p;

	unsigned char		rxbuf[255 + 5 + 1];
	unsigned		rxp;

	char			*version;
};

/*====================================================================*/

static unsigned
rmmsum(const unsigned char *p, unsigned len)
{
	unsigned h, l, m, n, i;

	h = l = 0xff;

	m = p[2] - 1;
	for (i = 2; i < len - 4; i++) {
		for (n = 0; n <= m; n++) {
			h -= p[i];
			if (h & 0x100)
				h += 0xff;
		}
		for (n = 0; n < m; n++) {
			l += p[i];
			if (l & 0x100)
				l -= 0xff;
		}
		m--;
	}
	return ((h << 8) | l);
}

static int
rmmchksum(const unsigned char *p, unsigned len)
{
	unsigned sum;

	sum = rmmsum(p, len);
	return (p[len - 4] == (sum >> 8) && p[len - 3] == (sum & 0xff));
}

static void
rmmsend(const struct fernau_rmm *sc, const char *msg)
{
	unsigned char buf[255 + 5 + 1];
	unsigned l, s;
	int i, fd;

	fd = serial_fd(sc->serial);
	l = strlen(msg);
	assert(l < 254);
	if (l < 3)
		return;
	buf[0] = 0x10;
	buf[1] = 0x82;
	buf[2] = l + 2;
	memcpy(buf + 3, msg, l);
	s = rmmsum(buf, l + 7);
	buf[3 + l] = s >> 8;
	buf[4 + l] = s & 0xff;
	buf[5 + l] = '\r';
	buf[6 + l] = '\n';
	buf[7 + l] = '\0';
	i = write(fd, buf, 7 + l);
	assert(i > 0);
	assert(i == (int)l + 7);
	Record(GI(sc->gp) + 2, 7 + l, buf);
	if (sc->debug & 4)
		SCDebug(sc, "send %V", buf);
}

static int
tst_packet(const unsigned char *p, unsigned l)
{

	if ((l > 0 && p[0] != 0x10) || (l > 1 && p[1] != 0x82))
		return (-1);
	if (l < p[2] + 5U)
		return (0);
	l = p[2] + 5;
	if (p[l - 2] != '\r')
		return (-1);
	if (p[l - 1] != '\n')
		return (-1);
	if (!rmmchksum(p, l))
		return (-1);
	return (l);
}

static int
recv_packet(struct fernau_rmm *sc)
{
	int i;
	unsigned l;

	while (1) {
		i = read(serial_fd(sc->serial),
		    &sc->rxbuf[sc->rxp], sizeof sc->rxbuf - (1 + sc->rxp));
		NSCDebug(sc, 2, "read %d", i);
		if (sc->debug & 4 && i > 0) {
			sc->rxbuf[sc->rxp + i] = '\0';
			SCDebug(sc, "read %V", &sc->rxbuf[sc->rxp], i);
		}
		if (i == 0) {
			sc->rxp = 0;
			GLog(sc->gp, "read timeout");
			return (0);
		}
		assert(i >= 1);
		sc->rxp += i;
		while (sc->rxp > 3) {
			i = tst_packet(sc->rxbuf, sc->rxp);
			if (i == 0)
				break;
			if (i > 0) {
				Record(GI(sc->gp), i, sc->rxbuf);
				return (i);
			}
			for (l = 1; l < sc->rxp; l++)
				if (sc->rxbuf[l] == 0x10)
					break;
			if (l < sc->rxp)  {
				Record(GI(sc->gp) + 1, l, sc->rxbuf);
				memmove(sc->rxbuf,
				    sc->rxbuf + l, sc->rxp - l);
			}
			sc->rxp -= l;
			continue;
		}
	}
}

/*====================================================================*/

static int
GetInt(char **p, int *r)
{
	char *q;
	int i;

	i = strtol(*p, &q, 0);
	if (q == *p)
		return (0);
	if (*q == ',')
		q++;
	*p = q;
	*r = i;
	return (1);
}

static int
GetDouble(char **p, double *r)
{
	char *q;
	double a;

	a = strtod(*p, &q);
	if (q == *p)
		return (0);
	if (*q == ',')
		q++;
	*p = q;
	*r = a;
	return (1);
}

static int
GetChar(char **p, int *r, const char *match)
{
	const char *q;

	q = strchr(match, **p);
	if (q == NULL)
		return (0);
	(*p)++;
	if (**p == '\0')
		;
	else if (**p == ',')
		(*p)++;
	*r = q - match;
	return (1);
}

/*====================================================================*/

/*
 *
 * Documentation from Sheet 9, Issue 2, Document FAS168B:
 * Corresponding to software version: "B06,C01,B02,A01,A01"
 *
 *	 4.2.1 PARAMETERS
 *	
 *	 command:
 *		request parameters, monitor identity
 *	 code:
 *		20, 1/2
 *	 response:
 *		20 (command id),
 *		1/2 (monitor id),
 *		1/2 (txp cnx),
 *		L/S (txp cnx),
 *		00..10235 (reply delay),
 *		0.0..5115 (reply spacing),
 *		00..635 (pulse width),
 *		00..315 (pulse rise time),
 *		0.0..3.15 (pulse fall time),
 *		0..8191 (pulse rate),
 *		0..8191 (decode rate),
 *		0..100 (efficiency),
 *		00..3276.800 (tx frequency),
 *		0..2000 (output power),
 *		P/A/S (ident presence)
 *	
 *		Numerical values will be negated for 'old data'.
 *
 * Example:
 * 	20,2,1,L,0.000,0.000,0.00,0.00,0.00,0,120,0,1159.003,29,A
 *
 * Documentation from Page 11, Issue 3G, Document FAS168B:
 * Corresponding to software version: "B08,C01,A01,B02"
 *
 *	 4.2.1 PARAMETERS
 *	
 *	 command:
 *		request parameters
 *	 code:
 *		20
 *	 response:
 *		20 (command id),
 *		1/2 (txp cnx)
 *		L/S (txp cnx),
 *		00..10235 (reply delay),
 *		0.0..5115 (reply spacing),
 *		00..635 (pulse width),
 *		00..315 (pulse rise time),
 *		0.0..3.15 (pulse fall time),
 *		0..8191 (pulse rate),
 *		0..8191 (decode rate),
 *		0..100 (efficiency),
 *		00..3276.800 (tx frequency),
 *		0..2000 (output power),
 *		P/A/S/O (ident presence)
 *	
 *		Numerical values will be negated for 'old data'.
 *
 * Example:
 * 	20,1,L,50.01,11.96,3.30,2.20,2.35,864,127,96,1162.006,1000,P
 * 
 * 
 * Corresponding to software version: "B09,A12,B03,B02"
 * 
 * Example:
 *      20,1,L,50.14,12.02,3.50,2.20,2.25,774,233,93,1163.991,636,P>
 *
 */

static int
msg_DA20(struct fernau_rmm *sc)
{
	char *p;
	int i, mon;
	double a;

	mon = 100 * (sc->rxbuf[4] - '0');
	p = sc->rxbuf + 8;	/* 0x10,0x82,len,dst,src,'D','A',',' */
	if (!GetInt(&p, &i))		return(__LINE__); /* command id */
	if (!GetInt(&p, &i))		return(__LINE__); /* monitor id */
	if (!strcmp(sc->version, "B06,C01,B02,A01,A01"))
		if (!GetInt(&p, &i))	return(__LINE__); /* txp cnx */
	if (!GetChar(&p, &i, "LS"))	return(__LINE__); /* txp cnx */
	if (!GetDouble(&p, &a))		return(__LINE__); /* reply delay */
	if (a >= 0)
		SetVal(sc->p[mon + 0], a);
	if (!GetDouble(&p, &a))		return(__LINE__); /* reply spacing */
	if (!GetDouble(&p, &a))		return(__LINE__); /* pulse width */
	if (!GetDouble(&p, &a))		return(__LINE__); /* pulse rise time */
	if (!GetDouble(&p, &a))		return(__LINE__); /* pulse fall time */
	if (!GetDouble(&p, &a))		return(__LINE__); /* pulse rate */
	if (!GetDouble(&p, &a))		return(__LINE__); /* decode rate */
	if (!GetDouble(&p, &a))		return(__LINE__); /* efficiency */
	if (a >= 0)
		SetVal(sc->p[mon + 1], a);
	if (!GetDouble(&p, &a))		return(__LINE__); /* tx frequency */
	if (!GetDouble(&p, &a))		return(__LINE__); /* output power */
	if (a >= 0)
		SetVal(sc->p[mon + 2], a);
	if (!GetChar(&p, &i, "OPAS"))	return(__LINE__); /* ident presence */
	if (*p)				return(__LINE__);
	return (0);
}

/*====================================================================*/

/*
 *
 * Documentation from Sheet 10, Issue 2, Document FAS168B:
 *
 *	 4.3.3 AMPLIFIER STATUS
 *	code:
 *		32
 *	response:
 *		32 (command id ),
 *		0..100 (fwd pwr),
 *		0.0..5.5 (VSWR) 0..15 (tx control profile),
 *		0..255 (pulse shape measurement),
 *		-273..224 (temperature1 C),
 *		- 273..224 (temperature2 C),
 *		- 273..224 (temperature3 C),
 *		00..4346 (+28 vdc reading),
 *		0..100 (lpa input pwr),
 *		100 (lpa output pwr),
 *		0..100 (hpa input pwr),
 *		0..100 (hpa output pwr),
 *		E/D (hvps en/dis),
 *		N/F (hvps),
 *		P/F (lpa input pwr),
 *		P/F/O (lpa output pwr),
 *		P/F (lpa temp.),
 *		P/F/O (hpa input pwr),
 *		P/F/O (hpa output pwr),
 *		P/F (hpa temp.),
 *		P/F (clock),
 *		P/F (pulse2),
 *		P/F (txc psu),
 *		P/F/O (o/p pulses),
 *		P/F (rev. pwr),
 *		P/F (txc temp),
 *		P/F (rem1 temp),
 *		P/F (rem2 temp),
 *		F/L/H (FAN STATUS Off,Low, High),
 *	
 * Example:
 * 	32,0,1.00,15,0,33,35,250,27.2,44,0,3,2,E,N,P,O,P,O,O,P,P,O,P,O,P,P,P,P,F
 */

static int
msg_DA32(struct fernau_rmm *sc)
{
	char *p;
	int i, mon;
	double a;

	switch(sc->rxbuf[4]) {
	case '1':	mon = 100 ; break;
	case '2':	mon = 200 ; break;
	default:	return (__LINE__);
	}

	p = sc->rxbuf + 8;	/* 0x10,0x82,len,dst,src,'D','A',',' */

	if (!GetInt(&p, &i))		return(__LINE__); /* command id */
	if (!GetInt(&p, &i))		return(__LINE__); /* forward power */
	if (!GetDouble(&p, &a))		return(__LINE__); /* VSWR */
	SetVal(sc->p[mon + 3], a);
	/* ... */
	return (0);
}

/*====================================================================*/

static int
msg_DA63(struct fernau_rmm *sc)
{
	char *p;
	int i, mon;

	switch(sc->rxbuf[4]) {
	case '1':	mon = 100 ; break;
	case '2':	mon = 200 ; break;
	default:	return (__LINE__);
	}
	p = sc->rxbuf + 8;	/* 0x10,0x82,len,dst,src,'D','A',',' */

	if (!GetInt(&p, &i))		return(__LINE__); /* command id */
	if (!GetInt(&p, &i))		return(__LINE__); /* command id */
	SetVal(sc->p[mon + 4], i);
	return (0);
}

/*====================================================================*/

/*
 *
 * Documentation from Sheet 19, Issue 2, Document FAS168B:
 *
 *	4.5.5 CHECK DATETIME
 *	code:
 *		64
 *	response:
 *		64,
 *		00..23 (hours)
 *		':'
 *		00..59 (minutes)
 *		':'
 *		00..59(secs),
 *		1..31 (day)
 *		'/'
 *		01..12'(month)
 *		/'
 *		94..34 (year)
 *
 *	Actual (SW version B08,C02,A01,B02)
 *		In practice there are at least two bugs in the
 *		transmitted data: the tens digit of the hours is
 *		transmitted in the ack-position (after the 'D')
 *		and the comma is missing after the 64:
 *		  \020\202\03101D0,649:45:08,22/11/06k\327\r\n
 *		is really 09:35:08, 22/11/06 and
 * 		  \020\202\03102D1,649:50:35,25/05/07K\362\r\n
 *		is 19:50:35, 25/05/07.
 *
 *		The code allows both as spec'ed and the buggy version.
 *
 */

static int
msg_DA64(struct fernau_rmm *sc)
{
	char *p, *q;
	int mon;
	struct tm tm;
	time_t tl, tr;

	/* We need to check here, since the distributor code doesn't */
	if (sc->rxbuf[7] != ',') return (__LINE__);
	if (sc->rxbuf[8] != '6') return (__LINE__);
	if (sc->rxbuf[9] != '4') return (__LINE__);

	time(&tl);
	switch(sc->rxbuf[4]) {
	case '1':	mon = 100 ; break;
	case '2':	mon = 200 ; break;
	default:	return (__LINE__);
	}
	if (sc->rxbuf[6] == 'A' && sc->rxbuf[10] == ',') {
		/* As documented */
		p = sc->rxbuf + 11;
	} else if (sc->rxbuf[6] >= '0' && sc->rxbuf[6] <= '2' &&
	    sc->rxbuf[10] != ',') {
		/* As seen */
		sc->rxbuf[9] = sc->rxbuf[6];
		p = sc->rxbuf + 9;
	} else
		return (__LINE__);
	memset(&tm, 0, sizeof tm);
	q = strptime(p, "%H:%M:%S,%d/%m/%y", &tm);
	if (q != (char*)sc->rxbuf + sc->rxbuf[2] + 1)
		return (__LINE__);
	tr = timegm(&tm);
	SetVal(sc->p[mon + 5], tr - tl);
	SetVal(sc->p[mon + 6], tr);
	return (0);
}

/*====================================================================*/

static int
msg_DA84(struct fernau_rmm *sc)
{

	if (sc->debug)
		printf("Version: %s\n", sc->rxbuf + 11);
	if (sc->version != NULL)
		free(sc->version);
	sc->version = strdup(sc->rxbuf + 11);
	return (0);
}

/*====================================================================*/

typedef int msg_func(struct fernau_rmm *sc);

static struct seq {
	const char	*send;
	const char	*expect;
	unsigned	elen;
	msg_func	*func;
} seqtable[] = {
	{ "10D80",	"01DA,80,", 0,	NULL },		/* logon */
	{ "20D80",	"02DA,80,", 0,	NULL },
	{ "10D84",	"01DA,84,", 0,	msg_DA84 },	/* sw version */
	{ "20D84",	"02DA,84,", 0,	msg_DA84 },
	{ "10D20,1",	"01DA,20,", 0,	msg_DA20 },	/* params */
	{ "20D20,2",	"02DA,20,", 0,	msg_DA20 },
	{ "10D32",	"01DA,32,", 0,	msg_DA32 },	/* amp status */
	{ "20D32",	"02DA,32,", 0,	msg_DA32 },
	{ "10D63",	"01DA,63,", 0,	msg_DA63 },	/* history status */
	{ "20D63",	"02DA,63,", 0,	msg_DA63 },
	{ "10D64",	"01D", 0,	msg_DA64 },	/* clock */
	{ "20D64",	"02D", 0,	msg_DA64 },
	{ NULL,		NULL, 0,	NULL }
};

/*====================================================================*/

static void *
fernau_rmm_thread(void *priv)
{
	struct termios t;
	struct fernau_rmm *sc;
	int i, j;
	struct seq *seq;

	CAST_OBJ_NOTNULL(sc, priv, FERNAU_RMM_MAGIC);

	AZ(tcgetattr(serial_fd(sc->serial), &t));
	AZ(cfsetspeed(&t, B9600));
	cfmakeraw(&t);
	t.c_cflag |= CLOCAL;
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 100;
	AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));

	for (seq = seqtable; seq->send != NULL; seq++)
		seq->elen = strlen(seq->expect);
	
	sc->rxp = 0;
	seq = seqtable;
	rmmsend(sc, seq->send);
	while (1) {
		i = recv_packet(sc);
		if (i == 0) {
			/* Receive problem, start over with login */
			seq = seqtable;
			rmmsend(sc, seq->send);
			continue;
		}
		sc->rxbuf[i - 4] = '\0';
		if (memcmp(sc->rxbuf + 3, seq->expect, seq->elen)) {
			/* Not what we expected, start over with login */
			seq = seqtable;
			rmmsend(sc, seq->send);
			GLog(sc->gp, "wrong answer %V", sc->rxbuf + 3);
			continue;
		}
		if (seq->func) {
			j = seq->func(sc);
			if (j) {
				GLog(sc->gp, "Parse failed in line %d <%V>",
				    j, sc->rxbuf + 3);
				SCDebug(sc, "Parse failed in line %d <%V>",
				    j, sc->rxbuf + 3);
			}
		}
		seq++;
		if (seq->send == NULL)
			seq = seqtable + 4;
		rmmsend(sc, seq->send);

		if (i != (int)sc->rxp)
			memmove(sc->rxbuf, sc->rxbuf + i, sc->rxp - i);
		sc->rxp -= i;
		sleep (1);
	}
}

static void
fernau_rmm_init(CK_ARG, struct fernau_rmm *sc)
{

	sc->debug = 0;
	CreatePoints(sc->gp, &sc->p, fernau_rmm_pt);
	serial_init(CK_PASS, &sc->serial, fernau_rmm_thread, sc);
}

static void
fernau_rmm_deinit(CK_ARG, struct fernau_rmm *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	free(sc->p);
}

WordVar(fernau_rmm, FERNAU_RMM_MAGIC, version, version, );
SerialFunc(fernau_rmm, FERNAU_RMM_MAGIC, serial);
UintVar(fernau_rmm, FERNAU_RMM_MAGIC, debug, debug, 0, );
MODULE_GROUP(fernau_rmm, FERNAU_RMM_MAGIC);
