/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: fifolog.c,v 1.56 2008/03/22 17:01:38 phk Exp $
 *
 */

#include <assert.h>
#include <stdio.h>
#include <configkit.h>
#include <time.h>
#include <pthread.h>
#include <zlib.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/sbuf.h>

#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"
#include "libfifolog.h"
#include "fifolog_write.h"

#define FIFO_DEFAULT_RATE		10
#define FIFO_DEFAULT_SYNC		60
#define FIFO_DEFAULT_COMPRESSION	Z_BEST_COMPRESSION

struct fifolog {
	unsigned			magic;
#define FIFOLOG_MAGIC			0xf1f0706
	struct group			*gp;
	struct ck_instance		cki;
	pthread_mutex_t			lock;

	LIST_ENTRY(fifolog)		list;

	struct fifolog_writer		*writer;

	struct point			*pt[FIFOLOG_NPOINT];

	u_int				writerate;
	u_int				syncrate;
	u_int				compression;
	char				*devicename;
	char				*match;
	pthread_t			tp;

	struct pattern			*pattern;
	struct event_subscription	*sub;
	u_char				*ibuf;

	int				cleanup;
};

/*
 * We need to be able to traverse the list of fifologs with a lock held.
 * Rather than teach configkit about this, we use our own locked list.
 */
static LIST_HEAD(, fifolog)	fifolist = LIST_HEAD_INITIALIZER(fifolist);
static pthread_mutex_t		fifolock;
static void fifo_record(unsigned point, unsigned len, const void *ptr);

static void *
fifolog_thread(void *arg)
{
	struct fifolog *sc;
	time_t now, last;
	int i;
	
	CAST_OBJ_NOTNULL(sc, arg, FIFOLOG_MAGIC);

	record_function = (void*)fifo_record;
	last = 0;
	while (1) {
		if (sc->cleanup)
			return(NULL);
		(void)time(&now);
		AZ(pthread_mutex_lock(&sc->lock));
		while (fifolog_write_poll(sc->writer, now))
			continue;
		AZ(pthread_mutex_unlock(&sc->lock));
		now /= 10;		/* XXX: param */
		if (now > last) {
			last = now;
			for (i = 0; i < FIFOLOG_NPOINT; i++)
				SetVal(sc->pt[i], sc->writer->cnt[i]);
		}
		usleep(100000);
	}
}

static void
fifo_bytes(struct fifolog *sc, uint32_t w, const void *p, u_int len)
{
	int i, j;

	for (i = 0; i < 20; i++) {
		AZ(pthread_mutex_lock(&sc->lock));
		j = fifolog_write_bytes(sc->writer, w, 0, p, len);
		AZ(pthread_mutex_unlock(&sc->lock));
		if (j)
			return;
		usleep(250000);
	}
}

/* Message logging ---------------------------------------------------*/

static void
fifo_record(unsigned point, unsigned len, const void *ptr)
{
	struct fifolog *sc;

	if (len == 0)
		return;
	assert(len < 256);
	pthread_cancelmutex_lock(&fifolock);
	LIST_FOREACH(sc, &fifolist, list)
		fifo_bytes(sc, point, ptr, len);
	pthread_cancelmutex_unlock(&fifolock);
}

/* Event logging -----------------------------------------------------*/

static void
fifo_event(void *priv, struct event_arg *e)
{
	struct sbuf *sb;
	struct point *pp;
	struct fifolog *sc;

	CAST_OBJ_NOTNULL(sc, priv, FIFOLOG_MAGIC);
	sb = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
	if (e->e == E_POINT_VALUE) {
		CAST_OBJ(pp, e->obj, POINT_MAGIC);
		sbuf_printf(sb, "%s %g", e->name, p_value(pp));
	} else 
		sbuf_printf(sb, "%s %s", e->name, Event_String(e));
	sbuf_finish(sb);
	fifo_bytes(sc, e->instance, sbuf_data(sb), 0);
	sbuf_delete(sb);
}

/* (Re)configuration -------------------------------------------------*/

static void
fifo_reconfig(CK_ARG)
{
	struct fifolog *sc;
	const char *s;
	void *vp;
	int i;

	CAST_OBJ_NOTNULL(sc, cfa->priv, FIFOLOG_MAGIC);
	if (sc->tp != NULL) {
		sc->cleanup = 1;
		AZ(pthread_join(sc->tp, &vp));
		sc->tp = NULL;
		sc->cleanup = 0;
	}
	if (sc->writer != NULL) {
		fifolog_write_flush(sc->writer);
		fifolog_write_close(sc->writer);
		fifolog_write_destroy(sc->writer);
		sc->writer = NULL;
	}
	if (sc->sub != NULL) {
		UnsubscribeEvents(sc->sub);
		sc->sub = NULL;
	}
	if (sc->pattern) {
		FreePattern(sc->pattern);
		sc->pattern = NULL;
	}

	for (i = 0; i < FIFOLOG_NPOINT; i++)
		SetVal(sc->pt[i], 0);

	if (sc->devicename == NULL)
		return;

	sc->writer = fifolog_write_new();
	s = fifolog_write_open(sc->writer, sc->devicename,
	    sc->writerate, sc->syncrate, sc->compression);
	if (s != NULL) {
		ck_printf(CK_PASS, "Cannot open fifolog: %s\n",
		    s);
		fifolog_write_destroy(sc->writer);
		sc->writer = NULL;
		return;
	}

	if (sc->match != NULL)
		sc->pattern = CompilePattern(sc->match);

	AZ(pthread_create(&sc->tp, NULL, fifolog_thread, sc));

	sc->sub = SubscribeEvents(fifo_event, sc);
	StartEvents(sc->sub, "all,msg");
}

/* ConfigKit functions -----------------------------------------------*/

static void
InitFifoLog(void)
{

	AZ(pthread_mutex_init(&fifolock, NULL));
}

static pthread_once_t	init_once;

static void
fifolog_init(CK_ARG, struct fifolog *sc)
{
	unsigned u;

	AZ(pthread_once(&init_once, InitFifoLog));
	sc->writerate = FIFO_DEFAULT_RATE;
	sc->syncrate = FIFO_DEFAULT_SYNC;
	sc->compression = FIFO_DEFAULT_COMPRESSION;
	pthread_cancelmutex_lock(&fifolock);
	LIST_INSERT_HEAD(&fifolist, sc, list);
	pthread_cancelmutex_unlock(&fifolock);

	for (u = 0; u < FIFOLOG_NPOINT; u++) {
		sc->pt[u] = CreatePoint(sc->gp, 1 + u, COUNT);
		assert(fifolog_write_statnames[u] != NULL);
		SetPointLabel(sc->pt[u], fifolog_write_statnames[u]);
	}
}

static void
fifolog_deinit(CK_ARG, struct fifolog *sc __unused)
{

	pthread_cancelmutex_lock(&fifolock);
	LIST_REMOVE(sc, list);
	pthread_cancelmutex_unlock(&fifolock);
	FREE(&sc->devicename);
	fifo_reconfig(CK_PASS);
}

WordVar(fifolog, FIFOLOG_MAGIC,
    match, match, fifo_reconfig(CK_PASS));
WordVar(fifolog, FIFOLOG_MAGIC,
    devicename, devicename, fifo_reconfig(CK_PASS));
UintVar(fifolog, FIFOLOG_MAGIC,
    writerate, writerate, FIFO_DEFAULT_RATE, fifo_reconfig(CK_PASS));
UintVar(fifolog, FIFOLOG_MAGIC,
    syncrate, syncrate, FIFO_DEFAULT_SYNC, fifo_reconfig(CK_PASS));
UintVar(fifolog, FIFOLOG_MAGIC,
    compression, compression, FIFO_DEFAULT_COMPRESSION, fifo_reconfig(CK_PASS));
MODULE_GROUP(fifolog, FIFOLOG_MAGIC);
