/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: mcast.c,v 1.15 2008/02/16 15:49:41 phk Exp $
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <configkit.h>
#include <assert.h>
#include <poll.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <md5.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <paths.h>
#include <sys/watchdog.h>

#include "measured.h"
#include "configkit_conf.h"

#define MAX_TX			20	/* <= floor((255-10) / 10) */
#define NPT			8

struct mcast_stat {
	uintmax_t		packets;
	uintmax_t		badpackets;
	uintmax_t		bytes;
	int			last_status;
	int			last_error;
	double 			last_good;
	double 			last_bad;
	uintmax_t		loops;
	struct point		*pt[NPT];
};

struct mcast {
	unsigned		magic;
#define MCAST_MAGIC 		0x83742834
	struct ck_instance	cki;
	struct group		*gp;
	pthread_t		tp;
	pthread_mutex_t		lock;
	struct mcast_tx		*tx[MAX_TX];
	char			*addr;
	char			*port;
	char			*txport;
	struct addrinfo		*ai;
	char			running;
	int			rxsock;
	int			txsock;
	double			statrate;
	double			laststat;

	unsigned		antiloop;
	unsigned char		*looplist;
	unsigned		loopidx;
	unsigned		looplen;

	unsigned		watchdog;
	unsigned		patpat;
	int			wfd;
	
	struct mcast_stat	stat;
};

struct mcast_tx {
	unsigned		magic;
#define MCAST_TX_MAGIC		0x1e062831
	struct ck_instance	cki;
	unsigned		tx;
	struct mcast		*sc;
	char			*addr;
	char			*port;
	struct addrinfo		*ai;

	struct mcast_stat	stat;
};

/*--------------------------------------------------------------------*/

static void
mcast_mkpoints(const struct mcast *sc, struct mcast_stat *st, unsigned base)
{
	unsigned u;

	st->pt[0] = CreatePoint(sc->gp, base + 0, COUNT);
	SetPointLabel(st->pt[0], "Good packets");

	st->pt[1] = CreatePoint(sc->gp, base + 1, COUNT);
	SetPointLabel(st->pt[1], "Good bytes");

	st->pt[2] = CreatePoint(sc->gp, base + 2, AIN);
	SetPointLabel(st->pt[2], "Last good packet ");
	SetPointFormat(st->pt[2], "%.1f");

	st->pt[3] = CreatePoint(sc->gp, base + 3, AIN);
	SetPointLabel(st->pt[3], "Last status");

	st->pt[4] = CreatePoint(sc->gp, base + 4, COUNT);
	SetPointLabel(st->pt[4], "Bad packets");

	st->pt[5] = CreatePoint(sc->gp, base + 5, AIN);
	SetPointLabel(st->pt[5], "Last bad packet ");
	SetPointFormat(st->pt[5], "%.1f");

	st->pt[6] = CreatePoint(sc->gp, base + 6, AIN);
	SetPointLabel(st->pt[6], "Last error");

	if (base == 0) {
		st->pt[7] = CreatePoint(sc->gp, 7, COUNT);
		SetPointLabel(st->pt[7], "Looped packets");
	}

	for (u = 0; u < NPT; u++) 
		if (st->pt[u] != NULL)
			SetPointTimeout(st->pt[u], sc->statrate * 2.0);
}

static void
mcast_updtimeout(struct mcast *sc)
{
	unsigned u, v;

	for (u = 0; u < NPT; u++) 
		if (sc->stat.pt[u] != NULL)
			SetPointTimeout(sc->stat.pt[u], sc->statrate * 2.0);
	AZ(pthread_mutex_lock(&sc->lock));
	for (v = 1; v < MAX_TX; v++) {
		if (sc->tx[v] == NULL)
			continue;
		for (u = 0; u < 7; u++) 
			SetPointTimeout(sc->tx[v]->stat.pt[u],
			    sc->statrate * 2.0);
	}
	AZ(pthread_mutex_unlock(&sc->lock));
}

static void
mcast_destroypoints(struct mcast_stat *st)
{
	unsigned u;

	for (u = 0; u < NPT; u++) {
		if (st->pt[u] != NULL) {
			DestroyPoint(st->pt[u]);
			st->pt[u] = NULL;
		}
	}
}

/*--------------------------------------------------------------------*/

static void
setup_antiloop(CK_ARG, struct mcast *sc)
{

	AZ(pthread_mutex_lock(&sc->lock));
	if (sc->looplist != NULL) {
		free(sc->looplist);
		sc->looplist = NULL;
		sc->loopidx = 0;
		sc->looplen = 0;
	}
	if (sc->antiloop > 0) {
		sc->looplist = calloc(sc->antiloop, MD5_DIGEST_LENGTH);
		if (sc->looplist == NULL) {
			ck_printf(CK_PASS, "mcast: not enough memory");
			AZ(pthread_mutex_unlock(&sc->lock));
			return;
		}
		sc->looplen = sc->antiloop * MD5_DIGEST_LENGTH;
	}
	AZ(pthread_mutex_unlock(&sc->lock));
}

static int
check_loop(struct mcast *sc, const void *pk, unsigned len)
{
	MD5_CTX ctx;
	unsigned char dig[MD5_DIGEST_LENGTH];
	unsigned u;

	if (sc->looplen == 0)
		return (0);
	MD5Init(&ctx);
	MD5Update(&ctx, pk, len);
	MD5Final(dig, &ctx);
	for (u = 0; u < sc->looplen; u += MD5_DIGEST_LENGTH)
		if (!memcmp(dig, sc->looplist + u, MD5_DIGEST_LENGTH))
			return (1);
	memcpy(sc->looplist + sc->loopidx, dig, MD5_DIGEST_LENGTH);
	sc->loopidx += MD5_DIGEST_LENGTH;
	sc->loopidx %= sc->looplen;
	return (0);
}

/*--------------------------------------------------------------------*/

static void
setup_watchdog(CK_ARG, struct mcast *sc)
{
	int i;

	AZ(pthread_mutex_lock(&sc->lock));
	if (sc->watchdog == 0) {
		sc->patpat = WD_TO_NEVER;
		if (sc->wfd >= 0) {
			i = ioctl(sc->wfd, WDIOCPATPAT, &sc->patpat);
			if (i)
				ck_printf(CK_PASS,
				    "Failed to disable watchdog: %s",
				    strerror(errno));
			close(sc->wfd);
			sc->wfd = -1;
		}
		AZ(pthread_mutex_unlock(&sc->lock));
		return;
	}
	if (sc->wfd < 0) {
		sc->wfd = open(_PATH_DEV _PATH_WATCHDOG, O_RDWR);
		if (sc->wfd < 0) {
			sc->watchdog = 0;
			sc->patpat = WD_TO_NEVER;
			AZ(pthread_mutex_unlock(&sc->lock));
			ck_printf(CK_PASS,
			    "mcast: cannot open watchdog: %s",
			    strerror(errno));
			return;
		}
	}
	sc->patpat = (unsigned)ceil(log(sc->watchdog * 1e9) / log(2.0));
	sc->patpat |= WD_ACTIVE;
	i = ioctl(sc->wfd, WDIOCPATPAT, &sc->patpat);
	if (i) {
		sc->watchdog = 0;
		sc->patpat = WD_TO_NEVER;
		ck_printf(CK_PASS,
		    "Patting watchdog failed: %s",
		    strerror(errno));
	}
	AZ(pthread_mutex_unlock(&sc->lock));
}

/*--------------------------------------------------------------------*/

static void
mcast_stats(struct mcast_stat *st, double now)
{
	SetVal(st->pt[0], st->packets);
	SetVal(st->pt[1], st->bytes);
	if (st->last_good == 0)
		st->last_good = now;
	SetVal(st->pt[2], now - st->last_good);
	SetVal(st->pt[3], st->last_status);
	SetVal(st->pt[4], st->badpackets);
	if (st->last_bad == 0)
		st->last_bad = now;
	SetVal(st->pt[5], now - st->last_bad);
	SetVal(st->pt[6], st->last_error);
	if (st->pt[7] != NULL)
		SetVal(st->pt[7], st->loops);
}

static void
mcast_dopacket(struct mcast *sc, double now) 
{
	ssize_t sz, tsz;
	char buf[65536];
	struct sockaddr sa[2];
	socklen_t salen;
	struct mcast_stat *st;
	int i;

	salen = sizeof sa;
	errno = 0;
	sz = recvfrom(sc->rxsock, buf, sizeof buf, 0, sa, &salen);
	st = &sc->stat;
	st->last_status = errno;
	if (sz < 0) {
		st->last_error = st->last_status;
		st->last_bad = now;
		st->badpackets++;
		usleep(100000);
		return;
	}
	if (sc == 0) {
		st->last_error = -1;
		st->last_bad = now;
		st->badpackets++;
		usleep(100000);
		return;
	}
	st->last_good = now;
	st->packets++;
	st->bytes += sz;

	if (sc->watchdog && ioctl(sc->wfd, WDIOCPATPAT, &sc->patpat))
		fprintf(stderr, "Watchdog not happy: %s\n",
		    strerror(errno));
	
	AZ(pthread_mutex_lock(&sc->lock));
	if (sc->looplen > 0 && check_loop(sc, buf, sz)) {
		st->loops++;
	} else {
		for (i = 1; i < MAX_TX; i++) {
			if (sc->tx[i] == NULL)
				continue;
			errno = 0;
			st = &sc->tx[i]->stat;
			tsz = sendto(sc->txsock, buf, sz, 0,
			    sc->tx[i]->ai->ai_addr,
			    sc->tx[i]->ai->ai_addrlen);
			st->last_status = errno;
			if (tsz == sz) {
				st->packets++;
				st->bytes += sz;
				st->last_good = now;
			} else if (tsz == 0) {
				st->last_bad = now;
				st->badpackets++;
				st->last_error = -1;
			} else {
				st->last_bad = now;
				st->badpackets++;
				st->last_error = st->last_status;
			}
		}
	}
	AZ(pthread_mutex_unlock(&sc->lock));
}


static void *
mcast_thread(void *priv)
{
	struct mcast *sc;
	struct timespec ts;
	struct pollfd pfd[1];
	double now;
	int i;

	CAST_OBJ_NOTNULL(sc, priv, MCAST_MAGIC);

	printf("Start mcast thread\n");
	while (1) {
		pfd[0].fd = sc->rxsock;
		pfd[0].events = POLLIN;
		i = poll(pfd, 1, 1000);
		clock_gettime(CLOCK_MONOTONIC, &ts);
		now = ts.tv_sec + ts.tv_nsec * 1e-9;
		if (i > 0)
			mcast_dopacket(sc, now);
		if (sc->laststat + sc->statrate > now)
			continue;
		mcast_stats(&sc->stat, now);
		AZ(pthread_mutex_lock(&sc->lock));
		for (i = 1; i < MAX_TX; i++) 
			if (sc->tx[i] != NULL)
				mcast_stats(&sc->tx[i]->stat, now);
		AZ(pthread_mutex_unlock(&sc->lock));
		sc->laststat = now;
	}
}

static void
mcast_reconf(struct mcast *sc)
{
	void *vp;

	if (sc->running) {
		AZ(pthread_mutex_lock(&sc->lock));
		AZ(pthread_cancel(sc->tp));
		AZ(pthread_join(sc->tp, &vp));
		AZ(pthread_mutex_unlock(&sc->lock));
		sc->running = 0;
	}
	if (sc->ai) {
		AZ(pthread_create(&sc->tp, NULL, mcast_thread, sc));
		sc->running = 1;
	}
}

static void
mcast_init(CK_ARG, struct mcast *sc)
{

	sc->rxsock = -1;
	sc->txsock = -1;
	sc->wfd = -1;
	sc->statrate = 1;
	mcast_mkpoints(sc, &sc->stat, 0);
}

static void
mcast_deinit(CK_ARG, struct mcast *sc)
{

	free(sc->addr);
	free(sc->port);
	if (sc->ai) {
		freeaddrinfo(sc->ai);
		sc->ai = NULL;
	}
	mcast_reconf(sc);
	if (sc->rxsock >= 0)
		close(sc->rxsock);
	if (sc->txsock >= 0)
		close(sc->txsock);
}

static void      
rx_mcast_on(CK_ARG, struct mcast *sc)
{
	struct ip_mreq m;
	struct in_addr ina;
	struct sockaddr_in *si;
        
	if (sc->ai->ai_family != PF_INET) {
		ck_printf(CK_PASS, "mcast: not PF_INET");
		return;
	}
	si = (void *)(sc->ai->ai_addr);
	ina = si->sin_addr;
	if (!IN_CLASSD(ntohl(ina.s_addr)))
		return;
	m.imr_multiaddr = ina;
	m.imr_interface.s_addr = htonl(INADDR_ANY);
	if (setsockopt(sc->rxsock,IPPROTO_IP, IP_ADD_MEMBERSHIP,
	    &m, sizeof(m)) < 0) {
		
		ck_printf(CK_PASS, "IP_ADD_MEMBERSHIP failed %s\n",
		    errno, strerror(errno));
	}
}

static int
setup_txsock(CK_ARG, struct mcast *sc)
{
	int txsock, i, error, j;
	struct addrinfo *ai,  hints;

	txsock = socket(sc->ai->ai_family, sc->ai->ai_socktype,
	    sc->ai->ai_protocol);
	if (txsock < 0) {
		ck_printf(CK_PASS, "Cannot open tx socket:\n\t%H\n\t%s\n",
		   sc->ai, sizeof *sc->ai, strerror(errno));
		return (1);
	}
	if (sc->txport != NULL) {
		memset(&hints, 0, sizeof hints);
		hints.ai_family = PF_INET;
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_flags = AI_PASSIVE;
		error = getaddrinfo(NULL, sc->txport, &hints, &ai);
		if (error) {
			ck_printf(CK_PASS, "getaddrinfo: %s\n",
			    gai_strerror(error));
			return (1);
		}
		i = 1;
		AZ(setsockopt(txsock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i));
		i = bind(txsock, ai->ai_addr, ai->ai_addrlen);
		if (i < 0) {
			ck_printf(CK_PASS, "Cannot bind socket:\n\t%H\n\t%s\n",
			    ai, sizeof *ai, strerror(errno));
			return (1);
		}
		freeaddrinfo(ai);
	}
	j = fcntl(txsock, F_GETFL);
	j |= O_NONBLOCK;
	i = fcntl(txsock, F_SETFL, j);
	AZ(i);
	AZ(pthread_mutex_lock(&sc->lock));
	if (sc->txsock >= 0)
		close (sc->txsock);
	sc->txsock = txsock;
	AZ(pthread_mutex_unlock(&sc->lock));
	return (0);
}

int
cfg_mcast_rx(CK_ARG, const char **arg_0, const char **arg_1)
{
	struct mcast *sc;
	struct addrinfo *ai, *ai2, hints;
	int error, i;

	CAST_OBJ_NOTNULL(sc, cfa->priv, MCAST_MAGIC);
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (sc->addr != NULL && sc->port != NULL)
			ck_printf(CK_PASS, "%s %Q %Q\n",
			    cfa->tree, sc->addr, sc->port);
		return (0);
	}
	memset(&hints, 0, sizeof hints);
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;
	error = getaddrinfo(*arg_0, *arg_1, &hints, &ai);
	if (error) {
		ck_printf(CK_PASS, "getaddrinfo: %s\n", gai_strerror(error));
		return (1);
	}
	AZ(pthread_mutex_lock(&sc->lock));
	ai2 = sc->ai;
	sc->ai = NULL;
	AZ(pthread_mutex_unlock(&sc->lock));
	mcast_reconf(sc);
	if (ai2 != NULL)
		freeaddrinfo(ai2);
	sc->ai = ai;
	free(sc->addr);
	free(sc->port);
	sc->addr = strdup(*arg_0);
	sc->port = strdup(*arg_1);

	if (setup_txsock(CK_PASS, sc))
		return (0);

	if (sc->rxsock >= 0)
		close(sc->rxsock);

	sc->rxsock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (sc->rxsock < 0) {
		ck_printf(CK_PASS, "Cannot open rx socket:\n\t%H\n\t%s\n",
		   ai, sizeof *ai, strerror(errno));
		return (0);
	}

	i = 1;
	AZ(setsockopt(sc->rxsock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i));
	
	i = bind(sc->rxsock, ai->ai_addr, ai->ai_addrlen);
	if (i < 0) {
		ck_printf(CK_PASS, "Cannot bind socket:\n\t%H\n\t%s\n",
		    ai, sizeof *ai, strerror(errno));
		return (0);
	}
	rx_mcast_on(CK_PASS, sc);

	mcast_reconf(sc);
	return (0);
}

int
cfg_mcast_tx_dest(CK_ARG, const char * *arg_0, const char * *arg_1)
{
	struct mcast_tx *txp;
	struct addrinfo *ai, hints;
	int error;

	CAST_OBJ_NOTNULL(txp, cfa->priv, MCAST_TX_MAGIC);
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (txp->addr != NULL && txp->port != NULL)
			ck_printf(CK_PASS, "%s %Q %Q\n",
			    cfa->tree, txp->addr, txp->port);
		return (0);
	}
	memset(&hints, 0, sizeof hints);
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	error = getaddrinfo(*arg_0, *arg_1, &hints, &ai);
	if (error) {
		ck_printf(CK_PASS, "getaddrinfo: %s\n", gai_strerror(error));
		return (1);
	}
	if (txp->ai != NULL)
		freeaddrinfo(txp->ai);
	txp->ai = ai;
	free(txp->addr);
	free(txp->port);
	txp->addr = strdup(*arg_0);
	txp->port = strdup(*arg_1);
	return (0);
}

struct ck_instance *
cfg_mcast_tx_new(CK_ARG, unsigned u)
{
	struct mcast *sc;
	struct mcast_tx *txp;

	CAST_OBJ_NOTNULL(sc, cfa->priv, MCAST_MAGIC);
	if (u < 1 || u >= MAX_TX) {
		ck_printf(CK_PASS, "Illegal Tx number %u [1...%u]",
		    u, MAX_TX - 1);
		return (NULL);
	}
	txp = sc->tx[u];
	if (txp == NULL) {
		ALLOC_OBJ(txp, MCAST_TX_MAGIC);
		txp->tx = u;
		sc->tx[u] = txp;
		mcast_mkpoints(sc, &txp->stat, u * 10);
	}
	txp->cki.obj = txp;
	txp->sc = sc;
	return (&txp->cki);
}

int
cfg_mcast_tx_destroy(CK_ARG)
{
	struct mcast_tx *txp;
	struct mcast *sc;

	CAST_OBJ(txp, cfa->priv, MCAST_TX_MAGIC);
	if (txp == NULL)
		return (0);

	sc = txp->sc;

	AZ(pthread_mutex_lock(&sc->lock));
	sc->tx[txp->tx] = NULL;
	AZ(pthread_mutex_unlock(&sc->lock));
	free(txp->addr);
	free(txp->port);
	if (txp->ai)
		freeaddrinfo(txp->ai);
	mcast_destroypoints(&txp->stat);
	return (0);
}

WordVar(mcast, MCAST_MAGIC, txport, txport,
    (void)setup_txsock(CK_PASS, sc));
UintVar(mcast, MCAST_MAGIC, antiloop, antiloop, 10,
    setup_antiloop(CK_PASS, sc));
UintVar(mcast, MCAST_MAGIC, watchdog, watchdog, 0,
    setup_watchdog(CK_PASS, sc));
RealVar(mcast, MCAST_MAGIC, statrate, statrate, 1.0, "%.1f", 
    mcast_updtimeout(sc));
	
MODULE_GROUP(mcast, MCAST_MAGIC);
