/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pc87366.c,v 1.9 2008/03/22 17:51:04 phk Exp $
 *
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <machine/cpufunc.h>

#include <configkit.h>

#include "measured.h"
#include "configkit_conf.h"

struct pc87366_gpio {
	unsigned			magic;
#define PC87366_GPIO_MAGIC		0x21e25173
	uint8_t				gpio;
	uint8_t				sel;
	struct ck_instance		cki;
	struct pc87366			*sc;
	struct point			*point;
	uint8_t				config;
	unsigned			port;
	uint8_t				bit;
};

struct pc87366 {
	unsigned			magic;
#define PC87366_MAGIC 			0x23676942
	struct ck_instance		cki;
	struct group			*gp;
	pthread_t			tp;
	pthread_mutex_t			lock;
	unsigned			debug;
	struct pc87366_gpio		*gpio[35];
	double				pollrate;
	int 				fd;
	unsigned			port_base;
};

static void *
pc87366_thread(void *priv)
{
	struct pc87366 *sc;
	struct pc87366_gpio *gp;
	unsigned u;

	CAST_OBJ_NOTNULL(sc, priv, PC87366_MAGIC);

	while (1) {
		AZ(pthread_mutex_lock(&sc->lock));
		for (u = 0; u < 35; u++) {
			if (sc->gpio[u] == NULL)
				continue;
			gp = sc->gpio[u];
			if (gp->point == NULL)
				continue;
			if (inb(gp->port + 1) & gp->bit)
				SetVal(gp->point, 1);
			else
				SetVal(gp->point, 0);
		}
		AZ(pthread_mutex_unlock(&sc->lock));
		usleep(sc->pollrate * 1e6);
	}
	return (NULL);
}

static void
pc87366_gpio_control(struct point *p __unused, void *priv1, void *priv2 __unused, double v)
{
	struct pc87366_gpio *gp;
	int fd;

	CAST_OBJ_NOTNULL(gp, priv1, PC87366_GPIO_MAGIC);
	assert(gp->config & 0x1);
	/*
	 * A fine point: this thread may not have inherited the /dev/io
	 * permissions from pc87366, so make sure we have what it takes.
	 */
	fd = open("/dev/io", O_RDWR);
	if (fd >= 0) {
		AZ(pthread_mutex_lock(&gp->sc->lock));
		if (v > .5)
			outb(gp->port, inb(gp->port) | gp->bit);
		else
			outb(gp->port, inb(gp->port) & ~gp->bit);
		if (inb(gp->port + 1) & gp->bit)
			SetVal(gp->point, 1);
		else
			SetVal(gp->point, 0);
		AZ(pthread_mutex_unlock(&gp->sc->lock));
		if (gp->sc->debug & 1)
			SCDebug(gp->sc, "port %x bit %x -> O%x I%x",
			    gp->port, gp->bit,
			    inb(gp->port), inb(gp->port + 1));
	} else
		GLog(gp->sc->gp, "Could not open /dev/io");
	AZ(close(fd));
}

struct ck_instance *
cfg_pc87366_gpio_new(CK_ARG, unsigned u)
{
	struct pc87366 *sc;
	struct pc87366_gpio *gp;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PC87366_MAGIC);
	if ((u % 10) > 7 || u > 34) {
		ck_printf(CK_PASS, "Illegal gpio pin %u, "
		    "legal: [0-7,10-17,20-27,30-34]", u);
		return (NULL);
	}
	AZ(pthread_mutex_lock(&sc->lock));
	gp = sc->gpio[u];
	if (gp == NULL) {
		ALLOC_OBJ(gp, PC87366_GPIO_MAGIC);
		gp->gpio = u & 0xff;
		sc->gpio[u] = gp;
	}
	gp->cki.obj = gp;
	gp->sc = sc;
	switch(gp->gpio / 10) {
	case 0: gp->port = gp->sc->port_base + 0x00; break;
	case 1: gp->port = gp->sc->port_base + 0x04; break;
	case 2: gp->port = gp->sc->port_base + 0x08; break;
	case 3: gp->port = gp->sc->port_base + 0x0a; break;
	default: assert(0 == __LINE__);
	}
	gp->sel = (gp->gpio / 10) << 4;
	gp->sel |= gp->gpio % 10;
	gp->bit = (1 << (gp->gpio % 10));
	AZ(pthread_mutex_unlock(&sc->lock));
	return (&gp->cki);
}

int
cfg_pc87366_gpio_destroy(CK_ARG)
{
	struct pc87366_gpio *gp;
	struct pc87366 *sc;

	CAST_OBJ_NOTNULL(gp, cfa->priv, PC87366_GPIO_MAGIC);

	sc = gp->sc;
	AZ(pthread_mutex_lock(&sc->lock));
	if (gp->point != NULL)
		DestroyPoint(gp->point);
	sc->gpio[gp->gpio] = NULL;
	AZ(pthread_mutex_unlock(&sc->lock));
	return (0);
}

int
cfg_pc87366_gpio_mode(CK_ARG, const char **arg_0, char **arg_1)
{
	struct pc87366_gpio *gp;
	enum point_type want;
	const char *p;
	int i;

	CAST_OBJ_NOTNULL(gp, cfa->priv, PC87366_GPIO_MAGIC);
	if (gp->sc->fd < 0)
		return (0);

	p = NULL;
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		switch (gp->config & 0x03) {
		case 0:
		case 2:
			p = "input";
			break;
		case 1:
			p = "open_drain";
			break;
		case 3:
			p = "push_pull";
			break;
		}
		assert(p != NULL);
		ck_printf(CK_PASS, "%s %s\n", cfa->tree, p);
		return (0);
	}
	if (!strcmp(*arg_0, "input")) {
		gp->config &= 0xfe;
		want = DIN;
	} else if (!strcmp(*arg_0, "push_pull")) {
		gp->config |= 0x03;
		want = DOUT;
	} else if (!strcmp(*arg_0, "open_drain")) {
		gp->config &= 0xfd;
		gp->config |= 0x01;
		want = DOUT;
	} else {
		want = 0;
		assert(0 == __LINE__);
	}
	
	AZ(pthread_mutex_lock(&gp->sc->lock));
	if (gp->point != NULL && p_type(gp->point) != want) {
		DestroyPoint(gp->point);
		gp->point = NULL;
	}
	if (gp->point == NULL)
		gp->point = CreatePoint(gp->sc->gp, gp->gpio, want);
	if (want == DOUT) {
		SetPointControl(gp->point,
		    pc87366_gpio_control, gp, NULL);
		if (*arg_1 != NULL) {
			if (atoi(*arg_1))
				outb(gp->port, inb(gp->port) | gp->bit);
			else
				outb(gp->port, inb(gp->port) & ~gp->bit);
		
		}
	}
	outb(0x2e, 0x07); outb(0x2f, 0x07);
	outb(0x2e, 0xf0); outb(0x2f, gp->sel);
	outb(0x2e, 0xf1); outb(0x2f, gp->config);
	outb(0x2e, 0xf1); i = inb(0x2f);
	if (gp->sc->debug & 1)
		SCDebug(gp->sc, "Config gpio %d config 0x%x sel 0x%x -> %x",
		    gp->gpio, gp->config, gp->sel, i);
	AZ(pthread_mutex_unlock(&gp->sc->lock));
	return (0);
}

int
cfg_pc87366_gpio_pull_up(CK_ARG)
{
	struct pc87366_gpio *gp;

	CAST_OBJ_NOTNULL(gp, cfa->priv, PC87366_GPIO_MAGIC);
	if (gp->sc->fd < 0)
		return (0);

	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (gp->config & 0x04)
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO)
		gp->config &= ~0x04;
	else
		gp->config |= 0x04;
	AZ(pthread_mutex_lock(&gp->sc->lock));
	outb(0x2e, 0x07); outb(0x2f, 0x07);
	outb(0x2e, 0xf0); outb(0x2f, gp->sel);
	outb(0x2e, 0xf1); outb(0x2f, gp->config);
	AZ(pthread_mutex_unlock(&gp->sc->lock));
	return (0);
}

static void
pc87366_init(CK_ARG, struct pc87366 *sc)
{
	int i;

	sc->pollrate = 10.0;
	sc->fd = open("/dev/io", O_RDWR);
	if (sc->fd < 0) {
		ck_printf(CK_PASS, "Could not open /dev/io: %s",
		    strerror(errno));
		return;
	}

	outb(0x2e, 0x20);
	if (inb(0x2f) != 0xe9) {
		AZ(close(sc->fd));
		sc->fd = -1;
		ck_printf(CK_PASS, "Found no PC87366 chip");
		return;
	}

	outb(0x2e, 0x07); outb(0x2f, 0x07);

	outb(0x2e, 0x60); sc->port_base = inb(0x2f) << 8;
	outb(0x2e, 0x61); sc->port_base |= inb(0x2f) << 8;

	outb(0x2e, 0x30); i = inb(0x2f); // check enable bit
	SCDebug(sc, "base %x 0x30=%x", sc->port_base, i);
	

	
	AZ(pthread_create(&sc->tp, NULL, pc87366_thread, sc));
}

static void
pc87366_deinit(CK_ARG, struct pc87366 *sc)
{
	void *vp;

	if (sc->fd < 0)
		return;
	AZ(close(sc->fd));
	sc->fd = -1;
	AZ(pthread_mutex_lock(&sc->lock));
	AZ(pthread_cancel(sc->tp));
	AZ(pthread_join(sc->tp, &vp));
	AZ(pthread_mutex_unlock(&sc->lock));
}

RealVar(pc87366, PC87366_MAGIC, pollrate, pollrate, 10.0, "%.3f", );
UintVar(pc87366, PC87366_MAGIC, debug, debug, 0, );
MODULE_GROUP(pc87366, PC87366_MAGIC);
