/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pic18.c,v 1.66 2008/02/16 16:27:39 phk Exp $
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>
#include "measured.h"
#include "configkit_conf.h"

/* Size of receive buffer.  PIC18 firmware is limited to 256 bytes */
#define NRCV_BUF 		256
/* Number of receive tokens */
#define NRCV 			(NRCV_BUF / 2)

#define MAX_PIN			65

#define DEFAULT_ADCON2		0x3e
#define DEFAULT_AVERAGING	5
#define DEFAULT_TEMPRES		3
#define DEFAULT_VREF		5.0
#define DEFAULT_POLLRATE	1

struct pic18 {
	unsigned		magic;
#define PIC18_MAGIC 0x3292932
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	struct serial		*serial;

	int			reconf;

	char			rcvbuf[NRCV_BUF];
	char			*rcv[NRCV];

	u_int			hwpins;
	struct pic18pin		*pins[MAX_PIN + 1];

	struct ck_instancehead  ckih;

	double			vref;
	unsigned		pollrate;
	unsigned		adcon2;
	unsigned		tempres;
	unsigned		averaging;
	unsigned		debug;
	
};

struct pic18pin {
	unsigned		magic;
#define PIC18PIN_MAGIC 0x2990428
	struct ck_instance	cki;
	unsigned		pin;
	struct pic18		*softc;
	enum point_type		type;
	unsigned char		subtype;
#define ST_TEMP			(1 << 0)
#define ST_PULLUP		(1 << 1)
	struct point		*point;
};


static int Ask(struct pic18 *sc, const char *rsp, const char *fmt, ...) __printflike(3, 4);

static void
Send(const struct pic18 *sc, const char *s)
{
	int i;

	/* XXX: use writev */
	if (sc->debug)
		SCDebug(sc, "PIC18 TX: %V", s);
	i = write(serial_fd(sc->serial), s, strlen(s));
	assert(i == (int)strlen(s));
	i = write(serial_fd(sc->serial), "\r", 1);
	assert(i == 1);
	/* XXX: return 0=success/1=fail */
}

/*
 * returns 0 == success, 1 == failue
 */
static int
Recv(struct pic18 *sc)
{
	int i;
	char *p;

	sc->rcvbuf[0] = '\0';
	while (1) {
		p = strchr(sc->rcvbuf, '\0');
		assert(p != NULL);
		/* XXX: use longer reads ? */
		i = read(serial_fd(sc->serial), p, 1);
		if (i != 1)
			return (1);
		p[1] = '\0';
		if (*p == '\r') {
			*p = '\0';
			continue;
		}
		if (*p == '\n') {
			*p = '\0';
			if (sc->debug)
				SCDebug(sc, "PIC18 RX: %V", sc->rcvbuf);
			return (0);
		}
		if ((*p < ' ' || *p > '~') ||
		    (p - sc->rcvbuf) + 1 > (int)sizeof sc->rcvbuf) {
			GLog(sc->gp, "PIC18 flushing input buffer");
			/* Garbage, wait for it to stop */
			do {
				i = read(serial_fd(sc->serial), sc->rcvbuf,
				    sizeof sc->rcvbuf);
				assert(i >= 0);
			} while (i != 0);
			sc->rcvbuf[0] = '\0';
			return (1);
		}
	}
}

static int
Ask(struct pic18 *sc, const char *rsp, const char *fmt, ...)
{
	va_list ap;
	char *s;
	char **rp;
	int retval;

	va_start(ap, fmt);
	vasprintf(&s, fmt, ap);
	va_end(ap);

	retval = 0;
	while (1) {
		Send(sc, s);
		/* XXX: retry only 3 times */
		if (!Recv(sc)) 
			break;
		GNotice(sc->gp, "PIC18 no response");
	}
	if (!memcmp(sc->rcvbuf, "RCON", 4)) {
		GNotice(sc->gp, "PIC18 \"%V\" (I sent: %s)", sc->rcvbuf, s);
		sc->reconf = 1;
		retval = 1;
	} else if (!memcmp(sc->rcvbuf, "ERR", 3)) {
		GNotice(sc->gp, "PIC18 \"%V\" (I sent: %s)", sc->rcvbuf, s);
		retval = 1;
	}
	free(s);
	if (retval) {
		while (!Recv(sc))
			continue;
		Send(sc, "CLR");
		(void)Recv(sc);
		return(retval);
	}
	s = sc->rcvbuf;
	for (rp = sc->rcv; (*rp = strsep(&s, " ")) != NULL;) {
		if (**rp != '\0') {
			if (++rp >= &sc->rcv[NRCV])
				break;
		}
	}
	if (sc->rcv[0] == NULL)
		return (1);
	if (strcmp(sc->rcv[0], rsp)) {
		GLog(sc->gp, "PIC18 expected \"%s\" got \"%s\"",
		    rsp, sc->rcv[0]);
		sc->reconf = 1;
		return (1);
	}
	return(0);
}

static void
pic18_setup(struct pic18 *sc)
{
	u_int u, v;
	const char *should;

	if (Ask(sc, "REV", "ID"))
		return;
	/* XXX: check firmware version */
	/* XXX: check TEMP capability */
	if (1 != sscanf(sc->rcv[3], "%u", &sc->hwpins))
		return;
	if (Ask(sc, "ADCON2_OK", "ADCON2 %u", sc->adcon2))
		SCDebug(sc, "Didn't like ADCON2");
	if (Ask(sc, "TEMPRES_OK", "TEMPRES %u", sc->tempres))
		SCDebug(sc, "Didn't like TEMPRES");
	if (Ask(sc, "AVG_OK", "AVG %u", sc->averaging))
		SCDebug(sc, "Didn't like AVG");
	for (u = 1; u < MAX_PIN && u <= sc->hwpins; u++) {
		if (Ask(sc, "PIN", "CONF %u", u))
			return;
		if (1 != sscanf(sc->rcv[1], "%u", &v))
			return;
		if (u != v)
			return;
		if (sc->pins[u] == NULL)
			continue;
		switch(sc->pins[u]->type) {
		case DOUT:
			should = "DOUT";
			break;
		case DIN:
			should = "DIN";
			break;
		case COUNT:
			should = "CNT";
			break;
		case AIN:
			if (sc->pins[u]->subtype == ST_TEMP)
				should = "TEMP";
			else
				should = "AIN";
			break;
		case UNDEF:
			should = NULL;
			break;
		default:
			should = "OFF";
			break;
		}
		if (should != NULL &&
		    strcmp(sc->rcv[2], should) &&
		    Ask(sc, "CHG", "DEF %u %s", u, should))
			return;
		if (sc->pins[u]->subtype == ST_PULLUP &&
		    Ask(sc, "STAT", "SET %u %u", u, 1))
			return;
	}
	sc->reconf = 0;
}

static void
pic18_poll(struct pic18 *sc)
{
	u_int u, v;

	if (Ask(sc, "DATA", "SHOW"))
		return;
	for (u = 1; u < MAX_PIN && u <= sc->hwpins; u++) {
		if (sc->reconf)
			break;
		if (sc->pins[u] == NULL)
			continue;
		if (sc->pins[u]->point == NULL)
			continue;
		if (1 != sscanf(sc->rcv[u], "%u", &v))
			return;
		if (sc->pins[u]->subtype == ST_TEMP) {
			if (v >= 0xf000) {
				SetVal(sc->pins[u]->point,
				    ((int)v - 65536) / 16.0);
			} else if (v >= 0xc000) {
				PLog(sc->pins[u]->point,
				    "PIC18/TEMP Wrong device family: 0x%x",
				    v - 0xc000);
			} else if (v == 0x8000) {
				/* timeout */
			} else if (v > 0x8000) {
				PLog(sc->pins[u]->point,
				    "PIC18/TEMP Error: 0x%x",
				    v - 0x8000);
			} else {
				SetVal(sc->pins[u]->point, (v / 16.0));
			}
		} else if (sc->pins[u]->type == AIN)
			SetVal(sc->pins[u]->point,
			    sc->vref * v / 65535.0);
		else
			SetVal(sc->pins[u]->point, (double)v);
	}
}

static void *
pic18_thread(void *priv)
{
	struct termios t;
	struct pic18 *sc;
	unsigned u;

	CAST_OBJ_NOTNULL(sc, priv, PIC18_MAGIC);
	SCLog(sc, "Thread Restarting");
	sc->reconf = 1;

	AZ(tcgetattr(serial_fd(sc->serial), &t));
	cfmakeraw(&t);
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 20;
	t.c_cflag |= CLOCAL;
	AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));

	if (sc->pollrate < 1) {
		SCLog(sc, "Pollrate adjusted up to 1s");
		sc->pollrate = 1;
	} else if (sc->pollrate > 3600) {
		SCLog(sc, "Pollrate adjusted up down 1h");
		sc->pollrate = 3600;
	}

	for (u = 0; u < MAX_PIN; u++)
		if (sc->pins[u] != NULL && sc->pins[u]->point != NULL)
			SetPointTimeout(
			    sc->pins[u]->point, 15.0 + sc->pollrate);

	(void)Recv(sc);	/* clear input buffer */
	while (1) {
		AZ(pthread_mutex_lock(&sc->lock));
		if (sc->reconf)
			pic18_setup(sc);
		if (!sc->reconf)
			pic18_poll(sc);
		AZ(pthread_mutex_unlock(&sc->lock));
		if (sc->reconf)
			sleep(10);
		else if (sc->pollrate >= 1 && sc->pollrate <= 3600)
			sleep(sc->pollrate);
		else
			sleep(1);
	}
}

static void
pic18_control(struct point *p __unused, void *priv1, void *priv2 __unused, double v)
{
	struct pic18pin *pp;
	struct pic18 *sc;

	CAST_OBJ_NOTNULL(pp, priv1, PIC18PIN_MAGIC);
	CAST_OBJ_NOTNULL(sc, pp->softc, PIC18_MAGIC);
	assert (pp->type == DOUT);
	AZ(pthread_mutex_lock(&sc->lock));
	(void)Ask(sc, "STAT", "SET %u %u", pp->pin, v > .5 ? 1 : 0);
	AZ(pthread_mutex_unlock(&sc->lock));
	SetVal(pp->point, v);
}

/* ConfigKit functions -----------------------------------------------*/

static void
pic18_init(CK_ARG, struct pic18 *sc) 
{
	serial_init(CK_PASS, &sc->serial, pic18_thread, sc);
	sc->vref = DEFAULT_VREF;
	sc->adcon2 = DEFAULT_ADCON2;
	sc->averaging = DEFAULT_AVERAGING;
	sc->tempres = DEFAULT_TEMPRES;
	sc->pollrate = DEFAULT_POLLRATE;
        TAILQ_INIT(&sc->ckih);
};

static void
pic18_deinit(CK_ARG, struct pic18 *sc)
{
	struct ck_instance *ip, *ip2;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18_MAGIC);
	serial_destroy(CK_PASS, &sc->serial);
	TAILQ_FOREACH_SAFE(ip, &sc->ckih, next, ip2) { 
		/* XXX: Configkit should do this ? */
		cfa->priv = ip->obj;
		(void)cfg_pic18_pin_destroy(CK_PASS);
	}
}

SerialFunc(pic18, PIC18_MAGIC, serial);

RealVar(pic18, PIC18_MAGIC, vref, vref, DEFAULT_VREF, "%.6f", );

UintVar(pic18, PIC18_MAGIC, adcon2, adcon2, DEFAULT_ADCON2,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18, PIC18_MAGIC, averaging, averaging, DEFAULT_AVERAGING,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18, PIC18_MAGIC, tempres, tempres, DEFAULT_TEMPRES,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18, PIC18_MAGIC, pollrate, pollrate, DEFAULT_POLLRATE, 
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18, PIC18_MAGIC, debug, debug, 0, );


struct ck_instancehead *
cfg_pic18_pin_head(CK_ARG)
{
	struct pic18 *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18_MAGIC);
	return (&sc->ckih);
}

struct ck_instance *
cfg_pic18_pin_new(CK_ARG, unsigned pin)
{
	struct pic18 *sc;
	struct pic18pin *pp;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18_MAGIC);
	if (pin == 0 || pin > MAX_PIN) {
		ck_printf(CK_PASS, "Illegal pin number %u [1...%u]",
		    pin, MAX_PIN);
		return (NULL);
	}
	pp = sc->pins[pin];
	if (pp == NULL) {
		ALLOC_OBJ(pp, PIC18PIN_MAGIC);
		sc->pins[pin] = pp;;
	}
	pp->cki.obj = pp;
	pp->softc = sc;
	pp->pin = pin;
	pp->type = UNDEF;
	sc->reconf = 1;
	return (&pp->cki);
}

int
cfg_pic18_pin_destroy(CK_ARG)
{
	struct pic18pin *sc;

	CAST_OBJ(sc, cfa->priv, PIC18PIN_MAGIC);
	if (sc == NULL)
		return (0);

	sc->type = UNDEF;
	if (sc->point != NULL)
		DestroyPoint(sc->point);
	sc->point = NULL;
	sc->softc->reconf = 1;
	return (0);
}

int
cfg_pic18_pin_type(CK_ARG, const char **type)	/*lint -e{818} */
{
	struct pic18pin *pp;
	enum point_type pt;

	CAST_OBJ_NOTNULL(pp, cfa->priv, PIC18PIN_MAGIC);
	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (pp->type == UNDEF)
			return (0);
		if (pp->subtype == ST_TEMP)
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, "temp");
		else if (pp->subtype == ST_PULLUP)
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, "pullup");
		else
			ck_printf(CK_PASS, "%s %s\n",
			    cfa->tree, p_Type(pp->point));
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (pp->point == NULL)
			return (0);
		DestroyPoint(pp->point);
		pp->point = NULL;
		pp->type = UNDEF;
		pp->subtype = 0;
		return (0);
	}
	assert(type != NULL);
	if (!strcmp(*type, "temp"))  {
		pt = AIN;
		pp->subtype = ST_TEMP;
	} else if (!strcmp(*type, "pullup"))  {
		pt = DOUT;
		pp->subtype = ST_PULLUP;
	} else {
		pt = TypePoint(*type);
		pp->subtype = 0;
	}
		
	if (pt == UNDEF) {
		ck_printf(CK_PASS, "Wrong point type (%s)\n", *type);
		return (0);
	}
	if (pp->point != NULL && pp->type == pt && pp->subtype != ST_PULLUP)
		return (0);
	if (pp->point != NULL)
		DestroyPoint(pp->point);
	if (pp->subtype == ST_PULLUP) {
		pp->type = pt;
		return (0);
	}
	pp->point = CreatePoint(pp->softc->gp, pp->pin, pt);
	if (pt == DOUT && pp->subtype != ST_PULLUP)
		SetPointControl(pp->point, pic18_control, pp, NULL);
	SetPointTimeout(pp->point, 15.0);
	/* XXX: set format %.3f vs %.0f ? */
	pp->type = pt;
	if (pp->subtype == ST_TEMP) {
		SetPointUnit(pp->point, "Celcius");
		SetPointFormat(pp->point, "%.3f");
	}
	return (0);
}

MODULE_GROUP(pic18, PIC18_MAGIC);
