/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pic18m.c,v 1.81 2008/05/23 22:26:39 phk Exp $
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <assert.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>
#include "measured.h"
#include "configkit_conf.h"
#include "phk_argv.h"

#include <sys/file.h>

/* Size of receive buffer.  PIC18M firmware is limited to 256 bytes */
#define NRCV_BUF 		256

#define MAX_PIN			65

#define DEFAULT_ADCON2		0x3e
#define DEFAULT_AVERAGING	5
#define DEFAULT_TEMPRES		3
#define DEFAULT_VREF		5.0
#define DEFAULT_POLLRATE	1

struct pic18m {
	unsigned		magic;
#define PIC18M_MAGIC 0x3292932
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	struct serial		*serial;

	int			reconf;

	char			rcvbuf[NRCV_BUF];
	char			**av;

	u_int			hwpins;
	struct pic18mpin	*pins[MAX_PIN + 1];

	struct ck_instancehead  pin_ckih;
	struct ck_instancehead  param_ckih;

	int			multicast;
	double			vref;
	unsigned		pollrate;
	unsigned		adcon2;
	unsigned		tempres;
	unsigned		averaging;
	unsigned		debug;
	
};

struct pic18mpin {
	unsigned		magic;
#define PIC18MPIN_MAGIC 0x2990428
	struct ck_instance	cki;
	unsigned		pin;
	struct pic18m		*softc;
	enum point_type		type;
	unsigned char		subtype;
#define ST_TEMP			(1 << 0)
#define ST_PULLUP		(1 << 1)
#define ST_ADC			(1 << 2)
	struct point		*point;
};

struct pic18mparam {
	unsigned		magic;
#define PIC18MPARAM_MAGIC 0x3cd278a1
	struct ck_instance	cki;
	struct pic18m		*softc;

	char			*name;
	char			*val_str;
	double			valf;
	unsigned		val;
};


static int Ask(struct pic18m *sc, const char *rsp, const char *fmt, ...) __printflike(3, 4);

static void
Send(const struct pic18m *sc, const char *s)
{
	int i;
	char c;

	/* XXX: use writev */
	if (sc->debug)
		SCDebug(sc, "PIC18M TX(%d): %V", sc->multicast, s);
	if (sc->multicast) {
		c = sc->multicast;
		write(serial_fd(sc->serial), &c, 1);
	} 
	i = write(serial_fd(sc->serial), s, strlen(s));
	assert(i == (int)strlen(s));
	i = write(serial_fd(sc->serial), "\r", 1);
	assert(i == 1);
	/* XXX: return 0=success/1=fail */
}

/*
 * Flush input
 */
static int
Flush(struct pic18m *sc)
{
	int i;

	GLog(sc->gp, "PIC18M flushing input buffer");
	/* Garbage, wait for it to stop */
	do {
		i = read(serial_fd(sc->serial), sc->rcvbuf, sizeof sc->rcvbuf);
		assert(i >= 0);
	} while (i != 0);
	sc->rcvbuf[0] = '\0';
	return (1);
}


/*
 * returns 0 == success, 1 == failue
 */
static int
Recv(struct pic18m *sc)
{
	int i;
	char *p, mc;

	sc->rcvbuf[0] = '\0';
	if (sc->multicast) {
		i = read(serial_fd(sc->serial), &mc, 1);
		if (i != 1) {
			GLog(sc->gp, "PIC18M no rx char");
			return (Flush(sc));
		}
		if (mc != sc->multicast) {
			GLog(sc->gp, "PIC18M wrong multicast %d", mc);
			return (Flush(sc));
		}
	}
	while (1) {
		p = strchr(sc->rcvbuf, '\0');
		assert(p != NULL);
		/* XXX: use longer reads ? */
		i = read(serial_fd(sc->serial), p, 1);
		if (i != 1)
			return (1);
		p[1] = '\0';
		if (*p == '\r') {
			*p = '\0';
			continue;
		}
		if (*p == '\n') {
			*p = '\0';
			if (sc->debug)
				SCDebug(sc, "PIC18M RX(%d): %V",
				    mc, sc->rcvbuf);
			return (0);
		}
		if ((*p < ' ' || *p > '~') ||
		    (p - sc->rcvbuf) + 1 > (int)sizeof sc->rcvbuf)
			return (Flush(sc));
	}
}

static int
Ask(struct pic18m *sc, const char *rsp, const char *fmt, ...)
{
	va_list ap;
	char *s;
	int retval;

	va_start(ap, fmt);
	vasprintf(&s, fmt, ap);
	va_end(ap);

	retval = 0;
	serial_lock(sc->serial);
	if (sc->reconf) {
		Send(sc, "CLR");
		if (Recv(sc)) {
			GNotice(sc->gp, "PIC18M no response");
			serial_unlock(sc->serial);
			return (1);
		}
	}
	Send(sc, s);
	retval = Recv(sc);
	serial_unlock(sc->serial);
	if (retval) {
		GNotice(sc->gp, "PIC18M no response");
		sc->reconf = 1;
	} else if (!memcmp(sc->rcvbuf, "RCON", 4)) {
		GNotice(sc->gp, "PIC18M \"%V\" (I sent: %s)", sc->rcvbuf, s);
		retval = 1;
	} else if (!memcmp(sc->rcvbuf, "ERR", 3)) {
		GNotice(sc->gp, "PIC18M \"%V\" (I sent: %s)", sc->rcvbuf, s);
		retval = 1;
	}
	free(s);
	if (retval) {
		sc->reconf = 1;
		return (retval);
	}

	if (sc->av)
		PhkArgvFree(sc->av);
	sc->av = PhkArgvParse(sc->rcvbuf, 0);
	if (sc->av[0] != NULL)
		return (1);
	if (strcmp(sc->av[1], rsp)) {
		GLog(sc->gp, "PIC18M expected \"%s\" got \"%s\"",
		    rsp, sc->av[1]);
		sc->reconf = 1;
		return (1);
	}
	return(0);
}

static void
pic18m_setup(struct pic18m *sc)
{
	u_int u, v;
	const char *should;
	struct pic18mparam *pm;
	struct ck_instance *cip;

	if (Ask(sc, "REV2", "ID"))
		return;
	if (1 != sscanf(sc->av[5], "%u", &sc->hwpins))
		return;
	for (u = 1; u < MAX_PIN && u <= sc->hwpins; u++) {
		/* Give other users of the serial port a chance */
		usleep(20000);
		if (Ask(sc, "PIN", "CONF %u", u))
			return;
		/*
		 * We must have at least:
		 *      [1]   [2] [3]  [4] [5]    [6]
		 * 	"PIN" %d %mode "|" %mode { %mode | "|" }
		 */
		for (v = 2; v < 7; v++)
			if (sc->av[v] == NULL)
				return;

		/* Check pin number */
		if (1 != sscanf(sc->av[2], "%u", &v))
			return;
		if (u != v)
			return;

		if (strcmp(sc->av[4], "|"))
			return;

		if (sc->pins[u] == NULL) {
			if (!strcmp(sc->av[7], "|")) {
				/* Pin can have multiple modes */
				continue;
			}
			if (!strcmp(sc->av[3], "OFF")) { 
				/* Not enabled */
				continue;
			}
			fprintf(stderr, "Missing: pic18m %d pin %d type %s\n",
				GROUP_OF(GI(sc->gp)), v, sc->av[3]);
			continue;
		}
		switch(sc->pins[u]->type) {
		case DOUT:
			should = "DOUT";
			break;
		case DIN:
			should = "DIN";
			break;
		case COUNT:
			should = "CNT";
			break;
		case AOUT:
			should = "AOUT";
			break;
		case AIN:
			if (sc->pins[u]->subtype == ST_TEMP)
				should = "TEMP";
			else
				should = "AIN";
			break;
		case UNDEF:
			should = NULL;
			break;
		default:
			should = "OFF";
			break;
		}
		/* Look for '|' after permissible modes, name follows */
		for (v = 6; sc->av[v] != NULL; v++) {
			if (!strcmp(sc->av[v], "|"))
				break;
		}
		v++;
		if (sc->av[v] != NULL && strlen(sc->av[v]) > 0)
			SetPointLabel(sc->pins[u]->point, sc->av[v]);

		if (should != NULL &&
		    strcmp(sc->av[3], should) &&
		    Ask(sc, "CHG", "DEF %u %s", u, should))
			return;
		if (sc->pins[u]->subtype == ST_PULLUP &&
		    Ask(sc, "STAT", "SET %u %u", u, 1))
			return;
	}
	TAILQ_FOREACH(cip, &sc->param_ckih, next) {
		CAST_OBJ_NOTNULL(pm, cip->obj, PIC18MPARAM_MAGIC);
		Ask(sc, "PARAM_OK", "PSET %s %u", pm->name, pm->val);
	}
	sc->reconf = 0;
}

static void
pic18m_poll(struct pic18m *sc)
{
	u_int u, v;

	if (Ask(sc, "DATA", "SHOW"))
		return;
	for (u = 1; u < MAX_PIN && u <= sc->hwpins; u++) {
		if (sc->reconf)
			break;
		if (sc->pins[u] == NULL)
			continue;
		if (sc->pins[u]->point == NULL)
			continue;
		if (1 != sscanf(sc->av[u + 1], "%u", &v))
			return;
		if (sc->pins[u]->subtype == ST_TEMP) {
			if (v >= 64656) {
				SetVal(sc->pins[u]->point,
				    ((int)v - 65536) / 16.0);
			} else if (v >= 40000) {
				PLog(sc->pins[u]->point,
				    "PIC18M/TEMP Wrong device family: 0x%x",
				    v - 40000);
			} else if (v == 0x8000) {
				/* timeout */
			} else if (v > 2000) {
				PLog(sc->pins[u]->point,
				    "PIC18M/TEMP Error: %d", v);
			} else {
				SetVal(sc->pins[u]->point, (v / 16.0));
			}
		} else if (sc->pins[u]->type == AIN &&
		    sc->pins[u]->subtype == ST_ADC)
			SetVal(sc->pins[u]->point,
			    sc->vref * v / 65535.0);
		else
			SetVal(sc->pins[u]->point, (double)v);
	}
}

static void *
pic18m_thread(void *priv)
{
	struct termios t;
	struct pic18m *sc;
	unsigned u;

	CAST_OBJ_NOTNULL(sc, priv, PIC18M_MAGIC);
	SCLog(sc, "Thread Restarting");
	sc->reconf = 1;

	AZ(tcgetattr(serial_fd(sc->serial), &t));
	cfmakeraw(&t);
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 2;
	t.c_cflag |= CLOCAL;
	AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));

	if (sc->pollrate < 1) {
		SCLog(sc, "Pollrate adjusted up to 1s");
		sc->pollrate = 1;
	} else if (sc->pollrate > 3600) {
		SCLog(sc, "Pollrate adjusted up down 1h");
		sc->pollrate = 3600;
	}

	for (u = 0; u < MAX_PIN; u++)
		if (sc->pins[u] != NULL && sc->pins[u]->point != NULL)
			SetPointTimeout(
			    sc->pins[u]->point, 15.0 + sc->pollrate);

	while (1) {
		AZ(pthread_mutex_lock(&sc->lock));
		if (sc->reconf)
			pic18m_setup(sc);
		if (!sc->reconf)
			pic18m_poll(sc);
		AZ(pthread_mutex_unlock(&sc->lock));
		if (sc->reconf)
			sleep(1);
		else if (sc->pollrate >= 1 && sc->pollrate <= 3600)
			sleep(sc->pollrate);
		else
			sleep(1);
	}
}

static void
pic18m_control(struct point *p __unused, void *priv1, void *priv2 __unused, double v)
{
	struct pic18mpin *pp;
	struct pic18m *sc;

	CAST_OBJ_NOTNULL(pp, priv1, PIC18MPIN_MAGIC);
	CAST_OBJ_NOTNULL(sc, pp->softc, PIC18M_MAGIC);
	assert (pp->type == DOUT || pp->type == AOUT);
	AZ(pthread_mutex_lock(&sc->lock));
	if (pp->type == DOUT)
		(void)Ask(sc, "STAT", "SET %u %u", pp->pin, v > .5 ? 1 : 0);
	else
		(void)Ask(sc, "STAT", "SET %u %.0f", pp->pin, v);
	AZ(pthread_mutex_unlock(&sc->lock));
	SetVal(pp->point, v);
}

/* ConfigKit functions -----------------------------------------------*/

static void
pic18m_init(CK_ARG, struct pic18m *sc) 
{
	serial_init(CK_PASS, &sc->serial, pic18m_thread, sc);
	serial_shared(CK_PASS, sc->serial);
	sc->vref = DEFAULT_VREF;
	sc->adcon2 = DEFAULT_ADCON2;
	sc->averaging = DEFAULT_AVERAGING;
	sc->tempres = DEFAULT_TEMPRES;
	sc->pollrate = DEFAULT_POLLRATE;
        TAILQ_INIT(&sc->param_ckih);
        TAILQ_INIT(&sc->pin_ckih);
}

static void
pic18m_deinit(CK_ARG, struct pic18m *sc)
{
	struct ck_instance *ip, *ip2;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18M_MAGIC);
	serial_destroy(CK_PASS, &sc->serial);
	TAILQ_FOREACH_SAFE(ip, &sc->pin_ckih, next, ip2) { 
		/* XXX: Configkit should do this ? */
		cfa->priv = ip->obj;
		(void)cfg_pic18m_pin_destroy(CK_PASS);
	}
}

SerialFunc(pic18m, PIC18M_MAGIC, serial);

RealVar(pic18m, PIC18M_MAGIC, vref, vref, DEFAULT_VREF, "%.6f", );

UintVar(pic18m, PIC18M_MAGIC, adcon2, adcon2, DEFAULT_ADCON2,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18m, PIC18M_MAGIC, averaging, averaging, DEFAULT_AVERAGING,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18m, PIC18M_MAGIC, tempres, tempres, DEFAULT_TEMPRES,
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18m, PIC18M_MAGIC, pollrate, pollrate, DEFAULT_POLLRATE, 
    serial_reconf(CK_PASS, sc->serial));

UintVar(pic18m, PIC18M_MAGIC, multicast, multicast, 0, );

UintVar(pic18m, PIC18M_MAGIC, debug, debug, 0, );

/**********************************************************************/

struct ck_instancehead *
cfg_pic18m_param_head(CK_ARG)
{
	struct pic18m *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18M_MAGIC);
	return (&sc->param_ckih);
}

struct ck_instance *
cfg_pic18m_param_new(CK_ARG, const char *nam)
{
	struct pic18m *sc;
	struct pic18mparam *pm;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18M_MAGIC);
	ALLOC_OBJ(pm, PIC18MPARAM_MAGIC);
	pm->cki.obj = pm;
	pm->softc = sc;
	pm->name = strdup(nam);
	assert(pm->name != NULL);
	return (&pm->cki);
}

int
cfg_pic18m_param_val(CK_ARG, const char **arg_0)
{
	struct pic18mparam *pm;

	CAST_OBJ_NOTNULL(pm, cfa->priv, PIC18MPARAM_MAGIC);

	if (ck_RealVar2(CK_PASS, arg_0, &pm->valf, &pm->val_str, 0, "%.0f")) {
		if (round(pm->valf) < 0 && round(pm->valf) > 65535) {
			ck_printf(CK_PASS, "Illegal param value %g",
			    round(pm->valf));
		} else {
			pm->val = round(pm->valf);
			pm->softc->reconf = 1;
		}
	}
	return (0);
}

int
cfg_pic18m_param_destroy(CK_ARG)
{
	struct pic18mparam *pm;

	CAST_OBJ_NOTNULL(pm, cfa->priv, PIC18MPARAM_MAGIC);
	return (0);
}

/**********************************************************************/

struct ck_instancehead *
cfg_pic18m_pin_head(CK_ARG)
{
	struct pic18m *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18M_MAGIC);
	return (&sc->pin_ckih);
}

struct ck_instance *
cfg_pic18m_pin_new(CK_ARG, unsigned pin)
{
	struct pic18m *sc;
	struct pic18mpin *pp;

	CAST_OBJ_NOTNULL(sc, cfa->priv, PIC18M_MAGIC);
	if (pin == 0 || pin > MAX_PIN) {
		ck_printf(CK_PASS, "Illegal pin number %u [1...%u]",
		    pin, MAX_PIN);
		return (NULL);
	}
	pp = sc->pins[pin];
	if (pp == NULL) {
		ALLOC_OBJ(pp, PIC18MPIN_MAGIC);
		sc->pins[pin] = pp;
	}
	pp->cki.obj = pp;
	pp->softc = sc;
	pp->pin = pin;
	pp->type = UNDEF;
	sc->reconf = 1;
	return (&pp->cki);
}

int
cfg_pic18m_pin_destroy(CK_ARG)
{
	struct pic18mpin *sc;

	CAST_OBJ(sc, cfa->priv, PIC18MPIN_MAGIC);
	if (sc == NULL)
		return (0);

	sc->type = UNDEF;
	if (sc->point != NULL)
		DestroyPoint(sc->point);
	sc->point = NULL;
	sc->softc->reconf = 1;
	return (0);
}

int
cfg_pic18m_pin_type(CK_ARG, const char **type)	/*lint -e{818} */
{
	struct pic18mpin *pp;
	enum point_type pt;

	CAST_OBJ_NOTNULL(pp, cfa->priv, PIC18MPIN_MAGIC);
	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (pp->type == UNDEF)
			return (0);
		if (pp->subtype == ST_TEMP)
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, "temp");
		else if (pp->subtype == ST_ADC)
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, "adc");
		else if (pp->subtype == ST_PULLUP)
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, "pullup");
		else
			ck_printf(CK_PASS, "%s %s\n",
			    cfa->tree, p_Type(pp->point));
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (pp->point == NULL)
			return (0);
		DestroyPoint(pp->point);
		pp->point = NULL;
		pp->type = UNDEF;
		pp->subtype = 0;
		return (0);
	}
	assert(type != NULL);
	if (!strcmp(*type, "temp"))  {
		pt = AIN;
		pp->subtype = ST_TEMP;
	} else if (!strcmp(*type, "adc"))  {
		pt = AIN;
		pp->subtype = ST_ADC;
	} else if (!strcmp(*type, "pullup"))  {
		pt = DOUT;
		pp->subtype = ST_PULLUP;
	} else {
		pt = TypePoint(*type);
		pp->subtype = 0;
	}
		
	if (pt == UNDEF) {
		ck_printf(CK_PASS, "Wrong point type (%s)\n", *type);
		return (0);
	}
	if (pp->point != NULL && pp->type == pt && pp->subtype != ST_PULLUP)
		return (0);
	if (pp->point != NULL)
		DestroyPoint(pp->point);
	if (pp->subtype == ST_PULLUP) {
		pp->type = pt;
		return (0);
	}
	pp->point = CreatePoint(pp->softc->gp, pp->pin, pt);
	if (pt == DOUT && pp->subtype != ST_PULLUP)
		SetPointControl(pp->point, pic18m_control, pp, NULL);
	if (pt == AOUT)
		SetPointControl(pp->point, pic18m_control, pp, NULL);
	SetPointTimeout(pp->point, 15.0);
	/* XXX: set format %.3f vs %.0f ? */
	pp->type = pt;
	if (pp->subtype == ST_TEMP) {
		SetPointUnit(pp->point, "Celcius");
		SetPointFormat(pp->point, "%.3f");
	}
	return (0);
}

MODULE_GROUP(pic18m, PIC18M_MAGIC);
