/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: prs10.c,v 1.25 2008/02/16 15:43:51 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <configkit.h>
#include <termios.h>
#include <errno.h>
#include <unistd.h>

#include "measured.h"
#include "configkit_conf.h"

static struct ptdesc prs10_pt[] = {
#define PRS10A(a, b, c, d, e)	{.idx = b, .type = c, .label = e, },
#define PRS10B(a, b, c, d)	{.idx = c, .type = DIN, .label = d},
#include "prs10.h"
#undef PRS10A
#undef PRS10B
    {.label = NULL, }
};

/* Softc -------------------------------------------------------------*/

struct prs10 {
	unsigned		magic;
#define PRS10_MAGIC		0x683486e5
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;

	unsigned		debug;
	struct serial		*serial;
	struct point		**pp;
	unsigned		np;
};

/* Answer parsing functions ------------------------------------------*/

static void
prs10_num(const struct prs10 *sc, unsigned pi, const char *buf)
{
	double a;

	if (1 != sscanf(buf, "%lf", &a))
		return;
	SetVal(sc->pp[pi], a);
}

static void
prs10_st(const struct prs10 *sc, unsigned pi __unused, const char *buf)
{
	unsigned u[6];

	if (6 != sscanf(buf, "%u,%u,%u,%u,%u,%u",
	    &u[0], &u[1], &u[2], &u[3], &u[4], &u[5]))
		return;
#define PRS10A(a, b, c, d, e)
#define PRS10B(a, b, c,d) SetVal(sc->pp[c], u[a] & b);
#include "prs10.h"
#undef PRS10A
#undef PRS10B
}

static void
prs10_fc(const struct prs10 *sc, unsigned pi, const char *buf)
{
	unsigned u[2];

	if (2 != sscanf(buf, "%u,%u", &u[0], &u[1]))
		return;
	SetVal(sc->pp[pi], u[0] * 1000.0 + u[1]);
}


static void
prs10_fcs(const struct prs10 *sc, unsigned pi, const char *buf)
{
	unsigned u[4];

	if (4 != sscanf(buf, "%u,%u,%u,%u", &u[0], &u[1], &u[2], &u[3]))
		return;
	SetVal(sc->pp[pi], u[0]);
	SetVal(sc->pp[pi + 1], u[1]);
	SetVal(sc->pp[pi + 2], u[2] * 1000.0 + u[3]);
}

static void
prs10_ds(const struct prs10 *sc, unsigned pi, const char *buf)
{
	int u[2];

	if (2 != sscanf(buf, "%d,%d", &u[0], &u[1]))
		return;
	SetVal(sc->pp[pi], (double)u[0]);
	SetVal(sc->pp[pi + 1], (double)u[1]);
}

static void
prs10_sp(const struct prs10 *sc, unsigned pi, const char *buf)
{
	unsigned u[3];

	if (3 != sscanf(buf, "%u,%u,%u", &u[0], &u[1], &u[2]))
		return;
	SetVal(sc->pp[pi], (double)u[0]);
	SetVal(sc->pp[pi + 1], (double)u[1]);
	SetVal(sc->pp[pi + 2], (double)u[2]);
}

/* Request/Point information -----------------------------------------*/

struct prs10;
typedef void prs10_rx_f(const struct prs10 *sc, unsigned pi, const char *buf);

static struct prs10q {
	const char	*q;
	unsigned	idx;
	enum point_type	pt;
	prs10_rx_f	*func;
} prs10q[] = {
#define PRS10A(a, b, c, d, e) { a, b, c, d },
#define PRS10B(a, b, c, d)
#include "prs10.h"
#undef PRS10A
#undef PRS10B
	{ "ST", 0, UNDEF, prs10_st },
	{ NULL,	0, UNDEF, NULL }
};

/* Serial port listening ---------------------------------------------*/

static int
ask(const struct prs10 *sc, const char *q, const char *r, char *b, unsigned len)
{
	int i;
	unsigned bp;

	sprintf(b, "%s%s\r", q, r);
	if (sc->debug)
		SCDebug(sc, "TX: %V", b);
	i = write(serial_fd(sc->serial), b, strlen(b));
	assert(i == (int)strlen(b));
	bp = 0;
	while (1) {
		b[bp] = '\0';
		i = read(serial_fd(sc->serial), b + bp, len - bp);
		if (i < 0) {
			GLog(sc->gp, "RX: Read error %V: %M", b, errno);
			return (0);
		}
		if (i == 0) {
			GLog(sc->gp, "RX: Timeout %V: %M", b, errno);
			return (0);
		}
		bp += (unsigned)i;
		if (b[bp - 1] == '\r') {
			b[--bp] = '\0';
			if (sc->debug)
				SCDebug(sc, "RX: %V", b);
			return (1);
		}
	} 
}

static void *
prs10_thread(void *foo)
{
	struct prs10 *sc;
	char buf[BUFSIZ];
	struct prs10q *qp;
	int j;
	struct termios t;

	CAST_OBJ_NOTNULL(sc, foo, PRS10_MAGIC);

	AZ(tcgetattr(serial_fd(sc->serial), &t));
	cfmakeraw(&t);
	AZ(cfsetspeed(&t, B9600));
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 50;
	t.c_cflag |= CLOCAL;
	AZ(tcsetattr(serial_fd(sc->serial), TCSANOW, &t));

	qp = &prs10q[0];
	while (1) {
		while (1) {
			usleep(100000);
			pthread_cancelmutex_lock(&sc->lock);
			j = ask(sc, qp->q, "?", buf, sizeof buf);
			pthread_cancelmutex_unlock(&sc->lock);
			if (!j) {
				sleep(1);
				continue;
			}
			if (qp->func)
				qp->func(sc, qp->idx, buf);
			for (qp++; qp->q != NULL && *qp->q == '\0'; qp++)
				continue;
			if (qp->q == NULL) {
				qp = &prs10q[0];
				sleep(1);
			}
		} 
	}
}

/* Control function --------------------------------------------------*/

static void
prs10_control(struct point *p, void *priv1, void *priv2, double v)
{
	struct prs10q *q;
	struct prs10 *sc;
	int i;
	char buf1[BUFSIZ], buf2[BUFSIZ];

	CAST_OBJ_NOTNULL(sc, priv1, PRS10_MAGIC);
	q = priv2;

	AZ(pthread_mutex_lock(&sc->lock));
	if (!strcmp(q->q, "SF"))
		(void)sprintf(buf1, "VB0\r%s %.0f\r%s?",
		    q->q, v, q->q);
	else
		(void)sprintf(buf1, "VB0\r%s %.0f\r%s!\r%s?",
		    q->q, v, q->q, q->q);
	PLog(p, "PRS10 control %g %s <%V>\n", v, q->q, buf1);
	i = ask(sc, buf1, "", buf2, sizeof buf2);
	PLog(p, "PRS10 control Resp %V i %d", buf2, i);
	AZ(pthread_mutex_unlock(&sc->lock));
}

/* ConfigKit routines ------------------------------------------------*/

static void
prs10_init(CK_ARG, struct prs10 *sc)
{
	struct prs10q *q;

	serial_init(CK_PASS, &sc->serial, prs10_thread, sc);
	sc->np = CreatePoints(sc->gp, &sc->pp, prs10_pt);

	for (q = prs10q; q->q != NULL; q++) {
		if (q->pt != AOUT && q->pt != DOUT)
			continue;
		SetPointControl(
		    sc->pp[q->idx], prs10_control, sc, q);
	}
}

static void
prs10_deinit(CK_ARG, struct prs10 *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	free(sc->pp);
}

SerialFunc(prs10, PRS10_MAGIC, serial);
UintVar(prs10, PRS10_MAGIC, debug, debug, 0, );
MODULE_GROUP(prs10, PRS10_MAGIC);
