/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: savs.c,v 1.11 2008/02/17 09:55:54 phk Exp $
 *
 * Receive and decode the serial SAVS2000 data stream
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>

#include "measured.h"
#include "configkit_conf.h"

#define NSAVSPOINT	25

static struct ptdesc savs_pt[NSAVSPOINT + 1] = {
    { .idx = 1,  .point = 1, .type = AIN, .label = "Runway in use landing" },
    { .idx = 2,  .point = 2, .type = AIN, .label = "Runway in use takeoff" },
    { .idx = 3,  .point = 3, .type = AIN, .label = "MET REPORT QNH" },
    { .idx = 10, .point = 10, .type = AIN, .label = "RVR: 22L-END, 04R-TDZ" },
    { .idx = 11, .point = 11, .type = AIN, .label = "RVR: 04R-MID" },
    { .idx = 12, .point = 12, .type = AIN, .label = "RVR: 22L-MID" },
    { .idx = 13, .point = 13, .type = AIN, .label = "RVR: 22L-TDZ, 04R-END" },
    { .idx = 14, .point = 14, .type = AIN, .label = "RVR: 12-END,  30-TDZ" },
    { .idx = 15, .point = 15, .type = AIN, .label = "RVR: 12-MID,  30-MID" },
    { .idx = 16, .point = 16, .type = AIN, .label = "RVR: 12-TDZ,  30-END" },
    { .idx = 17, .point = 17, .type = AIN, .label = "RVR: 22R-END, 04L-TDZ" },
    { .idx = 18, .point = 18, .type = AIN, .label = "RVR: 22R-MID, 04L-MID" },
    { .idx = 19, .point = 19, .type = AIN, .label = "RVR: 22R-TDZ, 04L-END" },
    {.label = NULL, }
};

/* Softc -------------------------------------------------------------*/

struct savs {
	unsigned		magic;
#define SAVS_MAGIC		0x4b892453
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;

	unsigned		debug;
	struct serial		*serial;
	struct point		**pp;
	unsigned		np;
};

static void
savs_record(const struct savs *sc, int offset, const char *ptr, unsigned len)
{
	while (len > 254) {
		Record(GI(sc->gp) + offset, 254, ptr);
		ptr += 254;
		len -= 254;
	}
	Record(GI(sc->gp) + offset, len, ptr);
}

static void
savs_rvr(const struct savs *sc, char *buf, int len)
{
	int i, j, k, l;
	double a[10];

	if (sc->debug & 8)
		SCDebug(sc, "<%V>", buf);
	if (len != 64) {
		savs_record(sc, 3, buf, len);
		return;
	}
	for (i = 0; i < 10; i++) {
		j = sscanf(buf + 2 + 6 * i, "%1d %4d", &k, &l);
		if (j != 2 || k != i) {
			savs_record(sc, 3, buf, len);
			return;
		}
		a[i] = l;
	}
	savs_record(sc, 4, buf, len);
	for (i = 0; i < 10; i++) 
		SetVal(sc->pp[i + 10], a[i]);
}

static struct landmark {
	int 		offset;
	const char	*txt;
} landmark[] = {
	{    4, "RWY LANDING" },
	{   22, "RWY TAKE-OFF" },
	{  739, "QNH" },
	{ 0, NULL }
};

/* XXX: how is "12" and "30" aligned in the field ? */
static double
savs_rwy(const char *p)
{
	if (!memcmp(p, "22R", 3))	return (0.);
	if (!memcmp(p, "22L", 3))	return (1.);
	if (!memcmp(p, "04R", 3))	return (2.);
	if (!memcmp(p, "04L", 3))	return (3.);
	if (!memcmp(p,  "12", 2))	return (4.);
	if (!memcmp(p,  " 12", 3))	return (4.);
	if (!memcmp(p,  "30", 2))	return (5.);
	if (!memcmp(p,  " 30", 3))	return (5.);
	return (NAN);
}


static void
savs_met_report(const struct savs *sc, char *buf, int len)
{
	int i, j, k;
	struct landmark *lm;
	double rwy[2];

	if (sc->debug & 16)
		SCDebug(sc, "%d <%V>\n", len, buf);
	if (len != 1054)
		return;
	for (i = 1; i < 22; i++) {
		if (buf[1 + 50 * i] != '\n') {
			savs_record(sc, 5, buf, len);
			return;
		}
	}
	for (lm = landmark; lm->offset; lm++) {
		if (memcmp(buf + lm->offset, lm->txt, strlen(lm->txt))) {
			SCDebug(sc, "Missed landmark: \"%s\" at %d: \"%.*s\"",
			    lm->txt, lm->offset, strlen(lm->txt), buf + lm->offset);
			savs_record(sc, 5, buf, len);
			return;
		}
	}
	rwy[0] = savs_rwy(buf + 16);
	rwy[1] = savs_rwy(buf + 35);
	j = sscanf(buf + 743, "%04d", &k);
	if (j != 1 || isnan(rwy[0]) || isnan(rwy[1])) {
		savs_record(sc, 5, buf, len);
		return;
	}
	SetVal(sc->pp[1], rwy[0]);
	SetVal(sc->pp[2], rwy[1]);
	SetVal(sc->pp[3], k);
	savs_record(sc, 6, buf, len);
}

static void
savs_rx(const struct savs *sc, char *buf, int len)
{
	int i;
	uint8_t sum, esum;
	char csb[3];


	/* Validate the Checksum */
	sum = 0;
	for (i = 0; i < len - 1; i++)
		sum += buf[i];

	esum = buf[i];

	if (esum != sum) {
		if (sc->debug & 4)
			SCDebug(sc, "%3d BADSUM <%x %x> [%V]", len, sum, esum, buf, csb);
		/* Checksum error */
		savs_record(sc, 1, buf, len);
		return;
	}
	if (sc->debug & 4)
		SCDebug(sc, "%3d [%V]", len, buf, csb);
	if (buf[1] == '7')
		savs_rvr(sc, buf, len);
	else if (buf[1] == '8')
		savs_met_report(sc, buf, len);
	else
		savs_record(sc, 0, buf, len);
}

/* Serial port listening ---------------------------------------------*/

static void *
savs_thread(void *foo)
{
	struct savs *sc;
	struct termios t;
	char buf[8192];
	int i, j, k;

	CAST_OBJ_NOTNULL(sc, foo, SAVS_MAGIC);

	if (isatty(serial_fd(sc->serial))) {
		AZ(tcgetattr(serial_fd(sc->serial), &t));
		cfmakeraw(&t);
		AZ(cfsetspeed(&t, B4800));
		t.c_cc[VMIN] = 0;
		t.c_cc[VTIME] = 50;
		t.c_cflag |= CLOCAL;
		AZ(tcsetattr(serial_fd(sc->serial), TCSANOW, &t));
	}

	j = 0;
	while (1) {
		k = sizeof buf - (1 + j);
		i = serial_read(sc->serial, buf + j, 1, k, 1000);
		if (i == 0)
			continue;
		assert(i > 0);
		buf[i + j] = '\0';
		if (sc->debug & 1)
			SCDebug(sc, "%3d %3d <%V>", j, i, buf + j);
		j += i;
		if (j > 0 && buf[0] != '\002') {
			for (i = 0; i < j; i++)
				if (buf[i] == '\002')
					break;
			savs_record(sc, 1, buf, i);
			if (i < j)
				memmove(buf, buf + i, j - i);
			j -= i;
		}
		if (!j)
			continue;
		for (i = 0; i < j; i++)
			if (buf[i] == '\003')
				break;
		if (i < j - 1) {
			i += 2;
			savs_rx(sc, buf, i);
			if (j > i)
				memmove(buf, buf + i, j - i);
			j -= i;
		} else if (j >= (int)sizeof buf - 2) {
			savs_record(sc, 1, buf, j);
			j = 0;
		} else {
			continue;
		}
	}
}

/* ConfigKit routines ------------------------------------------------*/

static void
savs_init(CK_ARG, struct savs *sc)
{

	serial_init(CK_PASS, &sc->serial, savs_thread, sc);
	sc->np = CreatePoints(sc->gp, &sc->pp, savs_pt);
}

static void
savs_deinit(CK_ARG, struct savs *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	free(sc->pp);
}

SerialFunc(savs, SAVS_MAGIC, serial);
UintVar(savs, SAVS_MAGIC, debug, debug, 0, );
MODULE_GROUP(savs, SAVS_MAGIC);
