/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: soekris.c,v 1.21 2008/02/16 20:10:03 phk Exp $
 *
 * XXX: config for vref
 * XXX: add GPIO in/out
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <configkit.h>
#include <pthread.h>
#include <sys/time.h>
#include <machine/cpufunc.h>
#include "measured.h"
#include "configkit_conf.h"

#define N_SOEKRIS_POINTS	11

struct soekris {
	unsigned		magic;
#define SOEKRIS_MAGIC		0x48014501
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;
	struct point		*p[N_SOEKRIS_POINTS];
	pthread_t		tp;
	double			pollrate;
};

static int devio = -1;

#define VREF	1.235
#define V245	2.45

#define VDIV(a, b)	(((a) + (b)) / (a))
static struct avi {
	const char 	*name;
	double		offset;
	double		scale;
} avi[N_SOEKRIS_POINTS] = {
	{ "VCORE",	0.0,	1.0 },
	{ "VCC",	0.0,	VDIV(10.0, 10.0) },
	{ "VPWR",	0.0,	VDIV(10.0, 191.0) },
	{ "+12V",	0.0,	VDIV(10.0, 38.3) },
	{ "-12V",	3.33,	VDIV(3.32, 60.4) },
	{ "GND",	0.0,	1.0 },
	{ "Vsb",	0.0,	2.0 },
	{ "Vdd",	0.0,	2.0 },
	{ "Vbat",	0.0,	1.0 },
	{ "AVdd",	0.0,	2.0 },
	{ "Temp",	0.0,	1.0 }
};

/*lint -e{732,734} */
static void *
net4801_thread(void *priv)
{
	int i, j, atms, avlm;
	struct avi *avp;
	double a, b;
	double vref = VREF;
	struct soekris *sc;

	CAST_OBJ_NOTNULL(sc, priv, SOEKRIS_MAGIC);
	if (devio < 0)
		devio = open("/dev/io", O_RDWR);
	if (devio < 0) {
		GNotice(sc->gp, "Cannot open /dev/io: %M", errno);
		pthread_exit(NULL);
	}
	outb(0x2e, 0x20);
	i = inb(0x2f);
	if (i != 0xe9) {
		GNotice(sc->gp, "Expected 0xe9 for SuperIO ID, got %x", i);
		pthread_exit(NULL);
	}

	outb(0x2e, 0x7);  outb(0x2f, 0xe);
	outb(0x2e, 0x60); atms = inb(0x2f) << 8;
	outb(0x2e, 0x61); atms |= inb(0x2f);
	outb(0x2e, 0x30); i = inb(0x2f);
	if (!(i & 1)) {
		GNotice(sc->gp, "TMS not enabled (0x%x)", i);
		pthread_exit(NULL);
	}

	outb(0x2e, 0x7);  outb(0x2f, 0xd);
	outb(0x2e, 0x60); avlm = inb(0x2f) * 256;
	outb(0x2e, 0x61); avlm |= inb(0x2f);
	outb(0x2e, 0x30); i = inb(0x2f);
	if (!(i & 1)) {
		GNotice(sc->gp, "VLM not enabled (0x%x)", i);
		pthread_exit(NULL);
	}

	/* initialization */

	/* Magic init sequence from page 208 */
	outb(atms + 0x8, 0x00);
	outb(atms + 0x9, 0x0f);
	outb(atms + 0xa, 0x08);
	outb(atms + 0xb, 0x04);
	outb(atms + 0xc, 0x35);
	outb(atms + 0xd, 0x05);
	outb(atms + 0xe, 0x05);

	outb(atms + 8, 0);
	for(i = 0; i < 3; i++) {
		outb(atms + 9, i);
		outb(atms + 0xa, 1);
	}
	outb(avlm + 7, 0x10);
	outb(avlm + 8, 0);
	for(i = 0; i < 10; i++) {
		outb(avlm + 9, i);
		outb(avlm + 0xa, 1);
	}
	for(; i < 14; i++) {
		outb(avlm + 9, i);
		outb(avlm + 0xa, 0);
	}

	while (1) {

		for(i = 2; i < 3; i++) {
			outb(atms + 9, i);
			/* u = inb(atms + 0xa); ??? */
			j = (char)inb(atms + 0xb);
			SetVal(sc->p[10], (double)j);
		}

#if 0
		/* The offset for the -12V is measured on AD10 */
		i = 10;
		avp = &avi[i];
		outb(avlm + 9, i);
		j = inb(avlm + 0x0b);
		a = vref * V245 * j / 256.0;
		avi[5].offset = a * avp->scale;
#endif

		for(i = 0; i < 10; i++) {
			avp = &avi[i];
			outb(avlm + 9, i + 1);
			j = inb(avlm + 0x0b);
			a = vref * V245 * j / 256.0;
			b = (a - avp->offset) * avp->scale + avp->offset;
			SetVal(sc->p[i], b);
		}
		if (sc->pollrate > 100)
			sleep((int)sc->pollrate);
		else
			usleep((int)(sc->pollrate * 1e6));
	}
}

static void
soekris_init(CK_ARG, struct soekris *sc)
{
	unsigned u;

	for (u = 0; u < N_SOEKRIS_POINTS; u++) {
		sc->p[u] = CreatePoint(sc->gp, u, AIN);
		SetPointLabel(sc->p[u], avi[u].name);
		SetPointFormat(sc->p[u], "%.3f");
		SetPointUnit(sc->p[u], "Volt");
	}
	SetPointFormat(sc->p[10], "%.0f");
	SetPointUnit(sc->p[10], "C");
	sc->pollrate = 10.0;
	AZ(pthread_create(&sc->tp, NULL, net4801_thread, sc));
}

static void
soekris_deinit(CK_ARG, struct soekris *sc)
{
	void *vp;

	AZ(pthread_cancel(sc->tp));
	AZ(pthread_join(sc->tp, &vp));
}

RealVar(soekris, SOEKRIS_MAGIC, pollrate, pollrate, 10.0, "%.3f", );
MODULE_GROUP(soekris, SOEKRIS_MAGIC);
