/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: tamiproto.c,v 1.15 2008/02/16 16:19:10 phk Exp $
 *
 * Receive and decode the serial TAMI2000 data stream
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>
#include <poll.h>

#include <err.h>		/* XXX */

#include "measured.h"
#include "configkit_conf.h"

#define NTAMIPOINT	8
static struct ptdesc tami_pt[NTAMIPOINT + 1] = {
    { .idx = 1,  .point = 1, .type = AIN, .label = "Tower" },
    { .idx = 2,  .point = 2, .type = AIN, .label = "Approach" },
    { .idx = 3,  .point = 3, .type = AIN, .label = "Supervisor" },
    { .idx = 4,  .point = 4, .type = AIN, .label = "Demo" },
    { .idx = 5,  .point = 5, .type = AIN, .label = "Meteo" },
    { .idx = 6,  .point = 6, .type = AIN, .label = "Offline" },
    { .idx = 7,  .point = 7, .type = AIN, .label = "Other" },
    { .idx = 8,  .point = 8, .type = AIN, .label = "Unreachable" },
    {.label = NULL, }
};

#define TAMIPROTO_MAGIC		0x1160c8e8

#define TAMI_SC_FIELDS							\
	unsigned		magic;					\
	struct group		*gp;					\
	struct ck_instance	cki;					\
	pthread_mutex_t		lock;					\
	pthread_t		tp;					\
	time_t			last_count;				\
									\
	char			*hint;					\
	char			*multicast;				\
	char			*netvar;				\
									\
	unsigned		np;					\
	struct point		**pp;					\

#include "../TclTamiProto/tamiproto.c"

static void
tamiproto_count(struct tamiproto *tp)
{
	struct partner *p;
	unsigned n[NTAMIPOINT + 1];
	int i;

	memset(n, 0, sizeof n);
	TAILQ_FOREACH(p, &tp->partners, list) {
		if (p->state == NULL) 				n[8]++;
		else if (!strcmp(p->state, "TOWER"))		n[1]++;
		else if (!strcmp(p->state, "APPROACH"))		n[2]++;
		else if (!strcmp(p->state, "SUP"))		n[3]++;
		else if (!strcmp(p->state, "DEMO"))		n[4]++;
		else if (!strcmp(p->state, "METEO"))		n[5]++;
		else if (!strcmp(p->state, "OFFLINE"))		n[6]++;
		else						n[7]++;
	}
	for (i = 1; i < NTAMIPOINT + 1; i++)
		SetVal(tp->pp[i], n[i]);
	time(&tp->last_count);
}

static void
tproto_update_partner(struct tamiproto *tp , struct partner *p)
{

	if (tp->debug & 0x8)
		printf("StateVar %p %s %s\n", p, p->ip, p->state);
	tamiproto_count(tp);
}

static void
tproto_setnetvar(struct tamiproto *tp, struct var *v, const char *val)
{

	if (tp->debug & 0x4)
		printf("NetVar %p %s %s\n", v, v->var, val);
}

static void *
tamiproto_thread(void *priv)
{
	struct tamiproto *sc;
	struct pollfd fds[1];
	int i;
	time_t t;

	CAST_OBJ_NOTNULL(sc, priv, TAMIPROTO_MAGIC);
	while (1) {
		fds[0].fd = sc->sock;
		fds[0].events = POLLIN;
		i = poll(fds, 1, 1000);
		if (i == 1)
			tproto_read(sc);
		(void)tproto_poll(sc);

		time(&t);
		if (sc->last_count + 10 < t)
			tamiproto_count(sc);
	}
}

static void
tamiproto_cfg_netvar(struct tamiproto *tp)
{

	if (tp->netvar != NULL)
		tproto_addvar(tp, tp->netvar);
}

static void
tamiproto_cfg_hint(struct tamiproto *tp)
{

	if (tp->hint != NULL)
		tproto_addhint(tp, tp->hint);
}

static void
tamiproto_cfg_multicast(struct tamiproto *tp)
{

	if (tp->multicast != NULL)
		tproto_addmulticast(tp, tp->multicast);
}

static void
tamiproto_cfg_mystate(struct tamiproto *tp)
{

	tproto_setmystate(tp, tp->mystate);
}

/* ConfigKit routines ------------------------------------------------*/

static void
tamiproto_init(CK_ARG, struct tamiproto *sc)
{

	sc->np = CreatePoints(sc->gp, &sc->pp, tami_pt);
	tproto_init(sc);
	AZ(pthread_create(&sc->tp, NULL, tamiproto_thread, sc));

	/* To keep FlexeLint (somewhat) happy */
	if (sc->gp == NULL) 
		tproto_setvar(sc, "foo", "bar");
}

static void
tamiproto_deinit(CK_ARG, struct tamiproto *sc)
{
	void *vp;

	AZ(pthread_cancel(sc->tp));
	AZ(pthread_join(sc->tp, &vp));
}

WordVar(tamiproto, TAMIPROTO_MAGIC, secret, secret, );
WordVar(tamiproto, TAMIPROTO_MAGIC, mystate, mystate,
    tamiproto_cfg_mystate(sc));
WordVar(tamiproto, TAMIPROTO_MAGIC, domain, domain, );
WordVar(tamiproto, TAMIPROTO_MAGIC, port, port, tproto_setport(sc););
WordVar(tamiproto, TAMIPROTO_MAGIC, hint, hint, tamiproto_cfg_hint(sc););
WordVar(tamiproto, TAMIPROTO_MAGIC, netvar, netvar, tamiproto_cfg_netvar(sc););
WordVar(tamiproto, TAMIPROTO_MAGIC, multicast, multicast,
	tamiproto_cfg_multicast(sc););
UintVar(tamiproto, TAMIPROTO_MAGIC, debug, debug, 0, );
MODULE_GROUP(tamiproto, TAMIPROTO_MAGIC);
