/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: vaillant.c,v 1.23 2008/02/16 15:47:08 phk Exp $
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <pthread.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>

#include "measured.h"
#include "configkit_conf.h"

struct vaillant {
	unsigned		magic;
#define VAILLANT_MAGIC 0x9328193
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	struct serial		*serial;
	unsigned		debug;
	struct point		*point[256];
};

static void f_temp(const u_char *buf, double *v);
static void f_word(const u_char *buf, double *v);
static void f_byte(const u_char *buf, double *v);
static void f_none(const u_char *buf, double *v);
static void f_a3(const u_char *buf, double *v);

#define FILL 

static struct datum {
	int			len;
	int			param;
	void			(*func)(const u_char *, double *);
	const char		*desc;
} datum[256] = {
	[0x6c] = { 1,  0, f_byte, "Dellast", FILL },
	[0x64] = { 1,  1, f_byte, "Pumpeefterlb varme", FILL },
	[0x21] = { 1,  2, f_byte, "Max spretid varme", FILL },
	[0x16] = { 3,  3, f_temp, "Solfanger temp", FILL },
	[0x17] = { 3,  4, f_temp, "Beholder temp [m]", FILL },
	[0x19] = { 2,  5, f_temp, "Fremlb temp [b]", FILL },
	[0x01] = { 2,  6, f_temp, "Udlb temp [b]", FILL },
	[0x04] = { 2,  7, f_temp, "Beholder temp [b]", FILL },
	[0x0e] = { 1,  8, f_byte, "Klemme 3-4[b]", FILL },
	[0x25] = { 2,  9, f_temp, "Rumtermostat", FILL },
	[0x44] = { 1, 10, f_byte, "Intern pumpe", FILL },
	[0x3f] = { 1, 11, f_byte, "Extern pumpe", FILL },
	[0x9e] = { 1, 12, f_byte, "Beholder pumpe", FILL },
	[0xaf] = { 1, 13, f_byte, "Circulations pumpe", FILL },
	[0xa1] = { 1, 14, f_byte, "Omdrejningstal [b]", FILL },
	[0x73] = { 2, 15, f_word, "Omdrejningstal [m]", FILL },
	[0xa0] = { 1, 16, f_byte, "2. pumpe", FILL },
	[0xb3] = { 1, 17, f_byte, "Reguleringsart", FILL },
	[0x58] = { 1, 22, f_byte, "Varmtvandsbehov", FILL },
	[0x08] = { 1, 23, f_byte, "Sommer/Vinter", FILL },
	[0xa3] = { 4, 24, f_a3,   "Lufttryk Pa [m]", FILL },
	[0x71] = { 1, 25, f_byte, "Varmtvandsfrigivelse", FILL },
	[0x24] = { 2, 33, f_word, "Blser [b]", FILL },
	[0x83] = { 2, 34, f_word, "Blser [m]", FILL },
	[0x42] = { 1, 35, f_byte, "3-vejsventil", FILL },
	[0x18] = { 3, 40, f_temp, "Fremlb [m]", FILL },
	[0x98] = { 5, 41, f_temp, "Returlb [m]", FILL },
	[0xa4] = { 2, 44, f_word, "Ioniseringsstrm [m]", FILL },
	[0x65] = { 1, 46, f_byte, "Udetemp korr.", FILL },
	[0x6a] = { 3, 47, f_temp, "Udetemp [m].", FILL },
	[0xa7] = { 1, 50, f_byte, "min rpm offset", FILL },
	[0xa8] = { 1, 51, f_byte, "max rpm offset", FILL },
	[0xad] = { 1, 52, f_byte, "min lufttr. offset", FILL },
	[0xae] = { 1, 53, f_byte, "max lufttr. offset", FILL },
	[0x20] = { 1, 60, f_byte, "antal overkog udkobling", FILL },
	[0x1f] = { 1, 61, f_byte, "antal GFA udkobling", FILL },
	[0x38] = { 1, 67, f_byte, "Restsprretid", FILL },
	[0x6e] = { 1, 68, f_byte, "antal ingen 1. start", FILL },
	[0x6f] = { 1, 69, f_byte, "antal ingen 2. start", FILL },
	[0xa5] = { 1, 71, f_byte, "Max fremlb varme", FILL },
	[0x63] = { 1, 72, f_byte, "Pumpeefterlb varmt vand", FILL },
	[0x9d] = { 1, 73, f_byte, "varmtvand temp-hysterese", FILL },
	[0x66] = { 1, 75, f_byte, "max opvarmnings tid varmtvand", FILL },
	[0x9a] = { 5, 76, f_byte, "kedel variant", FILL },
	[0xa9] = { 1, 77, f_byte, "dellast varmtvand", FILL },
	[0xa6] = { 1, 78, f_byte, "Max fremlb varmtvand", FILL },
	[0x28] = { 2, 80, f_word, "Brnderstarter varmeanlg", FILL },
	[0x29] = { 2, 82, f_word, "Driftstimer varmeanlg", FILL },
	[0x22] = { 2, 81, f_word, "Brnderstarter varmtvand", FILL },
	[0x23] = { 2, 83, f_word, "Driftstimer varmtvand", FILL },
	[0xac] = { 2, 84, f_word, "Timer til service", FILL },
	[0x74] = { 1, 90, f_word, "Digital styring", FILL },
	[0x69] = { 1, 91, f_word, "DCF77 status", FILL },
};

static void
mysend(const struct vaillant *sc, const u_char *buf, int len)
{
	u_char sbuf[100];
	int i, j, k;

	j = 0;
	k = sbuf[j++] = len + 2;
	for (i = 0; i < len; i++) {
		k += k;
		k ^= buf[i];
		if (k > 0xff)
			k ^= 0x119;
		sbuf[j++] = buf[i];
	}
	sbuf[j++] = k;
	i = write(serial_fd(sc->serial), sbuf, j);
	if (i != j)
		serial_reopen(NULL, sc->serial);
}

static int
myrecv(const struct vaillant *sc, u_char *buf, int len)
{
	int i, j, k, l;

	k = 5;
	for (j = 0; j < len; ) {
		if (j > 0)
			l = buf[0];
		else
			l = 5;
		if (l > 64) {
			SCDebug(sc, "Rx: abandon %H", buf, j);
			j = 0;
			l = 5;
		}
		l -= j;
		if (l <= 0)
			break;
		i = serial_read(sc->serial, buf + j, 1, l, 1000);
		if (i < 0) {
			serial_reopen(NULL, sc->serial);
			return (-1);
		}
		if (i == 0) {
			SCDebug(sc, "Rx: timeout %H", buf, j);
			return (-1);
		}
		if (i > 0)
			j += i;
		if (j > 1)
			len = buf[0];
		k = len - j;
	}
	k = 0;
	for (i = 0; i < j - 1; i++) {
		k += k;
		k ^= buf[i];
		if (k > 0xff)
			k ^= 0x119;
	}
	if (k != buf[i]) {
		SCDebug(sc, "Rx: bad checksum %H", buf, j);
		return (-2);
	}
	if (sc->debug)
		SCDebug(sc, "Rx: good %H", buf, j);
	return (j);
}

static void
f_temp(const u_char *buf, double *v)
{

	*v = buf[1] * 16.0 + buf[2] / 16.0;
}

static void
f_byte(const u_char *buf, double *v)
{

	*v = buf[1];
}

static void
f_word(const u_char *buf, double *v)
{

	*v =  buf[1] * 256 + buf[2];
}

static void
f_a3(const u_char *buf, double *v)
{
	*v = (buf[1] * 256 + buf[2]) / 16.0;
}

static void
f_none(const u_char *buf __unused, double *v)
{

	*v = 0;
}

static void *
vaillant_thread(void *priv)
{
	struct termios t;
	struct vaillant *sc;
	int i, j;
	u_char buf1[200];
	u_char buf2[200];
	struct datum *d;
	double v;


	sc = priv;

	if (isatty(serial_fd(sc->serial))) {
		AZ(tcgetattr(serial_fd(sc->serial), &t));
		cfmakeraw(&t);
		t.c_cflag |= CLOCAL;
		AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));
	}

	for(j = 0;;j = (j + 1) % 256) {
		d = &datum[j];
		if (d->func == NULL)
			continue;
		buf2[0] = 0x00;
		buf2[1] = 0x00;
		buf2[2] = 0x00;
		buf2[3] = j;
		buf2[4] = 1;
		mysend(sc, buf2, 5);
		i = myrecv(sc, buf1, sizeof buf1);
		if (i < 0)
			continue;
	
		if (d->func == f_none) {
			if (buf1[0] == 3 && buf1[1] == 3 && buf1[2] == 5)
				d->func = NULL;
			else if (buf1[0] == 4) 
				d->func = f_byte;
			else if (buf1[0] == 5)
				d->func = f_word;
			else if (buf1[0] == 6)
				d->func = f_a3;
			else {
				SCDebug(sc, "%02x(%d) has no function %H",
				    j, buf1[0], buf1, buf1[0]);
				d->func = NULL;
			}
		}
		if (d->func == NULL)
			continue;

		d->func(buf1 + 1, &v);

		if (sc->point[j] == NULL) {
			sc->point[j] = CreatePoint(sc->gp, j, AIN);
			SetPointLabel(sc->point[j], d->desc);
		}
		SetVal(sc->point[j], v);
	}
}

static void
vaillant_init(CK_ARG, struct vaillant *sc)
{
	int i;
	char *p;

	serial_init(CK_PASS, &sc->serial, vaillant_thread, sc);

	for (i = 0; i < 256; i++) {
		if (datum[i].len == 0) {
			datum[i].param = -1;
			datum[i].func = f_none;
			asprintf(&p, "Hex %02x", i);
			datum[i].desc = p;
		}
	}
}

static void
vaillant_deinit(CK_ARG, struct vaillant *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
}

SerialFunc(vaillant, VAILLANT_MAGIC, serial);
UintVar(vaillant, VAILLANT_MAGIC, debug, debug, 0, );
MODULE_GROUP(vaillant, VAILLANT_MAGIC);
