/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: vindpanel.c,v 1.13 2008/02/17 09:55:54 phk Exp $
 *
 * Receive and decode the serial SAVS2000->VindPanel data stream.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <stdarg.h>
#include <configkit.h>
#include <termios.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>

#include "measured.h"
#include "configkit_conf.h"

#define NVINDPANELPOINT	25

static struct ptdesc vindpanel_pt[NVINDPANELPOINT + 1] = {
    { .point = 10, .type = AIN, .label = "Vindretning, SW" },
    { .point = 11, .type = AIN, .label = "Vindhastiged, SW" },
    { .point = 12, .type = AIN, .label = "Max retningsvariation, SW" },
    { .point = 13, .type = AIN, .label = "Min retningsvariation, SW" },
    { .point = 14, .type = AIN, .label = "Max hastighedsvariation, SW" },
    { .point = 15, .type = AIN, .label = "Min hastighedsvariation, SW" },
    { .point = 20, .type = AIN, .label = "Vindretning, NE" },
    { .point = 21, .type = AIN, .label = "Vindhastiged, NE" },
    { .point = 22, .type = AIN, .label = "Max retningsvariation, NE" },
    { .point = 23, .type = AIN, .label = "Min retningsvariation, NE" },
    { .point = 24, .type = AIN, .label = "Max hastighedsvariation, NE" },
    { .point = 25, .type = AIN, .label = "Min hastighedsvariation, NE" },
    { .point = 30, .type = AIN, .label = "X-wind LN, 04L" },
    { .point = 31, .type = AIN, .label = "X-wind LN, 04R" },
    { .point = 32, .type = AIN, .label = "X-wind LN, 22L" },
    { .point = 33, .type = AIN, .label = "X-wind LN, 22R" },
    { .point = 34, .type = AIN, .label = "X-wind LN, 12" },
    { .point = 35, .type = AIN, .label = "X-wind LN, 30" },
    { .point = 40, .type = AIN, .label = "X-wind TO, 04L" },
    { .point = 41, .type = AIN, .label = "X-wind TO, 04R" },
    { .point = 42, .type = AIN, .label = "X-wind TO, 22L" },
    { .point = 43, .type = AIN, .label = "X-wind TO, 22R" },
    { .point = 44, .type = AIN, .label = "X-wind TO, 12" },
    { .point = 45, .type = AIN, .label = "X-wind TO, 30" },
    { .point = 50, .type = AIN, .label = "QNH" },
    {.label = NULL, }
};

/* Softc -------------------------------------------------------------*/

struct vindpanel {
	unsigned		magic;
#define VINDPANEL_MAGIC		0x52d4532d
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;

	unsigned		debug;
	unsigned		address;
	char			*version;
	unsigned		dip;
	struct serial		*serial;
	struct point		**pp;
	unsigned		np;
};

static void
vindpanel_tx(const struct vindpanel *sc, const char *fmt, ...) 
{
	va_list ap;
	char buf[64];
	int i, j, sum;

	buf[0] = '\002';
	va_start(ap, fmt);
	j = sizeof buf - 4; /* STX + CSUM + ETX */
	i = vsnprintf(buf + 1, j, fmt, ap);
	assert(i <= j);
	va_end(ap);
	sum = 0;
	for (i = 1; buf[i] != '\0'; i++)
		sum += buf[i];
	sprintf(buf + i, "%02X\003", sum & 0xff);
	if (sc->debug & 2)
		SCDebug(sc, "TX: <%V>", buf);
	i = strlen(buf);
	Record(GI(sc->gp) + 5, i, buf);
	j = write(serial_fd(sc->serial), buf, i);
	assert(i == j);	/* XXX */
}

static int vindpanel_V_field[NVINDPANELPOINT] = {
	3, 2, 3, 3, 2, 2,	/* SW wind */
	3, 2, 3, 3, 2, 2,	/* NW wind */
	2, 2, 2, 2, 2, 2,	/* X-wind Takeoff */
	2, 2, 2, 2, 2, 2,	/* X-wind Landing */
	4			/* QNH */
};

static void
vindpanel_rx(const struct vindpanel *sc, char *buf, int len)
{
	int i, j, f, sum;
	unsigned a;
	char csb[3], *p;
	double v[NVINDPANELPOINT];

	if (sc->debug & 4)
		SCDebug(sc, "%3d [%V]", len, buf, csb);

	/* Validate the Checksum */
	sum = 0;
	for (i = 1; i < len - 3; i++)
		sum += buf[i];
	sprintf(csb, "%02X", sum & 0xff);

	if (buf[i] != csb[0] || buf[i + 1] != csb[1]) {
		/* Checksum error */
		Record(GI(sc->gp) + 1, len, buf);
		return;
	}

	/* Figure out the address */
	if (buf[1] >= '0' && buf[1] <= '9')
		a = buf[1] - '0';
	else
		a = buf[1] + 10 - 'A';

	/* Broadcast or for us ? */
	if (a != 0 && a != sc->address) {
		/* Not for us */
		Record(GI(sc->gp) + 2, len, buf);
		return;
	}

	/* Transpond ? */
	if (buf[2] == 'T') {
		Record(GI(sc->gp) + 0, len, buf);
		if (a)
			vindpanel_tx(sc, "%cACK%c0", buf[1], buf[1]);
		return;
	}

	/* Version check, directed at us ? */
	if (buf[2] == 'M' && buf[3] == 'V') {
		Record(GI(sc->gp) + 0, len, buf);
		if (a)
			vindpanel_tx(sc, "%cDV%s", buf[1],
			    sc->version != NULL ? sc->version :
			    "$Id: vindpanel.c,v 1.13 2008/02/17 09:55:54 phk Exp $");
		return;
	}
	/* DIP Switch check */
	if (buf[2] == 'M' && buf[3] == 'S') {
		Record(GI(sc->gp) + 0, len, buf);
		if (a)
			vindpanel_tx(sc, "%cDS%d", buf[1], sc->dip);
		return;
	}
	/* Wind record */
	if (buf[2] == 'V') {
		p = buf + 2;
		for (f = 0; f < NVINDPANELPOINT; f++) {
			v[f] = 0;
			if (f > 0 && *p != ' ') {
				Record(GI(sc->gp) + 4, len, buf);
				return;
			}
			p++;
			for (j = 0; j < vindpanel_V_field[f]; j++) 
				if (p[j] != ' ')
					break;
			if (j == vindpanel_V_field[f]) {
				v[f]  = NAN;
			} else {
				for (; j < vindpanel_V_field[f]; j++)  {
					if (!isdigit(p[j])) {
						Record(GI(sc->gp) + 4,
						    len, buf);
						return;
					}
					v[f] *= 10;
					v[f] += (p[j] - '0');
				}
			}
			p += vindpanel_V_field[f];
		}
		Record(GI(sc->gp) + 0, len, buf);
		for (f = 0; f < NVINDPANELPOINT; f++) 
			SetVal(sc->pp[f], v[f]);
		if (a)
			vindpanel_tx(sc, "%cACK%c0", buf[1], buf[1]);
		return;
	}
	Record(GI(sc->gp) + 3, len, buf);
}

/* Serial port listening ---------------------------------------------*/

#define BUFLEN		100
static void *
vindpanel_thread(void *foo)
{
	struct vindpanel *sc;
	struct termios t;
	char buf[BUFLEN + 1];
	int i, j;

	CAST_OBJ_NOTNULL(sc, foo, VINDPANEL_MAGIC);

	if (isatty(serial_fd(sc->serial))) {
		AZ(tcgetattr(serial_fd(sc->serial), &t)); cfmakeraw(&t);
		AZ(cfsetspeed(&t, B2400));
		t.c_cc[VMIN] = 0;
		t.c_cc[VTIME] = 50;
		t.c_cflag |= CLOCAL;
		AZ(tcsetattr(serial_fd(sc->serial), TCSANOW, &t));
	}

	j = 0;
	while (1) {
		i = serial_read(sc->serial, buf + j, 1, BUFLEN - j, 1000);
		if (i == 0)
			continue;
		assert(i > 0);
		buf[i + j] = '\0';
		if (sc->debug & 1)
			SCDebug(sc, "%3d %3d <%V>", j, i, buf + j);
		j += i;
		if (j > 0 && buf[0] != '\002') {
			for (i = 0; i < j; i++)
				if (buf[i] == '\002')
					break;
			Record(GI(sc->gp) + 1, i, buf);
			if (i < j)
				memmove(buf, buf + i, j - i);
			j -= i;
		}
		if (!j)
			continue;
		for (i = 0; i < j; i++)
			if (buf[i] == '\003')
				break;
		if (i < j) {
			vindpanel_rx(sc, buf, j);
			i++;
			if (j > i)
				memmove(buf, buf + i, j - i);
			j -= i;
		} else if (j >= BUFLEN) {
			Record(GI(sc->gp) + 1, j, buf);
			j = 0;
		} else {
			continue;
		}
	}
}

/* ConfigKit routines ------------------------------------------------*/

static void
vindpanel_init(CK_ARG, struct vindpanel *sc)
{

	serial_init(CK_PASS, &sc->serial, vindpanel_thread, sc);
	sc->np = CreatePoints(sc->gp, &sc->pp, vindpanel_pt);
}

static void
vindpanel_deinit(CK_ARG, struct vindpanel *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	free(sc->pp);
}

SerialFunc(vindpanel, VINDPANEL_MAGIC, serial);
UintVar(vindpanel, VINDPANEL_MAGIC, debug, debug, 0, );
UintVar(vindpanel, VINDPANEL_MAGIC, address, address, 0, );
UintVar(vindpanel, VINDPANEL_MAGIC, dip, dip, 0, );
WordVar(vindpanel, VINDPANEL_MAGIC, version, version,  );
MODULE_GROUP(vindpanel, VINDPANEL_MAGIC);
