/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: vor4000.c,v 1.38 2008/02/17 09:55:54 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <termios.h>
#include <pthread.h>
#include <configkit.h>
#include "measured.h"
#include "configkit_conf.h"

#include "vor4000_proto.h"

#define POLL_COUNTDOWN		4	/*
					 * How many times to send command in
					 * total.  Should be even so both
					 * Xmitters get the same count.
					 */
#define POLL_TIMEOUT		3	/* 1/10ths of second */

#define LAST_POINT		223	/* Highest Point must match reality */

struct vor4000 {
	unsigned		magic;
#define VOR4000_MAGIC		0x13737443
	struct group		*gp;
	struct ck_instance	cki;
	pthread_mutex_t		lock;

	uint			isremote;
	uint			debug;
	double			azimuth1;
	double			azimuth2;

	struct serial		*serial;
	struct serial		*serial2;

	struct point		*pp[LAST_POINT + 1];	/* XXX: wasteful */

	/* Protocol buffers */
	u_char			ibuf[MSG_MAXLEN];
	u_int			iptr;
	u_char			obuf[MSG_MAXLEN];
	u_int			optr;

	/* Protocol state */
	u_char			tx_command;
	u_char			tx_command_countdown;
	const char		*tx_state;
	u_char			last_c2[2];
	u_int			last_s2;
	u_int			s2_state;
};

#define MON_OFFSET(foo)	((foo) ? 200 : 100)


/* Points in the "main" set (000...099) ------------------------------*/

struct pointspec {
	enum point_type		type;
	u_char			msg;
	u_char			byte;
	u_char			bit;
	double			scale;
	const char		*label;
};

#define PT_DOUT(CMD, DESC) \
	DOUT, CMD | C1B0_SWITCH_CONTROL, 0, 0, 0.0, DESC

#define PT_DIN(MSG, BYTE, NAME) \
	DIN, MSG, BYTE, S##MSG##B##BYTE##_##NAME, 0.0, #NAME

#define PT_AIN(MSG, MEAS, SPEC, DESC) \
	AIN, MSG, PARAM_##MEAS, SPEC, SCALE_##MEAS, DESC

static struct pointspec vor_points[] = {
	{ PT_DOUT(C1B1_TX1_ON_MAIN, "Switch to TX1") },
	{ PT_DOUT(C1B1_TX2_ON_MAIN, "Switch to TX2") },
	{ PT_DOUT(C1B1_TX1_TX2_OFF, "Switch TX1 and TX2 off") },
	{ PT_DIN(1, 0, REMOTE) },
	{ PT_DIN(1, 2, MODULATOR_ON) },
	{ PT_DIN(1, 4, 5V_TX1) },
	{ PT_DIN(1, 4, 5V_TX2) },
	{ PT_DIN(1, 4, BCPS_FAULT) },
	{ PT_DIN(1, 4, ACK_EGB) },
	{ PT_DIN(1, 4, REM_EN_DIS) },
	{ PT_DIN(1, 5, ALARM_SYS_OFF) },
	{ PT_DIN(1, 5, SHUT_MAIN_TX) },
	{ PT_DIN(1, 5, BITE_ALARM) },
	{ PT_DIN(1, 5, MON_BYPASS_OR) },
	{ PT_DIN(1, 5, MON_FAULT) },
	{ PT_DIN(1, 5, TX1_MAIN) },
	{ PT_DIN(1, 5, MAINS_OFF) },
	{ PT_DIN(1, 5, WARNING) },

	{ UNDEF, 0, 0, 0, 0.0, NULL }
};

/* Points in the "per monitor" set (100...199, 200...299) ------------*/

static struct pointspec mon_points[] = {
	{ PT_AIN(2, AZIMUTH,	0, "Azimuth") },
	{ PT_AIN(2, AZIMUTH,	1, "Azimuth error") },
	{ PT_AIN(2, RF_LEVEL,	0, "RF Level") },
	{ PT_AIN(2, 30HZ_MOD,	0, "30Hz Modulation") },
	{ PT_AIN(2, 9960HZ_MOD,	0, "9960Hz Modulation") },
	{ PT_AIN(2, 30HZ_FM_DEV,0, "30Hz FM Deviation") },

	{ PT_DIN(1, 1, BITE_WARNING) },
	{ PT_DIN(1, 1, OPERATION_OFF) },
	{ PT_DIN(1, 1, MONITOR_WARNING) },
	{ PT_DIN(1, 1, MONITOR_FAULT) },
	{ PT_DIN(1, 1, COMONITOR_FAULT) },
	{ PT_DIN(1, 2, OPER_ALARM_2ND_ORDER) },
	{ PT_DIN(1, 2, TX_ON_ANTENNA) },
	{ PT_DIN(1, 2, OPER_ALARM_1ST_ORDER) },
	{ PT_DIN(1, 3, AZIMUTH) },
	{ PT_DIN(1, 3, 30HZ_MOD) },
	{ PT_DIN(1, 3, 9960HZ_MOD) },
	{ PT_DIN(1, 3, FREQ) },
	{ PT_DIN(1, 3, IDENT) },
	{ PT_DIN(1, 3, RF_LEVEL) },
	{ PT_DIN(1, 3, 30HZ_FM) },
	{ PT_DIN(1, 3, RF_PHASE) },
	{ PT_DIN(1, 4, MON_ALARM) },
	{ PT_DIN(1, 4, MON_BYPASS) },

	{ UNDEF, 0, 0, 0, 0.0, NULL }
};

#undef PT


/* Transmit functions ------------------------------------------------*/

static void
cx_msg(struct vor4000 *sc __unused)
{

	sc->s2_state = 0;
}

static void
c2_msg(struct vor4000 *sc __unused)
{

	sc->last_c2[0] = sc->obuf[0];
	sc->last_c2[1] = sc->obuf[1];
	sc->s2_state++;
}


/* Receive functions -------------------------------------------------*/

static void
s0_msg(struct vor4000 *sc __unused)
{

	sc->s2_state = 0;
}

static void
s1_msg(struct vor4000 *sc)
{
	int off;
	struct point *pp;
	struct pointspec *ps;

	sc->s2_state = 0;
	sc->isremote = sc->ibuf[4] & S1B4_REM_EN_DIS;
	for (ps = vor_points; ps->type != UNDEF; ps++) {
		if (ps->msg != 1)
			continue;
		pp = sc->pp[ps - vor_points];
		SetValBool(pp, sc->ibuf[ps->byte] & ps->bit);
	}
	off = MON_OFFSET(sc->ibuf[0] & S1B0_DIFF_MON);
	for (ps = mon_points; ps->type != UNDEF; ps++) {
		if (ps->msg != 1)
			continue;
		pp = sc->pp[(ps - mon_points) + off];
		SetValBool(pp, sc->ibuf[ps->byte] & ps->bit);
	}
}

static void
s2_msg(struct vor4000 *sc)
{
	u_int s2;
	double val;
	int off;
	struct pointspec *ps;

	s2 = (sc->last_c2[0] << 8) | sc->last_c2[1];
	if (sc->s2_state == 1) {
		sc->last_s2 = s2;
		sc->s2_state++;
		return;
	}
	if (sc->s2_state != 3) {
		sc->s2_state = 0;
		return;
	}
	sc->s2_state = 0;
	if (s2 != sc->last_s2)
		return;

	off = MON_OFFSET(sc->last_c2[0] & C2B0_DIFF_MON);
	
	val = sc->ibuf[1] + sc->ibuf[2] * 256;
	for (ps = mon_points; ps->type != UNDEF; ps++) {
		if (ps->msg != 2)
			continue;
		if (ps->byte == sc->last_c2[1] && ps->bit == 0)
			break;
	}

	assert(ps->type == AIN);

	val *= ps->scale;
	SetVal(sc->pp[off + (ps - mon_points)], val);

	if (sc->last_c2[1] != PARAM_AZIMUTH)
		return;

	/* Find & update point for Azimuth deviation and update */
	for (ps = mon_points; ps->type != UNDEF; ps++) {
		if (ps->msg != 2)
			continue;
		if (ps->byte == sc->last_c2[1] && ps->bit == 1)
			break;
	}
	assert(ps->type == AIN);
	if (sc->last_c2[0] & C2B0_DIFF_MON)
		val -= sc->azimuth2;
	else
		val -= sc->azimuth1;
	if (val > 180.0)
		val -= 360.0;
	if (val < -180.0)
		val += 360.0;
	SetVal(sc->pp[off + (ps - mon_points)], val);
}


/* Message validation and dispatch -----------------------------------*/

typedef void msg_func_t(struct vor4000 *);

static const struct vor4000_msg {
	u_char		kind;
	u_int		type;
	u_int		len;
	msg_func_t	*func;
} vor4000_msg[] = {
	{ 1, C0B0_TYPE, C0_LEN, cx_msg },
	{ 1, C1B0_TYPE, C1_LEN, cx_msg },
	{ 1, C2B0_TYPE, C2_LEN, c2_msg },
	{ 1, C3B0_TYPE, C3_LEN, cx_msg },
	{ 0, S0B0_TYPE, C0_LEN, s0_msg },
	{ 0, S1B0_TYPE, S1_LEN, s1_msg },
	{ 0, S2B0_TYPE, S2_LEN, s2_msg },
	{ 0, 0,         1,      NULL   }
};

static unsigned char
msgxor(const u_char *p, u_int l)
{
	u_char x = 0;

	while (l--)
		x ^= *p++;
	return (x);
}

static void
attempt_msg(struct vor4000 *sc, u_int ctrl, u_char *p, u_int *lp)
{
	const struct vor4000_msg *mp;

	while (*lp > 0) {
		for(mp = vor4000_msg; mp->func != NULL; mp++)
			if (ctrl == mp->kind &&
			    (p[0] & TYPE_MASK) == mp->type)
				break;
		if (*lp < mp->len)
			return;
		if (sc->debug)
			SCDebug(sc, "attempt_msg %d <%H>",
			    mp->kind, p, mp->len);
		if (mp->func == NULL || msgxor(p, mp->len)) {
			GLog(sc->gp, "msg/xor error");
			Record(GI(sc->gp) + 0 + ctrl, mp->len, p);
		} else {
			Record(GI(sc->gp) + 2 + ctrl, mp->len, p);
			mp->func(sc);
		}
		memcpy(p, p + mp->len, *lp - mp->len);
		*lp -= mp->len;
	}
}


/* Serial port -------------------------------------------------------*/

static void
vor4000_stty(int fd)
{
	struct termios t;

	AZ(tcgetattr(fd, &t));
	cfmakeraw(&t);
	AZ(cfsetspeed(&t, B600));
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = POLL_TIMEOUT;
	t.c_cflag |= CLOCAL;
	t.c_cflag &= ~CRTSCTS;
	t.c_cflag &= ~MDMBUF;
	t.c_cflag &= ~HUPCL;
	t.c_cflag |= CREAD;
	t.c_cflag |= CS8;
	t.c_cflag |= CSTOPB;
	t.c_cflag |= PARENB;
	AZ(tcsetattr(fd, TCSANOW, &t));
}

/*
 * Define polling sequence:
 *	Lower-case == mon1, Upper-case == mon2
 *	'a' - Digital status
 *	'b' - Azimuth
 *	'c' - RF level
 *	'd' - 30 Hz modulation
 *	'e' - 9960Hz modulation
 *	'f' - 30 Hz FM deviation
 */

static const char *vor4000poll = "abbCCDDABBeeffabbccddABBEEFF";

static void
vor4000_tx_poll(struct vor4000 *sc)
{
	struct point *pp;
	struct pointspec *ps;
	int j;

	for (ps = vor_points; ps->type != UNDEF; ps++) {
		if (ps->type != DOUT)
			continue;
		pp = sc->pp[ps - vor_points];
		SetValBool(pp, sc->tx_command == ps->msg);
	}

	/*
	 * We can only ask about analogue values when the VOR is in remote
	 * mode, otherwise we block the front panel.
	 */
	while (1) {
		if (sc->tx_state == NULL || *sc->tx_state == '\0')
			sc->tx_state = vor4000poll;
		/*
		 * Restrict our selves to 'a' when we don't have the panel
		 * or when we try to control the transmitter
		 */
		if ((sc->isremote && sc->tx_command == 0)
		    || 'a' == tolower(*sc->tx_state))
			break;
		sc->tx_state++;
	}

	sc->obuf[0] = C2B0_TYPE | (isupper(*sc->tx_state) ? C2B0_DIFF_MON : 0);
	sc->optr = C2_LEN;
	switch(tolower(*sc->tx_state)) {
	case  'a':
		sc->obuf[0] = C1B0_TYPE;
		sc->obuf[0] |= (isupper(*sc->tx_state) ? C1B0_DIFF_MON : 0);
		if (sc->tx_command_countdown > 0)
			sc->tx_command_countdown--;
		else
			sc->tx_command = 0;
		sc->obuf[1] = sc->tx_command;
		sc->optr = C1_LEN;
		break;
	case  'b': sc->obuf[1] = PARAM_AZIMUTH; break;
	case  'c': sc->obuf[1] = PARAM_RF_LEVEL; break;
	case  'd': sc->obuf[1] = PARAM_30HZ_MOD; break;
	case  'e': sc->obuf[1] = PARAM_9960HZ_MOD; break;
	case  'f': sc->obuf[1] = PARAM_30HZ_FM_DEV; break;
	default:
		assert(tolower(*sc->tx_state) >= 'a');
		assert(tolower(*sc->tx_state) <= 'f');
	}
	sc->obuf[sc->optr - 1] = msgxor(sc->obuf, sc->optr - 1);
	if (sc->debug)
		SCDebug(sc, "Poll: %H", sc->obuf, sc->optr);
	j = TIOCM_RTS;
	j = TIOCM_RTS | TIOCM_DTR;
	AZ(ioctl(serial_fd(sc->serial), TIOCMBIS, &j));
	(void)write(serial_fd(sc->serial), sc->obuf, sc->optr);
	AZ(ioctl(serial_fd(sc->serial), TIOCDRAIN));
	AZ(ioctl(serial_fd(sc->serial), TIOCMBIC, &j));
	attempt_msg(sc, 1, sc->obuf, &sc->optr);
	sc->tx_state++;
}

static void *
vor4000_rx_thread(void *priv)
{
	struct vor4000 *sc;
	unsigned char c;
	int i;

	CAST_OBJ_NOTNULL(sc, priv, VOR4000_MAGIC);

	vor4000_stty(serial_fd(sc->serial));
	while (1) {
		i = read(serial_fd(sc->serial), &c, 1);
		assert(i >= 0);
		pthread_cancelmutex_lock(&sc->lock);
		if (i == 1) {
			sc->optr = 0;
			sc->ibuf[sc->iptr++] = c;
			attempt_msg(sc, 0, sc->ibuf, &sc->iptr);
		} else if (serial_fd(sc->serial2) == -1) {
			vor4000_tx_poll(sc);
		}
		pthread_cancelmutex_unlock(&sc->lock);
	}
}

static void *
vor4000_tx_thread(void *priv)
{
	struct vor4000 *sc;
	unsigned char c;
	int i;

	CAST_OBJ_NOTNULL(sc, priv, VOR4000_MAGIC);

	vor4000_stty(serial_fd(sc->serial2));
	while (1) {
		i = read(serial_fd(sc->serial2), &c, 1);
		assert(i >= 0);
		if (i == 1) {
			pthread_cancelmutex_lock(&sc->lock);
			sc->iptr = 0;
			sc->obuf[sc->optr++] = c;
			attempt_msg(sc, 1, sc->obuf, &sc->optr);
			pthread_cancelmutex_unlock(&sc->lock);
		}
	}
}


/* ConfigKit routines ------------------------------------------------*/

static void
vor4000_control(struct point *pp, void *priv1, void *priv2, double val)
{
	struct vor4000 *sc;
	struct pointspec *ps;

	CAST_OBJ_NOTNULL(sc, priv1, VOR4000_MAGIC);
	assert(priv2 != NULL);
	ps = priv2;
	assert(ps->type == DOUT);
	if (val < .5)
		return;
	if (sc->tx_command != 0)
		return;
	if (serial_fd(sc->serial2) != -1)
		return;
	AZ(pthread_mutex_lock(&sc->lock));
	sc->tx_command = ps->msg;
	sc->tx_command_countdown = POLL_COUNTDOWN;
	AZ(pthread_mutex_unlock(&sc->lock));
	SetValBool(pp, 1);
}

static void
vor4000_init(CK_ARG, struct vor4000 *sc)
{
	struct point *pp;
	unsigned u, pi;
	struct pointspec *ps;

	serial_init(CK_PASS, &sc->serial,  vor4000_rx_thread, sc);
	serial_init(CK_PASS, &sc->serial2, vor4000_tx_thread, sc);

	for (ps = vor_points; ps->type != UNDEF; ps++) {
		pi = (unsigned)(ps - vor_points);
		assert(pi <= LAST_POINT);
		pp = CreatePoint(sc->gp, pi, ps->type);
		sc->pp[pi] = pp;
		SetPointLabel(pp, ps->label);
		if (ps->type == DOUT)
			SetPointControl(pp, vor4000_control, sc, ps);
	}

	for (ps = mon_points; ps->type != UNDEF; ps++) {
		for (u = 100; u <= 200; u += 100) {
			pi = (unsigned)(ps - mon_points) + u;
			assert(pi <= LAST_POINT);
			pp = CreatePoint(sc->gp, pi, ps->type);
			sc->pp[pi] = pp;
			SetPointLabel(pp, ps->label);
			if (ps->type == AIN)
				SetPointFormat(pp, "%.2f");
		}
	}
}

static void
vor4000_deinit(CK_ARG, struct vor4000 *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
	serial_destroy(CK_PASS, &sc->serial2);
}

RealVar(vor4000, VOR4000_MAGIC, azimuth1, azimuth1, 0.0, "%.1f", );
RealVar(vor4000, VOR4000_MAGIC, azimuth2, azimuth2, 0.0, "%.1f", );
UintVar(vor4000, VOR4000_MAGIC, debug, debug, 0, );
SerialFunc(vor4000, VOR4000_MAGIC, serial);
SerialFunc(vor4000, VOR4000_MAGIC, serial2);

MODULE_GROUP(vor4000, VOR4000_MAGIC);
