/*-
 * Copyright (c) 2005 Poul-Henning Kamp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: pic18fcode.c,v 1.22 2008/01/04 22:08:53 phk Exp $
 *
 */

#ifdef FLEXELINT
typedef unsigned char __sfr;
#endif

#include "pic18fregs.h"

/*
 * Define TENMHZ if the 10->40 MHz PLL should be disabled.
 * It is not quite clear if this improves A/D performance or not.
 */
#undef TENMHZ


/* Chip specific setups ----------------------------------------------*/

/*
 * XXX: The 0xf7 is because we use the '8720 definitions for the '8722
 * XXX: bit3 if set controls internal clock modes.
 */
#ifdef TENMHZ
	code unsigned char at(__CONFIG1H) config1h = _OSC_HS_1H & 0xf7;
	#define SPBRG_VAL	15	/* (10000000/(9600*64) - 1) */
	#define SPEED_STRING	" 10 MHz"
#else
	code unsigned char at(__CONFIG1H) config1h = _OSC_HS_PLL_1H & 0xf7;
	#define SPBRG_VAL	64	/* (40000000/(9600*64) - 1) */
	#define SPEED_STRING	" 40 MHz"
#endif

#ifdef pic18f8720
	#pragma stack 0xe00 0xff
	/* Disable XINST - extended instruction set */
	code unsigned char at(__CONFIG4L) config4l = 0x85;
	#define DEVICE_STRING	"PIC18F8722 "
	#ifndef TENMHZ
		#define ONEWIRE
	#endif
#endif

#ifdef pic18f1320
	#pragma stack 0x20 28
	#define DEVICE_STRING	"PIC18F1320 "
#endif

/* Enable the watchdog at 4msec * 64 = .256 sec */
code unsigned char at(__CONFIG2H) config2h = (0x7 << 1) | 1;


/* Structures --------------------------------------------------------*/

struct cmd {
	unsigned char	val;
	const char 	*lbl;
};

struct pin {
	__sfr 		*tris;
	__sfr 		*port;
	unsigned char	bit;
	char		an;
	unsigned char	modes;
	unsigned char	eeadr;
};

/* Lookup table data -------------------------------------------------*/

#define DIN	(1 << 0)
#define AIN	(1 << 1)
#define DOUT	(1 << 2)
#define CNT	(1 << 3)
#define OFF	(1 << 7)

#ifdef ONEWIRE
#define TEMP	(1 << 4)
#else
#define TEMP	OFF
#endif

static const code struct cmd mode[] = {
	{ AIN,		"AIN"}, 
	{ DIN,		"DIN"},
	{ CNT,		"CNT"},
	{ DOUT,		"DOUT"},
#ifdef ONEWIRE
	{ TEMP,		"TEMP"},
#endif
	{ OFF,		"OFF"},
	{ 0, 0 }
};

#define	C_ID		1
#define	C_SHOW		2
#define	C_CONF		3
#define	C_DEF		4
#define	C_OK		5
#define	C_SET		6
#define	C_AVG		7
#define	C_ADCON2	8
#ifdef ONEWIRE
#define	C_TEMPRES	9
#endif
#ifdef pic18f8720
#define	C_PORTTEST	10
#define	C_BUZZ		11
#endif


static const struct cmd cmds[] = {
	{ C_ID, 	"ID" },
	{ C_SHOW,	"SHOW" },
	{ C_CONF,	"CONF" },
	{ C_DEF,	"DEF" },
	{ C_SET,	"SET" },
	{ C_AVG,	"AVG" },
	{ C_ADCON2,	"ADCON2" },
#ifdef ONEWIRE
	{ C_TEMPRES,	"TEMPRES" },
#endif
#ifdef pic18f8720
	{ C_PORTTEST,	"PORTTEST" },
	{ C_BUZZ,	"BUZZ" },
#endif
	{ C_OK,		"\r" },
	{ C_OK,		"\n" },
	{ 0, 0 }
};


/* Pin configuration -------------------------------------------------*/

#include "_pins.c"

static unsigned counts[NPINS];
static unsigned char last[NPINS];


/* Error handling ----------------------------------------------------*/

static unsigned char error;
#define NOERROR 0
 
static const char * const errs[] = {
	/*  0 */	"",
	/*  1 */	"CMD unknown",
	/*  2 */	"CNF invalid mode",
	/*  3 */	"CNF invalid pin",
	/*  4 */	"CNF invalid state",
	/*  5 */	"CNF not dout",
	/*  6 */	"NUM overflow",
	/*  7 */	"RST reset",
	/*  8 */	"RX linelength",
	/*  9 */	"RX overrun",
	/* 10 */	"TX busy",
};


/* Input and Output buffers ------------------------------------------*/

#define NIBUF	20
static unsigned char ibuf[NIBUF];
static unsigned char ibufi;

#ifdef pic18f1320
#define NOBUF	(6 + NPINS * 6)
#endif

#ifdef pic18f8720
#define NOBUF	255
#endif

static unsigned char obuf[NOBUF];
static unsigned char obufi;


/* Misc globals & prototypes -----------------------------------------*/

static unsigned char avg, rnd;
#ifdef ONEWIRE
static void one_wire(unsigned char pin, const struct pin *p);
static void one_wire_init(void);
static unsigned char tempres;
static unsigned char sp[9];
#endif

#ifdef pic18f8720
static void porttest(void);
static void buzz(unsigned char pin);
#endif

/* EEPROM functions --------------------------------------------------*/

static unsigned char
rd_eeprom(unsigned adr)
{

#ifdef pic18f8720
	EEADRH = adr >> 8;
#endif
	EEADR = adr;
	EECON1bits.EEPGD = 0;
	EECON1bits.RD = 1;
	return (EEDATA);
}

static void
wr_eeprom(unsigned adr, unsigned char dat)
{

#ifdef pic18f8720
	EEADRH = adr >> 8;
#endif
	EEADR = adr;
	EECON1bits.EEPGD = 0;
	EECON1bits.RD = 1;
	if (EEDATA == dat)
		return;
	EEDATA = dat;
	PIR2bits.EEIF = 0;
	EECON1bits.EEPGD = 0;
	EECON1bits.CFGS = 0;
	EECON1bits.WREN = 1;
	EECON2 = 0x55;
	EECON2 = 0xaa;
	EECON1bits.WR = 1;
	while (EECON1bits.WR)
		;
	EECON1bits.WREN = 0;
}

/*
 * These two are relics from earlier days, only DOUT state is stored
 * here these days
 */

static unsigned
rd_scale(unsigned eeadr)
{

	return ((rd_eeprom(eeadr + 2) << 8) | rd_eeprom(eeadr + 1));
}

static void
wr_scale(unsigned eeadr, unsigned k)
{

	wr_eeprom(eeadr + 1, k);
	wr_eeprom(eeadr + 2, k >> 8);
}


/* Serial input parsing functions ------------------------------------*/

static void
in_white(void)
{

	for (; ibufi < NIBUF; ibufi++)
		if (ibuf[ibufi] != ' ' && ibuf[ibufi] != '\t')
			return;
}

static char
in_cmd(const char *s)
{
	unsigned char i;

	for (i = ibufi; i < NIBUF; i++) {
		if (*s == '\0') {
			ibufi = i;
			return (1);
		}
		if (*s != (const char)ibuf[i])
			return (0);
		s++;
	}
	return (0);
}

static unsigned
in_dec(void)
{
	unsigned x;
	unsigned char i;

	in_white();
	i = ibufi;
	x = 0;
	while (i < NIBUF && ibuf[i] >= '0' && ibuf[i] <= '9') {
		if (x > 6553) {
			error = 6;
			return (0);
		}
		x *= 10;
		x += ibuf[i] - '0';
		i++;
	}
	ibufi = i;
	return (x);
}

static unsigned char
in_pin(void)
{
	unsigned int u;

	u = in_dec();
	if (error == NOERROR && (u == 0 || u > NPINS)) {
		error = 3;
		u = 1;
	}
	return (u - 1);
}

static unsigned char
in_tbl(const struct cmd *p)
{

	in_white();
	for (; p->lbl != 0; p++)
		if (in_cmd(p->lbl))
			break;
	return (p->val);
}


/* Serial output functions -------------------------------------------*/

static void
out_c(unsigned char c)
{
	if (obufi < NOBUF)
		obuf[obufi++] = c;
}

static void
out_s(const char *s)
{

	while (*s > '\1' && obufi < NOBUF)
		obuf[obufi++] = (unsigned char)*s++;
}

/* Output EOL sequence */
static void
out_end(void)
{

	out_c('\r');
	out_c('\n');
	out_c(0);
	obufi = 0;
}

static void
out_dec(unsigned u)
{
	unsigned d = 10000;
	unsigned char v;

	while (d > u && d > 1)
		d /= 10;
	while (d != 0) {
		v = u / d;
		out_c('0' + v);
		u -= v * d;
		d /= 10;
	}
}

#ifdef ONEWIRE
static const unsigned char hex[] = "0123456789abcdef";

static void
out_hex(unsigned char u)
{
	out_c(hex[u >> 4]);
	out_c(hex[u & 0xf]);
}
#endif


/* Serial hardware polling -------------------------------------------*/

static void
poll_serial_port_rx(void)
{
	int a;

	/* Tickle the watchdog */
	ClrWdt();

	/* Check and reset receive overruns */
	if (RCSTAbits.OERR) {
		error = 9;
		RCSTAbits.CREN = 0;
		RCSTAbits.CREN = 1;
	}

	/* Handle received chars */
	if(PIR1bits.RCIF) {
		a = RCREG;
		if (obuf[obufi] != '\0')
			error = 10;
		else if (ibufi < NIBUF)
			ibuf[ibufi++] = a;
		else
			error = 8;
		if (a == '\r' || a == '\n')
			ibufi = 255;
	}

}

static void
poll_serial_port(void)
{

	poll_serial_port_rx();

	/* Transmit chars */
	if(TXSTAbits.TRMT && obuf[0] != '\0') {
		TXREG = obuf[obufi];
		obufi++;
		if (obufi >= NOBUF || obuf[obufi] == '\0') {
			obufi = 0;
			obuf[obufi] = '\0';
		}
	}
}

/* High level command processing -------------------------------------*/

static void
conf_pin(unsigned char pin)
{
	unsigned char d;
	const struct pin *p = &pins[pin];
	
	counts[pin] = 0;
	d = rd_eeprom(p->eeadr);

	/* If recorded mode is illegal, use Off */
	if (!(d & p->modes) || (d & ~p->modes)) {
		d = OFF;
		wr_eeprom(p->eeadr, d);
	}

#ifdef pic18f1320
	if (p->an) {
		if (d & AIN)
			ADCON1 &= ~(unsigned char)p->an;
		else
			ADCON1 |= (unsigned char)p->an;
		ADCON1 |= 0x60;		/* B1 + B4 Not AIN */
	}
#endif
	if (d & DOUT) {
		*p->tris &= ~p->bit;
		if (rd_eeprom(p->eeadr + 1))
			*p->port |= p->bit;
		else
			*p->port &= ~p->bit;
	} else if (d & TEMP) {
		*p->tris |= p->bit;
		last[pin] = 0;
	} else {
		*p->tris |= p->bit;
	}

#ifdef pic18f8720
	for (pin = 0; pin < 17; pin++) {
		p = &pins[pin];
		d = rd_eeprom(p->eeadr);
		if (!(d & AIN)) {
			ADCON1 = (unsigned char)p->an;
#if 0
			break;
#else
			for (pin++; pin < 17; pin++) {
				p = &pins[pin];
				d = rd_eeprom(p->eeadr);
				if (d & AIN) {
					wr_eeprom(p->eeadr, DIN);
					*p->tris |= p->bit;
				}
			}
#endif
		}
	}
#endif
}

static void
report_config(const char *s, unsigned int pin)
{
	unsigned u;
	const struct pin *p;

	p = &pins[pin];
	u = rd_eeprom(p->eeadr);
	out_s(s);
	out_dec(pin + 1);
	out_c(' ');
	if (u & AIN)
		out_s ("AIN");
	else if (u & DIN)
		out_s ("DIN");
	else if (u & CNT)
		out_s ("CNT");
	else if (u & DOUT) {
		out_s ("DOUT");
		out_c(' ');
		u = rd_scale(p->eeadr);
		out_dec(u);
	}
#ifdef ONEWIRE
	else if (u & TEMP) {
		out_s ("TEMP ");
		last[pin] = 50;
		while (last[pin] >= 50)
			one_wire(pin, p);
	}
#endif
	else if (u & OFF)
		out_s ("OFF");
	else 
		out_s ("???");
}

static void
report_data(void)
{
	unsigned char i;
	const struct pin *p;
	unsigned char c0;
	unsigned v;

	out_s("DATA");
	for (i = 0; i < NPINS; i++) {
		p = &pins[i];
		c0 = rd_eeprom(p->eeadr);
		out_c(' ');
		v = counts[i];
		if (c0 & OFF) {
			out_c('-');
		} else if (p->an >= 0 && (c0 & AIN)) {
			out_dec(v);
		} else if (c0 & (DIN|DOUT)) {
			c0 = *p->port;
			if (c0 & p->bit)
				out_c('1');
			else
				out_c('0');
		} else if (c0 & CNT) {
			out_dec(v);
#ifdef ONEWIRE
		} else if (c0 & TEMP) {
			out_dec(v);
			/*
			 * If power is lost while a temperature conversion
			 * is in progress, we will never see the "I'm done"
			 * bit from the 1-wire device.  In order to catch
			 * this condition, erase the measurement whenever
			 * we have sent it, provided it is not an error code.
			 */
			v &= 0xf000;
			if (v == 0xf000 || v == 0x0000)
				counts[i] = 0x8000;
#endif
		}
	}
}

static void
handle_serial_line(void)
{
	unsigned char u;
	unsigned j, k;
	const struct pin *p;

	if (ibufi < NIBUF)
		ibuf[ibufi] = '\0';
	ibufi = 0;

	if (in_cmd("CLR")) {
		error = NOERROR;
		out_s("OK_CLR");
		return;
	}

	if (error != NOERROR)
		return;
	
	k = in_tbl(cmds);
	switch (k) {
	case C_ID:
		out_s("REV ");
		out_s(" $Revision: 1.22 $");
		out_s(DEVICE_STRING);
		out_dec(NPINS);
		out_s(" Pins");
		out_s(SPEED_STRING);
#ifdef ONEWIRE
		out_s(" TEMP");
#endif
		break;
	case C_SHOW:
		report_data();
		break;
	case C_ADCON2:
		ADCON2 = in_dec() & 0x3f;
		out_s("ADCON2_OK ");
		out_dec(ADCON2);
		break;
	case C_AVG:
		avg = in_dec();
		if (avg > 6)
			avg = 6;
		if (avg)
			rnd = (1 << (avg - 1)) - 1;
		else
			rnd = 0;
		out_s("AVG_OK ");
		out_dec(avg);
		out_c(' ');
		out_dec(rnd);
		break;
#ifdef ONEWIRE
	case C_TEMPRES:
		tempres = ((in_dec() & 3) << 5) | 0x1f;
		out_s("TEMPRES_OK ");
		out_dec(tempres);
		break;
#endif
	case C_CONF:
		u = in_pin();
		if (error != NOERROR)
			break;
		report_config("PIN ", u);
		break;
	case C_DEF:
		u = in_pin();
		if (error != NOERROR)
			break;
		p = &pins[u];
		j = in_tbl(mode);
		if (!(j & p->modes)) {
			error = 2;
			break;
		}
		wr_scale(p->eeadr, 0);
		wr_eeprom(p->eeadr, j);
		conf_pin(u);
		report_config("CHG ", u);
		break;
	case C_SET:
		u = in_pin();
		if (error != NOERROR)
			break;
		p = &pins[u];
		j = rd_eeprom(p->eeadr);
		if (j != DOUT) {
			error = 5;
			break;
		}
		k = in_dec();
		if (error != NOERROR)
			break;
		if (k > 1) {
			error = 4;
			break;
		}
		wr_scale(p->eeadr, k);
		conf_pin(u);
		report_config("STAT ", u);
		break;

#ifdef C_PORTTEST
	case C_PORTTEST:
		porttest();
		break;
	case C_BUZZ:
		u = in_pin();
		buzz(u);
		break;
#endif
	case C_OK:
		break;
	default:
		error = 1;
		break;
	}
}


/* Main function -----------------------------------------------------*/

void
main(void)
{
	const struct pin *p;
	unsigned char c0, i;
	unsigned int u, w;

	/* Set up async port */
	RCSTA = 0x90;
	TXSTA = 0x20;
	SPBRG = SPBRG_VAL;

	ADCON2 = 0x3e;

#ifdef ONEWIRE
	one_wire_init();
#endif

	avg = 6;
	rnd = 31;
		
	ADCON0bits.ADON = 1;

	for(i = 0; i < NPINS; i++)
		conf_pin(i);

	/* Initialize input buffer */
	ibufi = 0;

	/* Initialize output buffer */
	obufi = 0;
	obuf[obufi] = '\0';

	/* Output the RESET reason */
	out_s("\r\nRCON ");
	out_dec(RCON);
	out_end();
	RCON = 0xff;

	/* Initialize error state */
	error = 7;

	while (1) {

		/* Scan the pins */
		for (i = 0; i < NPINS; i++) {
			poll_serial_port();
			p = &pins[i];
			c0 = rd_eeprom(p->eeadr);
			if (p->an >= 0 && (c0 & AIN)) {
				ADCON0 = 0x03 | (i << 2);
				while (ADCON0bits.GO)
					continue;
				u = ((ADRESH << 8) | ADRESL);
				w = counts[i];
				if (u >= w)
					w += (rnd + u - w) >> avg;
				else
					w -= (rnd + w - u) >> avg;
				counts[i] = w;
			}
#ifdef ONEWIRE
			if (c0 & TEMP)
				one_wire(i, p);
#endif
			if (c0 & CNT) {
				c0 = *p->port;
				c0 &= p->bit;
				if (c0 == last[i]) 
					continue;
				last[i] = c0;
				if (c0)
					continue;
				counts[i]++;
			}
		}

		/* Process serial buffer if complete */
		if (ibufi == 255) {
			handle_serial_line();
			if (error != NOERROR) {
				out_s("ERR ");
				out_dec(error);
				out_c(' ');
				out_s(errs[error]);
			}
			out_end();
			ibufi = 0;
		}
	}
}

#ifdef ONEWIRE

/* Timer array -------------------------------------------------------*/

static unsigned ow_timer[NPINS];

/* CRC8 calculation --------------------------------------------------*/

static const code unsigned char crc8_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53
};

static unsigned char
crc8(const unsigned char *p, int len)
{
	unsigned char c;

	c = 0;
	while (len--)
		c = crc8_table[c ^ *p++];
	return (c);
}


/* 1Wire init --------------------------------------------------------*/

static void
one_wire_init(void)
{
	tempres	= 0x7f;

	T0CON = 0x87;	/* TRM0ON + prescale=256 */
}

/* 1Wire function ----------------------------------------------------*/

/*
 * Send a pulse as either zero or one.
 * When sending a one bit, read back the lack to see if the other end
 * wanted a zero here.
 *
 * Timing is highly magical for 40MHz.
 */

static unsigned char
pulse(const struct pin *p, unsigned char l, unsigned char strong)
{
	unsigned char u;

	if (l == 0) {
		*p->tris &= ~p->bit;
		for (u = 0; u < 220; u++)
			;
		if (strong) {
			*p->port |= p->bit;
		} else {
			*p->tris |= p->bit;
			for (u = 0; u < 220; u++)
				;
		}
		return (0);
	}
	
	l = 0;
	*p->tris &= ~p->bit;
	for (u = 0; u < 3; u++)
		;
	*p->tris |= p->bit;
	for (u = 0; u < 3; u++)
		l |= *p->port;
	for (u = 0; u < 100; u++)
		;
	if (l & p->bit)
		return (1);
	return (0);
}

/*
 * Send reset pulse
 * 
 * Timing is highly magical for 40MHz.
 */

static void
reset_pulse(unsigned char pin, const struct pin *p)
{
	unsigned u;
	unsigned char a;

	/* Give up right away if the line isn't pulled high */
	if (!(*p->port & p->bit)) {
		counts[pin] = 0x8003;
		last[pin] = 0;
		return;
	}
	*p->tris &= ~p->bit;
	for (u = 0; u < 300; u++)
		if (u == 150)
			poll_serial_port_rx();
	*p->tris |= p->bit;
	a = 0;
	for (u = 0; u < 5; u++)
		a |= *p->port;
	/* If the line doesn't go high, we're toast */
	if (!(a & p->bit)) {
		counts[pin] = 0x8004;
		last[pin] = 0;
		return;
	}
	a = 0;
	for (u = 0; u < 40; u++) {
		if (u == 20)
			poll_serial_port_rx();
		if (!(*p->port & p->bit))
			a = 1;
		else if (a) {
			*p->port |= p->bit;
			*p->tris &= ~p->bit;
			return;
		}
	}
	counts[pin] = 0x8005 + a;
	last[pin] = 0;
}

static void
send_byte(const struct pin *p, unsigned char b, unsigned char strong)
{

	(void)pulse(p, b & 0x01, 0);
	(void)pulse(p, b & 0x02, 0);
	(void)pulse(p, b & 0x04, 0);
	(void)pulse(p, b & 0x08, 0);
	(void)pulse(p, b & 0x10, 0);
	(void)pulse(p, b & 0x20, 0);
	(void)pulse(p, b & 0x40, 0);
	(void)pulse(p, b & 0x80, strong);
}

static unsigned char
read_byte(const struct pin *p)
{
	unsigned char r;

	r  = pulse(p, 1, 0);
	r |= pulse(p, 1, 0) << 1;
	r |= pulse(p, 1, 0) << 2;
	r |= pulse(p, 1, 0) << 3;
	r |= pulse(p, 1, 0) << 4;
	r |= pulse(p, 1, 0) << 5;
	r |= pulse(p, 1, 0) << 6;
	r |= pulse(p, 1, 0) << 7;
	return (r);
}

/*
 * Poll one pin for a DS18B20 device.
 *
 * Do only a little bit of work each time, we don't want to slow down
 * the polling too much.
 *
 * Record how far we got from poll to poll in last[].
 */

static void
one_wire(unsigned char pin, const struct pin *p)
{
	unsigned char u;
	unsigned v;

	if (last[pin] != 7) {
		*p->tris |= p->bit;
		*p->port &= ~p->bit;
	}
	switch(last[pin]) {
	/* Reset and read address to verify device type */
	case 1:
		*p->port &= ~p->bit;
		reset_pulse(pin, p);
		break;
	case 2:
		send_byte(p, 0x33, 0);	/* READ ROM */
		break;
	case 3:
		u = read_byte(p);
		if (0x28 != u) {	/* If not DS18B20 */
			counts[pin] = 0xc000 | u;
			last[pin] = 0;
		}
		break;

	/* Start temperature conversion */
	case 4:
		reset_pulse(pin, p);
		break;
	case 5:
		send_byte(p, 0xcc, 0);	/* SKIP ROM */
		break;
	case 6:
		send_byte(p, 0x44, 1);	/* CONVERT TEMPERATURE */
		ow_timer[pin] = TMR0L;
		ow_timer[pin] |= (TMR0H << 8);
		break;

	/* Wait for conversion to complete */
	case 7:
		v = TMR0L;
		v |= (TMR0H << 8);
		v -= ow_timer[pin];
		if (v < 0x8000)
			return;
		*p->tris |= p->bit;	/* end strong pullup */
		*p->port &= ~p->bit;
		break;

	/* Read back scratch pad registers */
	case 8:
		reset_pulse(pin, p);
		break;
	case 9:
		send_byte(p, 0xcc, 0);	/* SKIP ROM */
		break;
	case 10:
		send_byte(p, 0xbe, 0);	/* READ SCRATCH PAD */
		break;
	case 11:
		for (u = 0; u < 9; u++) {
			/* Make sure we do not loose serial input */
			poll_serial_port();
			sp[u] = read_byte(p);
		}
		if (sp[8] != crc8(sp, 8))
			counts[pin] = 0x8001;
		else
			counts[pin] = (sp[1] << 8) | sp[0];
		if (counts[pin] == 0x0550)
			counts[pin] = 0x8002;
		if (sp[4] == tempres)	/* If resolution is OK */
			last[pin] = 0;
		break;

	/* Write new configuration register value */
	case 12:
		reset_pulse(pin, p);
		break;
	case 13:
		send_byte(p, 0xcc, 0);	/* SKIP ROM */
		break;
	case 14:
		send_byte(p, 0x4e, 0);	/* WRITE SCRATCHPAD */
		break;
	case 15:
		send_byte(p, 0x7f, 0);	/* T(high) register */
		break;
	case 16:
		send_byte(p, 0x80, 0);	/* T(low) register */
		break;
	case 17:
		send_byte(p, tempres, 0);	/* Config register */
		last[pin] = 0;
		break;

	case 50:
		reset_pulse(pin, p);
		break;
	case 51:
		send_byte(p, 0x33, 0);	/* READ ROM */
		break;
	case 52:
		for (u = 0; u < 8; u++) {
			/* Make sure we do not loose serial input */
			sp[u] = read_byte(p);
		}
		if (sp[7] != crc8(sp, 7)) {
			out_s("bad crc");
		} else {
			for (u = 0; u < 8; u++)
				out_hex(sp[u]);
		}
		last[pin] = 0;
		break;

	default:
		last[pin] = 0;
	}
	last[pin]++;
}
#endif

#ifdef C_PORTTEST

static void
outwait(void) 
{

	out_end();
	ibufi = 0;
	error = 0;
	do {
		poll_serial_port();
	} while (obuf[0] != '\0');
}

static void
userwait(void) 
{

	out_end();
	ibufi = 0;
	error = 0;
	do {
		poll_serial_port();
	} while (ibufi != 255);
}

static void
wiggle(unsigned char pin1, unsigned char pin2)
{
	unsigned char c, e;
	const struct pin *p1, *p2;

	last[pin1] = 0;
	last[pin2] = 0;
	p1 = &pins[pin1];
	p2 = &pins[pin2];
	*p2->tris |= p2->bit;
	*p1->tris &= ~p1->bit;
	for (c = 0; c < 4; c++) {
		e = 0;

		*p1->port = 0;
		if (*p1->port & p1->bit)
			e |= 8;
		if (*p2->port & p2->bit)
			e |= 1;

		*p1->port = p1->bit;
		if (!(*p1->port & p1->bit))
			e |= 16;

		if (!(*p2->port & p2->bit))
			e |= 2;


		if (!e)
			e |= 4;
		last[pin1] |= e;
		last[pin2] |= e;
		counts[pin1] |= e;
		counts[pin2] |= e;
		out_c(e + '@');
	}
	*p1->tris |= p1->bit;
	*p2->tris |= p2->bit;
}

static void
rep_bit(unsigned char bit, unsigned char glyph)
{
	unsigned char i;

	for (i = 0; i < NPINS; i++) {
		if (last[i] & bit)
			out_c(glyph);
		else
			out_c('-');
		if ((i & 15) == 15)
			out_c(' ');
	}
	outwait();
}

static void
rep_bits(void)
{
	unsigned char i;

	rep_bit(1, 'h');
	rep_bit(2, 'l');
	rep_bit(8, 'H');
	rep_bit(16, 'L');
	rep_bit(4, '*');
	for (i = 0; i < NPINS; i++)
		last[i] = 0;
}

static void
onetest(const char *c1, const char *c2, unsigned char o1, unsigned char o2)
{
	unsigned char i;

	out_s("Connect cable between ");
	out_s(c1);
	out_s(" and ");
	out_s(c2);
	userwait();
	out_dec(o1);
	out_s(" -> ");
	out_dec(o2);
	outwait();

	for (i = 0; i < 16; i++)
		wiggle(i + o1, i + o2);
	outwait();
	rep_bits();

	out_dec(o2);
	out_s(" -> ");
	out_dec(o1);
	outwait();

	for (i = 0; i < 16; i++)
		wiggle(i + o2, i + o1);
	outwait();
	rep_bits();
}

static void
porttest(void)
{
	const struct pin *p1;
	unsigned char i;
	unsigned e;

	out_s("Remove all four connectors, press return");
	userwait();

	/* Define everything as input */
	for (i = 0; i < NPINS; i++) {
		p1 = &pins[i];
		*p1->tris |= p1->bit;
		wr_eeprom(p1->eeadr, DIN);
		last[i] = 0;
		counts[i] = 0;
	}
	out_s("EEPROM reset, all pins DIN");
	outwait();

	ADCON1 = 0xf;

	e = 0;

	onetest("J201", "J301",  0,  16);
	onetest("J201", "J302",  0,  32);
	onetest("J201", "J303",  0,  48);
	onetest("J302", "J303", 32,  48);
	onetest("J301", "J303", 16,  48);
	onetest("J301", "J302", 16,  32);

	out_s("Final result:");
	outwait();
	for (i = 0; i < NPINS; i++)
		last[i] = counts[i];
	rep_bits();
	for (i = 0; i < NPINS; i++) {
		if ((i & 15) == 0)
			out_s("Bad pins =");
		if (!(counts[i] & 4)) {
			out_c(' ');
			out_dec(i + 1);
		}
		if ((i & 15) == 15)
			outwait();
	}
	outwait();
	out_s("PORTTEST done");
	outwait();
}

static void
buzz(unsigned char pin)
{
	const struct pin *p;
	
	out_s("Buzzing pin ");
	out_dec(pin + 1);
	out_s(" press return to stop");
	outwait();
	p = &pins[pin];
	*p->tris &= ~p->bit;
	ibufi = 0;
	while (ibufi != 255) {
		*p->port &= ~p->bit;
		poll_serial_port();
		*p->port |= p->bit;
		poll_serial_port();
	}
	*p->tris |= p->bit;
	out_s("BUZZ done");
	outwait();
	ibufi = 0;
}
#endif
