/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: event.c,v 1.53 2007/08/21 21:13:14 phk Exp $
 *
 */

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <configkit.h>
#include <time.h>
#include <sys/sbuf.h>

#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

#define EM(e)	(1U << (e))

struct event_subscription {
	unsigned				magic;
#define EVENT_SUBSCRIPTION_MAGIC		0x5473366
	TAILQ_ENTRY(event_subscription)		list;
	unsigned				mask;
	event_handler_t				*func;
	void					*priv;
	unsigned				refcnt;
};

static TAILQ_HEAD(, event_subscription) subscribers =
    TAILQ_HEAD_INITIALIZER(subscribers);

struct evinfo {
	unsigned 		mask;
	unsigned char		site;
	unsigned char		group;
	unsigned char		point;
	unsigned char		msg;
	unsigned char		state;
	const char		*name;
};

static struct evinfo ev_names[] = {
#define EVD(a, b, s, g, p, c, m, t, x) \
	[b] = { EM(b), s, g, p, m, t, x },
#include "events.h"
#undef EVD
};

static unsigned ev_names_size = sizeof ev_names / sizeof ev_names[0];

static void
event(const struct event_subscription *sc, event_t ev, void *obj, unsigned instance, char **string_cache)
{
	struct event_arg e;

	memset(&e, 0, sizeof e);
	if (ev >= ev_names_size)
		Fatal("Event is illegal %d", ev);
	if (ev_names[ev].name == NULL)
		Fatal("Event has no name %d", ev);
	if (sc->mask & EM(ev)) {
		e.e = ev;
		e.obj = obj;
		e.instance = instance;
		e.name = ev_names[ev].name;
		e.string_cache = string_cache;
		e.free_me = NULL;
		sc->func(sc->priv, &e);
		if (e.free_me != NULL)
			free(e.free_me);
	}
}

struct event_subscription *
SubscribeEvents(event_handler_t *func, void *priv)
{
	struct event_subscription *sc;

	ALLOC_OBJ(sc, EVENT_SUBSCRIPTION_MAGIC);
	sc->func = func;
	sc->priv = priv;
	Lock();
	TAILQ_INSERT_TAIL(&subscribers, sc, list);
	Unlock();
	return (sc);
}

/*lint -e{715,818} */
static void
DummyEvent(void *priv __unused, struct event_arg *e __unused)
{
}

/*lint -e{641,777} */
static void
IntStartEvents(struct event_subscription *sc, unsigned mask)
{
	struct site *s;
	struct group *g;
	struct point *p;
	void *ptr;
	unsigned u;

	CHECK_OBJ_NOTNULL(sc, EVENT_SUBSCRIPTION_MAGIC);
	sc->refcnt = 1;
	
	TopoHold();
	sc->mask = mask;
	sc->mask &= ~(EM(E_POINT_VALUE)|EM(E_POINT_LIMIT));
	/* Now send any "catch-up" events */
	TAILQ_FOREACH(s, &all_sites, list) {
		if (mask & EM(E_NEW_SITE))
			event(sc, E_NEW_SITE, s, SI(s), NULL);
		TAILQ_FOREACH(g, &s->groups, list) {
			if (mask & EM(E_NEW_GROUP))
				event(sc, E_NEW_GROUP, g, GI(g), NULL);
			TAILQ_FOREACH(p, &g->points, list) {
				u = PI(p);
				event(sc, E_NEW_POINT, p, u, NULL);
				event(sc, E_POINT_LIMIT, p, u, NULL);
				event(sc, E_POINT_LABEL, p, u, NULL);
				event(sc, E_POINT_FORMAT, p, u, NULL);
				event(sc, E_POINT_UNIT, p, u, NULL);
				event(sc, E_POINT_LOW, p, u, NULL);
				event(sc, E_POINT_SAG, p, u, NULL);
				event(sc, E_POINT_RAISE, p, u, NULL);
				event(sc, E_POINT_HIGH, p, u, NULL);
				event(sc, E_POINT_LIMITDELAY, p, u, NULL);
				event(sc, E_POINT_HYSTERESIS, p, u, NULL);
				event(sc, E_POINT_TIMEOUT, p, u, NULL);
				event(sc, E_POINT_TRAP, p, u, NULL);
			}
		}
	}
	sc->mask = mask;
	if (mask & (EM(E_POINT_VALUE)|EM(E_POINT_LIMIT))) {
		TAILQ_FOREACH(s, &all_sites, list) {
			TAILQ_FOREACH(g, &s->groups, list) {
				TAILQ_FOREACH(p, &g->points, list)  {
					event(sc, E_POINT_VALUE,
					    p, PI(p), NULL);
					event(sc, E_POINT_LIMIT,
					    p, PI(p), NULL);
				}
			}
		}
	}
	if (mask & EM(E_STATE_CHG)) {
		ptr = NULL;
		while (StateDump(&ptr))
			event(sc, E_STATE_CHG, ptr, 0, NULL);
	}
	TopoRel();
	Lock();
	sc->refcnt--;
	if (sc->func == DummyEvent) {
		TAILQ_REMOVE(&subscribers, sc, list);
		FREE_OBJ(sc);
	}
	Unlock();
}

static int
islit(const char *b, const char *e, const char *t)
{

	assert(e >= b);
	if ((e - b) != (int)strlen(t))
		return (0);
	if (strncasecmp(b, t, e - b))
		return (0);
	return (1);
}

int
StartEvents(struct event_subscription *sc, const char *p)
{
	const char *q;
	unsigned m = 0, u;
	int f;

	while (*p != '\0') {
		if (!isalpha(*p)) {
			p++;
			continue;
		}
		for (q = p + 1; *q != '\0' && isalpha(*q); q++)
			continue;

		if (islit(p, q, "all")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].name != NULL &&
				    !ev_names[u].msg)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "site")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].site)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "group")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].group)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "point")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].point)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "config")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].point)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "msg")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].msg)
					m |= ev_names[u].mask;
		} else if (islit(p, q, "state")) {
			for (u = 0; u < ev_names_size; u++)
				if (ev_names[u].state)
					m |= ev_names[u].mask;
		} else {
			f = 0;
			for (u = 0; u < ev_names_size; u++) {
				if (ev_names[u].name != NULL &&
				    islit(p, q, ev_names[u].name + 2)) {
					m |= ev_names[u].mask;
					f++;
				}
			}
			if (!f) 
				return (1);
		}
		p = q;
	}
	IntStartEvents(sc, m);
	return (0);
}

void
UnsubscribeEvents(struct event_subscription *p)
{

	CHECK_OBJ_NOTNULL(p, EVENT_SUBSCRIPTION_MAGIC);
	Lock();
	if (p->refcnt == 0) {
		TAILQ_REMOVE(&subscribers, p, list);
		FREE_OBJ(p);
	} else {
		p->func = DummyEvent;
	}
	Unlock();
}

void
PostEvent(event_t ev, void *obj, unsigned instance)
{
	struct event_subscription *p, *p2;
	char *string_cache;

	string_cache = NULL;
	TAILQ_FOREACH_SAFE(p, &subscribers, list, p2) {
		CHECK_OBJ_NOTNULL(p, EVENT_SUBSCRIPTION_MAGIC);
		event(p, ev, obj, instance, &string_cache);
	}
	if (string_cache != NULL)
		free(string_cache);
}

static void
xEvent_String_Limit(struct sbuf *sb, const struct point *p, int lim)
{
	sbuf_printf(sb, "%g", p->limits[lim]);
}

static void
xEvent_String(struct sbuf *sb, const struct event_arg *e)
{
	struct group *g;
	struct point *p;

	switch (e->e) {
	case E_NEW_SITE:
	case E_DESTROY_SITE:
	case E_DESTROY_GROUP:
		return;
	case E_NEW_GROUP:
		CAST_OBJ_NOTNULL(g, e->obj, GROUP_MAGIC);
		sbuf_printf(sb, "%Q", g->name);
		return;
	case E_STATE_CHG:
		StateMaster(sb, e->obj);
		return;
	default:
		break;
	}
	CAST_OBJ_NOTNULL(p, e->obj, POINT_MAGIC);
	switch(e->e) {
	case E_NEW_POINT: sbuf_printf(sb, "%s",  p_Type(p)); return;
	case E_DESTROY_POINT: return;
	case E_POINT_VALUE:
		sbuf_printf(sb, p_format(p), p->value);
		sbuf_printf(sb, " %ld.%09d",
		    p->measured.tv_sec, p->measured.tv_nsec);
		return;
	case E_POINT_LIMIT:
		sbuf_printf(sb, "%s %u %d", LimitType(p->curlim),
		    p->trap, p->state);
		return;
	case E_POINT_LABEL: sbuf_printf(sb, "%Q", p_label(p)); return;
	case E_POINT_FORMAT: sbuf_printf(sb, "%Q", p_format(p)); return;
	case E_POINT_UNIT: sbuf_printf(sb, "%Q", p_unit(p)); return;

	case E_POINT_LOW: xEvent_String_Limit(sb, p, L_LOW); return;
	case E_POINT_SAG: xEvent_String_Limit(sb, p, L_SAG); return;
	case E_POINT_RAISE: xEvent_String_Limit(sb, p, L_RAISE); return;
	case E_POINT_HIGH: xEvent_String_Limit(sb, p, L_HIGH); return;
	case E_POINT_CHANGED: xEvent_String_Limit(sb, p, L_CHANGED); return;

	case E_POINT_TIMEOUT:
		sbuf_printf(sb, "%g", p_timeout(p));
		return;

	case E_POINT_LIMITDELAY: sbuf_printf(sb, "%g", p->limitdelay); return;
	case E_POINT_HYSTERESIS: sbuf_printf(sb, "%g", p->hysteresis); return;
	case E_POINT_TRAP: sbuf_printf(sb, "%u", p->trap); return;
	default:
		Fatal("event 0x%x instance %I unknown", e->e, e->instance);
		break;
	}
}

const char *
Event_String(struct event_arg *e)
{
	char *s;
	struct sbuf *sb;

	if (e->string_cache != NULL && *e->string_cache != NULL)
		return (*e->string_cache);

	if (ev_names[e->e].msg) {
		s = strdup(e->obj);
	} else {
		sb = sbuf_new(NULL, NULL, 0, SBUF_AUTOEXTEND);
		assert(sb != NULL);
		xEvent_String(sb, e);
		sbuf_finish(sb);
		s = strdup(sbuf_data(sb));
		assert(s != NULL);
		sbuf_delete(sb);
	}
	if (e->string_cache != NULL)
		*e->string_cache = s;
	else
		e->free_me = s;
	return (s);
}
