/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: httpd.c,v 1.53 2007/08/01 08:16:25 phk Exp $
 *
 */

#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "phk_argv.h"

#include "measured.h"
#include "measured_int.h"

struct url {
	unsigned		magic;
#define URL_MAGIC		0x3837282
	LIST_ENTRY(url)		next;
	const char		*verb;
	const char		*url;
	void			*priv;
	url_f			*func;
};

static LIST_HEAD(, url) urls = LIST_HEAD_INITIALIZER(urls);

static struct tcp_listener *tl;

void
RegisterUrl(const char *verb, const char *url, url_f *func, void *priv)
{
	struct url *up;

	ALLOC_OBJ(up, URL_MAGIC);
	up->verb = strdup(verb);
	up->url = strdup(url);
	up->func = func;
	up->priv = priv;
	LIST_INSERT_HEAD(&urls, up, next);
}

/* Unbugger an URL ---------------------------------------------------*/

/*
 * "What's the name of them small buggers there are hundred of in everything ?"
 * 		-- Granny Weathervax
 */

static void
Unbugger(char *url)
{
	char *p, *q;
	unsigned u;

	for (p = q = url; *p != '\0';) {
		if (*p == '%' && sscanf(p + 1, "%02x", &u) == 1) {
			*q++ = (unsigned char)u;
			p += 3;
		} else
			*q++ = *p++;
	}
	*q = '\0';
}


static void
httpd(int fd)
{
	char buf[BUFSIZ], *up, *ua, *p;
	FILE *f;
	struct url *uhp;
	char **av;
	int ihdr, ok;
	struct http h;
	struct linger linger;

	memset(&h, 0, sizeof h);
	linger.l_onoff = 0;
	linger.l_linger = 0;
	assert(0 == setsockopt(fd, SOL_SOCKET, SO_LINGER,
	    &linger, sizeof linger));
	f = fdopen(fd, "w+");
	av = NULL;

	ok = 1;
	do {
		if (fgets(buf, sizeof buf, f) == NULL) {
			Debug(0, "read error %M", errno);
			break;
		}
		ua = strchr(buf, '\0');
		assert(ua != NULL);
		if (ua == buf) {
			Debug(0, "first line empty");
			break;
		}
		if ((ua[-1] != '\n' && ua[-1] != '\r')) {
			Debug(0, "first line longer than %d", strlen(buf));
			break;
		}
		av = PhkArgvParse(buf, 0);
		if (av[0] != NULL) {
			Debug(0, "parse: %s", av[0]);
			break;
		}
		if (av[1] == NULL || av[2] == NULL) {
			Debug(0, "bad first line");
			break;
		}

		ihdr = 0;
		while (1) {
			if (fgets(buf, sizeof buf, f) == NULL) {
				ok = 0;
				break;
			}
			if (*buf == '\r' || *buf == '\n')
				break;
			if (ihdr + 1 < MAX_HTTP_PARTS)
				h.headers[ihdr++] = strdup(buf);
			else {
				Debug(0, "Too many headers");
				ok = 0;
			}
		}
		if (!ok)
			break;

		Unbugger(av[2]);
		
		/* Strip leading stuff off url */
		up = strrchr(av[2], '/');
		if (up == NULL)
			up = av[2];

		h.uri = strdup(up);

		/* Split args from url */
		ua = strchr(up, '?');
		if (ua != NULL) {
			*ua++ = '\0';
			if (*ua == '\0')
				ua = NULL;
		}
		if (ua != NULL) {
			p = strdup(ua);
			for (ihdr = 0; ihdr + 1 < MAX_HTTP_PARTS; ihdr++) {
				h.argv[ihdr] = p;
				if (p == NULL)
					break;
				p = strchr(p, '&');
				if (p != NULL)
					*p++ = '\0';
			}
		}
		h.req = av[1];
		h.url = up;
		h.args = ua;
		h.proto = av[3];
		Log(0, "HTTPD %d \"%V\" \"%V\" \"%V\" \"%V\"",
		    fd, av[1], up, ua ? ua : "(null)", av[3]);
		LIST_FOREACH(uhp, &urls, next)
			if (!strcmp(uhp->verb, av[1]) && !strcmp(uhp->url, up))
				break;
		if (uhp != NULL) {
			uhp->func(f, &h, uhp->priv);
		} else {
			fprintf(f, "HTTP/1.1 404 Not Found\r\n");
			fprintf(f, "Content-Type: text/html;");
			fprintf(f, " charset=iso-8859-1\r\n");
			fprintf(f, "\r\n");
			HtmlHead(f, "URL Not Found");
			fprintf(f, "<H1>URL not found</H1>\n");
			HtmlTail(f);
		}
	} while (0);
	FREE(&h.uri);
	FREE(&h.argv[0]);
	for (ihdr = 0; h.headers[ihdr] != NULL; ihdr++)
		FREE(&h.headers[ihdr]);
	if (av != NULL)
		PhkArgvFree(av);
	(void)fflush(f);
	(void)fclose(f);
}

char *
HttpHeader(const struct http *h, const char *hdr)
{
	int i;
	unsigned l;

	l = strlen(hdr);
	for (i = 0; h->headers[i] != NULL; i++) {
		if (memcmp(hdr, h->headers[i], l))
			continue;
		if (h->headers[i][l] == ':' && h->headers[i][l + 1] == ' ')
			return (h->headers[i] + l + 2);
	}
	return (NULL);
}

char *
HttpArg(const struct http *h, const char *hdr)
{
	int i;
	unsigned l;

	l = strlen(hdr);
	for (i = 0; h->argv[i] != NULL; i++) {
		if (memcmp(hdr, h->argv[i], l))
			continue;
		if (h->argv[i][l] == '=')
			return (h->argv[i] + l + 1);
	}
	return (NULL);
}

void
Httpd200(FILE *f)
{

	fprintf(f, "HTTP/1.1 200 OK\r\n");
	fprintf(f, "Content-Type: text/html; charset=iso-8859-1\r\n");
	fprintf(f, "Cache-control: no-cache; no-store;\r\n");
	fprintf(f, "Pragma: no-cache\r\n");
	fprintf(f, "\r\n");
}

void
HtmlHead(FILE *f, const char *title)
{
	fprintf(f, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"");
     	fprintf(f, "\t\"http://www.w3.org/TR/html4/strict.dtd\">\n");
	fprintf(f, "<HTML>\n");
	fprintf(f, "<HEAD>\n");
	fprintf(f, "  <TITLE>%s</TITLE>\n", title);
	fprintf(f, "  <STYLE type=\"text/css\">\n");
	fprintf(f, "  </STYLE>\n");
	fprintf(f, "</HEAD>\n");
	fprintf(f, "<BODY>\n");
	fprintf(f, "<TABLE cellpadding=\"2\" cellspacing=\"2\"><TR>\n");
	fprintf(f, "<TD><A href=\"/data\">Data</A></TD>\n");
	fprintf(f, "<TD><A href=\"/logfile\">Logfile(old->new)</A><BR>\n");
	fprintf(f, "    <A href=\"/rlogfile\">Logfile(new->old)</A></TD>\n");
	fprintf(f, "<TD><A href=\"/config\">Configuration</A></TD>\n");
	fprintf(f, "<TD><A href=\"/alarm\">Alarm-State</A></TD>\n");
	fprintf(f, "<TD><A href=\"/alarmlog\">Alarm-Log(old->new)</A><BR>\n");
	fprintf(f, "    <A href=\"/ralarmlog\">Alarm-Log(new->old)</A></TD>\n");
	fprintf(f, "</TR></TABLE><HR>\n");
}

void
HtmlTail(FILE *f)
{

	fprintf(f, "<HR>\n");
	fprintf(f, "<P><I>Measured</I></P>\n");
	fprintf(f, "</BODY>\n");
	fprintf(f, "</HTML>\n");
}

void
InitHttpd(void)
{

	tl = TcpListenTo("2040", httpd, "HTTPD");
	assert(tl != NULL);
}
