/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: httpd_alarm.c,v 1.13 2007/03/27 12:09:53 phk Exp $
 *
 */

#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>

#include "measured.h"
#include "measured_int.h"

#define NALARMS			10

struct alarmtxt {
	unsigned			magic;
#define ALARMTXT_MAGIC			0xa282894
	time_t				when;
	struct point			*pp;
	char				*txt;
	TAILQ_ENTRY(alarmtxt)		list;
};

struct alarm {
	unsigned			magic;
#define ALARM_MAGIC			0xa213838
	struct point			*point;
	enum limit_e			f_limit;
	time_t 				f_when;
	struct alarmtxt			*f_txt;
	enum limit_e			limit[NALARMS];
	time_t				when[NALARMS];
	double 				worst[NALARMS];
	struct alarmtxt			*txt[NALARMS];
	double				high, low;
	struct alarmtxt			*high_txt, *low_txt;
	unsigned			idx;
	unsigned			count;
	unsigned			ack;
	TAILQ_ENTRY(alarm)		list;
};

static TAILQ_HEAD(, alarm) alarms = TAILQ_HEAD_INITIALIZER(alarms);
static TAILQ_HEAD(alarmtxthead, alarmtxt) alarmtxts =
    TAILQ_HEAD_INITIALIZER(alarmtxts);
static pthread_mutex_t alarm_lock;

/* alarmtxt handling -------------------------------------------------*/

static void
freealarmtxt(struct alarmtxt **atp)
{
	struct alarmtxt *at;

	at = *atp;
	*atp = NULL;
	if (at != NULL) {
		TAILQ_REMOVE(&alarmtxts, at, list);
		free(at->txt);
		FREE_OBJ(at);
	}
}

static void
alarmtxt(time_t when, struct point *pp, struct alarmtxt **atp, const char *fmt, ...)
{
	struct alarmtxt *at;
	va_list ap;

	freealarmtxt(atp);
	ALLOC_OBJ(at, ALARMTXT_MAGIC);
	at->when = when;
	at->pp = pp;
	va_start(ap, fmt);
	vasprintf(&at->txt, fmt, ap);
	va_end(ap);
	TAILQ_INSERT_TAIL(&alarmtxts, at, list);
	*atp = at;
}

/* alarmevents -------------------------------------------------------*/

static void
alarmevents(void *priv __unused, struct event_arg *e)
{
	struct point *pp;
	struct alarm *ap;
	time_t now;
	char buf[100];
	int first = 0;

	CAST_OBJ_NOTNULL(pp, e->obj, POINT_MAGIC);
	if (e->e == E_POINT_VALUE &&
	    (pp->curlim == L_NORMAL || pp->curlim == L_TIMEOUT))
		return;
	time(&now);
	pthread_cancelmutex_lock(&alarm_lock);
	TAILQ_FOREACH(ap, &alarms, list)
		if (ap->point == pp)
			break;
	if (e->e == E_POINT_LIMIT || e->e == E_POINT_CHANGED) {
		if (ap == NULL) {
			first = 1;
			ALLOC_OBJ(ap, ALARM_MAGIC);
			if (e->e == E_POINT_CHANGED)
				ap->f_limit = L_CHANGED;
			else
				ap->f_limit = pp->curlim;
			ap->f_when = now;
			
			ap->point = pp;
			ap->high = NAN;
			ap->low = NAN;
			TAILQ_INSERT_TAIL(&alarms, ap, list);
			alarmtxt(now, pp, &ap->f_txt,
			    "New state: %s Value: %s", 
			    LimitType(ap->f_limit),
			    PointFmt(pp, p_value(pp), buf, sizeof buf));
		} 
		ap->idx = (ap->idx + 1) % NALARMS;
		if (ap->when[ap->idx] != 0)
			ap->count++;
		ap->when[ap->idx] = now;
		ap->worst[ap->idx] = pp->value;
		if (e->e == E_POINT_CHANGED) {
			ap->limit[ap->idx] = L_CHANGED;
		} else if (ap->limit[ap->idx] != pp->curlim) {
			ap->limit[ap->idx] = pp->curlim;
		}
		if (!first) 
			alarmtxt(now, pp, &ap->txt[ap->idx],
			    "Changed state: %s Value: %s",
			    LimitType(ap->limit[ap->idx]),
			    PointFmt(pp, p_value(pp), buf, sizeof buf));
	}
	if (ap != NULL) {
		switch(ap->limit[ap->idx]) {
		case L_RAISE:
		case L_HIGH:
			if (isnan(ap->high) || pp->value > ap->high) {
				ap->high = pp->value;
				if (!first)
					alarmtxt(now, pp, &ap->high_txt,
					    "Maximum: %s",
						PointFmt(pp, ap->high,
						    buf, sizeof buf));
			}
			if (pp->value > ap->worst[ap->idx])
				ap->worst[ap->idx] = pp->value;
			break;
		case L_SAG:
		case L_LOW:
			if (isnan(ap->low) || pp->value < ap->low) {
				ap->low = pp->value;
				if (!first)
					alarmtxt(now, pp, &ap->low_txt,
					    "Minimum: %s",
						PointFmt(pp, ap->low,
						    buf, sizeof buf));
			}
			if (pp->value < ap->worst[ap->idx])
				ap->worst[ap->idx] = pp->value;
			break;
		default:
			break;
		}
	}
	pthread_cancelmutex_unlock(&alarm_lock);
}

/*--------------------------------------------------------------------*/

static int
AckAlarm(unsigned instance)
{
	struct alarm *ap;

	pthread_cancelmutex_lock(&alarm_lock);
	TAILQ_FOREACH(ap, &alarms, list) {
		if (PI(ap->point) == instance)
			break;
	}
	if (ap != NULL)
		ap->ack = 1;
	pthread_cancelmutex_unlock(&alarm_lock);
	return (ap == NULL ? 1 : 0);
}

/*--------------------------------------------------------------------*/

static int
ClearAlarm(unsigned instance)
{
	struct alarm *ap;
	unsigned u;

	pthread_cancelmutex_lock(&alarm_lock);
	TAILQ_FOREACH(ap, &alarms, list) {
		if (PI(ap->point) == instance)
			break;
	}
	if (ap == NULL) {
		;
	} else if (ap->limit[ap->idx] != L_NORMAL &&
	    ap->limit[ap->idx] != L_CHANGED)
		ap = NULL;
	else {
		TAILQ_REMOVE(&alarms, ap, list);
		freealarmtxt(&ap->f_txt);
		freealarmtxt(&ap->high_txt);
		freealarmtxt(&ap->low_txt);
		for (u = 0; u < NALARMS; u++)
			freealarmtxt(&ap->txt[u]);
		FREE_OBJ(ap);
	}
	pthread_cancelmutex_unlock(&alarm_lock);
	return (ap == NULL ? 1 : 0);	/*lint !e449 */
}

/*--------------------------------------------------------------------*/

#if 0
static const char *bgcol[] = {
	[L_LOW] =	" bgcolor=\"#8080ff\"",
	[L_SAG] =	" bgcolor=\"#c0c0ff\"",
	[L_RAISE] =	" bgcolor=\"#ffc0c0\"",
	[L_HIGH] =	" bgcolor=\"#ff8080\"",
	[L_NORMAL] =	" bgcolor=\"#ffffff\"",
	[L_TIMEOUT] =	" bgcolor=\"#000000\""
};
#endif

static void
render_box(FILE *f, const struct alarm *ap, unsigned u, time_t t0)
{
	time_t t;

	fprintf(f, "<TD>");
	t = ap->when[u] - t0;
	fprintf(f, "%#T<BR>", &t);
	fprintf(f, "%s<BR>", LimitType(ap->limit[u]));
	if (ap->limit[u] != L_NORMAL && ap->limit[u] != L_TIMEOUT)
		fprintf(f, p_format(ap->point), ap->worst[u]);
	else
		fprintf(f, "&nbsp;");
	fprintf(f, "<BR>");
	fprintf(f, "</TD>");
}

static void
render_alarm(FILE *f, const struct alarm *ap)
{
	unsigned u, v;
	time_t t0;
	struct tm tm;
	char buf[100];

	fprintf(f, "<TR>");
	fprintf(f, "<TD>");
	if (!ap->ack) {
		fprintf(f, "<FORM ACTION=\"/alarm_ack?point=%P\"", ap->point);
		fprintf(f, " METHOD=\"POST\">\n");
		fprintf(f, "<P><INPUT TYPE=SUBMIT VALUE=ACK></P>\n");
		fprintf(f, "</FORM>\n");
	}
	if (ap->limit[ap->idx] == L_NORMAL ||
	    ap->limit[ap->idx] == L_CHANGED) {
		fprintf(f, "<FORM ACTION=\"/alarm_clear?point=%P\"", ap->point);
		fprintf(f, " METHOD=\"POST\">\n");
		fprintf(f, "<P><INPUT TYPE=SUBMIT VALUE=CLEAR></P>\n");
		fprintf(f, "</FORM>\n");
	}
	fprintf(f, "<BR>&nbsp;</TD>");
	fprintf(f, "<TD>%P<BR>%s<BR>%s</TD>\n",
	    ap->point, p_Type(ap->point), p_label(ap->point));
	gmtime_r(&ap->f_when, &tm);
	strftime(buf, sizeof buf,
	    "%Y-%m-%d<BR>%H:%M:%S UTC<BR>", &tm);
	fprintf(f, "<TD><A HREF=\"/alarmlog?pattern=%P\">%s%s</A></TD>",
	    ap->point, buf, LimitType(ap->f_limit));
	if (isnan(ap->high)) {
		fprintf(f, "<TD>&nbsp;");
	} else {
		fprintf(f, "<TD>Max seen:");
		fprintf(f, p_format(ap->point), ap->high);
	}
	if (isnan(ap->low)) {
		fprintf(f, "<BR>&nbsp;");
	} else {
		fprintf(f, "<BR>Min seen:");
		fprintf(f, p_format(ap->point), ap->low);
	}
	if (ap->count != 0)
		fprintf(f, "<BR>Older:%u</TD>\n", ap->count);
	else
		fprintf(f, "<BR>&nbsp;</TD>\n");
	t0 = ap->f_when;
	for (v = 0; v < NALARMS; v++) {
		u = (v + ap->idx + 1) % NALARMS;
		if (ap->when[u] == 0)
			continue;
		render_box(f, ap, u, t0);
	}
	fprintf(f, "</TR>\n");	
}

static void
render_head(FILE *f)
{
	unsigned u;

	fprintf(f, "<TABLE cellpadding=\"2\" cellspacing=\"2\" border=\"1\">");
	fprintf(f,
		"<TR>"
		"<TH>Control</TH>"
		"<TH>Point</TH>"
		"<TH>First</TH>"
		"<TH>Info</TH>\n");
	for (u = 0; u < NALARMS; u++)
		fprintf(f, "<TH>Event %u</TH>", u + 1);
	fprintf(f, "</TR>");
}

static void
url_alarm(URL_ARGS)
{
	struct alarm *ap;
	struct pattern *pt;

	pt = CompilePattern(HttpArg(h, "pattern"));
	Httpd200(f);
	HtmlHead(f, "Measured - Alarms");
	
	TopoHold();
	pthread_cancelmutex_lock(&alarm_lock);
	fprintf(f, "<H1>Alarms State</H1>\n");
	if (pt != NULL)
		fprintf(f, "<H2>Pattern=%s</H2>\n",
		    HttpArg(h, "pattern"));
	fprintf(f, "<H3>New Alarms</H3>\n");
	render_head(f);
	TAILQ_FOREACH(ap, &alarms, list)
		if (!ap->ack)
			if (MatchPattern(pt, PI(ap->point)))
				render_alarm(f, ap);
	fprintf(f, "</TABLE>");
	fprintf(f, "<H3>Ack'ed alarms</H3>\n");
	render_head(f);
	TAILQ_FOREACH(ap, &alarms, list)
		if (ap->ack)
			if (MatchPattern(pt, PI(ap->point)))
				render_alarm(f, ap);
	fprintf(f, "</TABLE>");
	pthread_cancelmutex_unlock(&alarm_lock);
	TopoRel();
	HtmlTail(f);
	FreePattern(pt);
}

static void
url_ralarmlog(URL_ARGS)
{
	struct alarmtxt *at;
	char buf[100];
	struct tm tm;
	struct pattern *pt;

	pt = CompilePattern(HttpArg(h, "pattern"));
	Httpd200(f);
	HtmlHead(f, "Measured - Alarm Log (new->old)");
	
	TopoHold();
	pthread_cancelmutex_lock(&alarm_lock);
	fprintf(f, "<H1>Alarm Log (new->old)</H1>\n");
	if (pt != NULL)
		fprintf(f, "<H2>Pattern=%s</H2>\n",
		    HttpArg(h, "pattern"));
	fprintf(f, "<PRE>\n");
	TAILQ_FOREACH_REVERSE(at, &alarmtxts, alarmtxthead, list) {
		if (!MatchPattern(pt, PI(at->pp)))
			continue;
		gmtime_r(&at->when, &tm);
		strftime(buf, sizeof buf,
		    "%Y-%m-%d %H:%M:%S UTC", &tm);
		fprintf(f,
		    "%s <A HREF=\"/ralarmlog?pattern=%P\">%P</A> %-20.20s %s\n",
		    buf, at->pp, at->pp, p_label(at->pp), at->txt);
	}
	fprintf(f, "</PRE>");
	pthread_cancelmutex_unlock(&alarm_lock);
	TopoRel();
	HtmlTail(f);
	FreePattern(pt);
}

static void
url_alarmlog(URL_ARGS)
{
	struct alarmtxt *at;
	char buf[100];
	struct tm tm;
	struct pattern *pt;

	pt = CompilePattern(HttpArg(h, "pattern"));
	Httpd200(f);
	HtmlHead(f, "Measured - Alarm Log (old->new)");
	
	TopoHold();
	pthread_cancelmutex_lock(&alarm_lock);
	fprintf(f, "<H1>Alarm Log (old->new)</H1>\n");
	if (pt != NULL)
		fprintf(f, "<H2>Pattern=%s</H2>\n",
		    HttpArg(h, "pattern"));
	
	fprintf(f, "<PRE>\n");
	TAILQ_FOREACH(at, &alarmtxts, list) {
		if (!MatchPattern(pt, PI(at->pp)))
			continue;
		gmtime_r(&at->when, &tm);
		strftime(buf, sizeof buf,
		    "%Y-%m-%d %H:%M:%S UTC", &tm);
		fprintf(f,
		    "%s <A HREF=\"/alarmlog?pattern=%P\">%P</A> %-20.20s %s\n",
		    buf, at->pp, at->pp, p_label(at->pp), at->txt);
	}
	fprintf(f, "</PRE>");
	pthread_cancelmutex_unlock(&alarm_lock);
	TopoRel();
	HtmlTail(f);
	FreePattern(pt);
}

static void
url_alarm_ack(URL_ARGS)
{
	char *n;
	unsigned instance;

	do {
		n = HttpArg(h, "point");
		if (n == NULL)
			break;
		if (ParseInstance(n, &instance))
			break;
		if (priv == NULL) {
			if (AckAlarm(instance))
				break;
		} else {
			if (ClearAlarm(instance))
				break;
		}
		n = HttpHeader(h, "Referer");
		if (n != NULL) {
			fprintf(f, "HTTP/1.1 303 See Other\r\n");
			fprintf(f, "Location: %s\r\n", n);
		} else {
			fprintf(f, "HTTP/1.1 200 OK\r\n");
			fprintf(f, "\r\n");
		}
		return;
	} while (0);
	fprintf(f, "HTTP/1.1 404 Not Found\r\n\r\n");
	fprintf(f, "Not Found\r\n");
}

void
InitUrlAlarm(void)
{
	struct event_subscription *es;

	AZ(pthread_mutex_init(&alarm_lock, NULL));
	es = SubscribeEvents(alarmevents, NULL);
	assert(es != NULL);
	StartEvents(es, "limit,value,changed");
	RegisterUrl("GET", "/alarm", url_alarm, NULL);
	RegisterUrl("GET", "/alarmlog", url_alarmlog, NULL);
	RegisterUrl("GET", "/ralarmlog", url_ralarmlog, NULL);
	RegisterUrl("POST", "/alarm_ack", url_alarm_ack, NULL);
	RegisterUrl("POST", "/alarm_clear", url_alarm_ack, es);
}
