/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: httpd_auth.c,v 1.12 2007/03/27 12:09:35 phk Exp $
 *
 */

#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <configkit.h>
#include <md5.h>
#include <math.h>
#include <sys/types.h>

#include "phk_argv.h"

#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

struct auth {
	unsigned		magic;
#define AUTH_MAGIC		0x26178481
	struct ck_instance	cki;

	char			*user;
	char			A1[33];
};

struct nonce {
	unsigned		magic;
#define NONCE_MAGIC		0x9090283
	TAILQ_ENTRY(nonce)	list;
	time_t			when;
	int			level;
	char			value[33];
};

static TAILQ_HEAD(, nonce) nonce_list = TAILQ_HEAD_INITIALIZER(nonce_list);

static char *realm;
#define DEFAULT_NONCE_AGE 600
static unsigned nonce_age = DEFAULT_NONCE_AGE;
#define DEFAULT_DEBUG 0
static unsigned debug = DEFAULT_DEBUG;
static struct ck_instancehead  auth_ckih;

static __inline const char *
get_realm(void)
{
	if (realm == NULL)
		return ("user@Measured");
	return (realm);
}

static __inline void
MD5UpdStr(MD5_CTX *c, const char *s)
{

	MD5Update(c, s, strlen(s));
}

/* Nonces ------------------------------------------------------------*/

static char *
CreateNonce(int level)
{
	struct nonce	*sc, *sc2;
	time_t now;
	unsigned i;
	
	time(&now);
	TAILQ_FOREACH_SAFE(sc, &nonce_list, list, sc2) {
		CHECK_OBJ(sc, NONCE_MAGIC);
		if (sc->when + (int)nonce_age < now) {
			TAILQ_REMOVE(&nonce_list, sc, list);
			FREE_OBJ(sc);
		}
	}
	ALLOC_OBJ(sc, NONCE_MAGIC);
	sc->when = now;
	sc->level = level;
	for (i = 0; i + 1 < sizeof sc->value; i += 2)
		sprintf(sc->value + i, "%02x", random());
	sc->value[i] = '\0';
	TAILQ_INSERT_TAIL(&nonce_list, sc, list);
	return (sc->value);
}

static int
CheckNonce(const char *nonce, int level, int *stale)
{
	struct nonce	*sc, *sc2;
	time_t now;
	
	time(&now);
	*stale = 0;
	TAILQ_FOREACH_SAFE(sc, &nonce_list, list, sc2) {
		CHECK_OBJ(sc, NONCE_MAGIC);
		if (sc->when + (int)nonce_age < now) {
			TAILQ_REMOVE(&nonce_list, sc, list);
			FREE_OBJ(sc);
			continue;
		}
		if (strcmp(sc->value, nonce))
			continue;
		if (level > sc->level) {
			TAILQ_REMOVE(&nonce_list, sc, list);
			FREE_OBJ(sc);
			return (0);
		}
		if (level >= 3) {
			TAILQ_REMOVE(&nonce_list, sc, list);
			FREE_OBJ(sc);
		}
		return (1);
	}
	return (0);
}

/* Nonces ------------------------------------------------------------*/
/* calculate request-digest/response-digest as per RFC2617 */
static void
DigestCalcResponse(
	const char * HA1,		/* H(A1) */
	const char * pszNonce,		/* nonce from server */
	const char * pszNonceCount,	/* 8 hex digits */
	const char * pszCNonce,		/* client nonce */
	const char * pszQop,		/* qop-value: "", "auth", "auth-int" */
	const char * pszMethod,		/* method from the request */
	const char * pszDigestUri,	/* requested URL */
	const char * HEntity,		/* H(entity body) if qop="auth-int" */
	char *Response			/* request-digest or response-digest */
	)
{
	MD5_CTX Md5Ctx;
	char HA2Hex[33];

	// calculate H(A2)
	MD5Init(&Md5Ctx);
	MD5UpdStr(&Md5Ctx, pszMethod);
	MD5UpdStr(&Md5Ctx, ":");
	MD5UpdStr(&Md5Ctx, pszDigestUri);
	if (strcasecmp(pszQop, "auth-int") == 0) {
		MD5UpdStr(&Md5Ctx, ":");
		MD5Update(&Md5Ctx, (const void *)HEntity, 32);
	};
	(void)MD5End(&Md5Ctx, HA2Hex);
	if (debug)
		Debug(0, "HA1 = %V", HA1);
	if (debug)
		Debug(0, "HA2 = %V", HA2Hex);
		

	// calculate response
	MD5Init(&Md5Ctx);
	MD5Update(&Md5Ctx, (const void *)HA1, 32);
	MD5UpdStr(&Md5Ctx, ":");
	MD5UpdStr(&Md5Ctx, pszNonce);
	MD5UpdStr(&Md5Ctx, ":");
	if (*pszQop) {
		MD5UpdStr(&Md5Ctx, pszNonceCount);
		MD5UpdStr(&Md5Ctx, ":");
		MD5UpdStr(&Md5Ctx, pszCNonce);
		MD5UpdStr(&Md5Ctx, ":");
		MD5UpdStr(&Md5Ctx, pszQop);
		MD5UpdStr(&Md5Ctx, ":");
	};
	MD5Update(&Md5Ctx, (const void *)HA2Hex, 32);
	(void)MD5End(&Md5Ctx, Response);
};

/* XXX: This may need more work */
static char *
find_item(const char *tag, char *const *av)
{
	int i;
	unsigned l;
	char *r, *p;

	l = strlen(tag);
	for (i = 1; av[i] != NULL; i++) {
		if (memcmp(av[i], tag, l))
			continue;
		if (av[i][l] != '=')
			continue;
		r = &av[i][l + 1];
		if (*r != '"') {
			p = strchr(r, '\0');
			assert(p != NULL);	/* XXX flint */
			p--;
			if (*p == ',')
				*p = '\0';
			if (debug)
				Debug(0, "AUTH(%s) = %V", tag, r);
			return (r);
		}
		r++;
		p = strchr(r, '"');
		if (p == NULL)
			return (NULL);
		*p = '\0';
		if (debug)
			Debug(0, "AUTH(%s) = %V", tag, r);
		return (r);
	}
	if (debug)
		Debug(0, "AUTH(%s) not found", tag);
	return (NULL);
}

int
Httpd_Auth(URL_ARGS, int level)
{
	struct auth *sc;
	struct ck_instance *ip;
	char *auth;
	char **av;
	char *response;
	const char *pszNonce;
	const char *pszCNonce;
	const char *pszUser;
	const char *szNonceCount;
	const char *pszMethod;
	const char *pszQop;
	const char *pszURI;
	char Response[33];
	int stale, i, j;

	/* If no users are defined, no authentication applies */
	if (TAILQ_EMPTY(&auth_ckih))
		return (0);
	sc = NULL;
	auth = HttpHeader(h, "Authorization");
	av = NULL;
	stale = 0;
	pszUser = NULL;
	while (auth != NULL) {
		if (debug)
			Debug(0, "Auth: %V", auth);
		av = PhkArgvParse(auth, 0);
		if (av[0] != NULL) {
			Log(0, "Auth: format error: %V", av[0]);
			break;
		}
		pszUser = find_item("username", av);
		response = find_item("response", av);
		pszNonce = find_item("nonce", av);
		pszCNonce = find_item("cnonce", av);
		pszURI = find_item("uri", av);
		szNonceCount = find_item("nc", av);
		if (pszUser == NULL ||
		    response == NULL ||
		    pszNonce == NULL ||
		    pszCNonce == NULL ||
		    szNonceCount == NULL) {
			Log(0, "Auth: missing fields");
			break;
		}

		/* Find user */
		TAILQ_FOREACH(ip, &auth_ckih, next) {
			CAST_OBJ_NOTNULL(sc, ip->obj, AUTH_MAGIC);
			if (!strcmp(pszUser, sc->user))
				break;
		}
		if (ip == NULL) {
			Log(0, "Auth(%s): user not found", pszUser);
			break;
		}
		assert(sc != NULL);

		/* Match Nonce */
		if (!CheckNonce(pszNonce, level, &stale)) {
			Log(0, "Auth(%s): stale nonce", sc->user);
			break;
		}

		pszMethod = h->req;
		pszQop = "auth";

		i = strlen(h->uri);
		j = strlen(pszURI);
		if (i == j && !memcmp(pszURI, h->uri, i)) 
			;
		else if (i > j &&
		    !memcmp(pszURI, h->uri, j) && h->uri[j] == '?')
			;
		else {
			Log(0, "Auth(%s): wrong URI", sc->user);
			break;
		}

		DigestCalcResponse(sc->A1, pszNonce, szNonceCount, pszCNonce,
		    pszQop, pszMethod, pszURI, "", Response);

		if (strcmp(Response, response)) {
			Log(0, "Auth(%s): wrong password", sc->user);
			break;
		}

		PhkArgvFree(av);
		Log(0, "Auth(%s): OK", sc->user);
		return (0);
	}
	
	if (av != NULL)
		PhkArgvFree(av);
	fprintf(f, "HTTP/1.1 401 Unauthorized\r\n");
	fprintf(f, "WWW-Authenticate: Digest");
	fprintf(f, " realm=\"%s\",", get_realm());
	if (stale)
		fprintf(f, " stale,");
	fprintf(f, " qop=\"auth\",");
	fprintf(f, " algorithm=\"MD5\",");
	fprintf(f, " nonce=\"%s\"", CreateNonce(level));
	fprintf(f, "\r\n");
	fprintf(f, "\r\n");
	return (1);
}

/* ConfigKit ---------------------------------------------------------*/

int
cfg_auth_realm(CK_ARG, const char **arg_0)
{

	(void)ck_WordVar(CK_PASS, arg_0, &realm);
	return (0);
}

int
cfg_auth_timeout(CK_ARG, unsigned *arg)
{

	(void)ck_UintVar(CK_PASS, arg, &nonce_age, DEFAULT_NONCE_AGE);
	return (0);
}

int
cfg_auth_debug(CK_ARG, unsigned *arg)
{

	(void)ck_UintVar(CK_PASS, arg, &debug, DEFAULT_DEBUG);
	return (0);
}

int
cfg_auth_user_password(CK_ARG, const char **arg_0)
{
	struct auth *sc;
	char *p;
	MD5_CTX Md5Ctx;

	CAST_OBJ_NOTNULL(sc, cfa->priv, AUTH_MAGIC);
	p = strdup("{password}");
	if (!ck_WordVar(CK_PASS, arg_0, &p)) {
		free(p);
		return (0);
	}
	if (p == NULL) {
		memset(sc->A1, 0, sizeof sc->A1);
		return (0);
	}
	/* Calculate the RFC2617 HA1 for algorithm = "md5" */
	MD5Init(&Md5Ctx);
	MD5UpdStr(&Md5Ctx, sc->user);
	MD5UpdStr(&Md5Ctx, ":");
	MD5UpdStr(&Md5Ctx, get_realm());
	MD5UpdStr(&Md5Ctx, ":");
	MD5UpdStr(&Md5Ctx, p);
	(void)MD5End(&Md5Ctx, sc->A1);
	free(p);
	return (0);
}

struct ck_instancehead *
cfg_auth_user_head(CK_ARG)
{

	return (&auth_ckih);
}

struct ck_instance *
cfg_auth_user_new(CK_ARG, const char *user)
{
	struct auth *sc;

	ALLOC_OBJ(sc, AUTH_MAGIC);
	sc->cki.obj = sc;
	sc->user = strdup(user);
	assert(sc->user != NULL);
	return (&sc->cki);
}

int
cfg_auth_user_destroy(CK_ARG)
{
	struct auth *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, AUTH_MAGIC);
	free(sc->user);
	return (0);
}

void
InitHttpdAuth(void)
{

	TAILQ_INIT(&auth_ckih);
}
