/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: httpd_pages.c,v 1.12 2006/09/09 12:32:26 phk Exp $
 *
 * XXX:
 *
 *	.print [opts] pattern
 *		-format <str>
 *		-{high,raise,sag,low,age,timeout,label...}
 *	.if [expr]
 *	.elif [expr]
 *	.else
 *	.endif
 *		expr:	expr1
 *			expr1 || expr
 *
 *		expr1:	expr2
 *			expr2 || expr
 *
 *		expr2:	expr3
 *			expr3 <rel> expr3
 *				<,>,<=,>=,=,!=,==
 *
 *		expr3:  numeric_constant
 *			function '(' <point> ')'
 *				value()
 *				high()
 *				low()
 *				alarm()
 *				--//--
 *
 *	Configure page_dir with configkit
 *	Configure debug with configkit
 *	Insert debug in HTML output as comments ?
 *
 *	What to do for point not found ?
 *
 *	Free memory
 *
 * DOC:
 * 	"" parsed by Argv, quoting necessary for nested calls.
 *
 */

#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>

#include "phk_argv.h"

#include "measured.h"
#include "measured_int.h"

static const char *page_dir = "/etc/measured_pages/";

/*--------------------------------------------------------------------*/

struct line {
	TAILQ_ENTRY(line)	list;
	char			*line;
};

struct macro {
	char			*name;
	unsigned		narg;
	TAILQ_ENTRY(macro)	list;
	TAILQ_HEAD(,line)	linelist;
};

struct pages {
	FILE			*fi;
	FILE			*fo;
	char			buf[BUFSIZ];
	char			**av;
	TAILQ_HEAD(,line)	linelist;
	TAILQ_HEAD(,macro)	macrolist;
	int			err;
};

/*--------------------------------------------------------------------*/

static void do_line(struct pages *p);

/*--------------------------------------------------------------------*/

static int
GetLine(struct pages *p)
{
	struct line *l;

	if (p->err)
		return (-1);

	if (TAILQ_EMPTY(&p->linelist)) {
		if (fgets(p->buf, sizeof p->buf, p->fi) == 0)
			return (-1);
	} else {
		l = TAILQ_FIRST(&p->linelist);
		TAILQ_REMOVE(&p->linelist, l, list);
		strcpy(p->buf, l->line);	
		free(l->line);
		free(l);
	}
	if (p->av != NULL)
		PhkArgvFree(p->av);
	p->av = PhkArgvParse(p->buf, 0);
	if (p->av[0] != NULL) {
		fprintf(p->fo, "Parse error: %s\n", p->av[0]);
		return (-1);
	}
	return (0);
}

/*====================================================================*/

static void
do_macro(struct pages *p)
{
	struct macro	*mp;
	struct line	*lp;

	if (p->av[2] == NULL) {
		fprintf(p->fo, "Error: .macro has no name\n");
		p->err = 1;
		return;
	}
	mp = calloc(sizeof *mp, 1);
	assert(mp != NULL);
	TAILQ_INIT(&mp->linelist);
	mp->name = strdup(p->av[2]);
	assert(mp->name != NULL);
	if (p->av[3] != NULL)
		mp->narg = strtoul(p->av[3], NULL, 0);
	while (1) {
		if (GetLine(p)) {
			/* XXX */
			free(mp->name);
			free(mp);
			return;
		}
		if (p->av[1] == NULL) {
			fprintf(p->fo, "Error: EOF in .macro\n");
			free(mp->name);
			free(mp);
			return;
		}
		if (!strcmp(p->av[1], ".endmacro"))
			break;
		lp = calloc(sizeof *lp, 1);
		assert(lp != NULL);
		lp->line = strdup(p->buf);
		assert(lp->line != NULL);
		TAILQ_INSERT_HEAD(&mp->linelist, lp, list);
	}
	TAILQ_INSERT_TAIL(&p->macrolist, mp, list);
}

/*--------------------------------------------------------------------*/

static void
do_call(struct pages *p)
{
	struct macro *mp;
	unsigned u;
	struct line *lp, *lp2;
	char buf[BUFSIZ];
	char *q, *r, *s;

	if (p->av[2] == NULL) {
		fprintf(p->fo, "Error: .call without name\n");
		p->err = 1;
		return;
	}
	TAILQ_FOREACH(mp, &p->macrolist, list) {
		if (!strcmp(mp->name, p->av[2]))
			break;
	}
	if (mp == NULL) {
		fprintf(p->fo, "Error: .call %s: macro not found\n", p->av[2]);
		p->err = 1;
		return;
	}
	for (u = 0; u < mp->narg; u++) {
		if (p->av[3 + u] != NULL)
			continue;
		fprintf(p->fo, "Error: .call %s: too few arguments\n",
		    mp->name);
		p->err = 1;
		return;
	}
	TAILQ_FOREACH(lp, &mp->linelist, list) {
		q = buf;
		for (s = lp->line; *s; s++) {
			if (*s != '$') {
				*q++ = *s;
				continue;
			}
			s++;
			if (*s == '$') {
				*q++ = *s;
				continue;
			}
			if (!isdigit(*s) || *s == '0') {
				fprintf(p->fo,
				    "Error: .call %s: '$' followed by junk\n",
				    mp->name);
				p->err = 1;
				return;
			}
			u = *s - '1';
			if (u > mp->narg) {
				fprintf(p->fo,
				    "Error: .call %s: '$%u' no such arg\n",
				    mp->name, u + 1);
				p->err = 1;
				return;
			}
			
			for (r = p->av[3 + u]; *r;)
				*q++ = *r++;
		}
		*q = '\0';
		lp2 = calloc(sizeof *lp2, 1);
		assert(lp2 != NULL);
		lp2->line = strdup(buf);
		assert(lp2->line != NULL);
		TAILQ_INSERT_HEAD(&p->linelist, lp2, list);
	}
}

/*====================================================================*/

struct exp {
	struct pages	*p;
	char		**av;
};

#define BAIL(e)	do { if ((e)->p->err) return (0); } while (0)

/*--------------------------------------------------------------------*/

static int
experr(const struct exp *e, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "Expression Error: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	fprintf(e->p->fo, "Expression Error: ");
	vfprintf(e->p->fo, fmt, ap);
	fprintf(e->p->fo, "\n");
	e->p->err = 1;
	va_end(ap);
	return (0);
}

/*--------------------------------------------------------------------*/

static double
expr3(struct exp *e)
{
	char *p, *q;
	double a;
	unsigned instance;
	struct point *pt;


	if (e->av[0] == NULL)
		return (experr(e, "Expected constant or function"));
	a = strtod(e->av[0], &q);
	if (q != e->av[0] && *q == '\0') {
		e->av++;
		return (a);
	}
	p = strchr(e->av[0], '(');
	if (p == NULL)
		return (experr(e,
		    "Expected constant or function, got \"%V\"", e->av[0]));
	q = strchr(p + 1, '\0');
	assert(q != NULL);
	if (q[-1] != ')')
		return (experr(e, "Missing ')' in \"%V\"\n", e->av[0]));
	q[-1] = '\0';
	*p++ = '\0';
	if (ParseInstance(p, &instance))
		return (experr(e, "Expected point got \"%V\"\n", p));

	/* XXX: locking, hold topo */
	pt = FindPoint(instance);
	if (pt == NULL)
		return (experr(e, "point \"%V\" Not found", p));
	
	if (!strcmp(e->av[0], "value")) {
		a = p_value(pt);
		e->av++;
	} else 
		return (experr(e, "Unknown function \"%V\"\n", e->av[0]));
	return (a);
}

/*--------------------------------------------------------------------*/

static int
expr2(struct exp *e)
{
	double a, b;
	char *rel;

	a = expr3(e);
	BAIL(e);
	if (e->av[0] == NULL) {
		return (a != 0.0);
	}
	rel = e->av[0];
	e->av++;
	b = expr3(e);
	BAIL(e);
	if (!strcmp(rel, ">"))  { return (a > b); }
	if (!strcmp(rel, ">=")) { return (a >= b); }
	if (!strcmp(rel, "<=")) { return (a <= b); }
	if (!strcmp(rel, "<")) { return (a < b); }
	if (!strcmp(rel, "=")) { return (a == b); }
	if (!strcmp(rel, "==")) { return (a == b); }
	if (!strcmp(rel, "!=")) { return (a != b); }
	return (experr(e, "Unknown relational got \"%V\"\n", rel));
}

/*--------------------------------------------------------------------*/

static int
expr1(struct exp *e)
{
	int i;

	i = expr2(e);
	BAIL(e);
	if (!i || (e->av[0] == NULL))
		return (i);
	if (strcmp(e->av[0], "&&"))
		return (experr(e, "expected \"&&\" got \"%V\"\n", e->av[0]));
	e->av++;
	if (e->av[0] == NULL)
		return (experr(e, "expected something after \"||\""));
	i = expr1(e);
	return (i);
}


/*--------------------------------------------------------------------*/

static int
expr0(struct exp *e)
{
	int i;

	i = expr1(e);
	BAIL(e);
	if (i || (e->av[0] == NULL))
		return (i);
	if (strcmp(e->av[0], "||"))
		return (experr(e, "expected \"||\" got \"%V\"\n", e->av[0]));
	e->av++;
	if (e->av[0] == NULL)
		return (experr(e, "expected something after \"||\""));
	i = expr0(e);
	return (i);
}

/*--------------------------------------------------------------------*/

static int
expr(struct pages *p)
{
	struct exp e;
	int i;

	memset(&e, 0, sizeof e);
	e.p = p;
	e.av = p->av + 2;
	i = expr0(&e);
	return (i);
}

/*--------------------------------------------------------------------*/

static void
do_if(struct pages *p)
{
	int i;

	i = expr(p);
	while (p->err == 0) {
		if (i) {
			while (!GetLine(p)) {
				if (!strcmp(p->av[1], ".elif"))
					break;
				if (!strcmp(p->av[1], ".else"))
					break;
				if (!strcmp(p->av[1], ".endif"))
					break;
				do_line(p);
			}
			while (strcmp(p->av[1], ".endif"))
				if (GetLine(p))
					break;
			return;
		}
		while (!GetLine(p)) {
			if (!strcmp(p->av[1], ".elif")) {
				i = expr(p);
				break;
			} else if (!strcmp(p->av[1], ".else")) {
				i = 1;
				break;
			} else if (!strcmp(p->av[1], ".endif"))
				return;
		}
	}
}

/*====================================================================*/

static void
do_print(struct pages *p)
{
	struct pattern *pt;
	struct site *sp;
	struct group *gp;
	struct point *pp;
	const char *fmt;

	if (p->av[2] == NULL) {
		fprintf(p->fo, "Error: Missing pattern argument\n");
		p->err = 1;
		return;
	}
	pt = CompilePattern(p->av[2]);
	if (pt == NULL) {
		fprintf(p->fo, "Error: illegal pattern \"%V\"\n", p->av[2]);
		p->err = 1;
		return;
	}
	TopoHold();
	TAILQ_FOREACH(sp, &all_sites, list) {
		TAILQ_FOREACH(gp, &sp->groups, list) {
			TAILQ_FOREACH(pp, &gp->points, list) {
				if (!MatchPattern(pt, pp->instance))
					continue;
				fmt = p_format(pp);
				if (p->av[3] != NULL)
					fmt = fmtcheck(p->av[3], fmt);
				fprintf(p->fo, fmt, pp->value);
			}
		}
	}
	TopoRel();
	FreePattern(pt);
}

/*--------------------------------------------------------------------*/

static void
do_line(struct pages *p)
{

	if (p->buf[0] != '.') {
		fprintf(p->fo, "%s", p->buf);
		return;
	}
	if (!strcmp(p->av[1], ".if")) {
		do_if(p);
	} else if (!strcmp(p->av[1], ".print")) {
		do_print(p);
	} else if (!strcmp(p->av[1], ".macro")) {
		do_macro(p);
	} else if (!strcmp(p->av[1], ".call")) {
		do_call(p);
	} else {
		fprintf(p->fo, "Unknown \"%V\"\n", p->av[1]);
		p->err = 1;
	}
}

/*--------------------------------------------------------------------*/

static void
url_pages(URL_ARGS)
{
	char *n, *m;
	FILE *fi;
	struct pages p;
	struct macro *mp;
	struct line *lp;


	Httpd200(f);
	HtmlHead(f, "Measured - Pages");

	n = HttpArg(h, "page");
	if (n == NULL) {
		/* XXX: could look for "index.html" here */
		fprintf(f, "no page specified\n");
		HtmlTail(f);
		return;
	}
	if (strchr(n, '/')) {
		fprintf(f, "Illegale page \"%V\"\n", n);
		HtmlTail(f);
		return;
	}
	asprintf(&m, "%s/%s", page_dir, n);
	assert(m != NULL);
	fi = fopen(m, "r");
	if (fi == NULL) {
		fprintf(f, "Page \"%V\" not found %M\n", n, errno);
		HtmlTail(f);
		free(m);
		return;
		
	}

	memset(&p, 0, sizeof p);
	TAILQ_INIT(&p.linelist);
	TAILQ_INIT(&p.macrolist);
	p.fi = fi;
	p.fo = f;
	free(m);
	while (!GetLine(&p))
		do_line(&p);
	PhkArgvFree(p.av);
	while (!TAILQ_EMPTY(&p.macrolist)) {
		mp = TAILQ_FIRST(&p.macrolist);
		TAILQ_REMOVE(&p.macrolist, mp, list);
		free(mp->name);
		while (!TAILQ_EMPTY(&mp->linelist)) {
			lp = TAILQ_FIRST(&mp->linelist);
			TAILQ_REMOVE(&mp->linelist, lp, list);
			free(lp->line);
			free(lp);
		}
		free(mp);
	}
	while (!TAILQ_EMPTY(&p.linelist)) {
		lp = TAILQ_FIRST(&p.linelist);
		TAILQ_REMOVE(&p.linelist, lp, list);
		free(lp->line);
		free(lp);
	}

	fclose(fi);
	
	HtmlTail(f);
}

void
InitUrlPages(void)
{

	RegisterUrl("GET", "/pages", url_pages, NULL);
}
