/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: log.c,v 1.18 2008/02/17 09:29:13 phk Exp $
 *
 */

#include <stdio.h>
#include <syslog.h>
#include <stdarg.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include "measured.h"
#include "measured_int.h"

static struct pattern *debug_pattern;
static struct pattern *log_pattern;
static pthread_mutex_t log_lock;

#define LOGSIZE		100

static time_t logtime[LOGSIZE];
static const char *logpfx[LOGSIZE];
static unsigned logid[LOGSIZE];
static char *logfifo[LOGSIZE];
static unsigned logidx;

static void
logit(unsigned id, const char *pfx, int syslogid, event_t e,
    const char *fmt, va_list ap)
{
	char *p, *q;

	vasprintf(&p, fmt, ap);
	assert(p != NULL);
	asprintf(&q, "%s %I %s", pfx, id, p);
	assert(q != NULL);
	AZ(pthread_mutex_lock(&log_lock));
	if (logfifo[logidx] != NULL)
		free(logfifo[logidx]);
	time(&logtime[logidx]);
	logpfx[logidx] = pfx;
	logfifo[logidx] = p;
	logid[logidx++] = id;
	logidx %= LOGSIZE;
	PostEvent(e, p, id);
	AZ(pthread_mutex_unlock(&log_lock));
	syslog(syslogid, "%s", q);
	if (d_flag) {
		printf("%s\n", q);
		AZ(fflush(stdout));
	}
	if (syslogid == LOG_CRIT)
		fprintf(stderr, "%s\n", q);
	free(q);
}

void
Debug(unsigned id, const char *fmt, ...)
{
	va_list ap;

	if (!MatchPattern(debug_pattern, id))
		return;
	va_start(ap, fmt);
	logit(id, "DEBUG", LOG_DEBUG, E_MSG_DEBUG, fmt, ap);
	va_end(ap);
}

void
Log(unsigned id, const char *fmt, ...)
{
	va_list ap;

	if (!MatchPattern(log_pattern, id))
		return;
	va_start(ap, fmt);
	logit(id, "LOG  ", LOG_INFO, E_MSG_LOG, fmt, ap);
	va_end(ap);
}

void
Notice(unsigned id, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	logit(id, "NOTE ", LOG_NOTICE, E_MSG_NOTICE, fmt, ap);
	va_end(ap);
}

void
Error(unsigned id, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	logit(id, "ERROR", LOG_CRIT, E_MSG_ERROR, fmt, ap);
	va_end(ap);
}

void
__Fatal(const char *func, const char *file, int line, const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf, fmt, ap);
	logit(0, "FATAL", LOG_CRIT, E_MSG_ERROR, buf, ap);
	va_end(ap);
	__assert(func, file, line, buf);
	while (1)
		;
}

/* -------------------------------------------------------------------*/

static void
url_logfile(URL_ARGS)
{
	unsigned u, v;
	char buf[100];
	char buf2[256];
	struct tm tm;

#if 0
	if (Httpd_Auth(f, h, priv, 1))
		return;
#endif
	Httpd200(f);
	HtmlHead(f, "Measured Logfile");
	fprintf(f, "<PRE>\n");
	for (u = 0; u < LOGSIZE; u++) {
		if (priv != NULL)
			v = (LOGSIZE + logidx - u) % LOGSIZE;
		else
			v = (u + logidx) % LOGSIZE;
		if (logfifo[v] == NULL)
			continue;
		(void)strftime(buf, sizeof buf, "%Y-%m-%d %H:%M:%S",
		    gmtime_r(&logtime[v], &tm));
		AZ(pthread_mutex_lock(&log_lock));
		snprintf(buf2, sizeof buf2, "%s %-5s %I %s\n",
		    buf, logpfx[v], logid[v], logfifo[v]);
		AZ(pthread_mutex_unlock(&log_lock));
		fprintf(f, "%s", buf2);
	}
	fprintf(f, "</PRE>\n");
	HtmlTail(f);
}

/* -------------------------------------------------------------------*/

void
InitLog(void)
{

	AZ(pthread_mutex_init(&log_lock, NULL));

	openlog("measured", LOG_NDELAY, LOG_LOCAL4);
	if (d_flag)
		debug_pattern = CompilePattern("*");
	else
		debug_pattern = CompilePattern("!*");
	log_pattern = CompilePattern("*");
	Log(0, "Starting pid=%d uid=%d", getpid(), getuid());
	RegisterUrl("GET", "/logfile", url_logfile, NULL);
	RegisterUrl("GET", "/rlogfile", url_logfile, url_logfile);
}
