/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: master.c,v 1.64 2007/09/05 13:15:10 phk Exp $
 *
 */

#define DO_VALGRIND
#undef DO_VALGRIND

#include <stdio.h>
#include <assert.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <configkit.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/endian.h>

#include "phk_argv.h"
#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

struct master_sc {
	unsigned			magic;
#define MASTER_SC_MAGIC			0x2924945
	int				fd;
	int				fd2;
	struct event_subscription	*sub;
	struct pattern			*pattern;
	char				**av;
};


static void
master_event(void *priv, struct event_arg *e)
{
	struct master_sc *sc;

	CAST_OBJ_NOTNULL(sc, priv, MASTER_SC_MAGIC);
	if (sc->pattern && !MatchPattern(sc->pattern, e->instance))
		return;
	fdprintf(sc->fd2, "E %I %s %s\n",
	    e->instance, e->name, Event_String(e));
}

static int
master_line(void *p, const char *l)
{
	struct master_sc *sc;
	unsigned u;

	CAST_OBJ_NOTNULL(sc, p, MASTER_SC_MAGIC);

	u = strlen(l);
	if (u == 0)
		return (0);

	if (sc->av != NULL)
		PhkArgvFree(sc->av);
	sc->av = PhkArgvParse(l, 0);
	if (sc->av[0] != NULL) {
		Log(0, "MASTER %d ERROR SYNTAX <%s>", sc->fd, sc->av[0]);
		fdprintf(sc->fd2, "ERROR syntax %s\n", sc->av[0]);
		return (-1);
	}

	if (sc->av[1] == NULL)
		return (0);

	if (!strcasecmp(sc->av[1], "PATTERN")) {
		if (sc->av[2] == NULL) {
			Log(0, "MASTER %d PATTERN missing arguments",
			    sc->fd);
			fdprintf(sc->fd2, "ERROR missing arguments\n");
			return (-1);
		}
		if (sc->pattern != NULL)
			FreePattern(sc->pattern);
		sc->pattern = CompilePattern(sc->av[2]);
		return (0);
	}
	if (!strcasecmp(sc->av[1], "EVENTS")) {
		if (sc->sub == NULL)
			sc->sub = SubscribeEvents(master_event, sc);
		if (sc->av[2] == NULL)
			StartEvents(sc->sub, "all");
		else if (StartEvents(sc->sub, sc->av[2])) {
			Log(0, "MASTER %d EVENT not recognized", sc->fd);
			fdprintf(sc->fd2, "ERROR wrong arguments\n");
			return (-1);
		}
		return (0);
	}

	if (!strcasecmp(sc->av[1], "CONTROL")) {
		if (sc->av[2] == NULL || sc->av[3] == NULL) {
			Log(0, "MASTER %d CONTROL missing arguments(s)",
			    sc->fd);
			fdprintf(sc->fd2, "ERROR missing arguments\n");
			return (-1);
		}
		if (ParseInstance(sc->av[2], &u)) {
			Log(0, "MASTER %d CONTROL bugus instance",
			    sc->fd);
			fdprintf(sc->fd2, "ERROR bogus argument \n");
			return (-1);
		}
		ControlPointI(u, strtod(sc->av[3], NULL));
		return (0);
	} 
	Log(0, "MASTER %d UNKNOWN ", sc->fd);
	fdprintf(sc->fd2, "ERROR unknown command\n");
	return (-1);
}

#ifdef DO_VALGRIND
#include "/usr/local/include/valgrind/memcheck.h"
#endif

static void
url_master(URL_ARGS)
{
	struct master_sc *sc;
	int i;
	struct lineup *lu;

	ALLOC_OBJ(sc, MASTER_SC_MAGIC);

	/* We can't read and write on the same fd with pthreads, so dup(2) */
	sc->fd = fileno(f);
	sc->fd2 = dup(sc->fd);
	if (sc->fd2 < 0) {
		FREE_OBJ(sc);
		return;
	}

	Httpd200(f);
	lu = NewLineup(sc, master_line);
	while (1) {
		if (fflush(f))
			break;
		i = LineUpFile(f, lu);
		if (i != 0)
			break;
	}
	if (sc->sub != NULL)
		UnsubscribeEvents(sc->sub);
	if (sc->av != NULL)
		PhkArgvFree(sc->av);
	(void)close(sc->fd2);
	(void)fclose(f);
	Log(0, "MASTER %d CLOSED", sc->fd);
	if (sc->pattern != NULL)
		FreePattern(sc->pattern);
	FREE_OBJ(sc);
#ifdef DO_VALGRIND
	VALGRIND_DO_LEAK_CHECK
#endif
}

void
InitMaster(void)
{

	RegisterUrl("GET", "/master", url_master, NULL);
}
