/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: measure_mib.c,v 1.32 2007/11/30 08:59:20 phk Exp $
 *
 * BSNMP module for MEASURE related mibs.
 */

/*lint -e713 */
/*lint -e737 */ /* Loss of sign in promotion from int to unsigned int */

#include <bsnmp/asn1.h>
#include <bsnmp/snmp.h>
#include <bsnmp/snmpmod.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <netdb.h>
#include <wchar.h>
#include <printf.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/endian.h>
#include <sys/queue.h>

#include "measure_tree.h"
#include "../measured_api.h"
#include "phk_obj.h"
#include "phk_argv.h"
#include "phk_tcp.h"
#include "measure_oid.h"
#include "oids.h"

static struct lmodule *lmmod;
static void *timer_hd;
static int slave_fd = -1;
static void *slave_fd_id;
static unsigned slave_state;
static uint32_t	trap_seq;
static struct lineup *slave_lup;
static char **slave_av;

/* This must match measureSystemState from the MIB */

static enum sys_state {
	ST_DISC	=	1,
	ST_CONN =	2,
	ST_ONL =	3,
	ST_AUTHERR =	4,
	ST_PROTOERR = 	5
	}		sysstate = ST_DISC;

struct point {
	unsigned		magic;
#define POINT_MAGIC		0x2383838
	TAILQ_ENTRY(point)	list;
	unsigned		instance;
	double			value;
	double			when;
	enum limit_e		limit;
};

static TAILQ_HEAD(,point)	pointlist =
    TAILQ_HEAD_INITIALIZER(pointlist);

struct state {
	unsigned		magic;
#define STATE_MAGIC		0x67b29bed
	TAILQ_ENTRY(state)	list;
	unsigned		stn;
	unsigned		pri;
	char			*msg;
};

static TAILQ_HEAD(,state)	statelist =
    TAILQ_HEAD_INITIALIZER(statelist);

static void send_trap_state(const struct state *st, const struct point *pt);
static void send_trap_sysstate(enum sys_state);

/* TCP connection management -----------------------------------------*/

static void
close_tcp(void)
{
	struct point *pt;

	fprintf(stderr, "close_tcp\n");
	fd_deselect(slave_fd_id);
	close(slave_fd);
	slave_fd = -1;

	if (slave_av != NULL)
		PhkArgvFree(slave_av);
	slave_av = NULL;
	if (slave_lup != NULL)
		DestroyLineup(slave_lup);
	slave_lup = NULL;
	while (1) {
		pt = TAILQ_FIRST(&pointlist);
		if (pt == NULL)
			break;
		TAILQ_REMOVE(&pointlist, pt, list);
		FREE_OBJ(pt);
	}
	send_trap_sysstate(ST_DISC);
}

static void
slave_err(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "Slave protocol error:");
	vfprintf(stderr, fmt, ap);	
	fprintf(stderr, "\n");
	va_end(ap);
	close_tcp();
	send_trap_sysstate(ST_PROTOERR);
}

static int
slave_proto(void *priv __unused, const char *l)
{
	unsigned u, v, in;
	struct point *pt, *pt2;
	enum limit_e lim;
	struct state *st, *st2;

	if (slave_av != NULL)
		PhkArgvFree(slave_av);

	slave_av = PhkArgvParse(l, 0);
	if (slave_av[0] != NULL) {
		slave_err("slave proto parse err: %s", slave_av[0]);
		return (-1);
	}
	if (slave_state == 0) {
		if (strcmp(slave_av[1], "HTTP/1.1")) {
			slave_err("Not HTTP reply", slave_av[1]);
			return (-1);
		}
		if (strcmp(slave_av[2], "200")) {
			slave_err("Not HTTP 200 reply", slave_av[2]);
			return (-1);
		}
		slave_state = 1;
		return (0);
	}

	if (slave_state == 1) {
		/* skip HTTP headers */
		if (slave_av[1] == NULL) {
			send_trap_sysstate(ST_ONL);
			slave_state = 2;
		}
		return (0);
	}

	if (strcmp(slave_av[1], "E")) {
		slave_err("Got non 'E' line %V", slave_av[1]);
		return (-1);
	}

	if (ParseInstance(slave_av[2], &in) < 0) {
		slave_err("invalid instance %V", slave_av[2]);
		return (-1);
	}
	if (!strcmp(slave_av[3], "P") && !strcmp(slave_av[4], "NEW")) {
		ALLOC_OBJ(pt, POINT_MAGIC);
		pt->instance = in;
		pt->limit = L_NORMAL;
		TAILQ_FOREACH(pt2, &pointlist, list)
			if (pt2->instance > in)
				break;
		if (pt2 != NULL)
			TAILQ_INSERT_BEFORE(pt2, pt, list);
		else
			TAILQ_INSERT_TAIL(&pointlist, pt, list);
		return (0);
	}
	if (!strcmp(slave_av[3], "P") && !strcmp(slave_av[4], "DESTROY")) {
		TAILQ_FOREACH(pt, &pointlist, list)
			if (pt->instance == in)
				break;
		assert(pt != NULL);
		TAILQ_REMOVE(&pointlist, pt, list);
		free(pt);
		return (0);
	}
	if (!strcmp(slave_av[3], "P") && !strcmp(slave_av[4], "VALUE")) {
		TAILQ_FOREACH(pt, &pointlist, list)
			if (pt->instance == in)
				break;
		assert(pt != NULL);
		pt->value = strtod(slave_av[5], NULL);
		pt->when = strtod(slave_av[6], NULL);
		return (0);
	}
	if (!strcmp(slave_av[3], "P") && !strcmp(slave_av[4], "LIMIT")) {
		TAILQ_FOREACH(pt, &pointlist, list)
			if (pt->instance == in)
				break;
		assert(pt != NULL);
		lim = TypeLimit(slave_av[5]);
		if (lim == L_INVALID) {
			slave_err("invalid limit \"%V\"", slave_av[5]);
			return (-1);
		}
		if (pt->limit == lim)
			return (0);
		pt->limit = lim;
		return (0);
	}
	if (!strcmp(slave_av[3], "T") && !strcmp(slave_av[4], "CHG")) {
		u = strtoul(slave_av[5], NULL, 0);
		v = strtoul(slave_av[6], NULL, 0);
		TAILQ_FOREACH(st, &statelist, list)
			if (st->stn >= u)
				break;
		if (st == NULL || st->stn > u) {
			ALLOC_OBJ(st2, STATE_MAGIC);
			st2->stn = u;
			if (st != NULL)
				TAILQ_INSERT_BEFORE(st, st2, list);
			else
				TAILQ_INSERT_TAIL(&statelist, st2, list);
			st2->msg = strdup(slave_av[7]);
			assert(st2->msg != NULL);
			st = st2;
		}
		if (strcmp(st->msg, slave_av[7])) {
			free(st->msg);
			st->msg = strdup(slave_av[7]);
			assert(st->msg != NULL);
		}
		TAILQ_FOREACH(pt2, &pointlist, list)
			if (pt2->instance == in)
				break;
		st->pri = v;
		send_trap_state(st, pt2);
		return (0);
	}
	slave_err("unexpected request %V %V", slave_av[3], slave_av[4]);
	return (-1);
}

static void
slave_func(int fd, void *argu __unused)
{

	if (LineUpFd(fd, slave_lup) != 0)
		close_tcp();
}

static void
measure_timer(void *arg)
{
	int i;
	char *buf;

	(void)arg;
	timer_hd = timer_start(1 * 100, measure_timer, NULL, lmmod);

	if (slave_fd >= 0)
		return;

	slave_fd = PhkTcpOpen("127.0.0.1", "2040");
	if (slave_fd <= 0)
		return;

	send_trap_sysstate(ST_CONN);

	slave_lup = NewLineup(NULL, slave_proto);
	slave_fd_id = fd_select(slave_fd, slave_func, NULL, lmmod);

	slave_state = 0;

	/*
	 * XXX: regretable workaround, fix in ../httpd.c by giving each
	 * fd a lineup from the start.
	 */
	asprintf(&buf, "GET /master HTTP/1.1\n\n");
	assert(buf != NULL);
	i = write(slave_fd, buf, strlen(buf));
	if (i != strlen(buf)) {
		close_tcp();
		return;
	}
	free(buf);

	usleep(100000);
	asprintf(&buf, 
	    // "GET /master HTTP/1.1\r\n\r\n"
	    "PATTERN *\r\n"
	    "EVENTS \"point state\"\n");
	assert(buf != NULL);
	i = write(slave_fd, buf, strlen(buf));
	if (i != strlen(buf)) {
		close_tcp();
		return;
	}
	free(buf);
}

/* SNMP PDU handling -------------------------------------------------*/


static void
DmpReq(u_int sub, u_int iidx, const struct snmp_value *v)
{
	u_int u;

	return;
	printf("sub %d, iidx %d  {", sub, iidx);
	for (u = 0; u < v->var.len; u++)
		printf(".%d", v->var.subs[u]);
	printf("}\n");
	
}

/*
 * Increment the trap-count and return a snmp_value for it.
 */
static struct snmp_value *
trap_seq_oid(void)
{
	static struct snmp_value trapseq;

	trapseq.var = oid_measureSystemTrapSeq;
	trapseq.syntax = SNMP_SYNTAX_GAUGE;
	trapseq.v.integer = ++trap_seq;

	return(&trapseq);
}

static void
send_trap_sysstate(enum sys_state st)
{
	static struct snmp_value trapoid, *seqoid;

	sysstate = st;
	trapoid.var = oid_measureSystemState;
	trapoid.syntax = SNMP_SYNTAX_INTEGER;
	trapoid.v.integer = (int)st;

	seqoid = trap_seq_oid();
	snmp_send_trap(&oid_measureTrapConn, seqoid, &trapoid, NULL);
}

static void
send_trap_state(const struct state *st, const struct point *pt)
{
	static struct snmp_value trapoid[9], *seqoid;
	unsigned n;
	time_t t;

	CHECK_OBJ_NOTNULL(st, STATE_MAGIC);
	if (pt != NULL)
		CHECK_OBJ(pt, POINT_MAGIC);

	n = 0;
	trapoid[n].var = oid_measureStateTrap;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = st->stn;
	n++;
	trapoid[n].var = oid_measureStatePriority;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = st->pri;
	n++;
	trapoid[n].var = oid_measureStateDesc;
	trapoid[n].syntax = SNMP_SYNTAX_OCTETSTRING;
	trapoid[n].v.octetstring.octets = strdup(st->msg);
	assert(trapoid[n].v.octetstring.octets != NULL);
	trapoid[n].v.octetstring.len = strlen(st->msg);
	n++;
	trapoid[n].var = oid_measurePointSite;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = pt == NULL ? 0 : SITE_OF(pt->instance);
	n++;
	trapoid[n].var = oid_measurePointGroup;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = pt == NULL ? 0 : GROUP_OF(pt->instance);
	n++;
	trapoid[n].var = oid_measurePointPoint;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = pt == NULL ? 0 : POINT_OF(pt->instance);
	n++;
	trapoid[n].var = oid_measurePointValue;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = pt == NULL ? 0 : (int)round(pt->value * 1000.0);
	n++;
	trapoid[n].var = oid_measurePointAlarm;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	trapoid[n].v.integer = pt == NULL ? 0 : (int)pt->limit;
	n++;
	trapoid[n].var = oid_measurePointAge;
	trapoid[n].syntax = SNMP_SYNTAX_INTEGER;
	time(&t);
	trapoid[n].v.integer = pt == NULL ? 0 : t - (int)pt->when;

	seqoid = trap_seq_oid();

	snmp_send_trap(&oid_measurePointStateTrap, seqoid,
	    &trapoid[0],
	    &trapoid[1],
	    &trapoid[2],
	    &trapoid[3],
	    &trapoid[4],
	    &trapoid[5],
	    &trapoid[6],
	    &trapoid[7],
	    &trapoid[8],
	    NULL);
}

int
MSystem(struct snmp_context *c __unused, struct snmp_value *v, u_int x, u_int y __unused, enum snmp_op o)
{

	DmpReq(x, y, v);

	if (o != SNMP_OP_GET)
		return (SNMP_ERR_NOT_WRITEABLE);

	switch (v->var.subs[x - 1]) {
	case LEAF_measureSystemMajor:
		v->v.integer = MEASURED_MAJOR;
		break;
	case LEAF_measureSystemMinor:
		v->v.integer = MEASURED_MINOR;
		break;
	case LEAF_measureSystemTrapSeq:
		v->v.integer = trap_seq;
		break;
	case LEAF_measureSystemState:
		v->v.integer = (int)sysstate;
		break;
	default:
		return(SNMP_ERR_NOSUCHNAME);
	}
	return (SNMP_ERR_NOERROR);
}

/* measurePointTable -------------------------------------------------*/

int
MPointTbl(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx __unused, enum snmp_op o)
{
	struct point *pt;
	unsigned instance;
	time_t t;

	DmpReq(sub, iidx, v);

	instance = 0;
	switch (o) {
	case SNMP_OP_GETNEXT:
		if (v->var.len == sub)
			pt = TAILQ_FIRST(&pointlist);
		else {
			if (v->var.len >= sub + 1)
				instance |= v->var.subs[sub + 0] << SITE_SHIFT;
			if (v->var.len >= sub + 2)
				instance |= v->var.subs[sub + 1] << GROUP_SHIFT;
			if (v->var.len >= sub + 3)
				instance |= v->var.subs[sub + 2] << POINT_SHIFT;
			TAILQ_FOREACH(pt, &pointlist, list)
				if (pt->instance > instance)
					break;
		}
		if (pt != NULL) {
			v->var.subs[sub + 0] = SITE_OF(pt->instance);
			v->var.subs[sub + 1] = GROUP_OF(pt->instance);
			v->var.subs[sub + 2] = POINT_OF(pt->instance);
			v->var.len = sub + 3;
		}
		break;
	case SNMP_OP_GET:
		if (v->var.len != sub + 3) 
			return(SNMP_ERR_NOSUCHNAME);
		instance |= v->var.subs[sub + 0] << SITE_SHIFT;
		instance |= v->var.subs[sub + 1] << GROUP_SHIFT;
		instance |= v->var.subs[sub + 2] << POINT_SHIFT;
		TAILQ_FOREACH(pt, &pointlist, list)
			if (pt->instance == instance)
				break;
		break;
	case SNMP_OP_SET:
		if (v->var.len != sub + 3) 
			return(SNMP_ERR_NOSUCHNAME);
		if (v->var.subs[sub - 1] != LEAF_measurePointValue)
			return(SNMP_ERR_NOSUCHNAME);
		instance |= v->var.subs[sub + 0] << SITE_SHIFT;
		instance |= v->var.subs[sub + 1] << GROUP_SHIFT;
		instance |= v->var.subs[sub + 2] << POINT_SHIFT;
		TAILQ_FOREACH(pt, &pointlist, list)
			if (pt->instance == instance)
				break;
		if (pt == NULL)
			return(SNMP_ERR_NOSUCHNAME);
#if 0
		{
		double d;

		d = v->v.integer * 1e-3;
		i = msproto_send(slave_fd, SPROTO_MAGIC,
		    (unsigned)SL_CONTROLPOINT, 0, 
		    instance, sizeof d, &d);
		assert(i == 0);
		}
#endif
		return (SNMP_ERR_NOERROR);
	default:
		return(SNMP_ERR_NOSUCHNAME);
	}
	
	if (pt == NULL)
		return(SNMP_ERR_NOSUCHNAME);
	switch (v->var.subs[sub - 1]) {
	case LEAF_measurePointSite:
		v->v.integer = SITE_OF(pt->instance);
		break;
	case LEAF_measurePointGroup:
		v->v.integer = GROUP_OF(pt->instance);
		break;
	case LEAF_measurePointPoint:
		v->v.integer = POINT_OF(pt->instance);
		break;
	case LEAF_measurePointValue:
		v->v.integer = (int)round(pt->value * 1000.0);
		break;
	case LEAF_measurePointAge:
		time(&t);
		v->v.integer = t - (int)pt->when;
		break;
	case LEAF_measurePointAlarm:
		v->v.integer = (int)pt->limit;
		break;
	default:
		return(SNMP_ERR_NOSUCHNAME);
	}
	return (SNMP_ERR_NOERROR);
}


/* measureStateTable -------------------------------------------------*/

int
MStateTbl(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx __unused, enum snmp_op o)
{
	struct state *st;
	unsigned u;

	DmpReq(sub, iidx, v);
	switch (o) {
	case SNMP_OP_GETNEXT:
		u = 0;
		if (v->var.len == sub)
			st = TAILQ_FIRST(&statelist);
		else {
			if (v->var.len >= sub + 1)
				u = v->var.subs[sub + 0];
			TAILQ_FOREACH(st, &statelist, list)
				if (st->stn > u)
					break;
		}
		if (st != NULL) {
			v->var.subs[sub + 0] = st->stn;
			v->var.len = sub + 1;
		}
		break;
	case SNMP_OP_GET:
		if (v->var.len != sub + 1) 
			return(SNMP_ERR_NOSUCHNAME);
		u = v->var.subs[sub + 0];
		TAILQ_FOREACH(st, &statelist, list)
			if (st->stn == u)
				break;
		break;
	default:
		return(SNMP_ERR_READONLY);
	}
	if (st == NULL)
		return(SNMP_ERR_NOSUCHNAME);
	switch (v->var.subs[sub - 1]) {
	case LEAF_measureStateTrap:
		v->v.integer = st->stn;
		break;
	case LEAF_measureStateDesc:
		v->v.octetstring.octets = (void*)strdup(st->msg);
		assert(v->v.octetstring.octets != NULL);
		v->v.octetstring.len = strlen(st->msg);
		break;
	case LEAF_measureStatePriority:
		v->v.integer = st->pri;
		break;
	default:
		return(SNMP_ERR_NOSUCHNAME);
	}
	return (SNMP_ERR_NOERROR);

}

/* snmpmod setup -----------------------------------------------------*/

static int
measure_init(struct lmodule *mod, int argc __unused, char *argv[] __unused)
{

	register_printf_render_std("V");
	lmmod = mod;
	return (0);
}

static void
measure_start(void)
{

	measure_timer(NULL);
}

const struct snmp_module config = {
        .comment =      "This module implements the MEASURE MIB.",
        .init =         measure_init,
        .start =        measure_start,
        .tree =         measure_ctree,
        .tree_size =    measure_CTREE_SIZE,
};
