/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: measured.c,v 1.142 2008/02/17 09:55:50 phk Exp $
 *
 */

#include <stdio.h>
#include <assert.h>
#include <signal.h>
#include <errno.h>
#include <err.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <printf.h>
#include <string.h>
#include <time.h>
#include <configkit.h>
#include <configkit_src.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/socket.h>

#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

struct ck_tree *cft;

struct sitehead all_sites = TAILQ_HEAD_INITIALIZER(all_sites);

struct site *site_zero;

pthread_mutex_t glbmtx;

int d_flag = 0;

/**********************************************************************
 *
 * UTILITY
 *
 */

struct timespec
Now(void)
{
	struct timespec ts;

	AZ(clock_gettime(CLOCK_MONOTONIC, &ts));
	return (ts);
}

struct timespec
Utc(void)
{
	struct timespec ts;

	AZ(clock_gettime(CLOCK_REALTIME, &ts));
	return (ts);
}

double
delta_t(const struct timespec *t1, const struct timespec *t2)
{
	double d;

	d = t1->tv_sec - t2->tv_sec;
	d += 1e-9 * (t1->tv_nsec - t2->tv_nsec);
	return (d);
}

int
show_all(CK_ARG, char **arg)
{
	struct site *sp;
	struct group *gp;
	struct point *pp;
	struct timespec t, t1;
	struct pattern *pt;

	if (!(cfa->flag & CK_SHOW))
		return (0);

	if (arg[0] == NULL)
		pt = CompilePattern("*");
	else
		pt = CompilePattern(arg[0]);
	t = Utc();
	TAILQ_FOREACH(sp, &all_sites, list) {
		TAILQ_FOREACH(gp, &sp->groups, list) {
			TAILQ_FOREACH(pp, &gp->points, list) {
				if (!MatchPattern(pt, pp->instance))
					continue;

				ck_printf(CK_PASS, "%P ", pp);
				ck_printf(CK_PASS, " %7s", p_Type(pp));
				ck_printf(CK_PASS, " %10.3f", pp->value);

				t1 = t;
				ck_printf(CK_PASS, " %5.0f %5.0f", 
				    delta_t(&t1, &pp->measured),
				    p_timeout(pp));
				ck_printf(CK_PASS, " %s", p_label(pp));
				if (pp->curlim != L_NORMAL) {
					ck_printf(CK_PASS, " limit=%d", 
					    pp->curlim);
				}
				ck_printf(CK_PASS, "\n");
			}
		}
	}
	return (0);
}

void
pthread_cancelmutex_lock(pthread_mutex_t *mp)
{
	AZ(pthread_mutex_lock(mp));
	pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock, mp);
}

void
pthread_cancelmutex_unlock(pthread_mutex_t *mp)
{
	pthread_cleanup_pop(0);
	AZ(pthread_mutex_unlock(mp));
}

/* -------------------------------------------------------------------*/
static int
url_config_printf(CK_ARG, const char *fmt, va_list ap)
{
	FILE *f;

	f = cfa->src;
	if (fmt == NULL)
		return (0);
	return (vfprintf(f, fmt, ap));
}

static void
url_config(URL_ARGS)
{
	struct ck_src *cf;

	Httpd200(f);
	HtmlHead(f, "Measured Configuration");
	fprintf(f, "<PRE>\n");
	cf = ck_SrcCreate(cft, url_config_printf, f);
	ck_SrcFeed(cf, "conf\n", 5);
	ck_SrcDestroy(cf);
	fprintf(f, "</PRE>\n");
	HtmlTail(f);
}

/*********************************************************************/
static int
printf_arginfo_instance(const struct printf_info *pi __unused, size_t n, int *argt)
{

	assert(n >= 1);
	argt[0] = PA_INT;
	return (1);
}

static int
printf_render_instance(struct __printf_io *io, const struct printf_info *pi __unused, const void *const *arg)
{
	const unsigned *u;
	char buf[12];
	int i, ret = 0;

	u = ((const unsigned *)arg[0]);
	i = sprintf(buf, "%03u.%03u.%03u",
	    SITE_OF(*u), GROUP_OF(*u), POINT_OF(*u));
	ret += __printf_out(io, pi, buf, i);
	__printf_flush(io);
	return (ret);
}

/*********************************************************************/
static int
printf_arginfo_point(const struct printf_info *pi __unused, size_t n, int *argt)
{

	assert(n >= 1);
	argt[0] = PA_POINTER;
	return (1);
}

static int
printf_render_point(struct __printf_io *io, const struct printf_info *pi __unused, const void *const *arg)
{
	const struct point *pp;
	char buf[12];
	int i, ret = 0;

	pp = *(const struct point * const *)arg[0];
	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	i = sprintf(buf, "%03u.%03u.%03u",
	    SITE_OF(pp->instance),
	    GROUP_OF(pp->instance),
	    POINT_OF(pp->instance));
	ret += __printf_out(io, pi, buf, i);
	__printf_flush(io);
	return(ret);
}

/* printf(%Q) --------------------------------------------------------*/

static int
printf_arginfo_quote(const struct printf_info *pi __unused, size_t n, int *argt)
{

	assert(n >= 1);
	argt[0] = PA_POINTER;
	return (1);
}

static int
printf_render_quote(struct __printf_io *io, const struct printf_info *pi __unused, const void *const *arg)
{
	const char *str, *p, *t, *o;
	char *q, *r;
	int i, ret;

	str = *((const char *const *)arg[0]);
	if (str == NULL)
		return (__printf_out(io, pi, "\"(null)\"", 8));
	i = 0;
	if (*str == '\0')
		i++;
	else if (*str == '#')
		i++;
	else {
		for (p = str; *p; p++)
			if (isspace(*p) || *p == '\\' || *p == '"')
				i++;
	}
	if (!i) 
		return (__printf_out(io, pi, str, strlen(str)));
	
	q = malloc(strlen(str) * 5);
	assert(q != NULL);
	r = q;
	ret = __printf_out(io, pi, "\"", 1);
	t = str;
	for (p = str; *p; p++) {
		o = NULL;
		if (*p == '\\')
			o = "\\\\";
		else if (*p == '\n')
			o = "\\n";
		else if (*p == '\r')
			o = "\\r";
		else if (*p == '\t')
			o = "\\t";
		else if (*p == ' ')
			o = " ";
		else if (*p == '"')
			o = "\\\"";
		else if (isspace(*p)) {
			sprintf(r, "\\%03o", *p);
			o = r;
			r += strlen(r) + 1;
		} else
			continue;
		if (p != t) {
			ret += __printf_out(io, pi, t, p - t);
			t = p + 1;
		}
		ret += __printf_out(io, pi, o, strlen(o));
	}
	if (p != t)
		ret += __printf_out(io, pi, t, p - t);
	ret += __printf_out(io, pi, "\"", 1);
	__printf_flush(io);
	return(ret);
}

/* printf(%M) --------------------------------------------------------*/

static int
printf_arginfo_errno(const struct printf_info *pi __unused, size_t n, int *argt)
{

	assert(n >= 1);
	argt[0] = PA_INT;
	return (1);
}

static int
printf_render_errno(struct __printf_io *io, const struct printf_info *pi __unused, const void *const *arg)
{
	int ret, error;
	char buf[64];
	const char *p;

	ret = 0;
	error = *((const int *)arg[0]);
	if (error >= 0 && error < sys_nerr) {
		p = strerror(error);
		return (__printf_out(io, pi, p, strlen(p)));
	}
	sprintf(buf, "errno=%d/0x%x", error, error);
	ret += __printf_out(io, pi, buf, strlen(buf));
	__printf_flush(io);
	return(ret);
}

/*********************************************************************/

volatile record_f *record_function;

void Record(unsigned point, unsigned len, const void *ptr)
{
	volatile record_f *rf;

	rf = record_function;
	if (rf != NULL)
		rf(point, len, ptr);
}

/*********************************************************************/

static int
LogCk(CK_ARG, const char *fmt, va_list ap)
{
	char *p;

	if (fmt == NULL)
		return (0);
	vasprintf(&p, fmt, ap);
	Notice(1, "Config \"%s\": %s", cfa->line, p);
	return (0);
}

static void
RdFile(const char *fn)
{
	int fd, i;
	struct ck_src *cf;
	char buf[BUFSIZ];

	fd = open(fn, O_RDONLY);
	if (fd < 0) {
		Notice(0, "Could not open %s: %M", fn, errno);
		return;
	}
	Log(0, "Reading %s", fn);

	cf = ck_SrcCreate(cft, LogCk, NULL);
	while (1) {
		i = read(fd, buf, sizeof buf - 1);
		if (i < 0) {
			Notice(0, "Read error %s: %M", fn, errno);
			break;
		}
		if (i == 0)
			break;
		buf[i] = '\0';
		ck_SrcFeed(cf, buf, i);
	}
	ck_SrcDestroy(cf);
	close (fd);
}

static void
url_cli(URL_ARGS)
{

	fprintf(f, "CLI ready...\n\r");
	AZ(fflush(f));
	ck_SrcFd(cft, NULL, fileno(f));
	AZ(fclose(f));
}

int
main(int argc, char **argv) /*lint -e{818,715} */
{
	int ch;
	char *f_arg = NULL;
	int C_flag = 0;


	signal(SIGPIPE, SIG_IGN);

	while ((ch = getopt(argc, argv, "Cdf:")) != -1) {
		switch (ch) {
		case 'C':
			C_flag = 1;
			break;
		case 'd':
			d_flag = 1;
			break;
		case 'f':
			f_arg = strdup(optarg);
			break;
		default:
			errx(1, "Usage");
		}
	}

	if (!d_flag)
		AZ(daemon(0, 0));

	AZ(register_printf_render('I',
	    printf_render_instance, printf_arginfo_instance));
	AZ(register_printf_render('P',
	    printf_render_point, printf_arginfo_point));
	AZ(register_printf_render('Q',
	    printf_render_quote, printf_arginfo_quote));
	AZ(register_printf_render('M',
	    printf_render_errno, printf_arginfo_errno));

	AZ(register_printf_render_std("HVT"));

       	AZ(pthread_mutex_init(&glbmtx, NULL));

	/* Order is important */
	InitLog();
	InitTopo();
	InitPoint();
	InitHttpd();
	InitHttpdAuth();
	InitUrlAlarm();
	InitUrlFavicon();
	InitUrlPattern();
	InitUrlPages();
	InitMaster();

	RegisterUrl("GET", "/config", url_config, NULL);

	if (C_flag)
		RegisterUrl("GET", "/cli", url_cli, NULL);

	cft = ck_TreeNew();
	ck_TreeAdd(cft, &cfg_all);
	ck_TreeAdd(cft, &cfg_auth);
	ck_TreeAdd(cft, &cfg_slave);
	ck_TreeAdd(cft, &cfg_modbus);
	ck_TreeAdd(cft, &cfg_state);
	ck_TreeAdd(cft, &cfg_load);

	RdFile("/etc/measured.conf");
	if (f_arg != NULL) {
		RdFile(f_arg);
		free(f_arg);
	}
	if (d_flag) {
		ck_SrcFd(cft, stdout, 0);
		fprintf(stderr, "Finished stdin config\n");
	}
	while (1)
		sleep(10);
	return (0);
}

void
__assert(const char *func, const char *file, int line, const char *failedexpr)
{
	int e;
	int *ip = NULL;
	static int already = 0;

	if (already++)
		abort();
	e = errno;
	Error(0, 
	    "Assertion failed: (%s), function %s, file %s, line %d error %M.\n",
	    failedexpr,
	    func == NULL ? "?" : func,
	    file, line, e);
	*ip = 0; /*lint !e413 */
	abort();
	/* NOTREACHED */
}

