/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: modbus.c,v 1.14 2008/03/22 17:28:25 phk Exp $
 *
 */

#include <stdio.h>
#include <pthread.h>
#include <configkit.h>
#include <termios.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/endian.h>

#include "measured.h"
#include "configkit_conf.h"

#include "phk_modbus.h"

#include "modbus.h"

struct modbus_req {
	unsigned		magic;
#define REQ_MAGIC		0x06e14da8
	unsigned		bus;
	unsigned		unit;
	unsigned		r0;
	unsigned		rlen;
	modbus_cb_t		*cb;
	void			*priv1;
	void			*priv2;
	TAILQ_ENTRY(modbus_req)	list;
	
	unsigned		busy;
	time_t			last;
};

struct modbus {
	unsigned		magic;
#define MODBUS_MAGIC 		0x83742834
	struct ck_instance	cki;
	struct group		*gp;
	pthread_mutex_t		lock;

	unsigned		debug;
	struct serial		*serial;

	unsigned		dtr;
	unsigned		rts;
	unsigned		bus;
	unsigned		rate;
	unsigned		delay;
};

static TAILQ_HEAD(,modbus_req) req_head = TAILQ_HEAD_INITIALIZER(req_head);

static pthread_mutex_t modbus_mtx;

/*--------------------------------------------------------------------*/

struct modbus_req *
ModBus_Register(unsigned bus, unsigned unit, unsigned r0, unsigned rlen, modbus_cb_t *cb, void *priv1, void *priv2)
{
	struct modbus_req *r;

	ALLOC_OBJ(r, REQ_MAGIC);
	r->bus = bus;
	r->unit = unit;
	r->r0 = r0;
	r->rlen = rlen;
	r->cb = cb;
	r->priv1 = priv1;
	r->priv2 = priv2;
	AZ(pthread_mutex_lock(&modbus_mtx));
	TAILQ_INSERT_TAIL(&req_head, r, list);
	AZ(pthread_mutex_unlock(&modbus_mtx));
	return (r);
}

void
ModBus_Unregister(struct modbus_req *r)
{

	CHECK_OBJ_NOTNULL(r, REQ_MAGIC);
	AZ(pthread_mutex_lock(&modbus_mtx));
	r->cb = NULL;
	if (!r->busy)
		TAILQ_REMOVE(&req_head, r, list);
	else
		r = NULL;
	AZ(pthread_mutex_unlock(&modbus_mtx));
	if (r)
		FREE_OBJ(r);
}

static int
RdReg(struct modbus *mb, struct serial *sc, unsigned id, unsigned start, unsigned n, unsigned char *p)
{
	unsigned char msg[10];
	unsigned l;
	int i, j;

	msg[0] = id;
	msg[1] = 0x03;
	be16enc(msg + 2, start);
	be16enc(msg + 4, n);
	j = 0;
	if (mb->rts)
		j |= TIOCM_RTS;
	if (mb->dtr)
		j |= TIOCM_DTR;
	if (j != 0) {
		AZ(ioctl(serial_fd(sc), TIOCMBIS, &j));
		if (mb->delay)
			usleep(mb->delay);
	}
	Record(GI(mb->gp) + 1, 6, msg);
	PhkModbusRtuSend(serial_fd(sc), msg, 6);
	if (mb->debug)
		printf("Snd:\n%#H\n", msg, 6);
	if (j != 0) {
		AZ(tcdrain(serial_fd(sc)));
		if (mb->delay)
			usleep(mb->delay);
		AZ(ioctl(serial_fd(sc), TIOCMBIC, &j));
	}
	/* XXX: reconf serial on failure (?) */
	i = PhkModbusRtuRecv(serial_fd(sc), p, &l, n * 2 + 5, 1000);
	if (i != 0)
		Record(GI(mb->gp) + 0, sizeof i, &i);
	Record(GI(mb->gp) + 2, l, p);
	if (mb->debug)
		printf("Rcv: (%d)\n%#H\n", i, p, l);
	if (i == -1) 
		serial_reopen(NULL, sc);
	return (i);
}

static void *
modbus_thread(void *priv)
{
	struct termios t;
	struct modbus *sc;
	struct modbus_req *r, *r2;
	time_t now, low;
	unsigned char rx[BUFSIZ];
	int i;

	CAST_OBJ_NOTNULL(sc, priv, MODBUS_MAGIC);

	if (isatty(serial_fd(sc->serial))) {
		AZ(tcgetattr(serial_fd(sc->serial), &t));
		cfmakeraw(&t);
		t.c_cflag |= CLOCAL;
		t.c_cc[VMIN] = 0;
		t.c_cc[VTIME] = 10;
		AZ(tcsetattr(serial_fd(sc->serial), TCSAFLUSH, &t));
	}
	
	r = NULL;
	while (1) {
		if (serial_fd(sc->serial) < 0) {
			sleep (1);
			serial_reopen(NULL, sc->serial);
			continue;
		}
		low = now = time(NULL);
		AZ(pthread_mutex_lock(&modbus_mtx));
		TAILQ_FOREACH(r2, &req_head, list)
			if (r2->bus == sc->bus && r2->last < low &&
			    r2->cb != NULL) {
				low = r2->last;
				r = r2;
			}
		if (low + (int)sc->rate > now)
			r = NULL;
		if (r != NULL)
			r->busy = 1;
		AZ(pthread_mutex_unlock(&modbus_mtx));
		if (r == NULL) {
			usleep (100000);
			continue;
		}
		i = RdReg(sc, sc->serial, r->unit, r->r0, r->rlen, rx);
		AZ(pthread_mutex_lock(&modbus_mtx));
		r->busy = 0;
		if (r->cb == NULL) {
			TAILQ_REMOVE(&req_head, r, list);
			FREE_OBJ(r);
			r = NULL;
		} else if (i == 0) {
			r->cb(r->priv1, r->priv2, rx, r->rlen);
		} else {
			printf("modbus result: i = %d\n", i);
		}
		if (r != NULL)
			r->last = now;
		AZ(pthread_mutex_unlock(&modbus_mtx));
	}
}

static pthread_once_t once_only = PTHREAD_ONCE_INIT;

static void
modbus_initmtx(void)
{
	AZ(pthread_mutex_init(&modbus_mtx, NULL));
}

static void
modbus_init(CK_ARG, struct modbus *sc)
{

	sc->bus = GI(sc->gp) >> GROUP_SHIFT;
	sc->rate = 3;
	serial_init(CK_PASS, &sc->serial, modbus_thread, sc);
	AZ(pthread_once(&once_only, modbus_initmtx));
}

static void
modbus_deinit(CK_ARG, struct modbus *sc)
{

	serial_destroy(CK_PASS, &sc->serial);
}

SerialFunc(modbus, MODBUS_MAGIC, serial);
UintVar(modbus, MODBUS_MAGIC, debug, debug, 0, );
UintVar(modbus, MODBUS_MAGIC, rate, rate, 3, );
UintVar(modbus, MODBUS_MAGIC, rts, rts, 0, );
UintVar(modbus, MODBUS_MAGIC, dtr, dtr, 0, );
UintVar(modbus, MODBUS_MAGIC, delay, delay, 0, );
GROUP(modbus, MODBUS_MAGIC);
