/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pattern.c,v 1.6 2006/02/08 08:14:54 phk Exp $
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/queue.h>

#include "measured.h"

struct part {
	unsigned		from;
	unsigned		to;
};

struct match {
	unsigned		magic;
#define MATCH_MAGIC		0x281738
	TAILQ_ENTRY(match)	next;
	unsigned		not;
	struct part		part[3];
};

struct pattern {
	unsigned		magic;
#define PATTERN_MAGIC		0x1837484
	TAILQ_HEAD(,match)	matches;
};

static void
skip_space(char **i)
{

	if (*i == NULL)
		return;
	while (isspace(**i))
		(*i)++;
	if (**i == '\0')
		*i = NULL;
}

static void
part_compile(char **inp, struct part *p)
{
	char *s, *q;
	int i;

	p->from = 0;
	p->to = 999;
	skip_space(inp);
	if (*inp == NULL)
		return;
	s = *inp;
	q = strchr(s, '.');
	if (q != NULL)
		*q++ = '\0';
	*inp = q;
	if (*s == '*') 
		return;
	i = sscanf(s, "%u-%u", &p->from, &p->to);
	if (i == 1)
		p->to = p->from;
	return;
}

static struct match *
match_compile(char *inp, int pointonly)
{
	struct match *m;

	ALLOC_OBJ(m, MATCH_MAGIC);

	skip_space(&inp);

	if (*inp == '!') {
		m->not = 1;
		inp++;
	}

	if (pointonly) {
		m->part[0].from = m->part[1].from = 0;
		m->part[0].to = m->part[1].to = 999;
	} else {
		part_compile(&inp, &m->part[0]);
		part_compile(&inp, &m->part[1]);
	}
	part_compile(&inp, &m->part[2]);
	return (m);
}

struct pattern *
CompilePattern(const char *inp)
{
	struct pattern *p;
	struct match *m;
	char *l, *ll, *n;

	if (inp == NULL)
		return (NULL);
	ll = strdup(inp);
	assert(ll != NULL);
	ALLOC_OBJ(p, PATTERN_MAGIC);
	TAILQ_INIT(&p->matches);
	for (l = ll; l != NULL; l = n) {
		n = strchr(l, ';');
		if (n == NULL)
			n = strchr(l, ',');
		if (n != NULL) {
			*n = '\0';
			n++;
		}
		m = match_compile(l, 0);
		TAILQ_INSERT_TAIL(&p->matches, m, next);
	}
	free(ll);
	return (p);
}

struct pattern *
CompilePointPattern(const char *inp)
{
	struct pattern *p;
	struct match *m;
	char *l, *ll, *n;

	ll = strdup(inp);
	assert(ll != NULL);
	ALLOC_OBJ(p, PATTERN_MAGIC);
	TAILQ_INIT(&p->matches);
	for (l = ll; l != NULL; l = n) {
		n = strchr(l, ';');
		if (n != NULL) {
			*n = '\0';
			n++;
		}
		m = match_compile(l, 1);
		TAILQ_INSERT_TAIL(&p->matches, m, next);
	}
	free(ll);
	return (p);
}

void
FreePattern(struct pattern *p)
{

	struct match *m;

	if (p == NULL)
		return;
	CHECK_OBJ(p, PATTERN_MAGIC);
	while (!TAILQ_EMPTY(&p->matches)) {
		m = TAILQ_FIRST(&p->matches);
		CHECK_OBJ_NOTNULL(m, MATCH_MAGIC);
		TAILQ_REMOVE(&p->matches, m, next);
		free(m);
	}
	free(p);
}

int
MatchPattern(const struct pattern *p, unsigned id)
{
	struct match *m;

	if (p == NULL)
		return (1);
	CHECK_OBJ(p, PATTERN_MAGIC);
	TAILQ_FOREACH(m, &p->matches, next) {
		CHECK_OBJ_NOTNULL(m, MATCH_MAGIC);
		if (m->part[0].from > SITE_OF(id) || 
		    m->part[0].to < SITE_OF(id))
			continue;
		if (m->part[1].from > GROUP_OF(id) ||
		    m->part[1].to < GROUP_OF(id))
			continue;
		if (m->part[2].from > POINT_OF(id) ||
		    m->part[2].to < POINT_OF(id))
			continue;
		break;
	}
	if (m == NULL)
		return (0);
	return (!m->not);
}

#if 0

void
PrintPattern(const struct pattern *p, FILE *f)
{
	struct match *m;
	const char *pfx = "";
	const char *p2;
	int i;

	TAILQ_FOREACH(m, &p->matches, next) {
		fprintf(f, "%s%s", pfx, m->not ? "!" : "");
		pfx = ";";
		p2 = "";
		for (i = 0; i < 3; i++) {
			if (m->part[i].from == 0 &&
			    m->part[i].to == 999 && i == 2) {
				/* nothing */
			} else if (i == 1 && m->part[i].from == 0 &&
			    m->part[i].to == 999 &&
			    m->part[2].from == 0 && m->part[2].to == 999) {
				/* nothing */
			} else if (m->part[i].from == 0 &&
			    m->part[i].to == 999) {
				fprintf(f, "%s*", p2);
			} else {
				fprintf(f, "%s%d", p2, m->part[i].from);
				if (m->part[i].to != m->part[i].from)
					fprintf(f, "-%d", m->part[i].to);
			}
			p2 = ".";
		}
	}
}

int
main(int argc __unused, const char **argv __unused)
{
	struct pattern *p;
	unsigned x;


	setbuf(stdout, NULL);
	p = CompilePattern("1.2.3;!2.3.4;1.2-4.3;1;1.2.*;4;5.6");
	PrintPattern(p, stdout);
	printf("\n");

	x = 4 << SITE_SHIFT;
	x |= 3 << GROUP_SHIFT;
	x |= 5 << POINT_SHIFT;
	printf("Match(%03u.%03u.%03u) = %d\n",
	    SITE_OF(x), GROUP_OF(x), POINT_OF(x), MatchPattern(p, x));
	FreePattern(p);
	return(0);
}

#endif
