/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: point.c,v 1.81 2008/04/29 19:45:14 phk Exp $
 *
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <configkit.h>
#include <time.h>

#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

static pthread_t point_tp;

/* Point search functions --------------------------------------------*/

static struct point *
FindPointLocked(unsigned instance)
{
	struct site *s;
	struct group *g;
	struct point *p;

	TAILQ_FOREACH(s, &all_sites, list) {
		if (s->instance < (instance & SITE_MASK))
			continue;
		if (s->instance > (instance & SITE_MASK))
			break;
		TAILQ_FOREACH(g, &s->groups, list) {
			if (g->instance < 
			    (instance & (SITE_MASK | GROUP_MASK)))
				continue;
			if (g->instance > 
			    (instance & (SITE_MASK | GROUP_MASK)))
				break;
			TAILQ_FOREACH(p, &g->points, list) {
				if (p->instance == instance)
					return (p);
				if (p->instance > instance)
					break;
			}
		}
	}
	return (NULL);
}

struct point *
FindPoint(unsigned instance)
{
	struct point *p;

	TopoHold();
	p = FindPointLocked(instance);
	TopoRel();
	return (p);
}

/* Point control -----------------------------------------------------*/

static void
ControlPoint(struct point *pp, double val)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->type != DOUT && pp->type != AOUT) {
		PNotice(pp, "Failed attempt to control %s %P = %g",
		    PointType(pp->type), pp, val);
	} else if (pp->control == NULL) {
		PNotice(pp, "No control possible for %P = %g", pp, val);
	} else {
		PLog(pp, "Controlling %P = %g", pp, val);
		pp->control(pp, pp->control_priv1, pp->control_priv2,
		    (val - pp->offset) / pp->scale);
	}
}

void
ControlPointI(unsigned instance, double val)
{
	struct point *pp;

	TopoHold();
	pp = FindPointLocked(instance);
	if (pp != NULL)
		ControlPoint(pp, val);
	TopoRel();
	if (pp == NULL)
		Notice(0, "ControlPointI: point %I not found", instance);
}

void
SetPointControl(struct point *pp, control_t *func, void *priv1, void *priv2)
{
	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	Lock();
	pp->control = func;
	pp->control_priv1 = priv1;
	pp->control_priv2 = priv2;
	Unlock();
}

/* Point values ------------------------------------------------------*/

static void
limitis(struct point *pp, enum limit_e limit)
{

	if (pp->curlim == limit) {
		/*
		 * If we have a timer and it expired, but we're back in the
		 * same limit bracket, so don't tell anybody.
		 */
		if (pp->limtimer.tv_sec != 0 &&
		     delta_t(&pp->measured, &pp->limtimer) > pp->limitdelay)
			pp->limtimer.tv_sec = 0;
		return;
	}
	if (pp->limtimer.tv_sec == 0 && pp->limitdelay > 0) {
		/* Have delay, will wait */
		pp->limtimer = pp->measured;
		return;
	}
	if (pp->limtimer.tv_sec != 0 &&
	     delta_t(&pp->measured, &pp->limtimer) < pp->limitdelay) {
		/* Still ticking */
		return;
	}
	pp->limtimer.tv_sec = 0;
	StateLimit(pp, limit);
}

static int
istimeout(const struct point *pp, struct timespec ts)
{
	double to, dt;

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->type == AOUT || pp->type == DOUT)
		return (0);
	
	to = p_timeout(pp);
	if (to == 0.0)
		return (0);

	dt = ts.tv_sec - pp->measured.tv_sec;
	dt += 1e-9 * (ts.tv_nsec - pp->measured.tv_nsec);
	if (dt > to)
		return (1);
	return (0);
}

/*
 * Once every second we check if any of our local points have timed out
 */

static void *
point_thread(void *priv __unused)
{
	struct timespec ts;
	struct group *gp;
	struct point *pp;

	while (1) {
		sleep(1);
		TopoHold();
		ts = Utc();
		TAILQ_FOREACH(gp, &site_zero->groups, list) {
			TAILQ_FOREACH(pp, &gp->points, list) {
				if (istimeout(pp, ts))
					limitis(pp, L_TIMEOUT);
			}
		}
		TopoRel();
	}
}

/*
 * Timeout and limit check with hysteresis.
 *
 * A timeout has highest priority.
 *
 * We have four limit values and possibly a hysteresis.  The hysteresis
 * marks a "guard-band" symmetrically around the limit value which is
 * considered part of whatever band the previous state inhabited.
 */

static void
CheckLimit(struct point *pp, double old)
{ /*lint -e{777} */

	double h;

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	h = pp->hysteresis * .5;
	if (istimeout(pp, Utc()))
		limitis(pp, L_TIMEOUT);
	else if (!isnan(pp->limits[L_LOW]) &&
	    (pp->value < pp->limits[L_LOW] - h ||
	    (pp->curlim == L_LOW && pp->value < pp->limits[L_LOW] + h)))
		limitis(pp, L_LOW);
	else if (!isnan(pp->limits[L_SAG]) &&
	    (pp->value < pp->limits[L_SAG] - h ||
	    (pp->curlim == L_SAG && pp->value < pp->limits[L_SAG] + h)))
		limitis(pp, L_SAG);
	else if (!isnan(pp->limits[L_HIGH]) &&
	    (pp->value > pp->limits[L_HIGH] + h ||
	    (pp->curlim == L_HIGH && pp->value > pp->limits[L_HIGH] - h)))
		limitis(pp, L_HIGH);
	else if (!isnan(pp->limits[L_RAISE]) &&
	    (pp->value > pp->limits[L_RAISE] + h ||
	    (pp->curlim == L_RAISE && pp->value > pp->limits[L_RAISE] - h)))
		limitis(pp, L_RAISE);
	else if (!isnan(pp->limits[L_CHANGED]) && !isnan(old) &&
	    fabs(pp->value - old) > pp->limits[L_CHANGED])
		limitis(pp, L_CHANGED);
	else
		limitis(pp, L_NORMAL);
}

void
SetVal(struct point *pp, double d)
{
	struct timespec when;
	double o;

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	when = Utc();
	Lock();
	pp->measured = when;
	pp->raw = d;
	o = pp->value;
	pp->value = d * pp->scale + pp->offset;
	PostEvent(E_POINT_VALUE, pp, PI(pp));
	CheckLimit(pp, o);
	Unlock();
}

void
SetValBool(struct point *pp, int v)
{

	SetVal(pp, v ? 1.0 : 0.0);
}

double
p_value(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->curlim == L_TIMEOUT)
		return (NAN);
	return (pp->value);
}

/* Point label -------------------------------------------------------*/

void
SetPointLabel(struct point *pp, const char *label)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	Lock();
	FREE(&pp->d_label);
	if (label != NULL)
		pp->d_label = strdup(label);
	if (pp->label == NULL)
		PostEvent(E_POINT_LABEL, pp, PI(pp));	/*lint !e641 */
	Unlock();
}

const char *
p_label(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->label != NULL)
		return (pp->label);
	if (pp->d_label != NULL)
		return (pp->d_label);
	return ("");
}

/* Point unit --------------------------------------------------------*/

void
SetPointUnit(struct point *pp, const char *unit)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	Lock();
	FREE(&pp->d_unit);
	if (unit != NULL)
		pp->d_unit = strdup(unit);
	if (pp->unit == NULL)
		PostEvent(E_POINT_UNIT, pp, PI(pp));	/*lint !e641 */
	Unlock();
}

const char *
p_unit(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->unit != NULL)
		return (pp->unit);
	if (pp->d_unit != NULL)
		return (pp->d_unit);
	return ("");
}

/* Point format ------------------------------------------------------*/

void
SetPointFormat(struct point *pp, const char *format)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	Lock();
	FREE(&pp->d_format);
	if (format != NULL)
		pp->d_format = strdup(format);
	if (pp->format == NULL)
		PostEvent(E_POINT_FORMAT, pp, PI(pp));	/*lint !e641 */
	Unlock();
}

const char *
p_format(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (pp->format != NULL)
		return (pp->format);
	if (pp->d_format != NULL)
		return (pp->d_format);
	return ("%g");
}

const char *
PointFmt(const struct point *pp, double v, char *buf, unsigned buflen)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	snprintf(buf, buflen, p_format(pp), v);
	return (buf);
}

/* Point timeout -----------------------------------------------------*/

void
SetPointTimeout(struct point *pp, double timeout)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	Lock();
	pp->d_timeout = timeout;
	Unlock();
}

double
p_timeout(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (!isnan(pp->limits[L_TIMEOUT]))
		return (pp->limits[L_TIMEOUT]);
	return (pp->d_timeout);
}

/* Point type --------------------------------------------------------*/

enum point_type
p_type(const struct point *pp)
{

	return (pp->type);
}

const char *
p_Type(const struct point *pp)
{

	return (PointType(pp->type));
}

/* -------------------------------------------------------------------*/

static void
url_point_set(URL_ARGS)
{
	unsigned instance;
	double a;
	char *n, *e;

	if (Httpd_Auth(f, h, priv, 3))
		return;
	do {
		n = HttpArg(h, "val");
		if (n == NULL)
			break;
		a = strtod(n, &e);
		if (e == NULL || *e != '\0')
			break;
		n = HttpArg(h, "point");
		if (n == NULL)
			break;
		if (ParseInstance(n, &instance))
			break;
		ControlPointI(instance, a);
		n = HttpHeader(h, "Referer");
		if (n != NULL) {
			fprintf(f, "HTTP/1.1 303 See Other\r\n");
			fprintf(f, "Location: %s\r\n", n);
		} else {
			fprintf(f, "HTTP/1.1 200 OK\r\n");
			fprintf(f, "\r\n");
		}
		return;
	} while (0);
	fprintf(f, "HTTP/1.1 404 Not Found\r\n\r\n");
	fprintf(f, "Not Found\r\n");
}

/* ConfigKit ---------------------------------------------------------*/

/*
 * XXX: Locking for all these ??
 * XXX: ck_FooVar may call printf :-(
 */

WordVar(point, POINT_MAGIC, label, label, 
	PostEvent(E_POINT_LABEL, sc, PI(sc));
	);

WordVar(point, POINT_MAGIC, format, format, 
	PostEvent(E_POINT_FORMAT, sc, PI(sc));
	);

WordVar(point, POINT_MAGIC, unit, unit, 
	PostEvent(E_POINT_UNIT, sc, PI(sc));
	);

int
cfg_point_offset(CK_ARG, const char **offset)
{
	struct point *pp;

	CAST_OBJ(pp, cfa->priv, POINT_MAGIC);
	if (pp == NULL)
		return (0);
	if (ck_RealVar2(CK_PASS, offset, &pp->offset, &pp->offset_str,
	    0.0, p_format(pp))) {
		if (!isnan(pp->raw)) {
			pp->value = pp->raw * pp->scale + pp->offset;
			PostEvent(E_POINT_VALUE, pp, PI(pp));	/*lint !e641 */
			CheckLimit(pp, NAN);
		}
	}
	return (0);
}

int
cfg_point_scale(CK_ARG, const char **scale)
{
	struct point *pp;

	CAST_OBJ(pp, cfa->priv, POINT_MAGIC);
	if (pp == NULL)
		return (0);
	if (ck_RealVar2(CK_PASS, scale, &pp->scale, &pp->scale_str,
	    1.0, p_format(pp))) {
		if (!isnan(pp->raw)) {
			pp->value = pp->raw * pp->scale + pp->offset;
			PostEvent(E_POINT_VALUE, pp, 0);	/*lint !e641 */
			CheckLimit(pp, NAN);
		}
	}
	return (0);
}

RealVar2(point, POINT_MAGIC, hysteresis, hysteresis, 0.0, p_format(sc),
	PostEvent(E_POINT_HYSTERESIS, sc, PI(sc));
	CheckLimit(sc, NAN);
	);

RealVar(point, POINT_MAGIC, limitdelay, limitdelay, 0.0, p_format(sc),
	PostEvent(E_POINT_LIMITDELAY, sc, PI(sc));
	CheckLimit(sc, NAN);
	);

static int									
cfg_point_lim (CK_ARG, struct point *sc, const char **val, event_t e, unsigned idx, const char *fmt)
{
	char buf[BUFSIZ];

	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (isnan(sc->limits[idx]))
			return (0);
		sprintf(buf, fmt, sc->limits[idx]);
		ck_printf(CK_PASS, "%s %s\n", cfa->tree, buf);
		return (0);
	}
	if (ck_RealVar2(CK_PASS, val, &sc->limits[idx], &sc->limits_str[idx],
	    NAN, fmt)) {
		PostEvent(e, sc, PI(sc));
		CheckLimit(sc, NAN);
	} 
	return (0);
}

#define LimitFunc(name, uname, format) 					\
int									\
cfg_point_##name (CK_ARG, const char **val)				\
{									\
	struct point *sc;						\
									\
	CAST_OBJ_NOTNULL(sc, cfa->priv, POINT_MAGIC);			\
	return(cfg_point_lim(CK_PASS, sc, val, 				\
	    E_POINT_##uname, L_##uname, format));			\
}

LimitFunc(low,	   LOW,	    p_format(sc))
LimitFunc(sag,	   SAG,	    p_format(sc))
LimitFunc(raise,   RAISE,   p_format(sc))
LimitFunc(high,	   HIGH,    p_format(sc))
LimitFunc(change,  CHANGED, p_format(sc))
LimitFunc(timeout, TIMEOUT, "%g")

#undef LimitFunc

int
cfg_point_state(CK_ARG, unsigned *state, const char **pri)
{
	struct point *pp;

	CAST_OBJ_NOTNULL(pp, cfa->priv, POINT_MAGIC);
	if (cfa->flag & (CK_SHOW | CK_CONF)) {
		if (pp->state) {
			ck_printf(CK_PASS, "%s %u ",
			    cfa->tree, pp->state);
			switch(pp->state_pri) {
			case SP_MAJOR:	ck_printf(CK_PASS, "major\n"); break;
			case SP_MINOR:	ck_printf(CK_PASS, "minor\n"); break;
			default: ck_printf(CK_PASS, "%u\n", pp->state_pri);
			}
		}
		return (0);
	}
	if (!strcasecmp(*pri, "major")) {
		pp->state_pri = SP_MAJOR;
	} else if (!strcasecmp(*pri, "minor")) {
		pp->state_pri = SP_MINOR;
	} else {
		ck_printf(CK_PASS, "Illegal priority \"%s\"\n", *pri);
		return (0);
	}
	(void)ck_UintVar(CK_PASS, state, &pp->state, 0);
	StatePoint(pp);
	return (0);
}

/* XXX: check <= MAXTRAP */
UintVar(point, POINT_MAGIC, trap, trap, 0, 
	PostEvent(E_POINT_TRAP, sc, PI(sc));
	);

int
ConfigPoint(CK_ARG, struct group *gp, const char **spec, char **argv)
{
	char *tree;
	struct point *pp;
	int retval = 0;
	struct pattern *pt;

	CHECK_OBJ_NOTNULL(gp, GROUP_MAGIC);
	if (spec != NULL) {
		pt = CompilePointPattern(*spec);
		if (pt == NULL) {
			ck_printf(CK_PASS, "Pattern error\n");
			return (0);
		}
		TAILQ_FOREACH(pp, &gp->points, list) {
			if (MatchPattern(pt, pp->instance)) {
				cfa->priv = pp;
				retval = ck_continue(CK_PASS, &cfg_point, argv);
			}
		}
		FreePattern(pt);
		return (retval);
	}

	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		tree = cfa->tree;
		TAILQ_FOREACH(pp, &gp->points, list) {
			cfa->priv = pp;
			asprintf(&cfa->tree, "%s %d",
			    tree, POINT_OF(pp->instance));
			retval = ck_continue(CK_PASS, &cfg_point, argv);
			free(cfa->tree);
		}
		cfa->tree = tree;
		return (retval);
	}
	return (0);
}

void
InitPoint()
{

	AZ(pthread_create(&point_tp, NULL, point_thread, NULL));
	RegisterUrl("GET", "/point_set", url_point_set, NULL);
	RegisterUrl("POST", "/point_set", url_point_set, NULL);
}
