/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: serial.c,v 1.10 2008/03/09 09:58:40 phk Exp $
 *
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <poll.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <configkit.h>
#include "measured.h"

#include "phk_tcp.h"

/*********************************************************************/

struct sserial;

struct serial {
	unsigned		magic;
#define	SERIAL_MAGIC		0x28718558
	char			*name;
	pthread_t 		tp;
	int			ser_fd;
	void 			*(*func)(void *);
	void			*priv;
	char			*stty;
	int			isshared;
	struct sserial		*shared;
};

struct sserial {
	unsigned		magic;
#define	SSERIAL_MAGIC		0x4b0a126f
	TAILQ_ENTRY(sserial)	list;
	pthread_mutex_t		mtx;
	char			*name;
	int			fd;
	int			refcnt;
	struct serial		*base;
};

static TAILQ_HEAD(,sserial)	sserial_head =
    TAILQ_HEAD_INITIALIZER(sserial_head);

static pthread_mutex_t sserial_mtx = PTHREAD_MUTEX_INITIALIZER;

/*********************************************************************/

static void
sserial_deref(CK_ARG, struct serial *sc)
{
	struct sserial *ss;

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	ss = sc->shared;
	sc->shared = NULL;
	CHECK_OBJ_NOTNULL(ss, SSERIAL_MAGIC);
	AZ(pthread_mutex_lock(&sserial_mtx));
	ss->refcnt--;
	if (ss->refcnt == 0) 
		TAILQ_REMOVE(&sserial_head, ss, list);
	else
		ss = NULL;
	AZ(pthread_mutex_unlock(&sserial_mtx));
	if (ss == NULL)
		return;
	serial_destroy(CK_PASS, &ss->base);
	free(ss);
	sc->ser_fd = -1;
}

static void
sserial_ref(CK_ARG, struct serial *sc)
{
	struct sserial *ss, *ss2;

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	assert(sc->name != NULL);
	AZ(sc->shared);
	ALLOC_OBJ(ss2, SSERIAL_MAGIC);
	serial_init(CK_PASS, &ss2->base, NULL, NULL);
	AZ(pthread_mutex_lock(&sserial_mtx));
	TAILQ_FOREACH(ss, &sserial_head, list)
		if (!strcmp(ss->name, sc->name))
			break;
	if (ss == NULL) {
		ss = ss2;		
		ss2 = NULL;
		TAILQ_INSERT_TAIL(&sserial_head, ss, list);
		ss->name = strdup(sc->name);
		AZ(pthread_mutex_init(&ss->mtx, NULL));
	}
	ss->refcnt++;
	AZ(pthread_mutex_lock(&ss->mtx));
	AZ(pthread_mutex_unlock(&sserial_mtx));
	if (ss2 != NULL)
		free(ss2);
	sc->shared = ss;
}

/*********************************************************************/


void
serial_init(CK_ARG, struct serial **ssc, void *(*func)(void *), void *priv)
{

	assert(*ssc == NULL);
	ALLOC_OBJ(*ssc, SERIAL_MAGIC);
	(*ssc)->func = func;
	(*ssc)->priv = priv;
	(*ssc)->ser_fd = -1;
}

void
serial_shared(CK_ARG, struct serial *sc)
{
	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	sc->isshared = 1;
}

int
serial_fd(const struct serial *sc)
{

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	return (sc->ser_fd);
}

static void
serial_doreopen(CK_ARG, struct serial *sc)
{

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);

	if (sc->ser_fd >= 0)
		AZ(close(sc->ser_fd));
	sc->ser_fd = -1;

	if (sc->name == NULL)
		return;

	if (sc->name[0] != '/') {
		sc->ser_fd = PhkTcpOpen(sc->name, NULL);
		if (sc->ser_fd < 0) {
			ck_printf(CK_PASS, "Could not connect to %s: %M\n",
			    sc->name, errno);
			sc->ser_fd = -2;
		}
		return;
	}
	sc->ser_fd = open(sc->name, O_RDWR);
	if (sc->ser_fd < 0) {
		ck_printf(CK_PASS, "Could not open %s: %M\n",
		    sc->name, errno);
		return;
	}
	if (!isatty(sc->ser_fd)) {
		ck_printf(CK_PASS, "%s is not a serial port\n",
		    sc->name);
		AZ(close(sc->ser_fd));
		sc->ser_fd = -1;
		return;
	}
	if (sc->stty != NULL) 
		serial_stty(CK_PASS, sc, sc->stty);
}

void
serial_reopen(CK_ARG, struct serial *sc)
{
	struct sserial *ss;

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	if (sc->shared != NULL)
		CHECK_OBJ_NOTNULL(sc->shared, SSERIAL_MAGIC);

	if (!sc->isshared) {
		serial_doreopen(CK_PASS, sc);
		return;
	}

	if (sc->shared != NULL)
		sserial_deref(CK_PASS, sc);
	if (sc->name == NULL)
		return;
	sserial_ref(CK_PASS, sc);
	CHECK_OBJ_NOTNULL(sc->shared, SSERIAL_MAGIC);
	ss = sc->shared;
	if (ss->refcnt == 1) {
		AZ(ss->base->name);
		ss->base->name = strdup(sc->name);
		assert(ss->base->name != NULL);
		AZ(ss->base->stty);
		if (sc->stty != NULL) {
			ss->base->stty = strdup(sc->stty);
			assert(ss->base->stty != NULL);
		}
		serial_doreopen(CK_PASS, ss->base);
	}
	sc->ser_fd = ss->base->ser_fd;
	AZ(pthread_mutex_unlock(&ss->mtx));
}

void
serial_reconf(CK_ARG, struct serial *sc)
{
	void *vp;

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	if (sc->tp) {
		AZ(pthread_cancel(sc->tp));
		AZ(pthread_join(sc->tp, &vp));
		sc->tp = NULL;
	}

	serial_reopen(CK_PASS, sc);
	if (sc->ser_fd >= 0 || sc->ser_fd == -2)
		AZ(pthread_create(&sc->tp, NULL, sc->func, sc->priv));
}

int
serial_params(CK_ARG, struct serial *sc, const char **port, char **argv)
{
	
	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	if (cfa->flag & (CK_CONF|CK_SHOW)) {
		if (sc->name != NULL) {
			ck_printf(CK_PASS, "%s %s", cfa->tree, sc->name);
			if (sc->stty != NULL) 
				ck_printf(CK_PASS, " %Q", sc->stty);
			ck_printf(CK_PASS, "\n");
		}
		return (0);
	}
	if (argv[0] != NULL && argv[1] != NULL) {
		ck_printf(CK_PASS, "Too many arguments\n");
		return (-1);
	}
	if (sc->stty != NULL)
		free(sc->stty);
	sc->stty = NULL;
	if (argv[0] != NULL)
		sc->stty = strdup(argv[0]);
	if (sc->name != NULL)
		free(sc->name);
	sc->name = strdup(port[0]);
	serial_reconf(CK_PASS, sc);
	return (0);
}

void
serial_destroy(CK_ARG, struct serial **ssc)
{
	struct serial *sc;

	assert(ssc != NULL);
	sc = *ssc;
	*ssc = NULL;
	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	if (sc->stty != NULL) {
		free(sc->stty);
		sc->stty = NULL;
	}
	if (sc->name != NULL) {
		free(sc->name);
		sc->name = NULL;
		serial_reconf(CK_PASS, sc);
	}
	FREE_OBJ(sc);
}

void
serial_write(const struct serial *sc, const char *str)
{
	unsigned u;
	int i;

	u = strlen(str);
	i = write(sc->ser_fd, str, u);
	assert(i == (int)u);
}

int
serial_read(const struct serial *sc, void *dst, int cmin, int cmax, int tmo)
{
	int i, j;
	unsigned char *p;
	struct pollfd pfd[1];

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	assert(dst != NULL);
	assert(tmo > 0);
	p = dst;

	if (cmax == 0)
		cmax = cmin;
	assert(cmin >= 0);
	assert(cmax >= cmin);

	for (j = 0; j < cmin; ) {
		pfd[0].fd = sc->ser_fd;
		pfd[0].events = POLLIN;
		i = poll(pfd, 1, tmo);
		if (i == 0) 
			return (j);
		i = read(sc->ser_fd, p + j, cmax - j);
		if (i == 0)
			return (j);
		if (i > 0)
			j += i;
		if (i < 0) 
			return (i);
	}
	return (j);
}

int
serial_stty(CK_ARG, struct serial *sc, const char *stty)
{
	char *p;
	int i, j;

	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	if (!isatty(sc->ser_fd))
		return (0);
	if (stty == NULL)
		return (0);
	asprintf(&p, "/bin/stty -f %s %s", sc->name, stty);
	i = system(p);
	j = errno;
	if (i)
		ck_printf(CK_PASS, "<%s>: %d %d\n", p, i, j);
	free(p);
	return (0);
}

void
serial_lock(struct serial *sc)
{
	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	CHECK_OBJ_NOTNULL(sc->shared, SSERIAL_MAGIC);
	AZ(pthread_mutex_lock(&sc->shared->mtx));
}

void
serial_unlock(struct serial *sc)
{
	CHECK_OBJ_NOTNULL(sc, SERIAL_MAGIC);
	CHECK_OBJ_NOTNULL(sc->shared, SSERIAL_MAGIC);
	AZ(pthread_mutex_unlock(&sc->shared->mtx));
}
