/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: sitegrp.c,v 1.42 2008/04/07 20:11:16 phk Exp $
 *
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>

#include "measured.h"
#include "measured_int.h"

static pthread_rwlock_t	topo_lock;

static void iDestroyGroup(struct group *);
static void iDestroyPoint(struct point *pp);

/* Locking -----------------------------------------------------------*/

void
TopoHold(void)
{

	AZ(pthread_rwlock_rdlock(&topo_lock));
}

void
TopoRel(void)
{

	AZ(pthread_rwlock_unlock(&topo_lock));
}

static void
TopoLock(void)
{

	AZ(pthread_rwlock_wrlock(&topo_lock));
}

/* Sites -------------------------------------------------------------*/

struct site *
CreateSite(unsigned instance)
{
	struct site *sp, *sp2;

	assert(instance < 1000);
	ALLOC_OBJ(sp, SITE_MAGIC);
	instance <<= SITE_SHIFT;
	sp->instance = instance;
	TAILQ_INIT(&sp->groups);
	TopoLock();
	TAILQ_FOREACH(sp2, &all_sites, list) {
		if (sp2->instance == instance) {
			TopoRel();
			free(sp);
			return (NULL);
		}
		if (sp2->instance > instance)
			break;
	}
	if (sp2 == NULL)
		TAILQ_INSERT_TAIL(&all_sites, sp, list);
	else
		TAILQ_INSERT_BEFORE(sp2, sp, list);
	PostEvent(E_NEW_SITE, sp, SI(sp));
	TopoRel();
	return (sp);
}

void
DestroySite(struct site *sp)
{

	CHECK_OBJ_NOTNULL(sp, SITE_MAGIC);
	TopoLock();
	while (!TAILQ_EMPTY(&sp->groups))
		iDestroyGroup(TAILQ_FIRST(&sp->groups));
	TAILQ_REMOVE(&all_sites, sp, list);
	PostEvent(E_DESTROY_SITE, sp, SI(sp));
	TopoRel();
	free(sp);
}

void
CleanSite(const struct site *sp)
{

	CHECK_OBJ_NOTNULL(sp, SITE_MAGIC);
	TopoLock();
	while (!TAILQ_EMPTY(&sp->groups))
		iDestroyGroup(TAILQ_FIRST(&sp->groups));
	TopoRel();
}

unsigned
SI(const struct site *sp)
{

	CHECK_OBJ_NOTNULL(sp, SITE_MAGIC);
	return (sp->instance);
}

/* Groups ------------------------------------------------------------*/

struct group *
CreateGroup(struct site *sp, const char *name, unsigned instance)
{
	struct group *gp, *gp2;

	assert(instance < 1000);
	if (sp == NULL)
		sp = site_zero;
	ALLOC_OBJ(gp, GROUP_MAGIC);
	gp->site = sp;
	gp->name = strdup(name);
	instance <<= GROUP_SHIFT;
	instance |= sp->instance;
	gp->instance = instance;
	TAILQ_INIT(&gp->points);
	TopoLock();
	TAILQ_FOREACH(gp2, &sp->groups, list) {
		if (gp2->instance == instance) {
			TopoRel();
			free(gp->name);
			free(gp);
			return (NULL);
		}
		if (gp2->instance > instance)
			break;
	}
	if (gp2 == NULL)
		TAILQ_INSERT_TAIL(&sp->groups, gp, list);
	else
		TAILQ_INSERT_BEFORE(gp2, gp, list);
	PostEvent(E_NEW_GROUP, gp, GI(gp));
	TopoRel();
	return (gp);
}

static void
iDestroyGroup(struct group *gp)
{

	CHECK_OBJ_NOTNULL(gp, GROUP_MAGIC);
	while (!TAILQ_EMPTY(&gp->points))
		iDestroyPoint(TAILQ_FIRST(&gp->points));
	TAILQ_REMOVE(&gp->site->groups, gp, list);
	PostEvent(E_DESTROY_GROUP, gp, GI(gp));
	free(gp->name);
	free(gp);
}

void
CleanGroup(const struct group *gp)
{

	CHECK_OBJ_NOTNULL(gp, GROUP_MAGIC);
	TopoLock();
	while (!TAILQ_EMPTY(&gp->points))
		iDestroyPoint(TAILQ_FIRST(&gp->points));
	TopoRel();
}

void
DestroyGroup(struct group *gp)
{

	CHECK_OBJ_NOTNULL(gp, GROUP_MAGIC);
	TopoLock();
	iDestroyGroup(gp);
	TopoRel();
}

unsigned
GI(const struct group *gp)
{

	CHECK_OBJ_NOTNULL(gp, GROUP_MAGIC);
	return (gp->instance);
}

/* Points ------------------------------------------------------------*/

struct point *
CreatePoint(struct group *gp, unsigned instance, enum point_type type)
{
	struct point *pp, *pp2;

	assert(instance < 1000);
	if (gp == NULL)
		return (NULL);
	ALLOC_OBJ(pp, POINT_MAGIC);
	pp->group = gp;
	instance |= gp->instance;
	pp->instance = instance;
	pp->type = type;
	pp->d_timeout = DEFAULT_POINT_TIMEOUT;
	pp->scale = 1.0;
	pp->hysteresis = 0;
	pp->limitdelay = 0;
	pp->limits[L_LOW] = NAN;
	pp->limits[L_SAG] = NAN;
	pp->limits[L_RAISE] = NAN;
	pp->limits[L_HIGH] = NAN;
	pp->limits[L_TIMEOUT] = NAN;
	pp->limits[L_CHANGED] = NAN;
	pp->prevlim = L_NORMAL;
	pp->curlim = L_NORMAL;
	pp->measured = Utc();
	pp->raw = NAN;
	pp->value = NAN;
	TopoLock();
	TAILQ_FOREACH(pp2, &gp->points, list) {
		if (pp2->instance == instance) {
			TopoRel();
			free(pp);
			return (NULL);
		}
		if (pp2->instance > instance)
			break;
	}
	if (pp2 == NULL)
		TAILQ_INSERT_TAIL(&gp->points, pp, list);
	else
		TAILQ_INSERT_BEFORE(pp2, pp, list);
	PostEvent(E_NEW_POINT, pp, PI(pp));
	TopoRel();
	return (pp);
}

static void
iDestroyPoint(struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	TAILQ_REMOVE(&pp->group->points, pp, list);
	PostEvent(E_DESTROY_POINT, pp, PI(pp));
	free(pp->label);
	free(pp->format);
	free(pp->unit);
	StateLimit(pp, L_NORMAL);
	pp->state = 0;
	StatePoint(pp);
	free(pp);
}

void
DestroyPoint(struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	TopoLock();
	iDestroyPoint(pp);
	TopoRel();
}

unsigned
PI(const struct point *pp)
{

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	return (pp->instance);
}

/*--------------------------------------------------------------------*/

unsigned
CreatePoints(struct group *gp, struct point ***pp, struct ptdesc *pt)
{
	struct ptdesc *p;
	unsigned u, v, w, retval;

	u = 0; v = 0;
	for(p = pt; p->label != NULL; p++) {
		if (p->idx >= u)
			u = p->idx + 1;
		v++;
	}
	if (v > u)
		u = v;
	*pp = calloc(u, sizeof *pp);
	assert(pp != NULL);
	retval = u;
	u = 0; v = 0;
	for(p = pt; p->label != NULL; p++) {
		u = p->idx;
		if (u == 0)
			u = v;	
		w = p->point;
		if (w == 0)
			w = v + 1;
		(*pp)[u] = CreatePoint(gp, w, p->type);
		SetPointLabel((*pp)[u], p->label);
		v++;
	}
	return (retval);
}

/*--------------------------------------------------------------------*/

void
InitTopo(void)
{

	AZ(pthread_rwlock_init(&topo_lock, NULL));
	site_zero = CreateSite(0);
}
