/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: slave.c,v 1.29 2007/06/28 11:17:51 phk Exp $
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <configkit.h>
#include <sys/socket.h>

#include "phk_argv.h"
#include "phk_tcp.h"
#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

struct slave {
	unsigned		magic;
#define SLAVE_MAGIC		0x2849253
	unsigned		instance;
	struct site		*site;
	struct ck_instance	cki;

	double			timeout;
	double			retry;
	char			*hostname;
	char			*port;
	pthread_t 		tp;
	int			fd;
	int			state;
};

static void
slave_printf(struct slave *sc, const char *fmt, ...)
{
	va_list ap;
	char *p = NULL;
	int l, i = 0;

	CHECK_OBJ_NOTNULL(sc, SLAVE_MAGIC);
	if (sc->fd < 0)
		return;
	va_start(ap, fmt);
	l = vasprintf(&p, fmt, ap);
	va_end(ap);
	if (l > 0) 
		i = write(sc->fd, p, l);
	free(p);
	if (l <= 0 || i != l) {
		Log(SI(sc->site), "write error: %s", strerror(errno));
		close(sc->fd);
		sc->fd = -1;
	}
}

static void
slave_control(struct point *p, void *p1, void *p2 __unused, double val)
{
	struct slave *sc;

	CAST_OBJ_NOTNULL(sc, p1, SLAVE_MAGIC);

	PDebug(p, "SlaveControl(%g)", val);
	slave_printf(sc, "CONTROL 0.%u.%u %g\n",
	    GROUP_OF(PI(p)), POINT_OF(PI(p)), val);
}

static int
slave_event_site(const struct slave *sc, unsigned instance, char **av)
{

	if (SITE_OF(instance) != 0)
		return (1);
	if (!strcasecmp(av[0], "NEW"))
		return(0);
	if (!strcasecmp(av[0], "DESTROY"))
		return(0);
	Debug(SI(sc->site), "%I ev_site(%Q)", instance, av[0]);
	return (__LINE__);
}

static int
slave_event_group(const struct slave *sc, unsigned instance, char **av)
{
	struct group *g;

	if (SITE_OF(instance) != 0)
		return (__LINE__);
	TAILQ_FOREACH(g, &sc->site->groups, list)
		if (GROUP_OF(instance) == GROUP_OF(g->instance))
			break;
	if (!strcasecmp(av[0], "NEW")) {
		if (g != NULL) {
			GDebug(g, "Group already exists");
			return (__LINE__);
		}
		if (av[1] == NULL) {
			Debug(SI(sc->site), "No name ??");
			return (__LINE__);
		}
		g = CreateGroup(sc->site, av[1], GROUP_OF(instance));
		assert(g != NULL);
		return (0);
	}
	if (!strcasecmp(av[0], "DESTROY")) {
		if (g == NULL) {
			GDebug(g, "Group does not exists");
			return (1);
		}
		DestroyGroup(g);
		return (0);
	}
	Debug(SI(sc->site), "%I ev_group(%Q)", instance, av[0]);
	return (__LINE__);
}

static int
slave_event_point(struct slave *sc, unsigned instance, char **av)
{
	struct group *g;
	struct point *p;
	enum point_type e;
	unsigned u;
	double a;
	int i;
	struct timespec ts;

	if (SITE_OF(instance) != 0)
		return (__LINE__);
	TAILQ_FOREACH(g, &sc->site->groups, list)
		if (GROUP_OF(instance) == GROUP_OF(g->instance))
			break;
	if (g == NULL) {
		Debug(SI(sc->site), "No group");
		return(__LINE__);
	}
	TAILQ_FOREACH(p, &g->points, list)
		if (POINT_OF(instance) == POINT_OF(p->instance))
			break;
	if (!strcasecmp(av[0], "NEW")) {
		if (p != NULL) {
			PDebug(p, "Point already exists");
			return (__LINE__);
		}
		if (av[1] == NULL) {
			Debug(SI(sc->site), "No type ??");
			return (__LINE__);
		}
		e = TypePoint(av[1]);
		if (e == UNDEF) {
			Debug(SI(sc->site), "Wrong pointtype");
			return (__LINE__);
		}
		p = CreatePoint(g, POINT_OF(instance), e);
		assert(p != NULL);
		if (e == DOUT || e == AOUT) 
			SetPointControl(p, slave_control, sc, NULL);
		return (0);
	}
	if (p == NULL) {
		Debug(SI(sc->site), "No point");
		return (__LINE__);
	}
	if (!strcasecmp(av[0], "LABEL")) {
		SetPointLabel(p, av[1]);
		return (0);
	}
	if (!strcasecmp(av[0], "CHANGED")) {
		PostEvent(E_POINT_CHANGED, p, PI(p));
		SetPointUnit(p, av[1]);
		return (0);
	}
	if (!strcasecmp(av[0], "UNIT")) {
		SetPointUnit(p, av[1]);
		return (0);
	}
	if (!strcasecmp(av[0], "FORMAT")) {
		SetPointFormat(p, av[1]);
		return (0);
	}
	if (!strcasecmp(av[0], "LIMIT")) { 
		p->curlim = TypeLimit(av[1]);
		PostEvent(E_POINT_LIMIT, p, PI(p));
		return (0);
	}
	a = 0.0;
	if (av[1] != NULL) {
		if (!strcasecmp(av[1], "nan"))
			a = NAN;
		else
			a = strtod(av[1], NULL);
	}

#define FP_HANDLE(foo, field)						\
	do {								\
		if (!strcasecmp(av[0], #foo)) {				\
			p->field = a;					\
			PostEvent(E_POINT_##foo, p, PI(p));		\
			return (0);					\
		}							\
	} while (0)		

	FP_HANDLE(LOW, limits[L_LOW]);
	FP_HANDLE(SAG, limits[L_SAG]);
	FP_HANDLE(RAISE, limits[L_RAISE]);
	FP_HANDLE(HIGH, limits[L_HIGH]);
	FP_HANDLE(TIMEOUT, limits[L_TIMEOUT]);
	FP_HANDLE(CHANGED, limits[L_CHANGED]);
	FP_HANDLE(LIMITDELAY, limitdelay);
	FP_HANDLE(TRAP, trap);		/*lint !e524 !e732 double->uint */
	FP_HANDLE(HYSTERESIS, hysteresis); 	/*lint !e732 double->uint */
	if (!strcasecmp(av[0], "VALUE")) {
		i = sscanf(av[2], "%ld.%ld", &ts.tv_sec, &ts.tv_nsec);
		if (i != 2) {
			PDebug(p, "Timespec parse");
			return (0);
		}
		p->raw = a;
		p->value = a;
		p->measured = ts;
		PostEvent(E_POINT_VALUE, p, PI(p));
		return (0);
	}
	for (u = 0; av[u] != NULL; u++)
		Debug(SI(sc->site),
		    "ev_point(av[%u] %V)", u, av[u]);
	return (__LINE__);
}

static int
slave_event(struct slave *sc, char **av)
{
	unsigned u;

	if (av[0] == NULL || av[1] == NULL || av[2] == NULL)
		return (1);
	if (ParseInstance(av[0], &u))
		return (1);
	if (!strcasecmp(av[1], "p"))
		return (slave_event_point(sc, u, &av[2]));
	if (!strcasecmp(av[1], "g"))
		return (slave_event_group(sc, u, &av[2]));
	if (!strcasecmp(av[1], "s"))
		return (slave_event_site(sc, u, &av[2]));
	return (1);
}

static void
slave_update_timeout(struct slave *sc)
{
	struct timeval tv;

	CHECK_OBJ_NOTNULL(sc, SLAVE_MAGIC);
	if (sc->fd < 0)
		return;
	tv.tv_sec = (int)sc->timeout;
	tv.tv_usec = (sc->timeout - tv.tv_sec) * 1e6;
	assert(0 == setsockopt(sc->fd, SOL_SOCKET, SO_RCVTIMEO,
	    &tv, sizeof tv));
}

static int
slave_lineup(void *priv, const char *p)
{
	struct slave *sc;
	char **av;
	int i;
	unsigned u;

	CAST_OBJ_NOTNULL(sc, priv, SLAVE_MAGIC);
	if (sc->fd < 0 || sc->state < 0)
		return (-1);
	av = PhkArgvParse(p, 0);
	if (av[0] != NULL) {
		Log(SI(sc->site), "PROTO error %s", av[0]);
		slave_printf(sc, "ERROR %s\n", av[0]);
		PhkArgvFree(av);
		return (-1);
	}
	
	if (av[1] != NULL && !strcmp(av[1], "DEBUG")) {
		PhkArgvFree(av);
		return(0);
	}
	switch (sc->state) {
	case 0:
		if (av[1] == NULL)
			break;
		if (strcmp(av[1], "HTTP/1.1"))
			sc->state = -1;
		else if (strcmp(av[2], "200"))
			sc->state = -1;
		else
			sc->state++;
		break;
	case 1:
		if (av[1] == NULL) {
			slave_printf(sc, "PATTERN 0\n");
			slave_printf(sc, "EVENTS\n");
			sc->state++;
		}
		break;
	case 2:
		if (av[1] == NULL)
			break;
		if (!strcasecmp(av[1], "E")) {
			i = slave_event(sc, &av[2]);
			if (i) {
				Debug(0, "PROTO err line %d",
				    i);
				sc->state = -1;
			}
		} else {
			Debug(SI(sc->site), "state %d", sc->state);
			for (u = 1; av[u] != NULL; u++)
				Debug(SI(sc->site),
				    "Argv[%u] %V", u, av[u]);
			sc->state = -1;
		}
		break;
		
	default:
		Debug(SI(sc->site), "state %d", sc->state);
		for (u = 1; av[u] != NULL; u++)
			Debug(SI(sc->site),
			    "Argv[%u] %V", u, av[u]);
		break;
	}
	PhkArgvFree(av);
	return (0);
}

static void *
slave_thread(void *priv)
{
	struct slave *sc;
	struct lineup *lup;
	struct linger linger;
	int i, e;

	CAST_OBJ_NOTNULL(sc, priv, SLAVE_MAGIC);

	while (1) {
		/* XXX: backoff timing */
		sc->fd = PhkTcpOpen(sc->hostname, sc->port);
		if (sc->fd < 0) {
			Log(SI(sc->site),
			    "Connect failed, trying again in %g sec",
			    sc->retry);
			usleep((int)(sc->retry * 1e6));
			continue;
		}
		linger.l_onoff = 0;
		linger.l_linger = 0;
		assert(0 == setsockopt(sc->fd, SOL_SOCKET, SO_LINGER,
		    &linger, sizeof linger));
		slave_update_timeout(sc);
		Log(SI(sc->site), "Connected to slave %d (%s %s) on fd %d",
		    sc->instance, sc->hostname, sc->port, sc->fd);
		lup = NewLineup(sc, slave_lineup);
		slave_printf(sc, "GET /master HTTP/1.1\n\n");
		sc->state = 0;
		e = 0;
		do {
			i = LineUpFd(sc->fd, lup);
			e = errno;
		} while (sc->fd >= 0 && i == 0);
		Log(SI(sc->site), "Lineup failed: %d, fd %d, errno %d: %s",
		    i, sc->fd, e, strerror(e));
		DestroyLineup(lup);
		Log(SI(sc->site), "Dropping connection to slave %d (%s %s)",
		    sc->instance, sc->hostname, sc->port);
		CleanSite(sc->site);
		if (sc->fd >= 0)
			close(sc->fd);
		sc->fd = -1;
		sleep(1);
	}
}

static void
slave_reconf(CK_ARG, struct slave *sc)
{
	void *vp;

	if (sc->tp) {
		AZ(pthread_cancel(sc->tp));
		AZ(pthread_join(sc->tp, &vp));
		sc->tp = NULL;
	}

	if (sc->fd >= 0) {
		close(sc->fd);
		sc->fd = -1;
	}

	CleanSite(sc->site);

	if (sc->hostname == NULL || sc->port == NULL)
		return;

	AZ(pthread_create(&sc->tp, NULL, slave_thread, sc));
}

/* ConfigKit ---------------------------------------------------------*/

struct ck_instance *
cfg_slave_new(CK_ARG, unsigned site)
{
	struct slave *sc;
	struct site *sp;

	if (site < 1) {
		ck_printf(CK_PASS, "Remote site must have number 1...");
		return (NULL);
	}
	sp = CreateSite(site);
	if (sp == NULL) {
		ck_printf(CK_PASS, "Already a site numbered %u\n", site);
		return (NULL);
	}
	ALLOC_OBJ(sc, SLAVE_MAGIC);
	sc->cki.obj = sc;
	sc->fd = -1;
	sc->instance = site;
	sc->site = sp;
	sc->timeout = 30.0;
	sc->retry = 10.0;
	sc->port = strdup("2040");
	assert(sc->port != NULL);
	return (&sc->cki);
}

int
cfg_slave_destroy(CK_ARG)
{
	struct slave *sc;

	CAST_OBJ_NOTNULL(sc, cfa->priv, SLAVE_MAGIC);
	free(sc->hostname);
	sc->hostname = NULL;
	free(sc->port);
	sc->port = NULL;
	slave_reconf(CK_PASS, sc);
	DestroySite(sc->site);
	free(sc);
	return (0);
}

WordVar(slave, SLAVE_MAGIC, hostname, hostname, slave_reconf(CK_PASS, sc););
WordVar(slave, SLAVE_MAGIC, port, port, slave_reconf(CK_PASS, sc););
RealVar(slave, SLAVE_MAGIC, timeout, timeout,
    30.0, "%g", slave_update_timeout(sc));
RealVar(slave, SLAVE_MAGIC, retry, retry, 10.0, "%g", );
