/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: state.c,v 1.4 2007/06/12 20:48:06 phk Exp $
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/sbuf.h>
#include <configkit.h>


#include "measured.h"
#include "measured_int.h"
#include "configkit_conf.h"

struct state {
	unsigned		magic;
#define STATE_MAGIC		0x4c38cdd5
	unsigned		stn;
	char			*string;
	struct ck_instance	cki;
	TAILQ_ENTRY(state)	list;
	unsigned		ref;
	unsigned		nmajor;
	unsigned		nminor;
	unsigned		nfailed;
	enum statepri_e		pri;
};

static TAILQ_HEAD(,state)	statehead = TAILQ_HEAD_INITIALIZER(statehead);

static struct state *
NewState(unsigned stn)
{
	struct state *st;

	assert (stn >= 1 && stn < MAX_TRAP);
	TAILQ_FOREACH(st, &statehead, list) {
		if (st->stn == stn)
			return (st);
	}
	ALLOC_OBJ(st, STATE_MAGIC);
	st->stn = stn;
	st->pri = SP_UNKNOWN;
	TAILQ_INSERT_TAIL(&statehead, st, list);
	return (st);
}

static void
DelState(struct state *st)
{

	if (st->ref > 0)
		return;
	if (st->string != NULL)
		return;
	TAILQ_REMOVE(&statehead, st, list);
	FREE_OBJ(st);
}

static void
UpdStateLimit(struct point *pp, enum limit_e lim, int new)
{
	struct state *st;
	unsigned *up;
	enum statepri_e npri;
	enum limit_e oldlim = L_INVALID;

	CHECK_OBJ_NOTNULL(pp, POINT_MAGIC);
	if (!new) {
		if (lim == pp->curlim)
			return;
		oldlim = pp->curlim;
		pp->curlim = lim;
		PostEvent(E_POINT_LIMIT, pp, PI(pp));	/*lint !e641 */
	}
	if (pp->state_p == NULL)
		return;
	st = pp->state_p;
	CHECK_OBJ_NOTNULL(st, STATE_MAGIC);
	if (pp->state_pri == SP_MAJOR)
		up = &st->nmajor;
	else if (pp->state_pri == SP_MINOR)
		up = &st->nminor;
	else {
		fprintf(stderr, "StateLimit, illegal state_pri %u\n",
		    pp->state_pri);
		abort();
	}
	if (!new) {
		switch (oldlim) {
		case L_NORMAL: break;
		case L_TIMEOUT: st->nfailed--; break;
		default: (*up)--; break;
		}
	}
	switch (lim) {
	case L_NORMAL: break;
	case L_TIMEOUT: st->nfailed++; break;
	default: (*up)++; break;
	}
	if (st->ref == 0)
		npri = SP_UNKNOWN;
	else if (st->nfailed)
		npri = SP_FAILED;
	else if (st->nmajor)
		npri = SP_MAJOR;
	else if (st->nminor)
		npri = SP_MINOR;
	else
		npri = SP_OK;
	if (st->pri == npri)
		return;
	st->pri = npri;
	if (st->string != NULL) {
		PostEvent(E_STATE_CHG, st, PI(pp));
	}
}

void
StateLimit(struct point *pp, enum limit_e lim)
{
	
	UpdStateLimit(pp, lim, 0);
}

void
StatePoint(struct point *pp)
{
	struct state *st;

	if (pp->state_p == NULL && pp->state == 0)
		return;

	/* Delete from current state */
	TAILQ_FOREACH(st, &statehead, list) {
		if (st == pp->state_p) {
			assert(st->ref > 0);
			st->ref--;
			pp->state_p = NULL;
			DelState(st);
			break;
		}
	}
	if (pp->state_p != NULL) {
		fprintf(stderr, "State_Point with invalid state_p\n");
		abort();
	}
	if (pp->state == 0)
		return;
	/* Add */
	TAILQ_FOREACH(st, &statehead, list) {
		if (st->stn == pp->state) 
			break;
	}
	if (st == NULL)
		st = NewState(pp->state);
	st->ref++;
	pp->state_p = st;
	UpdStateLimit(pp, pp->curlim, 1);
}

int
StateDump(void **p)
{
	struct state *st;

	CAST_OBJ(st, *p, STATE_MAGIC);
	if (st == NULL) {
		st = TAILQ_FIRST(&statehead);
	} else {
		st = TAILQ_NEXT(st, list);
	}
	while (st != NULL && st->string == NULL)
		st = TAILQ_NEXT(st, list);
	*p = st;
	return (st != NULL);
}

void
StateMaster(struct sbuf *sb, void *arg)
{
	struct state *st;

	CAST_OBJ_NOTNULL(st, arg, STATE_MAGIC);
	if (st->string != NULL)
		sbuf_printf(sb, "%u %u %Q", st->stn, st->pri, st->string);
}

struct ck_instance *
cfg_state_new(CK_ARG, unsigned stn)
{
	struct state *st;

	if (stn < 1 || stn > MAX_TRAP) {
		ck_printf(CK_PASS,
		    "state site must have number 1...%d", MAX_TRAP);
		return (NULL);
	}
	st = NewState(stn);
	st->cki.obj = st;
	return (&st->cki);
}

int
cfg_state_destroy(CK_ARG)
{
        struct state *st;

        CAST_OBJ_NOTNULL(st, cfa->priv, STATE_MAGIC);
	free(st->string);
	st->string = NULL;
	DelState(st);
        return (0);
}

WordVar(state, STATE_MAGIC, string, string,
	PostEvent(E_STATE_CHG, sc, sc->stn));
