/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: tcp.c,v 1.17 2007/03/27 09:31:26 phk Exp $
 *
 */

#include <assert.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "measured.h"
#include "measured_int.h"


struct tcp_listener {
	unsigned		magic;
#define TCP_LISTENER_MAGIC	0x588223
	int			fd;
	pthread_t		thread;
	char			*port;
	tcp_handler_t		*func;
	char			*name;
};

struct tcp_instance {
	unsigned		magic;
#define TCP_INSTANCE_MAGIC	0x2885738
	int			fd;
	tcp_handler_t		*func;
};

static void *
tcp_thread(void *p)
{
	struct tcp_instance *sc;

	CAST_OBJ_NOTNULL(sc, p, TCP_INSTANCE_MAGIC);
	sc->func(sc->fd);
	FREE_OBJ(sc);
	pthread_exit(NULL);
}

static void *
my_thread(void *p)
{
	int i, j, s;
	unsigned l;
	struct addrinfo ai, *r0;
	struct tcp_listener *tl;
	struct sockaddr in;
	char buf1[64], buf2[20];
	struct tcp_instance *ti;
	pthread_t thread;


	CAST_OBJ_NOTNULL(tl, p, TCP_LISTENER_MAGIC);
	memset(&ai, 0, sizeof ai);
	ai.ai_family = PF_INET;
	ai.ai_socktype = SOCK_STREAM;
	ai.ai_flags = AI_PASSIVE;
	i = getaddrinfo(NULL, tl->port, &ai, &r0);

	if (i) {
		Notice(0, "TCP port %s: %s", gai_strerror(i));
		freeaddrinfo(r0);
		return (NULL);
	}

	s = socket(r0->ai_family, r0->ai_socktype, r0->ai_protocol);
	assert(s >= 0);
	tl->fd = s;

	i = 1;
	AZ(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i));

	i = bind(s, r0->ai_addr, r0->ai_addrlen);
	if (i < 0) {
		Notice(0, "TCP port %s: %M", tl->port, errno);
		freeaddrinfo(r0);
		return (NULL);
	}

	freeaddrinfo(r0);

	AZ(listen(s, 5));	/* XXX: 5 ? */

	while (1) {
		i = accept(s, NULL, NULL);
		l = sizeof in;
		AZ(getpeername(i, &in, &l));

		j = getnameinfo(&in, l, buf1, sizeof buf1,
		    buf2, sizeof buf2, NI_NUMERICSERV | NI_NUMERICHOST);
		if (j == 0) 
			Log(0, "Accept %s from %s:%s on port %s fd %d",
			    tl->name, buf1, buf2, tl->port, i);
		else {
			Log(0, "getnameinfo: %s", gai_strerror(j));
			Log(0, "Accept %s from [%H] on port %s fd %d",
			    tl->name, &in, sizeof in, tl->port, i);
		}
		ALLOC_OBJ(ti, TCP_INSTANCE_MAGIC);
		ti->fd = i;
		ti->func = tl->func;
		i = 1;
		AZ(setsockopt(ti->fd, SOL_SOCKET, SO_NOSIGPIPE, &i, sizeof i));
		AZ(pthread_create(&thread, NULL, tcp_thread, ti));
		AZ(pthread_detach(thread));
	}
}

struct tcp_listener *
TcpListenTo(const char *port, tcp_handler_t *func, const char *name)
{
	struct tcp_listener *tl;

	ALLOC_OBJ(tl, TCP_LISTENER_MAGIC);
	tl->port = strdup(port);
	tl->func = func;
	tl->fd = -1;
	tl->name = strdup(name);

	AZ(pthread_create(&tl->thread, NULL, my_thread, tl));
	return (tl);
}

void
TcpUnlistenTo(struct tcp_listener *tl)
{
	void *vp;

	CAST_OBJ_NOTNULL(tl, tl, TCP_LISTENER_MAGIC);
	AZ(pthread_cancel(tl->thread));
	AZ(pthread_join(tl->thread, &vp));
	close(tl->fd);
	free(tl->port);
	free(tl->name);
	free(tl);
}
