/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: utils.c,v 1.7 2007/09/05 13:15:10 phk Exp $
 *
 * This file contains functions which can be shared with other programs,
 * like for instance the MIB module.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include "measured_api.h"
#include "phk_obj.h"

int
ParseInstance(const char *str, unsigned *instance)
{
	int i;
	unsigned s, g, p;

	i = sscanf(str, "%u.%u.%u", &s, &g, &p);
	if (i != 3)
		return (-1);
	if (s & ~POINT_MASK)
		return (-1);
	if (g & ~POINT_MASK)
		return (-1);
	if (p & ~POINT_MASK)
		return (-1);
	*instance =
	    (s << SITE_SHIFT) | (g << GROUP_SHIFT) | (p << POINT_SHIFT);
	return (0);
}

const char *
LimitType(enum limit_e e)
{

	switch(e) {
	case L_LOW:	return ("low");
	case L_SAG:	return ("sag");
	case L_RAISE:	return ("raise");
	case L_HIGH:	return ("high");
	case L_NORMAL:	return ("normal");
	case L_TIMEOUT:	return ("timeout");
	case L_CHANGED:	return ("changed");
	default:
		return (NULL);
	}
}

enum limit_e 
TypeLimit(const char *pp)
{

	if (!strcasecmp(pp, "low"))		return (L_LOW);
	if (!strcasecmp(pp, "sag"))		return (L_SAG);
	if (!strcasecmp(pp, "raise"))		return (L_RAISE);
	if (!strcasecmp(pp, "high"))		return (L_HIGH);
	if (!strcasecmp(pp, "normal"))		return (L_NORMAL);
	if (!strcasecmp(pp, "timeout"))		return (L_TIMEOUT);
	if (!strcasecmp(pp, "changed"))		return (L_CHANGED);
	return (L_INVALID);
}

const char *
PointType(enum point_type t)
{

	switch(t) {
	case UNDEF:	return ("undef");
	case AIN:	return ("ain");
	case DIN:	return ("din");
	case AOUT:	return ("aout");
	case DOUT:	return ("dout");
	case COUNT:	return ("counter");
	default:	return ("inval");
	}
}

enum point_type 
TypePoint(const char *pp)
{

	if (!strcasecmp(pp, "ain"))		return AIN;
	if (!strcasecmp(pp, "din"))		return DIN;
	if (!strcasecmp(pp, "dout"))		return DOUT;
	if (!strcasecmp(pp, "aout"))		return AOUT;
	if (!strcasecmp(pp, "count"))		return COUNT;
	if (!strcasecmp(pp, "counter"))		return COUNT;
	return (UNDEF);
}

/* LineUp ------------------------------------------------------------*/

struct lineup {
	unsigned	magic;
#define LINEUP_MAGIC	0x8286661
	char		*buf;
	unsigned	bufl;
	unsigned	bufp;
	void		*priv;
	lineup_f	*func;
};

struct lineup *
NewLineup(void *priv, lineup_f *func)
{
	struct lineup *l;

	ALLOC_OBJ(l, LINEUP_MAGIC);
	l->func = func;
	l->priv = priv;
	l->bufl = BUFSIZ;
	l->buf = malloc(l->bufl + 1);
	assert(l->buf != NULL);
	return (l);
}

void
DestroyLineup(struct lineup *l)
{

	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
	free(l->buf);
	FREE_OBJ(l);
}

static int
LineUpProcess(struct lineup *l)
{
	char *p, *q;
	int i;

	l->buf[l->bufp] = '\0';
	for (p = l->buf; *p != '\0'; p = q) {
		q = strchr(p, '\n');
		if (q == NULL)
			break;
		*q++ = '\0';
		i = l->func(l->priv, p);
		if (i != 0)
			return (i);
	}
	if (*p != '\0') {
		q = strchr(p, '\0');
		assert(q != NULL);
		l->bufp = (unsigned)(q - p);
		memmove(l->buf, p, l->bufp);
		l->buf[l->bufp] = '\0';
	} else
		l->bufp = 0;
	return (0);
}

int
LineUpFd(int fd, struct lineup *l)
{
	int i;

	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
	i = read(fd, l->buf + l->bufp, l->bufl - l->bufp);
	if (i <= 0)
		return (-1);
	l->bufp += i;
	return (LineUpProcess(l));
}

int
LineUpFile(FILE *f, struct lineup *l)
{
	char *p;

	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
	p = fgets(l->buf + l->bufp, l->bufl - l->bufp, f);
	if (p == NULL)
		return (-1);
	l->bufp = strlen(l->buf);
	return (LineUpProcess(l));
}

int
fdprintf(int fd, const char *format, ...)
{
	char *p;
	va_list ap;
	int i, l;

	va_start(ap, format);
	p = NULL;
	l = vasprintf(&p, format, ap);
	va_end(ap);
	i = write(fd, p, l);
	free(p);
	if (i < 0 || i != l)
		return (-1);
	return (0);
}
